// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

// Title: axi4_top
//

package mgc_axi4_pkg;
import QUESTA_MVC::*;

`ifdef QUESTA
// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

// Title: AXI4 Enumeration Types
//

//------------------------------------------------------------------------------
//
// Enum: axi4_prot_e
//
//------------------------------------------------------------------------------
//  Protection type mapped to the AxPROT signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.7))
//  
// AXI4_NORM_SEC_DATA - Normal/Secure/Data
// AXI4_PRIV_SEC_DATA - Privileged/Secure/Data
// AXI4_NORM_NONSEC_DATA - Normal/Non-secure/Data
// AXI4_PRIV_NONSEC_DATA - Privileged/Non-secure/Data
// AXI4_NORM_SEC_INST - Normal/Secure/Instruction
// AXI4_PRIV_SEC_INST - Privileged/Secure/Instruction
// AXI4_NORM_NONSEC_INST - Normal/Non-secure/Instruction
// AXI4_PRIV_NONSEC_INST - Privileged/Non-secure/Instruction
//  
typedef enum bit [2:0]
{
    AXI4_NORM_SEC_DATA    = 3'h0,
    AXI4_PRIV_SEC_DATA    = 3'h1,
    AXI4_NORM_NONSEC_DATA = 3'h2,
    AXI4_PRIV_NONSEC_DATA = 3'h3,
    AXI4_NORM_SEC_INST    = 3'h4,
    AXI4_PRIV_SEC_INST    = 3'h5,
    AXI4_NORM_NONSEC_INST = 3'h6,
    AXI4_PRIV_NONSEC_INST = 3'h7
} axi4_prot_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_response_e
//
//------------------------------------------------------------------------------
//  Response type mapped to the xRESP signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)
// 
// AXI4_OKAY   - Normal access has been successful, or exclusive access has failed (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4).
// AXI4_EXOKAY - Exclusive acess okay (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)) OR Reserved value if QVIP BFM is configured as axi4lite interface (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.1.
// AXI4_SLVERR - Slave signals an error to the originating master.
// AXI4_DECERR - Decode error (generally used to indicate that there is no slave at the transaction address).
//  
typedef enum bit [1:0]
{
    AXI4_OKAY   = 2'h0,
    AXI4_EXOKAY = 2'h1,
    AXI4_SLVERR = 2'h2,
    AXI4_DECERR = 2'h3
} axi4_response_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_rw_e
//
//------------------------------------------------------------------------------
//  This is used as the <read_or_write> argument to the <axi4_master_rw_transaction> transaction to indicate read or write.
// 
// AXI4_TRANS_READ - read transaction
// AXI4_TRANS_WRITE - write transaction
//  
typedef enum bit [0:0]
{
    AXI4_TRANS_READ  = 1'h0,
    AXI4_TRANS_WRITE = 1'h1
} axi4_rw_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_size_e
//
//------------------------------------------------------------------------------
//  Word-size encoding mapped to the AxSIZE signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1). 
// 
// AXI4_BYTES_1 - 4'b0000 i.e. 1 byte
// AXI4_BYTES_2 - 4'b0001 i.e. 2 bytes
// AXI4_BYTES_4 - 4'b0010 i.e. 4 bytes
// AXI4_BYTES_8 - 4'b0011 i.e. 8 bytes
// AXI4_BYTES_16 - 4'b0100 i.e. 16 bytes
// AXI4_BYTES_32 - 4'b0101 i.e. 32 bytes
// AXI4_BYTES_64 - 4'b0110 i.e. 64 bytes
// AXI4_BYTES_128 - 4'b0111 i.e. 128 bytes
// AXI4_BYTES_256 - 4'b1000 i.e. 256 bytes
// 
typedef enum bit [3:0]
{
    AXI4_BYTES_1   = 4'h0,
    AXI4_BYTES_2   = 4'h1,
    AXI4_BYTES_4   = 4'h2,
    AXI4_BYTES_8   = 4'h3,
    AXI4_BYTES_16  = 4'h4,
    AXI4_BYTES_32  = 4'h5,
    AXI4_BYTES_64  = 4'h6,
    AXI4_BYTES_128 = 4'h7,
    AXI4_BYTES_256 = 4'h8
} axi4_size_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_cache_e
//
//------------------------------------------------------------------------------
//  Cache behaviour type, mapped to the AxCACHE signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4).
// 
// This has identical values to those in AXI3, but the AxCACHE[1] bit is now named 'Modifiable' (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.3.1)
// 
// AXI4_NONMODIFIABLE_NONBUF - Non-modifiable and non-bufferable
// AXI4_BUF_ONLY - Bufferable only
// AXI4_CACHE_2 - Bit pattern 0010
// AXI4_CACHE_3 - Bit pattern 0011
// AXI4_CACHE_RSVD_4 - Reserved
// AXI4_CACHE_RSVD_5 - Reserved
// AXI4_CACHE_6 - Bit pattern 0110
// AXI4_CACHE_7 - Bit pattern 0111
// AXI4_CACHE_RSVD_8 - Reserved
// AXI4_CACHE_RSVD_9 - Reserved
// AXI4_CACHE_10 - Bit pattern 1010
// AXI4_CACHE_11 - Bit pattern 1011
// AXI4_CACHE_RSVD_12 - Reserved
// AXI4_CACHE_RSVD_13 - Reserved
// AXI4_CACHE_14 - Bit pattern 1110
// AXI4_CACHE_15 - Bit pattern 1111
// 
typedef enum bit [3:0]
{
    AXI4_NONMODIFIABLE_NONBUF = 4'h0,
    AXI4_BUF_ONLY             = 4'h1,
    AXI4_CACHE_2              = 4'h2,
    AXI4_CACHE_3              = 4'h3,
    AXI4_CACHE_RSVD_4         = 4'h4,
    AXI4_CACHE_RSVD_5         = 4'h5,
    AXI4_CACHE_6              = 4'h6,
    AXI4_CACHE_7              = 4'h7,
    AXI4_CACHE_RSVD_8         = 4'h8,
    AXI4_CACHE_RSVD_9         = 4'h9,
    AXI4_CACHE_10             = 4'ha,
    AXI4_CACHE_11             = 4'hb,
    AXI4_CACHE_RSVD_12        = 4'hc,
    AXI4_CACHE_RSVD_13        = 4'hd,
    AXI4_CACHE_14             = 4'he,
    AXI4_CACHE_15             = 4'hf
} axi4_cache_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_burst_e
//
//------------------------------------------------------------------------------
//  Burst type mapped to the AxBURST signals to determine address calculation (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4). 
// 
// AXI4_FIXED - Fixed
// AXI4_INCR - Incrementing
// AXI4_WRAP - Wrap
// AXI4_RESERVED - Reserved value. May be enabled by setting the configuration parameter <config_enable_burst_reserved_value>.
// 
typedef enum bit [1:0]
{
    AXI4_FIXED    = 2'h0,
    AXI4_INCR     = 2'h1,
    AXI4_WRAP     = 2'h2,
    AXI4_RESERVED = 2'h3
} axi4_burst_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_lock_e
//
//------------------------------------------------------------------------------
//  Access type mapped to the AxLOCK signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.4).
// 
// AXI4_NORMAL - Normal Access, respecting barriers
// AXI4_EXCLUSIVE - Exclusive Access, respecting barriers
// 
typedef enum bit [0:0]
{
    AXI4_NORMAL    = 1'h0,
    AXI4_EXCLUSIVE = 1'h1
} axi4_lock_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_interface_type_e
//
//------------------------------------------------------------------------------
//  Type to configure the category of the interface (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2) as read/write, read-only or write-only by setting the value of <config_interface_type> configuration parameter.
// 
// AXI4_READWRITE - Read and Write
// AXI4_READONLY - Read Only
// AXI4_WRITEONLY - Write Only
// 
typedef enum bit [1:0]
{
    AXI4_READWRITE = 2'h0,
    AXI4_READONLY  = 2'h1,
    AXI4_WRITEONLY = 2'h2
} axi4_interface_type_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_assertion_e
//
//------------------------------------------------------------------------------
//  Type defining the error messages which can be produced by the <mgc_axi4> Questa Verification IP.
// 
// Individual error messages can be disabled using the <config_enable_assertion> array of configuration bits.
// 
// AXI4_ADDRESS_WIDTH_EXCEEDS_64                               -  60000 -  AXI4 supports up to 64-bit addressing (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_ARADDR_CHANGED_BEFORE_ARREADY                          -  60001 -  The value of <ARADDR> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARADDR_UNKN                                            -  60002 -  <ARADDR> has an X value/<ARADDR> has an Z value
// AXI4_ARPROT_CHANGED_BEFORE_ARREADY                          -  60003 -  The value of <ARPROT> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARPROT_UNKN                                            -  60004 -  <ARPROT> has an X value/<ARPROT> has an Z value
// AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID                     -  60005 -  Once <ARVALID> has been asserted, <ARREADY> should be asserted within <config_max_latency_ARVALID_assertion_to_ARREADY> clock periods
// AXI4_ARREADY_UNKN                                           -  60006 -  <ARREADY> has an X value/<ARREADY> has a Z value
// AXI4_ARVALID_DEASSERTED_BEFORE_ARREADY                      -  60007 -  <ARVALID> has been de-asserted before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARVALID_HIGH_ON_FIRST_CLOCK                            -  60008 -  A master interface must begin driving <ARVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_ARVALID_UNKN                                           -  60009 -  <ARVALID> has an X value/<ARVALID> has an Z value
// AXI4_AWADDR_CHANGED_BEFORE_AWREADY                          -  60010 -  The value of <AWADDR> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWADDR_UNKN                                            -  60011 -  <AWADDR> has an X value/<AWADDR> has an Z value
// AXI4_AWPROT_CHANGED_BEFORE_AWREADY                          -  60012 -  The value of <AWPROT> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWPROT_UNKN                                            -  60013 -  <AWPROT> has an X value/<AWPROT> has an Z value
// AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID                     -  60014 -  Once <AWVALID> has been asserted, <AWREADY> should be asserted within <config_max_latency_AWVALID_assertion_to_AWREADY> clock periods
// AXI4_AWREADY_UNKN                                           -  60015 -  <AWREADY> has an X value/<AWREADY> has an Z value
// AXI4_AWVALID_DEASSERTED_BEFORE_AWREADY                      -  60016 -  <AWVALID> has been de-asserted before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWVALID_HIGH_ON_FIRST_CLOCK                            -  60017 -  A master interface must begin driving <AWVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_AWVALID_UNKN                                           -  60018 -  <AWVALID> has an X value/<AWVALID> has an Z value
// AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID                       -  60019 -  Once <BVALID> has been asserted, <BREADY> should be asserted within <config_max_latency_BVALID_assertion_to_BREADY> clock periods
// AXI4_BREADY_UNKN                                            -  60020 -  <BREADY> has an X value/<BREADY> has an Z value
// AXI4_BRESP_CHANGED_BEFORE_BREADY                            -  60021 -  The value of <BRESP> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BRESP_UNKN                                             -  60022 -  <BRESP> has an X value/<BRESP> has a Z value
// AXI4_BVALID_DEASSERTED_BEFORE_BREADY                        -  60023 -  <BVALID> has been de-asserted before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BVALID_HIGH_EXITING_RESET                              -  60024 -  <BVALID> should have been driven low when exiting reset (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_BVALID_UNKN                                            -  60025 -  <BVALID> has an X value/<BVALID> has a Z value
// AXI4_DEC_ERR_RESP_FOR_READ                                  -  60026 -  DEPRECATED - This assertion is now DEPRECATED, as DECERR response for a read transaction is not a protocol violation.
// AXI4_DEC_ERR_RESP_FOR_WRITE                                 -  60027 -  DEPRECATED - This assertion is now DEPRECATED, as DECERR response for a write transaction is not a protocol violation.
// AXI4_EXOKAY_RESPONSE_NORMAL_READ                            -  60028 -  Slave has responded <AXI4_EXOKAY> to a non exclusive read transfer
// AXI4_EXOKAY_RESPONSE_NORMAL_WRITE                           -  60029 -  Slave has responded <AXI4_EXOKAY> to a non exclusive write transfer 
// AXI4_INVALID_WRITE_STROBES_ON_UNALIGNED_WRITE_TRANSFER      -  60030 -  Write strobe(s) incorrect for address/size of an unaligned transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.3)
// AXI4_MINIMUM_SLAVE_ADDRESS_SPACE_VIOLATION                  -  60031 -  The minimum address space occupied by a single slave device is 4 kilobytes (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.2)
// AXI4_PARAM_READ_DATA_BUS_WIDTH                              -  60032 -  The value of <AXI4_RDATA_WIDTH> must be either 32 or 64 (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.2)
// AXI4_PARAM_WRITE_DATA_BUS_WIDTH                             -  60033 -  The value of <AXI4_WDATA_WIDTH> must be either 32 or 64 (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.2)
// AXI4_READ_DATA_BEFORE_ADDRESS                               -  60034 -  An unexpected read response has occurred (there are no outstanding read transactions with this id) (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.3)
// AXI4_READ_DATA_CHANGED_BEFORE_RREADY                        -  60035 -  The value of <RDATA> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_READ_DATA_UNKN                                         -  60036 -  <RDATA> has an X value/<RDATA> has a Z value
// AXI4_READ_RESP_CHANGED_BEFORE_RREADY                        -  60037 -  The value of <RRESP> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID                       -  60038 -  Once <RVALID> has been asserted, <RREADY> should be asserted within <config_max_latency_RVALID_assertion_to_RREADY> clock periods
// AXI4_RREADY_UNKN                                            -  60039 -  <RREADY> has an X value/<RREADY> has a Z value
// AXI4_RRESP_UNKN                                             -  60040 -  <RRESP> has an X value/<RRESP> has a Z value
// AXI4_RVALID_DEASSERTED_BEFORE_RREADY                        -  60041 -  <RVALID> has been de-asserted before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RVALID_HIGH_EXITING_RESET                              -  60042 -  <RVALID> should have been driven low when exiting reset (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_RVALID_UNKN                                            -  60043 -  <RVALID> has an X value/<RVALID> has a Z value
// AXI4_SLV_ERR_RESP_FOR_READ                                  -  60044 -  DEPRECATED - This assertion is now DEPRECATED, as SLVERR response for a read transaction is not a protocol violation.
// AXI4_SLV_ERR_RESP_FOR_WRITE                                 -  60045 -  DEPRECATED - This assertion is now DEPRECATED, as SLVERR response for a write transaction is not a protocol violation.
// AXI4_TIMEOUT_WAITING_FOR_READ_RESPONSE                      -  60046 -  Timed-out waiting for a read response (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_TIMEOUT_WAITING_FOR_WRITE_RESPONSE                     -  60047 -  Timed-out waiting for a write response (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_INVALID_LANE            -  60048 -  On a lane whose strobe is 0, the value of <WDATA> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_VALID_LANE              -  60049 -  On a lane whose strobe is 1, the value of <WDATA> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID                       -  60050 -  Once <WVALID> has been asserted, <WREADY> should be asserted within <config_max_latency_WVALID_assertion_to_WREADY> clock periods
// AXI4_WREADY_UNKN                                            -  60051 -  <WREADY> has an X value/<WREADY> has a Z value
// AXI4_WRITE_DATA_BEFORE_ADDRESS                              -  60052 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol write data beat can occurred before the corresponding address phase
// AXI4_WRITE_DATA_UNKN_ON_INVALID_LANE                        -  60053 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol Byte lanes for which strobe is 0 could take unknown value.
// AXI4_WRITE_DATA_UNKN_ON_VALID_LANE                          -  60054 -  On a lane whose strobe is 1, <WDATA> has an X value/<WDATA> has a Z value
// AXI4_WRITE_RESPONSE_WITHOUT_ADDR_DATA                       -  60055 -  An unexpected write response has occurred (there are no outstanding write transactions with this id) 
// AXI4_WSTRB_CHANGED_BEFORE_WREADY                            -  60056 -  The value of <WSTRB> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WSTRB_UNKN                                             -  60057 -  <WSTRB> has an X value/<WSTRB> has an Z value
// AXI4_WVALID_DEASSERTED_BEFORE_WREADY                        -  60058 -  <WVALID> has been de-asserted before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WVALID_HIGH_ON_FIRST_CLOCK                             -  60059 -  A master interface must begin driving <WVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_WVALID_UNKN                                            -  60060 -  <WVALID> has an X value/<WVALID> has an Z value
// MVC_FAILED_POSTCONDITION                                    -  60061 -  A postcondition failed
// MVC_FAILED_RECOGNITION                                      -  60062 -  An item failed to be recognized  
// AXI4_DEC_ERR_ILLEGAL_FOR_MAPPED_SLAVE_ADDR                  -  60063 -  Slave receives a burst to a mapped address but responds with DECERR (signaled by <AXI4_DECERR>) (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)
// AXI4_AWVALID_HIGH_DURING_RESET                              -  60064 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol AWVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                     
// AXI4_WVALID_HIGH_DURING_RESET                               -  60065 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol WVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                       
// AXI4_BVALID_HIGH_DURING_RESET                               -  60066 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol BVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                        
// AXI4_ARVALID_HIGH_DURING_RESET                              -  60067 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol ARVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                          
// AXI4_RVALID_HIGH_DURING_RESET                               -  60068 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol RVALID timing during the reset state  is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                         
// AXI4_TIMEOUT_WAITING_FOR_WRITE_DATA                         -  60069 -  Timed-out waiting for a data phase in write data burst (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_TIMEOUT_WAITING_FOR_WRITE_ADDR_AFTER_DATA              -  60070 -  Timed-out waiting for a write address phase to be coming after data see AMBA AXI and ACE Protocol Specification IHI0022D section A2.2
// AXI4_ARESETn_SIGNAL_Z                                       -  60071 -  AXI4 Reset signal has z value
// AXI4_ARESETn_SIGNAL_X                                       -  60072 -  AXI4 Reset signal has x value
// AXI4_ADDR_FOR_READ_BURST_ACROSS_4K_BOUNDARY                 -  60073 -  This read transaction has crossed a 4KB boundary (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1) 
// AXI4_ADDR_FOR_WRITE_BURST_ACROSS_4K_BOUNDARY                -  60074 -  This write transaction has crossed a 4KB boundary (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ARADDR_FALLS_IN_REGION_HOLE                            -  60075 -  The <ARADDR> value cannot be decoded to a region in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_ARBURST_CHANGED_BEFORE_ARREADY                         -  60076 -  The value of <ARBURST> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARBURST_UNKN                                           -  60077 -  <ARBURST> has an X value/<ARBURST> has an Z value
// AXI4_ARCACHE_CHANGED_BEFORE_ARREADY                         -  60078 -  The value of <ARCACHE> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARCACHE_UNKN                                           -  60079 -  <ARCACHE> has an X value/<ARCACHE> has an Z value
// AXI4_ARID_CHANGED_BEFORE_ARREADY                            -  60080 -  The value of <ARID> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARID_UNKN                                              -  60081 -  <ARID> has an X value/<ARID> has an Z value
// AXI4_ARLEN_CHANGED_BEFORE_ARREADY                           -  60082 -   The value of <ARLEN> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARLEN_UNKN                                             -  60083 -  <ARLEN> has an X value/<ARLEN> has an Z value
// AXI4_ARLOCK_CHANGED_BEFORE_ARREADY                          -  60084 -  The value of <ARLOCK> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARLOCK_UNKN                                            -  60085 -  <ARLOCK> has an X value/<ARLOCK> has an Z value
// AXI4_ARQOS_CHANGED_BEFORE_ARREADY                           -  60086 -  The value of <ARQOS>  has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARQOS_UNKN                                             -  60087 -  <ARQOS>  has an X value/<ARQOS> has an Z value
// AXI4_ARREGION_CHANGED_BEFORE_ARREADY                        -  60088 -  The value of <ARREGION> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARREGION_MISMATCH                                      -  60089 -  The <ARREGION> value does not match the value defined in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_ARREGION_UNKN                                          -  60090 -  <ARREGION> has an X value/<ARREGION> has an Z value
// AXI4_ARSIZE_CHANGED_BEFORE_ARREADY                          -  60091 -  The value of <ARSIZE> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARSIZE_UNKN                                            -  60092 -  <ARSIZE> has an X value/<ARSIZE> has an Z value
// AXI4_ARUSER_CHANGED_BEFORE_ARREADY                          -  60093 -  The value of <ARUSER> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARUSER_UNKN                                            -  60094 -  <ARUSER> has an X value/<ARUSER> has an Z value
// AXI4_AWADDR_FALLS_IN_REGION_HOLE                            -  60095 -  The addr value cannot be decoded to a region in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_AWBURST_CHANGED_BEFORE_AWREADY                         -  60096 -  The value of <AWBURST> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWBURST_UNKN                                           -  60097 -  <AWBURST> has an X value/<AWBURST> has an Z value
// AXI4_AWCACHE_CHANGED_BEFORE_AWREADY                         -  60098 -  The value of <AWCACHE> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWCACHE_UNKN                                           -  60099 -  <AWCACHE> has an X value/AWCACHE has an Z value
// AXI4_AWID_CHANGED_BEFORE_AWREADY                            -  60100 -  The value of <AWID> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWID_UNKN                                              -  60101 -  <AWID> has an X value/<AWID> has an Z value
// AXI4_AWLEN_CHANGED_BEFORE_AWREADY                           -  60102 -  The value of <AWLEN> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWLEN_UNKN                                             -  60103 -  <AWLEN> has an X value/<AWLEN> has an Z value
// AXI4_AWLOCK_CHANGED_BEFORE_AWREADY                          -  60104 -  The value of <AWLOCK> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWLOCK_UNKN                                            -  60105 -  <AWLOCK> has an X value/<AWLOCK> has an Z value
// AXI4_AWQOS_CHANGED_BEFORE_AWREADY                           -  60106 -  The value of <AWQOS> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWQOS_UNKN                                             -  60107 -  <AWQOS> has an X value/<AWQOS> has an Z value
// AXI4_AWREGION_CHANGED_BEFORE_AWREADY                        -  60108 -  The value of <AWREGION> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWREGION_MISMATCH                                      -  60109 -  The <AWREGION> value does not match the value defined in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_AWREGION_UNKN                                          -  60110 -  <AWREGION> has an X value/<AWREGION> has an Z value
// AXI4_AWSIZE_CHANGED_BEFORE_AWREADY                          -  60111 -  The value of <AWSIZE> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWSIZE_UNKN                                            -  60112 -  <AWSIZE> has an X value/<AWSIZE> has an Z value
// AXI4_AWUSER_CHANGED_BEFORE_AWREADY                          -  60113 -  The value of <AWUSER> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWUSER_UNKN                                            -  60114 -  <AWUSER> has an X value/<AWUSER> has an Z value
// AXI4_BID_CHANGED_BEFORE_BREADY                              -  60115 -  The value of <BID> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BID_UNKN                                               -  60116 -  <BID> has an X value/<BID> has a Z value
// AXI4_BUSER_CHANGED_BEFORE_BREADY                            -  60117 -  The value of <BUSER> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BUSER_UNKN                                             -  60118 -  <BUSER> has an X value/<BUSER> has a Z value
// AXI4_EXCLUSIVE_READ_ACCESS_MODIFIABLE                       -  60119 -  Exclusive read access must not have AxCACHE value that indicates that the transaction is cacheable (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_128              -  60120 -  Number of bytes in an exclusive read transaction must be less than or equal to 128 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_128             -  60121 -  Number of bytes in an exclusive write transaction must be less than or equal to 128 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_NOT_POWER_OF_2           -  60122 -  Number of bytes of an exclusive read transaction is not a power of 2 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_NOT_POWER_OF_2          -  60123 -  Number of bytes of an exclusive write transaction is not a power of 2 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_LENGTH_EXCEEDS_16                       -  60124 -  Exclusive read accesses are not permitted to use a burst length greater than 16 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_ADDRESS_NOT_SAME_AS_RD                    -  60125 -  Exclusive write does not match the address of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_BURST_NOT_SAME_AS_RD                      -  60126 -  Exclusive write does not match the burst setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_CACHE_NOT_SAME_AS_RD                      -  60127 -  Exclusive write does not match the cache setting of the previous exclusive read to this id (see the ARM AXI4 compliance-checker AXI4_RECM_EXCL_MATCH assertion code)
// AXI4_EXCLUSIVE_WRITE_ACCESS_MODIFIABLE                      -  60128 -  Exclusive write access must not have AxCACHE value that indicates that the transaction is cacheable (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_LENGTH_NOT_SAME_AS_RD                     -  60129 -  Exclusive write does not match the length of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_PROT_NOT_SAME_AS_RD                       -  60130 -  Exclusive write does not match the prot setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_REGION_NOT_SAME_AS_RD                     -  60131 -  Exclusive write does not match the region setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_SIZE_NOT_SAME_AS_RD                       -  60132 -  Exclusive write does not match the size of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_EXOKAY_RESP_EXPECTED_OKAY                        -  60133 -  Expected <AXI4_OKAY> response to this exclusive read (because the parameters did not meet the the restrictions), but got <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS       -  60134 -  Response for an exclusive read to a slave which does not support exclusive access should be <AXI4_OKAY>, but it returned <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.5)
// AXI4_EX_RD_OKAY_RESP_EXPECTED_EXOKAY                        -  60135 -  Expected <AXI4_EXOKAY> response to this exclusive read (because the parameters met the restrictions), but got <AXI4_OKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_WHEN_EX_NOT_ENABLED                              -  60136 -  An exclusive read should not be issued when exclusive transactions are not enabled
// AXI4_EX_WRITE_BEFORE_EX_READ_RESPONSE                       -  60137 -  Exclusive write has occurred, with no previous exclusive read
// AXI4_EX_WRITE_EXOKAY_RESP_EXPECTED_OKAY                     -  60138 -  Exclusive write has not been successful, yet slave has responded with <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.2)
// AXI4_EX_WRITE_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS    -  60139 -  Response for an exclusive write to a slave which does not support exclusive access should be <AXI4_OKAY>, but it returned <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.5)
// AXI4_EX_WRITE_OKAY_RESP_EXPECTED_EXOKAY                     -  60140 -  An <AXI4_OKAY> response to an exclusive write occurred, but an <AXI4_EXOKAY> response had been expected (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.2). If the slave has multiple interfaces to the system, this check should be disabled, as it is possible for this response to occur as a result of activity on another port.
// AXI4_EX_WR_WHEN_EX_NOT_ENABLED                              -  60141 -  An exclusive write should not be issued when exclusive transactions are not enabled
// AXI4_ILLEGAL_ARCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     -  60142 -  For a read from a cacheable address region, one of bits 2 or 3 of the cache parameter must be HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_ARCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION -  60143 -  For a read from a non-cacheable address region, bits 2 and 3 of the cache parameter must be LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_AWCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     -  60144 -  For a write to a cacheable address region, one of bits 2 or 3 of the cache parameter must be HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_AWCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION -  60145 -  For a write to a non-cacheable address region, bits 2 and 3 of the cache parameter must be LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_LENGTH_FIXED_READ_BURST                        -  60146 -  In the last read address phase burst_length has an illegal value for a burst of type AXI4_FIXED (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_LENGTH_FIXED_WRITE_BURST                       -  60147 -  In the last write address phase burst_length has an illegal value for a burst of type AXI4_FIXED (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1) 
// AXI4_ILLEGAL_LENGTH_WRAPPING_READ_BURST                     -  60148 -  In the last read address phase burst_length has an illegal value for a burst of type AXI4_WRAP (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_LENGTH_WRAPPING_WRITE_BURST                    -  60149 -  In the last write address phase burst_length has an illegal value for a burst of type AXI4_WRAP (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_READ                        -  60150 -  Error response (<AXI4_SLVERR>/<AXI4_DECERR>) received for an exclusive read
// AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_WRITE                       -  60151 -  Error response (<AXI4_SLVERR>/<AXI4_DECERR>) received for an exclusive write
// AXI4_INVALID_REGION_CARDINALITY                             -  60152 -  The configuration parameter <config_slave_regions> does not lie in the range 1-16, inclusive (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1).
// AXI4_INVALID_WRITE_STROBES_ON_ALIGNED_WRITE_TRANSFER        -  60153 -  Write strobe(s) incorrect for address/size of an aligned transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.3)
// AXI4_NON_INCREASING_REGION_SPECIFICATION                    -  60154 -  A region address-range has an upper bound smaller than the lower bound.
// AXI4_NON_ZERO_ARQOS                                         -  60155 -  The master is configured to not participate in the Quality-of-Service scheme, but <ARQOS>  is not 4'b0000, as it should be (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.1.2)
// AXI4_NON_ZERO_AWQOS                                         -  60156 -  The master is configured to not participate in the Quality-of-Service scheme, but <AWQOS> is not 4'b0000, as it should be (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.1.2)
// AXI4_OVERLAPPING_REGION                                     -  60157 -  An address-range in the region map overlaps with another address in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1). 
// AXI4_PARAM_READ_REORDERING_DEPTH_EQUALS_ZERO                -  60158 -  The user-supplied config_read_data_reordering_depth should be greater than zero (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_PARAM_READ_REORDERING_DEPTH_EXCEEDS_MAX_ID             -  60159 -  The user-supplied config_read_data_reordering_depth exceeds the maximum possible value, as defined by the AXI4_ID_WIDTH parameter (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_12                   -  60160 -  The RA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_13                   -  60161 -  The RA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_4                    -  60162 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_5                    -  60163 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_8                    -  60164 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_9                    -  60165 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_BURST_LENGTH_VIOLATION                            -  60166 -  The burst_length implied by the number of beats actually read does not match the burst_length defined by the <axi4_master_read_addr_channel_phase>
// AXI4_READ_BURST_MAXIMUM_LENGTH_VIOLATION                    -  60167 -  256 read data beats were seen without <RLAST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_READ_BURST_SIZE_VIOLATION                              -  60168 -  In this read transaction, size has been set too high for the defined data buswidth
// AXI4_READ_EXCLUSIVE_ENCODING_VIOLATION                      -  60169 -  A read-only interface does not support exclusive accesses (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2.2)
// AXI4_READ_REORDERING_VIOLATION                              -  60170 -  The arrival of a read response has exceeded the read reordering depth (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_TRANSFER_EXCEEDS_ADDRESS_SPACE                    -  60171 -  This read transfer runs off the edge of the address space defined by AXI4_ADDRESS_WIDTH (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_REGION_SMALLER_THAN_4KB                                -  60172 -  An address-range in the region map is smaller than 4kB (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_RESERVED_ARBURST_ENCODING                              -  60173 -  The reserved encoding of 2'b11 should not be used for <ARBURST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_RESERVED_AWBURST_ENCODING                              -  60174 -  The reserved encoding of 2'b11 should not be used for <AWBURST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_RID_CHANGED_BEFORE_RREADY                              -  60175 -  The value of <RID> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RID_UNKN                                               -  60176 -  <RID> has an X value/<RID> has a Z value
// AXI4_RLAST_CHANGED_BEFORE_RREADY                            -  60177 -  The value of <RLAST> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RLAST_UNKN                                             -  60178 -  <RLAST> has an X value/<RLAST> has a Z value
// AXI4_RUSER_CHANGED_BEFORE_RREADY                            -  60179 -  The value of <RUSER> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RUSER_UNKN                                             -  60180 -  <RUSER> has an X value/<RUSER> has a Z value
// AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_READ                   -  60181 -  Exclusive read accesses must have address aligned to the total number of bytes in the transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_WRITE                  -  60182 -  Exclusive write accesses must have address aligned to the total number of bytes in the transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_UNALIGNED_ADDR_FOR_WRAPPING_READ_BURST                 -  60183 -  Wrapping bursts must have address aligned to the start of the read transfer (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_UNALIGNED_ADDR_FOR_WRAPPING_WRITE_BURST                -  60184 -  Wrapping bursts must have address aligned to the start of the write transfet (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)   
// AXI4_WLAST_CHANGED_BEFORE_WREADY                            -  60185 -  The value of <WLAST> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WLAST_UNKN                                             -  60186 -  <WLAST> has an X value/<WLAST> has an Z value
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_12                  -  60187 -  The WA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_13                  -  60188 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_4                   -  60189 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_5                   -  60190 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_8                   -  60191 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_9                   -  60192 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_BURST_LENGTH_VIOLATION                           -  60193 -  The number of data beats in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_STROBES_LENGTH_VIOLATION                         -  60194 -  The size of the write_strobes array in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_USER_DATA_LENGTH_VIOLATION                       -  60195 -  The size of the wdata_user_data array in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_BURST_MAXIMUM_LENGTH_VIOLATION                   -  60196 -  256 write data beats were seen without <WLAST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1 )
// AXI4_WRITE_BURST_SIZE_VIOLATION                             -  60197 -  In this write transaction, size has been set too high for the defined data buswidth
// AXI4_WRITE_EXCLUSIVE_ENCODING_VIOLATION                     -  60198 -  A write-only interface does not support exclusive accesses (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2.3)
// AXI4_WRITE_STROBE_FIXED_BURST_VIOLATION                     -  60199 -  Write strobe(s) incorrect for the address/size of a fixed transfer 
// AXI4_WRITE_TRANSFER_EXCEEDS_ADDRESS_SPACE                   -  60200 -  This write transfer runs off the edge of the address space defined by <AXI4_ADDRESS_WIDTH> (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_WRONG_ARREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    -  60201 -  The region value should be 4'b0000 for a read from a slave with a single address decode in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_WRONG_AWREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    -  60202 -  The region value should be 4'b0000 for a write to a slave with a single address decode in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1) 
// AXI4_WUSER_CHANGED_BEFORE_WREADY                            -  60203 -  The value of <WUSER> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WUSER_UNKN                                             -  60204 -  <WUSER> has an X value/<WUSER> has an Z value
// AXI4_EXCL_RD_WHILE_EXCL_WR_IN_PROGRESS_SAME_ID              -  60205 -  Master starts an exclusive read burst while exclusive write burst with same ID tag is in progress (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCL_WR_WHILE_EXCL_RD_IN_PROGRESS_SAME_ID              -  60206 -  Master starts an exclusive write burst while exclusive read burst with same ID tag is in progress (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_RLAST_VIOLATION                                        -  60207 -  RLAST signal should be asserted along with the final transfer of the read data burst (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_WLAST_ASSERTED_DURING_DATA_PHASE_OTHER_THAN_LAST       -  60208 -  Wlast must only be asserted during the last data phase (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_READ_ISSUING_CAPABILITY_VIOLATION                      -  60209 -  Number of outstanding Read transactions exceeded maximum Read issuing capability 
// AXI4_WRITE_ISSUING_CAPABILITY_VIOLATION                     -  60210 -  Number of outstanding Write transactions exceeded maximum Write issuing capability
// AXI4_COMBINED_ISSUING_CAPABILITY_VIOLATION                  -  60211 -  Number of outstanding Read and Write transactions exceeded maximum combined issuing capability
// AXI4_READ_ACCEPTANCE_CAPABILITY_VIOLATION                   -  60212 -  Number of outstanding Read transactions exceeded maximum Read acceptance capability
// AXI4_WRITE_ACCEPTANCE_CAPABILITY_VIOLATION                  -  60213 -  Number of outstanding Write transactions exceeded maximum Write acceptance capability
// AXI4_COMBINED_ACCEPTANCE_CAPABILITY_VIOLATION               -  60214 -  Number of outstanding Read and Write transactions exceeded maximum combined acceptance capability
// AXI4_READ_INTERLEAVING_VIOLATION                            -  60215 -  The number of read transactions being interleaved has crossed the interleaving depth (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_BURST_SIZE_ILL                                    -  60216 -  In this read transaction, ARSIZE has been set to an illegal value (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_WRITE_BURST_SIZE_ILL                                   -  60217 -  In this write transaction, AWSIZE has been set to an illegal value (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_MAX              -  60218 -  Number of bytes in an exclusive read transaction must be less than or equal to maximum allowed count
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_MAX             -  60219 -  Number of bytes in an exclusive write transaction must be less than or equal to maximum allowed count
typedef enum bit [7:0]
{
    AXI4_ADDRESS_WIDTH_EXCEEDS_64                               = 8'h00,
    AXI4_ARADDR_CHANGED_BEFORE_ARREADY                          = 8'h01,
    AXI4_ARADDR_UNKN                                            = 8'h02,
    AXI4_ARPROT_CHANGED_BEFORE_ARREADY                          = 8'h03,
    AXI4_ARPROT_UNKN                                            = 8'h04,
    AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID                     = 8'h05,
    AXI4_ARREADY_UNKN                                           = 8'h06,
    AXI4_ARVALID_DEASSERTED_BEFORE_ARREADY                      = 8'h07,
    AXI4_ARVALID_HIGH_ON_FIRST_CLOCK                            = 8'h08,
    AXI4_ARVALID_UNKN                                           = 8'h09,
    AXI4_AWADDR_CHANGED_BEFORE_AWREADY                          = 8'h0a,
    AXI4_AWADDR_UNKN                                            = 8'h0b,
    AXI4_AWPROT_CHANGED_BEFORE_AWREADY                          = 8'h0c,
    AXI4_AWPROT_UNKN                                            = 8'h0d,
    AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID                     = 8'h0e,
    AXI4_AWREADY_UNKN                                           = 8'h0f,
    AXI4_AWVALID_DEASSERTED_BEFORE_AWREADY                      = 8'h10,
    AXI4_AWVALID_HIGH_ON_FIRST_CLOCK                            = 8'h11,
    AXI4_AWVALID_UNKN                                           = 8'h12,
    AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID                       = 8'h13,
    AXI4_BREADY_UNKN                                            = 8'h14,
    AXI4_BRESP_CHANGED_BEFORE_BREADY                            = 8'h15,
    AXI4_BRESP_UNKN                                             = 8'h16,
    AXI4_BVALID_DEASSERTED_BEFORE_BREADY                        = 8'h17,
    AXI4_BVALID_HIGH_EXITING_RESET                              = 8'h18,
    AXI4_BVALID_UNKN                                            = 8'h19,
    AXI4_DEC_ERR_RESP_FOR_READ                                  = 8'h1a,
    AXI4_DEC_ERR_RESP_FOR_WRITE                                 = 8'h1b,
    AXI4_EXOKAY_RESPONSE_NORMAL_READ                            = 8'h1c,
    AXI4_EXOKAY_RESPONSE_NORMAL_WRITE                           = 8'h1d,
    AXI4_INVALID_WRITE_STROBES_ON_UNALIGNED_WRITE_TRANSFER      = 8'h1e,
    AXI4_MINIMUM_SLAVE_ADDRESS_SPACE_VIOLATION                  = 8'h1f,
    AXI4_PARAM_READ_DATA_BUS_WIDTH                              = 8'h20,
    AXI4_PARAM_WRITE_DATA_BUS_WIDTH                             = 8'h21,
    AXI4_READ_DATA_BEFORE_ADDRESS                               = 8'h22,
    AXI4_READ_DATA_CHANGED_BEFORE_RREADY                        = 8'h23,
    AXI4_READ_DATA_UNKN                                         = 8'h24,
    AXI4_READ_RESP_CHANGED_BEFORE_RREADY                        = 8'h25,
    AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID                       = 8'h26,
    AXI4_RREADY_UNKN                                            = 8'h27,
    AXI4_RRESP_UNKN                                             = 8'h28,
    AXI4_RVALID_DEASSERTED_BEFORE_RREADY                        = 8'h29,
    AXI4_RVALID_HIGH_EXITING_RESET                              = 8'h2a,
    AXI4_RVALID_UNKN                                            = 8'h2b,
    AXI4_SLV_ERR_RESP_FOR_READ                                  = 8'h2c,
    AXI4_SLV_ERR_RESP_FOR_WRITE                                 = 8'h2d,
    AXI4_TIMEOUT_WAITING_FOR_READ_RESPONSE                      = 8'h2e,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_RESPONSE                     = 8'h2f,
    AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_INVALID_LANE            = 8'h30,
    AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_VALID_LANE              = 8'h31,
    AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID                       = 8'h32,
    AXI4_WREADY_UNKN                                            = 8'h33,
    AXI4_WRITE_DATA_BEFORE_ADDRESS                              = 8'h34,
    AXI4_WRITE_DATA_UNKN_ON_INVALID_LANE                        = 8'h35,
    AXI4_WRITE_DATA_UNKN_ON_VALID_LANE                          = 8'h36,
    AXI4_WRITE_RESPONSE_WITHOUT_ADDR_DATA                       = 8'h37,
    AXI4_WSTRB_CHANGED_BEFORE_WREADY                            = 8'h38,
    AXI4_WSTRB_UNKN                                             = 8'h39,
    AXI4_WVALID_DEASSERTED_BEFORE_WREADY                        = 8'h3a,
    AXI4_WVALID_HIGH_ON_FIRST_CLOCK                             = 8'h3b,
    AXI4_WVALID_UNKN                                            = 8'h3c,
    MVC_FAILED_POSTCONDITION                                    = 8'h3d,
    MVC_FAILED_RECOGNITION                                      = 8'h3e,
    AXI4_DEC_ERR_ILLEGAL_FOR_MAPPED_SLAVE_ADDR                  = 8'h3f,
    AXI4_AWVALID_HIGH_DURING_RESET                              = 8'h40,
    AXI4_WVALID_HIGH_DURING_RESET                               = 8'h41,
    AXI4_BVALID_HIGH_DURING_RESET                               = 8'h42,
    AXI4_ARVALID_HIGH_DURING_RESET                              = 8'h43,
    AXI4_RVALID_HIGH_DURING_RESET                               = 8'h44,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_DATA                         = 8'h45,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_ADDR_AFTER_DATA              = 8'h46,
    AXI4_ARESETn_SIGNAL_Z                                       = 8'h47,
    AXI4_ARESETn_SIGNAL_X                                       = 8'h48,
    AXI4_ADDR_FOR_READ_BURST_ACROSS_4K_BOUNDARY                 = 8'h49,
    AXI4_ADDR_FOR_WRITE_BURST_ACROSS_4K_BOUNDARY                = 8'h4a,
    AXI4_ARADDR_FALLS_IN_REGION_HOLE                            = 8'h4b,
    AXI4_ARBURST_CHANGED_BEFORE_ARREADY                         = 8'h4c,
    AXI4_ARBURST_UNKN                                           = 8'h4d,
    AXI4_ARCACHE_CHANGED_BEFORE_ARREADY                         = 8'h4e,
    AXI4_ARCACHE_UNKN                                           = 8'h4f,
    AXI4_ARID_CHANGED_BEFORE_ARREADY                            = 8'h50,
    AXI4_ARID_UNKN                                              = 8'h51,
    AXI4_ARLEN_CHANGED_BEFORE_ARREADY                           = 8'h52,
    AXI4_ARLEN_UNKN                                             = 8'h53,
    AXI4_ARLOCK_CHANGED_BEFORE_ARREADY                          = 8'h54,
    AXI4_ARLOCK_UNKN                                            = 8'h55,
    AXI4_ARQOS_CHANGED_BEFORE_ARREADY                           = 8'h56,
    AXI4_ARQOS_UNKN                                             = 8'h57,
    AXI4_ARREGION_CHANGED_BEFORE_ARREADY                        = 8'h58,
    AXI4_ARREGION_MISMATCH                                      = 8'h59,
    AXI4_ARREGION_UNKN                                          = 8'h5a,
    AXI4_ARSIZE_CHANGED_BEFORE_ARREADY                          = 8'h5b,
    AXI4_ARSIZE_UNKN                                            = 8'h5c,
    AXI4_ARUSER_CHANGED_BEFORE_ARREADY                          = 8'h5d,
    AXI4_ARUSER_UNKN                                            = 8'h5e,
    AXI4_AWADDR_FALLS_IN_REGION_HOLE                            = 8'h5f,
    AXI4_AWBURST_CHANGED_BEFORE_AWREADY                         = 8'h60,
    AXI4_AWBURST_UNKN                                           = 8'h61,
    AXI4_AWCACHE_CHANGED_BEFORE_AWREADY                         = 8'h62,
    AXI4_AWCACHE_UNKN                                           = 8'h63,
    AXI4_AWID_CHANGED_BEFORE_AWREADY                            = 8'h64,
    AXI4_AWID_UNKN                                              = 8'h65,
    AXI4_AWLEN_CHANGED_BEFORE_AWREADY                           = 8'h66,
    AXI4_AWLEN_UNKN                                             = 8'h67,
    AXI4_AWLOCK_CHANGED_BEFORE_AWREADY                          = 8'h68,
    AXI4_AWLOCK_UNKN                                            = 8'h69,
    AXI4_AWQOS_CHANGED_BEFORE_AWREADY                           = 8'h6a,
    AXI4_AWQOS_UNKN                                             = 8'h6b,
    AXI4_AWREGION_CHANGED_BEFORE_AWREADY                        = 8'h6c,
    AXI4_AWREGION_MISMATCH                                      = 8'h6d,
    AXI4_AWREGION_UNKN                                          = 8'h6e,
    AXI4_AWSIZE_CHANGED_BEFORE_AWREADY                          = 8'h6f,
    AXI4_AWSIZE_UNKN                                            = 8'h70,
    AXI4_AWUSER_CHANGED_BEFORE_AWREADY                          = 8'h71,
    AXI4_AWUSER_UNKN                                            = 8'h72,
    AXI4_BID_CHANGED_BEFORE_BREADY                              = 8'h73,
    AXI4_BID_UNKN                                               = 8'h74,
    AXI4_BUSER_CHANGED_BEFORE_BREADY                            = 8'h75,
    AXI4_BUSER_UNKN                                             = 8'h76,
    AXI4_EXCLUSIVE_READ_ACCESS_MODIFIABLE                       = 8'h77,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_128              = 8'h78,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_128             = 8'h79,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_NOT_POWER_OF_2           = 8'h7a,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_NOT_POWER_OF_2          = 8'h7b,
    AXI4_EXCLUSIVE_READ_LENGTH_EXCEEDS_16                       = 8'h7c,
    AXI4_EXCLUSIVE_WR_ADDRESS_NOT_SAME_AS_RD                    = 8'h7d,
    AXI4_EXCLUSIVE_WR_BURST_NOT_SAME_AS_RD                      = 8'h7e,
    AXI4_EXCLUSIVE_WR_CACHE_NOT_SAME_AS_RD                      = 8'h7f,
    AXI4_EXCLUSIVE_WRITE_ACCESS_MODIFIABLE                      = 8'h80,
    AXI4_EXCLUSIVE_WR_LENGTH_NOT_SAME_AS_RD                     = 8'h81,
    AXI4_EXCLUSIVE_WR_PROT_NOT_SAME_AS_RD                       = 8'h82,
    AXI4_EXCLUSIVE_WR_REGION_NOT_SAME_AS_RD                     = 8'h83,
    AXI4_EXCLUSIVE_WR_SIZE_NOT_SAME_AS_RD                       = 8'h84,
    AXI4_EX_RD_EXOKAY_RESP_EXPECTED_OKAY                        = 8'h85,
    AXI4_EX_RD_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS       = 8'h86,
    AXI4_EX_RD_OKAY_RESP_EXPECTED_EXOKAY                        = 8'h87,
    AXI4_EX_RD_WHEN_EX_NOT_ENABLED                              = 8'h88,
    AXI4_EX_WRITE_BEFORE_EX_READ_RESPONSE                       = 8'h89,
    AXI4_EX_WRITE_EXOKAY_RESP_EXPECTED_OKAY                     = 8'h8a,
    AXI4_EX_WRITE_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS    = 8'h8b,
    AXI4_EX_WRITE_OKAY_RESP_EXPECTED_EXOKAY                     = 8'h8c,
    AXI4_EX_WR_WHEN_EX_NOT_ENABLED                              = 8'h8d,
    AXI4_ILLEGAL_ARCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     = 8'h8e,
    AXI4_ILLEGAL_ARCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION = 8'h8f,
    AXI4_ILLEGAL_AWCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     = 8'h90,
    AXI4_ILLEGAL_AWCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION = 8'h91,
    AXI4_ILLEGAL_LENGTH_FIXED_READ_BURST                        = 8'h92,
    AXI4_ILLEGAL_LENGTH_FIXED_WRITE_BURST                       = 8'h93,
    AXI4_ILLEGAL_LENGTH_WRAPPING_READ_BURST                     = 8'h94,
    AXI4_ILLEGAL_LENGTH_WRAPPING_WRITE_BURST                    = 8'h95,
    AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_READ                        = 8'h96,
    AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_WRITE                       = 8'h97,
    AXI4_INVALID_REGION_CARDINALITY                             = 8'h98,
    AXI4_INVALID_WRITE_STROBES_ON_ALIGNED_WRITE_TRANSFER        = 8'h99,
    AXI4_NON_INCREASING_REGION_SPECIFICATION                    = 8'h9a,
    AXI4_NON_ZERO_ARQOS                                         = 8'h9b,
    AXI4_NON_ZERO_AWQOS                                         = 8'h9c,
    AXI4_OVERLAPPING_REGION                                     = 8'h9d,
    AXI4_PARAM_READ_REORDERING_DEPTH_EQUALS_ZERO                = 8'h9e,
    AXI4_PARAM_READ_REORDERING_DEPTH_EXCEEDS_MAX_ID             = 8'h9f,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_12                   = 8'ha0,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_13                   = 8'ha1,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_4                    = 8'ha2,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_5                    = 8'ha3,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_8                    = 8'ha4,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_9                    = 8'ha5,
    AXI4_READ_BURST_LENGTH_VIOLATION                            = 8'ha6,
    AXI4_READ_BURST_MAXIMUM_LENGTH_VIOLATION                    = 8'ha7,
    AXI4_READ_BURST_SIZE_VIOLATION                              = 8'ha8,
    AXI4_READ_EXCLUSIVE_ENCODING_VIOLATION                      = 8'ha9,
    AXI4_READ_REORDERING_VIOLATION                              = 8'haa,
    AXI4_READ_TRANSFER_EXCEEDS_ADDRESS_SPACE                    = 8'hab,
    AXI4_REGION_SMALLER_THAN_4KB                                = 8'hac,
    AXI4_RESERVED_ARBURST_ENCODING                              = 8'had,
    AXI4_RESERVED_AWBURST_ENCODING                              = 8'hae,
    AXI4_RID_CHANGED_BEFORE_RREADY                              = 8'haf,
    AXI4_RID_UNKN                                               = 8'hb0,
    AXI4_RLAST_CHANGED_BEFORE_RREADY                            = 8'hb1,
    AXI4_RLAST_UNKN                                             = 8'hb2,
    AXI4_RUSER_CHANGED_BEFORE_RREADY                            = 8'hb3,
    AXI4_RUSER_UNKN                                             = 8'hb4,
    AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_READ                   = 8'hb5,
    AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_WRITE                  = 8'hb6,
    AXI4_UNALIGNED_ADDR_FOR_WRAPPING_READ_BURST                 = 8'hb7,
    AXI4_UNALIGNED_ADDR_FOR_WRAPPING_WRITE_BURST                = 8'hb8,
    AXI4_WLAST_CHANGED_BEFORE_WREADY                            = 8'hb9,
    AXI4_WLAST_UNKN                                             = 8'hba,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_12                  = 8'hbb,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_13                  = 8'hbc,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_4                   = 8'hbd,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_5                   = 8'hbe,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_8                   = 8'hbf,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_9                   = 8'hc0,
    AXI4_WRITE_BURST_LENGTH_VIOLATION                           = 8'hc1,
    AXI4_WRITE_STROBES_LENGTH_VIOLATION                         = 8'hc2,
    AXI4_WRITE_USER_DATA_LENGTH_VIOLATION                       = 8'hc3,
    AXI4_WRITE_BURST_MAXIMUM_LENGTH_VIOLATION                   = 8'hc4,
    AXI4_WRITE_BURST_SIZE_VIOLATION                             = 8'hc5,
    AXI4_WRITE_EXCLUSIVE_ENCODING_VIOLATION                     = 8'hc6,
    AXI4_WRITE_STROBE_FIXED_BURST_VIOLATION                     = 8'hc7,
    AXI4_WRITE_TRANSFER_EXCEEDS_ADDRESS_SPACE                   = 8'hc8,
    AXI4_WRONG_ARREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    = 8'hc9,
    AXI4_WRONG_AWREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    = 8'hca,
    AXI4_WUSER_CHANGED_BEFORE_WREADY                            = 8'hcb,
    AXI4_WUSER_UNKN                                             = 8'hcc,
    AXI4_EXCL_RD_WHILE_EXCL_WR_IN_PROGRESS_SAME_ID              = 8'hcd,
    AXI4_EXCL_WR_WHILE_EXCL_RD_IN_PROGRESS_SAME_ID              = 8'hce,
    AXI4_RLAST_VIOLATION                                        = 8'hcf,
    AXI4_WLAST_ASSERTED_DURING_DATA_PHASE_OTHER_THAN_LAST       = 8'hd0,
    AXI4_READ_ISSUING_CAPABILITY_VIOLATION                      = 8'hd1,
    AXI4_WRITE_ISSUING_CAPABILITY_VIOLATION                     = 8'hd2,
    AXI4_COMBINED_ISSUING_CAPABILITY_VIOLATION                  = 8'hd3,
    AXI4_READ_ACCEPTANCE_CAPABILITY_VIOLATION                   = 8'hd4,
    AXI4_WRITE_ACCEPTANCE_CAPABILITY_VIOLATION                  = 8'hd5,
    AXI4_COMBINED_ACCEPTANCE_CAPABILITY_VIOLATION               = 8'hd6,
    AXI4_READ_INTERLEAVING_VIOLATION                            = 8'hd7,
    AXI4_READ_BURST_SIZE_ILL                                    = 8'hd8,
    AXI4_WRITE_BURST_SIZE_ILL                                   = 8'hd9,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_MAX              = 8'hda,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_MAX             = 8'hdb
} axi4_assertion_e;


//------------------------------------------------------------------------------
//
// Struct: axi4_en_perf_stats_s
//
//------------------------------------------------------------------------------
// This struct specifies the control over axi4 performance statistics
// 
// enable - Enables performance statistics collection
// 
// Default: 0 (perf stats collection is disabled)
// 
// step - Sets the step size of the sliding window for performance monitoring.
// The step size within the sliding window is measured in terms of clock cycle
// 
// Default: 100 (clock cycles)
// 
// multiple - Sets the number of steps within the sliding window. 
// The minimum value can be 1 and the maximum 10.
// 
// Default: 5
//------------------------------------------------------------------------------

typedef struct
{
    bit enable;
    int step;
    int multiple;
} axi4_en_perf_stats_s;


//------------------------------------------------------------------------------
//
// Struct: axi4_debug_type_s
//
//------------------------------------------------------------------------------
// The field of this struct if set to 1 enables printing of
//  debug statements.
//  master_seq - Enables master_seq debug
//  slave_seq - Enables slave_seq debug
// 
// Default: 0 (Debug statements printing is disabled)
//------------------------------------------------------------------------------

typedef struct packed
{
    bit master_seq;
    bit slave_seq;
    bit vip_config;
} axi4_debug_type_s;


//-------------------------------------------------------------------------
// Private Types
//-------------------------------------------------------------------------

typedef bit [1023:0] axi4_max_bits_t;

// enum: axi4_config_e
//
// An enum which fields corresponding to each configuration parameter of the VIP
//    AXI4_CONFIG_AXI4LITE_axi4 - 
//          This configuration can be enabled to use the AXI4 QVIP as AXI4-Lite.
//         Setting this variable to 1 enables the user to use the AXI4 BFM and all as AXI4-Lite
//         Setting this variable to 0 ensures usual usage as AXI4.
//          
//         Default: false(behaves as AXI4 interface)
//         
//    AXI4_CONFIG_INTERFACE_TYPE - 
//          
//         A configuration parameter (of type <axi4_interface_type_e>) defining the
//         interface category of the interface: read/write, read only or write only.
//         
//         Default: <AXI4_READWRITE>.
//         
//    AXI4_CONFIG_ENABLE_RLAST - 
//          
//         Enables use of optional RLAST signal.
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ID - 
//          
//         Enables use of optional AWID, BID, ARID, RID signals.
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ALL_ASSERTIONS - 
//         
//         Enables all protocol assertions. 
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ASSERTION - 
//         
//         Enables individual protocol assertion.
//         This variable controls whether specific assertion within QVIP (of type <axi4_assertion_e>) is enabled or disabled.
//         Individual assertion can be disabled as follows:-
//         //-----------------------------------------------------------------------
//         // < BFM interface>.config_enable_assertion[<name of assertion>] = 1'b0;
//         //-----------------------------------------------------------------------
//         
//         For example, the assertion AXI4_READ_DATA_UNKN is disabled as follows:
//         <bfm>.config_enable_assertion[AXI4_READ_DATA_UNKN] =  1'b0; 
//         
//         Here bfm is the AXI4 interface instance name for which the assertion to be disabled. 
//         
//         Default: All assertions are enabled
//           
//         
//    AXI4_CONFIG_ENABLE_USER_SIDEBAND - 
//         
//         Enables user sideband signals. 
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_SLAVE_EXCLUSIVE - 
//         
//         Enables exclusive transactions support for slave.
//         If disabled, every exclusive read/write returns an OKAY response,
//         and exclusive write updates memory. 
//         
//         Default: 1  
//         
//    AXI4_CONFIG_READ_DATA_REORDERING_DEPTH - 
//         
//         Sets the maximum number of different read transaction addresses for which read 
//         data(response) can be sent in any order from slave. 
//         
//         Default: 2 ** AXI4_ID_WIDTH
//         
//    AXI4_CONFIG_MAX_TRANSACTION_TIME_FACTOR - 
//          
//         Sets maximum timeout period (in terms of ACLK) for any complete read or write transaction, which
//         includes time period for all individual phases of transaction. 
//         
//         Default: 100000 clock cycles
//         
//    AXI4_CONFIG_BURST_TIMEOUT_FACTOR - 
//          
//         Sets maximum timeout period (in terms of ACLK) between individual phases of a transaction. 
//         
//         Default: 10000 clock cycles 
//         
//    AXI4_CONFIG_MAX_LATENCY_AWVALID_ASSERTION_TO_AWREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of AWVALID to assertion of AWREADY.
//         An error message AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID is generated if AWREADY is not asserted
//         after assertion of AWVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_ARVALID_ASSERTION_TO_ARREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of ARVALID to assertion of ARREADY.
//         An error message AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID is generated if ARREADY is not asserted
//         after assertion of ARVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_RVALID_ASSERTION_TO_RREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of RVALID to assertion of RREADY.
//         An error message AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID is generated if RREADY is not asserted
//         after assertion of RVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_BVALID_ASSERTION_TO_BREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of BVALID to assertion of BREADY.
//         An error message AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID is generated if BREADY is not asserted
//         after assertion of BVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_WVALID_ASSERTION_TO_WREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of WVALID to assertion of WREADY.
//         An error message AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID is generated if WREADY is not asserted
//         after assertion of WVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_ENABLE_QOS - 
//          
//         Enables quality-of-service support. 
//         
//         Default: 1 
//         
//    AXI4_CONFIG_SETUP_TIME - 
//         
//         Sets number of simulation time units from the setup time to the active 
//         clock edge of ACLK. The setup time will always be less than the time period
//         of the clock. 
//         
//         Default: 0
//         
//    AXI4_CONFIG_HOLD_TIME - 
//         
//         Sets number of simulation time units from the hold time to the active 
//         clock edge of ACLK. 
//         
//         Default: 0
//         
//    AXI4_CONFIG_NUM_MAX_OUTSTANDING_READS - 
//         
//         Configures maximum number of read outstanding transfers allowed on the bus.
//         
//         Default: -1
//         
//    AXI4_CONFIG_NUM_MAX_OUTSTANDING_WRITES - 
//                                                                                   
//         Configures maximum number of write outstanding transfers allowed on the bus. 
//                                                                                      
//         Default: -1                                                                 
//         
//    AXI4_CONFIG_MAX_OUTSTANDING_WR -  Configures maximum possible outstanding Write transactions
//    AXI4_CONFIG_MAX_OUTSTANDING_RD -  Configures maximum possible outstanding Read transactions
//    AXI4_CONFIG_MAX_OUTSTANDING_RW -  Configures maximum possible outstanding Combined (Read and Write) transactions
//    AXI4_CONFIG_NUM_OUTSTANDING_WR_PHASE -  Provides current number of outstanding Write address phases 
//    AXI4_CONFIG_NUM_OUTSTANDING_RD_PHASE -  Provides current number of outstanding Read address phases
//    AXI4_CONFIG_IS_ISSUING -  Enables Master component to use "config_max_outstanding_wr/config_max_outstanding_rd/config_max_outstanding_rw" variables for transaction issuing capability when set to true

typedef enum bit [7:0]
{
    AXI4_CONFIG_AXI4LITE_axi4           = 8'd0,
    AXI4_CONFIG_INTERFACE_TYPE               = 8'd1,
    AXI4_CONFIG_ENABLE_RLAST                 = 8'd2,
    AXI4_CONFIG_ENABLE_ID                    = 8'd3,
    AXI4_CONFIG_ENABLE_ALL_ASSERTIONS        = 8'd4,
    AXI4_CONFIG_ENABLE_ASSERTION             = 8'd5,
    AXI4_CONFIG_ENABLE_USER_SIDEBAND         = 8'd6,
    AXI4_CONFIG_AXI4LITE_TR_ID               = 8'd7,
    AXI4_CONFIG_ENABLE_SLAVE_EXCLUSIVE       = 8'd8,
    AXI4_CONFIG_READ_DATA_REORDERING_DEPTH   = 8'd9,
    AXI4_CONFIG_SLAVE_START_ADDR             = 8'd10,
    AXI4_CONFIG_SLAVE_END_ADDR               = 8'd11,
    AXI4_CONFIG_MAX_TRANSACTION_TIME_FACTOR  = 8'd12,
    AXI4_CONFIG_BURST_TIMEOUT_FACTOR         = 8'd13,
    AXI4_CONFIG_MAX_LATENCY_AWVALID_ASSERTION_TO_AWREADY = 8'd14,
    AXI4_CONFIG_MAX_LATENCY_ARVALID_ASSERTION_TO_ARREADY = 8'd15,
    AXI4_CONFIG_MAX_LATENCY_RVALID_ASSERTION_TO_RREADY = 8'd16,
    AXI4_CONFIG_MAX_LATENCY_BVALID_ASSERTION_TO_BREADY = 8'd17,
    AXI4_CONFIG_MAX_LATENCY_WVALID_ASSERTION_TO_WREADY = 8'd18,
    AXI4_CONFIG_ENABLE_QOS                   = 8'd19,
    AXI4_CONFIG_SETUP_TIME                   = 8'd20,
    AXI4_CONFIG_HOLD_TIME                    = 8'd21,
    AXI4_CONFIG_NUM_MAX_OUTSTANDING_READS    = 8'd22,
    AXI4_CONFIG_NUM_MAX_OUTSTANDING_WRITES   = 8'd23,
    AXI4_CONFIG_MAX_OUTSTANDING_WR           = 8'd24,
    AXI4_CONFIG_MAX_OUTSTANDING_RD           = 8'd25,
    AXI4_CONFIG_MAX_OUTSTANDING_RW           = 8'd26,
    AXI4_CONFIG_NUM_OUTSTANDING_WR_PHASE     = 8'd27,
    AXI4_CONFIG_NUM_OUTSTANDING_RD_PHASE     = 8'd28,
    AXI4_CONFIG_IS_ISSUING                   = 8'd29
} axi4_config_e;

// enum: axi4_vhd_if_e
//
// For VHDL use only
typedef enum int
{
    AXI4_VHD_SET_CONFIG                         = 32'd0,
    AXI4_VHD_GET_CONFIG                         = 32'd1,
    AXI4_VHD_CREATE_WRITE_TRANSACTION           = 32'd2,
    AXI4_VHD_CREATE_READ_TRANSACTION            = 32'd3,
    AXI4_VHD_SET_READ_OR_WRITE                  = 32'd4,
    AXI4_VHD_GET_READ_OR_WRITE                  = 32'd5,
    AXI4_VHD_SET_ADDR                           = 32'd6,
    AXI4_VHD_GET_ADDR                           = 32'd7,
    AXI4_VHD_SET_PROT                           = 32'd8,
    AXI4_VHD_GET_PROT                           = 32'd9,
    AXI4_VHD_SET_REGION                         = 32'd10,
    AXI4_VHD_GET_REGION                         = 32'd11,
    AXI4_VHD_SET_SIZE                           = 32'd12,
    AXI4_VHD_GET_SIZE                           = 32'd13,
    AXI4_VHD_SET_BURST                          = 32'd14,
    AXI4_VHD_GET_BURST                          = 32'd15,
    AXI4_VHD_SET_LOCK                           = 32'd16,
    AXI4_VHD_GET_LOCK                           = 32'd17,
    AXI4_VHD_SET_CACHE                          = 32'd18,
    AXI4_VHD_GET_CACHE                          = 32'd19,
    AXI4_VHD_SET_QOS                            = 32'd20,
    AXI4_VHD_GET_QOS                            = 32'd21,
    AXI4_VHD_SET_ID                             = 32'd22,
    AXI4_VHD_GET_ID                             = 32'd23,
    AXI4_VHD_SET_BURST_LENGTH                   = 32'd24,
    AXI4_VHD_GET_BURST_LENGTH                   = 32'd25,
    AXI4_VHD_SET_ADDR_USER                      = 32'd26,
    AXI4_VHD_GET_ADDR_USER                      = 32'd27,
    AXI4_VHD_SET_DATA_WORDS                     = 32'd28,
    AXI4_VHD_GET_DATA_WORDS                     = 32'd29,
    AXI4_VHD_SET_WRITE_STROBES                  = 32'd30,
    AXI4_VHD_GET_WRITE_STROBES                  = 32'd31,
    AXI4_VHD_SET_DATA_USER                      = 32'd32,
    AXI4_VHD_GET_DATA_USER                      = 32'd33,
    AXI4_VHD_SET_RESP                           = 32'd34,
    AXI4_VHD_GET_RESP                           = 32'd35,
    AXI4_VHD_SET_RESP_USER                      = 32'd36,
    AXI4_VHD_GET_RESP_USER                      = 32'd37,
    AXI4_VHD_SET_ADDRESS_VALID_DELAY            = 32'd38,
    AXI4_VHD_GET_ADDRESS_VALID_DELAY            = 32'd39,
    AXI4_VHD_SET_DATA_VALID_DELAY               = 32'd40,
    AXI4_VHD_GET_DATA_VALID_DELAY               = 32'd41,
    AXI4_VHD_SET_WRITE_RESPONSE_VALID_DELAY     = 32'd42,
    AXI4_VHD_GET_WRITE_RESPONSE_VALID_DELAY     = 32'd43,
    AXI4_VHD_SET_ADDRESS_READY_DELAY            = 32'd44,
    AXI4_VHD_GET_ADDRESS_READY_DELAY            = 32'd45,
    AXI4_VHD_SET_DATA_READY_DELAY               = 32'd46,
    AXI4_VHD_GET_DATA_READY_DELAY               = 32'd47,
    AXI4_VHD_SET_WRITE_RESPONSE_READY_DELAY     = 32'd48,
    AXI4_VHD_GET_WRITE_RESPONSE_READY_DELAY     = 32'd49,
    AXI4_VHD_SET_GEN_WRITE_STROBES              = 32'd50,
    AXI4_VHD_GET_GEN_WRITE_STROBES              = 32'd51,
    AXI4_VHD_SET_OPERATION_MODE                 = 32'd52,
    AXI4_VHD_GET_OPERATION_MODE                 = 32'd53,
    AXI4_VHD_SET_WRITE_DATA_MODE                = 32'd54,
    AXI4_VHD_GET_WRITE_DATA_MODE                = 32'd55,
    AXI4_VHD_SET_DATA_BEAT_DONE                 = 32'd56,
    AXI4_VHD_GET_DATA_BEAT_DONE                 = 32'd57,
    AXI4_VHD_SET_TRANSACTION_DONE               = 32'd58,
    AXI4_VHD_GET_TRANSACTION_DONE               = 32'd59,
    AXI4_VHD_EXECUTE_TRANSACTION                = 32'd60,
    AXI4_VHD_GET_RW_TRANSACTION                 = 32'd61,
    AXI4_VHD_EXECUTE_WRITE_DATA_BURST           = 32'd62,
    AXI4_VHD_GET_WRITE_DATA_BURST               = 32'd63,
    AXI4_VHD_EXECUTE_READ_ADDR_PHASE            = 32'd64,
    AXI4_VHD_GET_READ_ADDR_PHASE                = 32'd65,
    AXI4_VHD_EXECUTE_READ_DATA_BURST            = 32'd66,
    AXI4_VHD_GET_READ_DATA_BURST                = 32'd67,
    AXI4_VHD_EXECUTE_READ_DATA_PHASE            = 32'd68,
    AXI4_VHD_GET_READ_DATA_PHASE                = 32'd69,
    AXI4_VHD_EXECUTE_WRITE_ADDR_PHASE           = 32'd70,
    AXI4_VHD_GET_WRITE_ADDR_PHASE               = 32'd71,
    AXI4_VHD_EXECUTE_WRITE_DATA_PHASE           = 32'd72,
    AXI4_VHD_GET_WRITE_DATA_PHASE               = 32'd73,
    AXI4_VHD_EXECUTE_WRITE_RESPONSE_PHASE       = 32'd74,
    AXI4_VHD_GET_WRITE_RESPONSE_PHASE           = 32'd75,
    AXI4_VHD_GET_READ_ADDR_CYCLE                = 32'd76,
    AXI4_VHD_EXECUTE_READ_ADDR_READY            = 32'd77,
    AXI4_VHD_GET_READ_ADDR_READY                = 32'd78,
    AXI4_VHD_GET_READ_DATA_CYCLE                = 32'd79,
    AXI4_VHD_EXECUTE_READ_DATA_READY            = 32'd80,
    AXI4_VHD_GET_READ_DATA_READY                = 32'd81,
    AXI4_VHD_GET_WRITE_ADDR_CYCLE               = 32'd82,
    AXI4_VHD_EXECUTE_WRITE_ADDR_READY           = 32'd83,
    AXI4_VHD_GET_WRITE_ADDR_READY               = 32'd84,
    AXI4_VHD_GET_WRITE_DATA_CYCLE               = 32'd85,
    AXI4_VHD_EXECUTE_WRITE_DATA_READY           = 32'd86,
    AXI4_VHD_GET_WRITE_DATA_READY               = 32'd87,
    AXI4_VHD_GET_WRITE_RESPONSE_CYCLE           = 32'd88,
    AXI4_VHD_EXECUTE_WRITE_RESP_READY           = 32'd89,
    AXI4_VHD_GET_WRITE_RESP_READY               = 32'd90,
    AXI4_VHD_CREATE_MONITOR_TRANSACTION         = 32'd91,
    AXI4_VHD_CREATE_SLAVE_TRANSACTION           = 32'd92,
    AXI4_VHD_PUSH_TRANSACTION_ID                = 32'd93,
    AXI4_VHD_POP_TRANSACTION_ID                 = 32'd94,
    AXI4_VHD_GET_WRITE_ADDR_DATA                = 32'd95,
    AXI4_VHD_GET_READ_ADDR                      = 32'd96,
    AXI4_VHD_SET_READ_DATA                      = 32'd97,
    AXI4_VHD_PRINT                              = 32'd98,
    AXI4_VHD_DESTRUCT_TRANSACTION               = 32'd99,
    AXI4_VHD_WAIT_ON                            = 32'd100
} axi4_vhd_if_e;


typedef enum bit [7:0]
{
    AXI4_CLOCK_POSEDGE = 8'd0,
    AXI4_CLOCK_NEGEDGE = 8'd1,
    AXI4_CLOCK_ANYEDGE = 8'd2,
    AXI4_CLOCK_0_TO_1  = 8'd3,
    AXI4_CLOCK_1_TO_0  = 8'd4,
    AXI4_RESET_POSEDGE = 8'd5,
    AXI4_RESET_NEGEDGE = 8'd6,
    AXI4_RESET_ANYEDGE = 8'd7,
    AXI4_RESET_0_TO_1  = 8'd8,
    AXI4_RESET_1_TO_0  = 8'd9
} axi4_wait_e;

`ifndef MAX_AXI4_ADDRESS_WIDTH
  `define MAX_AXI4_ADDRESS_WIDTH 64
`endif

`ifndef MAX_AXI4_RDATA_WIDTH
  `define MAX_AXI4_RDATA_WIDTH 1024
`endif

`ifndef MAX_AXI4_WDATA_WIDTH
  `define MAX_AXI4_WDATA_WIDTH 1024
`endif

`ifndef MAX_AXI4_ID_WIDTH
  `define MAX_AXI4_ID_WIDTH 18
`endif

`ifndef MAX_AXI4_USER_WIDTH
  `define MAX_AXI4_USER_WIDTH 8
`endif

`ifndef MAX_AXI4_REGION_MAP_SIZE
  `define MAX_AXI4_REGION_MAP_SIZE 16
`endif

// enum: axi4_operation_mode_e
//
typedef enum int
{
    AXI4_TRANSACTION_NON_BLOCKING = 32'd0,
    AXI4_TRANSACTION_BLOCKING     = 32'd1
} axi4_operation_mode_e;

// enum: axi4_write_data_mode_e
//
typedef enum int
{
    AXI4_DATA_AFTER_ADDRESS = 32'd0,
    AXI4_DATA_WITH_ADDRESS  = 32'd1
} axi4_write_data_mode_e;

// Global Transaction Class
class axi4_transaction;
    // Protocol 
    axi4_rw_e read_or_write;
    bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0] addr;
    axi4_prot_e prot;
    bit [3:0] region;
    axi4_size_e size;
    axi4_burst_e burst;
    axi4_lock_e lock;
    axi4_cache_e cache;
    bit [3:0] qos;
    bit [((`MAX_AXI4_ID_WIDTH) - 1):0] id;
    bit [7:0] burst_length;
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] addr_user;
    bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0] data_words [];
    bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [];
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] data_user [];
    axi4_response_e resp[];
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] resp_user [];
    int address_valid_delay;
    int data_valid_delay[];
    int write_response_valid_delay;
    int address_ready_delay;
    int data_ready_delay[];
    int write_response_ready_delay;

    // Housekeeping
    bit gen_write_strobes = 1'b1;
    axi4_operation_mode_e  operation_mode  = AXI4_TRANSACTION_BLOCKING;
    axi4_write_data_mode_e write_data_mode = AXI4_DATA_WITH_ADDRESS;
    bit data_beat_done[];
    bit transaction_done;

    // This varaible is for printing component name and should not be visible/documented
    string driver_name;

    function void set_read_or_write( input axi4_rw_e lread_or_write );
      read_or_write = lread_or_write;
    endfunction

    function axi4_rw_e get_read_or_write();
      return read_or_write;
    endfunction

    function void set_addr( input bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0] laddr );
      addr = laddr;
    endfunction

    function bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0]  get_addr();
      return addr;
    endfunction

    function void set_prot( input axi4_prot_e lprot );
      prot = lprot;
    endfunction

    function axi4_prot_e get_prot();
      return prot;
    endfunction

    function void set_region( input bit [3:0] lregion );
      region = lregion;
    endfunction

    function bit [3:0]  get_region();
      return region;
    endfunction

    function void set_size( input axi4_size_e lsize );
      size = lsize;
    endfunction

    function axi4_size_e get_size();
      return size;
    endfunction

    function void set_burst( input axi4_burst_e lburst );
      burst = lburst;
    endfunction

    function axi4_burst_e get_burst();
      return burst;
    endfunction

    function void set_lock( input axi4_lock_e llock );
      lock = llock;
    endfunction

    function axi4_lock_e get_lock();
      return lock;
    endfunction

    function void set_cache( input axi4_cache_e lcache );
      cache = lcache;
    endfunction

    function axi4_cache_e get_cache();
      return cache;
    endfunction

    function void set_qos( input bit [3:0] lqos );
      qos = lqos;
    endfunction

    function bit [3:0]  get_qos();
      return qos;
    endfunction

    function void set_id( input bit [((`MAX_AXI4_ID_WIDTH) - 1):0] lid );
      id = lid;
    endfunction

    function bit [((`MAX_AXI4_ID_WIDTH) - 1):0]  get_id();
      return id;
    endfunction

    function void set_burst_length( input bit [7:0] lburst_length );
      burst_length = lburst_length;
      data_words           = new[(lburst_length + 1)];
      write_strobes        = new[(lburst_length + 1)];
      data_user            = new[(lburst_length + 1)];
      resp                 = new[(lburst_length + 1)];
      resp_user            = new[(lburst_length + 1)];
      data_valid_delay     = new[(lburst_length + 1)];
      data_ready_delay     = new[(lburst_length + 1)];
      data_beat_done       = new[(lburst_length + 1)];
    endfunction

    function bit [7:0]  get_burst_length();
      return burst_length;
    endfunction

    function void set_addr_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] laddr_user );
      addr_user = laddr_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_addr_user();
      return addr_user;
    endfunction

    function void set_data_words( input bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0] ldata_words, input int index = 0 );
      data_words[index] = ldata_words;
    endfunction

    function bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0]  get_data_words( input int index = 0 );
      return data_words[index];
    endfunction

    function void set_write_strobes( input bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0] lwrite_strobes, input int index = 0 );
      write_strobes[index] = lwrite_strobes;
    endfunction

    function bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0]  get_write_strobes( input int index = 0 );
      return write_strobes[index];
    endfunction

    function void set_data_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] ldata_user, input int index = 0 );
      data_user[index] = ldata_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_data_user( input int index = 0 );
      return data_user[index];
    endfunction

    function void set_resp( input axi4_response_e lresp, input int index = 0 );
      resp[index] = lresp;
    endfunction

    function axi4_response_e get_resp( input int index = 0 );
      return resp[index];
    endfunction

    function void set_resp_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] lresp_user, input int index = 0 );
      resp_user[index] = lresp_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_resp_user( input int index = 0 );
      return resp_user[index];
    endfunction

    function void set_address_valid_delay( input int laddress_valid_delay );
      address_valid_delay = laddress_valid_delay;
    endfunction

    function int get_address_valid_delay();
      return address_valid_delay;
    endfunction

    function void set_data_valid_delay( input int ldata_valid_delay, input int index = 0 );
      data_valid_delay[index] = ldata_valid_delay;
    endfunction

    function int get_data_valid_delay( input int index = 0 );
      return data_valid_delay[index];
    endfunction

    function void set_write_response_valid_delay( input int lwrite_response_valid_delay );
      write_response_valid_delay = lwrite_response_valid_delay;
    endfunction

    function int get_write_response_valid_delay();
      return write_response_valid_delay;
    endfunction

    function void set_address_ready_delay( input int laddress_ready_delay );
      address_ready_delay = laddress_ready_delay;
    endfunction

    function int get_address_ready_delay();
      return address_ready_delay;
    endfunction

    function void set_data_ready_delay( input int ldata_ready_delay, input int index = 0 );
      data_ready_delay[index] = ldata_ready_delay;
    endfunction

    function int get_data_ready_delay( input int index = 0 );
      return data_ready_delay[index];
    endfunction

    function void set_write_response_ready_delay( input int lwrite_response_ready_delay );
      write_response_ready_delay = lwrite_response_ready_delay;
    endfunction

    function int get_write_response_ready_delay();
      return write_response_ready_delay;
    endfunction

    function void set_gen_write_strobes( input bit lgen_write_strobes);
      gen_write_strobes = lgen_write_strobes;
    endfunction

    function bit get_gen_write_strobes();
      return gen_write_strobes;
    endfunction

    function void set_operation_mode( input axi4_operation_mode_e loperation_mode );
      operation_mode = loperation_mode;
    endfunction

    function axi4_operation_mode_e get_operation_mode();
      return operation_mode;
    endfunction

    function void set_write_data_mode( input axi4_write_data_mode_e lwrite_data_mode );
      write_data_mode = lwrite_data_mode;
    endfunction

    function axi4_write_data_mode_e get_write_data_mode();
      return write_data_mode;
    endfunction

    function void set_data_beat_done( input int ldata_beat_done, input int index = 0 );
      data_beat_done[index] = ldata_beat_done;
    endfunction

    function int get_data_beat_done( input int index = 0 );
      return data_beat_done[index];
    endfunction

    function void set_transaction_done( input int ltransaction_done );
      transaction_done = ltransaction_done;
    endfunction

    function int get_transaction_done();
      return transaction_done;
    endfunction

    // Function: do_print
    //
    // Prints axi4_transaction transaction attributes
    function void print (bit print_delays = 1'b0);
      $display("------------------------------------------------------------------------");
      $display("%0t: %s axi4_transaction", $time, driver_name);
      $display("------------------------------------------------------------------------");
      $display("read_or_write : %s", read_or_write.name());
      $display("addr : 'h%h", addr);
      $display("prot : %s", prot.name());
      $display("region : 'h%h", region);
      $display("size : %s", size.name());
      $display("burst : %s", burst.name());
      $display("lock : %s", lock.name());
      $display("cache : %s", cache.name());
      $display("qos : 'h%h", qos);
      $display("id : 'h%h", id);
      $display("burst_length : 'h%h", burst_length);
      $display("addr_user : 'h%h", addr_user);
      foreach( data_words[i0_1] )
        $display("data_words[%0d] : 'h%h", i0_1, data_words[i0_1]);
      foreach( write_strobes[i0_1] )
        $display("write_strobes[%0d] : 'h%h", i0_1, write_strobes[i0_1]);
      foreach( data_user[i0_1] )
        $display("data_user[%0d] : 'h%h", i0_1, data_user[i0_1]);
      foreach( resp[i0_1] )
        $display("resp[%0d] : %s", i0_1, resp[i0_1].name());
      foreach( resp_user[i0_1] )
        $display("resp_user[%0d] : 'h%h", i0_1, resp_user[i0_1]);
      $display("gen_write_strobes : 'b%b", gen_write_strobes );
      $display("operation_mode   : %s", operation_mode.name() );
      $display("write_data_mode  : %s", write_data_mode.name() );
      foreach( data_beat_done[i0_1] )
        $display("data_beat_done[%0d] : 'b%b", i0_1, data_beat_done[i0_1] );
      $display("transaction_done : 'b%b", transaction_done );
      if ( print_delays == 1'b1 )
      begin
        $display("address_valid_delay : %0d", address_valid_delay);
        foreach( data_valid_delay[i0_1] )
          $display("data_valid_delay[%0d] : %0d", i0_1, data_valid_delay[i0_1]);
        $display("write_response_valid_delay : %0d", write_response_valid_delay);
        $display("address_ready_delay : %0d", address_ready_delay);
        foreach( data_ready_delay[i0_1] )
          $display("data_ready_delay[%0d] : %0d", i0_1, data_ready_delay[i0_1]);
        $display("write_response_ready_delay : %0d", write_response_ready_delay);
      end
    endfunction
endclass


//------------------------------------------------------------------------------
//
// Enum: axi4_call_back_e
//
//------------------------------------------------------------------------------
//
// The types of callback that can be registered with this <dvc_axi4> interface.
//
// AXI4_REPORTER_CB                         - Callback used to redirect BFM assertion and debug messages to the UVM messaging system
//
typedef enum
{
    AXI4_REPORTER_CB                         = 0
} axi4_call_back_e;
`else
// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

// Title: AXI4 Enumeration Types
//

//------------------------------------------------------------------------------
//
// Enum: axi4_prot_e
//
//------------------------------------------------------------------------------
//  Protection type mapped to the AxPROT signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.7))
//  
// AXI4_NORM_SEC_DATA - Normal/Secure/Data
// AXI4_PRIV_SEC_DATA - Privileged/Secure/Data
// AXI4_NORM_NONSEC_DATA - Normal/Non-secure/Data
// AXI4_PRIV_NONSEC_DATA - Privileged/Non-secure/Data
// AXI4_NORM_SEC_INST - Normal/Secure/Instruction
// AXI4_PRIV_SEC_INST - Privileged/Secure/Instruction
// AXI4_NORM_NONSEC_INST - Normal/Non-secure/Instruction
// AXI4_PRIV_NONSEC_INST - Privileged/Non-secure/Instruction
//  
typedef enum bit [2:0]
{
    AXI4_NORM_SEC_DATA    = 3'h0,
    AXI4_PRIV_SEC_DATA    = 3'h1,
    AXI4_NORM_NONSEC_DATA = 3'h2,
    AXI4_PRIV_NONSEC_DATA = 3'h3,
    AXI4_NORM_SEC_INST    = 3'h4,
    AXI4_PRIV_SEC_INST    = 3'h5,
    AXI4_NORM_NONSEC_INST = 3'h6,
    AXI4_PRIV_NONSEC_INST = 3'h7
} axi4_prot_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_response_e
//
//------------------------------------------------------------------------------
//  Response type mapped to the xRESP signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)
// 
// AXI4_OKAY   - Normal access has been successful, or exclusive access has failed (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4).
// AXI4_EXOKAY - Exclusive acess okay (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)) OR Reserved value if QVIP BFM is configured as axi4lite interface (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.1.
// AXI4_SLVERR - Slave signals an error to the originating master.
// AXI4_DECERR - Decode error (generally used to indicate that there is no slave at the transaction address).
//  
typedef enum bit [1:0]
{
    AXI4_OKAY   = 2'h0,
    AXI4_EXOKAY = 2'h1,
    AXI4_SLVERR = 2'h2,
    AXI4_DECERR = 2'h3
} axi4_response_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_rw_e
//
//------------------------------------------------------------------------------
//  This is used as the <read_or_write> argument to the <axi4_master_rw_transaction> transaction to indicate read or write.
// 
// AXI4_TRANS_READ - read transaction
// AXI4_TRANS_WRITE - write transaction
//  
typedef enum bit [0:0]
{
    AXI4_TRANS_READ  = 1'h0,
    AXI4_TRANS_WRITE = 1'h1
} axi4_rw_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_size_e
//
//------------------------------------------------------------------------------
//  Word-size encoding mapped to the AxSIZE signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1). 
// 
// AXI4_BYTES_1 - 4'b0000 i.e. 1 byte
// AXI4_BYTES_2 - 4'b0001 i.e. 2 bytes
// AXI4_BYTES_4 - 4'b0010 i.e. 4 bytes
// AXI4_BYTES_8 - 4'b0011 i.e. 8 bytes
// AXI4_BYTES_16 - 4'b0100 i.e. 16 bytes
// AXI4_BYTES_32 - 4'b0101 i.e. 32 bytes
// AXI4_BYTES_64 - 4'b0110 i.e. 64 bytes
// AXI4_BYTES_128 - 4'b0111 i.e. 128 bytes
// AXI4_BYTES_256 - 4'b1000 i.e. 256 bytes
// 
typedef enum bit [3:0]
{
    AXI4_BYTES_1   = 4'h0,
    AXI4_BYTES_2   = 4'h1,
    AXI4_BYTES_4   = 4'h2,
    AXI4_BYTES_8   = 4'h3,
    AXI4_BYTES_16  = 4'h4,
    AXI4_BYTES_32  = 4'h5,
    AXI4_BYTES_64  = 4'h6,
    AXI4_BYTES_128 = 4'h7,
    AXI4_BYTES_256 = 4'h8
} axi4_size_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_cache_e
//
//------------------------------------------------------------------------------
//  Cache behaviour type, mapped to the AxCACHE signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4).
// 
// This has identical values to those in AXI3, but the AxCACHE[1] bit is now named 'Modifiable' (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.3.1)
// 
// AXI4_NONMODIFIABLE_NONBUF - Non-modifiable and non-bufferable
// AXI4_BUF_ONLY - Bufferable only
// AXI4_CACHE_2 - Bit pattern 0010
// AXI4_CACHE_3 - Bit pattern 0011
// AXI4_CACHE_RSVD_4 - Reserved
// AXI4_CACHE_RSVD_5 - Reserved
// AXI4_CACHE_6 - Bit pattern 0110
// AXI4_CACHE_7 - Bit pattern 0111
// AXI4_CACHE_RSVD_8 - Reserved
// AXI4_CACHE_RSVD_9 - Reserved
// AXI4_CACHE_10 - Bit pattern 1010
// AXI4_CACHE_11 - Bit pattern 1011
// AXI4_CACHE_RSVD_12 - Reserved
// AXI4_CACHE_RSVD_13 - Reserved
// AXI4_CACHE_14 - Bit pattern 1110
// AXI4_CACHE_15 - Bit pattern 1111
// 
typedef enum bit [3:0]
{
    AXI4_NONMODIFIABLE_NONBUF = 4'h0,
    AXI4_BUF_ONLY             = 4'h1,
    AXI4_CACHE_2              = 4'h2,
    AXI4_CACHE_3              = 4'h3,
    AXI4_CACHE_RSVD_4         = 4'h4,
    AXI4_CACHE_RSVD_5         = 4'h5,
    AXI4_CACHE_6              = 4'h6,
    AXI4_CACHE_7              = 4'h7,
    AXI4_CACHE_RSVD_8         = 4'h8,
    AXI4_CACHE_RSVD_9         = 4'h9,
    AXI4_CACHE_10             = 4'ha,
    AXI4_CACHE_11             = 4'hb,
    AXI4_CACHE_RSVD_12        = 4'hc,
    AXI4_CACHE_RSVD_13        = 4'hd,
    AXI4_CACHE_14             = 4'he,
    AXI4_CACHE_15             = 4'hf
} axi4_cache_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_burst_e
//
//------------------------------------------------------------------------------
//  Burst type mapped to the AxBURST signals to determine address calculation (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4). 
// 
// AXI4_FIXED - Fixed
// AXI4_INCR - Incrementing
// AXI4_WRAP - Wrap
// AXI4_RESERVED - Reserved value. May be enabled by setting the configuration parameter <config_enable_burst_reserved_value>.
// 
typedef enum bit [1:0]
{
    AXI4_FIXED    = 2'h0,
    AXI4_INCR     = 2'h1,
    AXI4_WRAP     = 2'h2,
    AXI4_RESERVED = 2'h3
} axi4_burst_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_lock_e
//
//------------------------------------------------------------------------------
//  Access type mapped to the AxLOCK signals (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.4).
// 
// AXI4_NORMAL - Normal Access, respecting barriers
// AXI4_EXCLUSIVE - Exclusive Access, respecting barriers
// 
typedef enum bit [0:0]
{
    AXI4_NORMAL    = 1'h0,
    AXI4_EXCLUSIVE = 1'h1
} axi4_lock_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_interface_type_e
//
//------------------------------------------------------------------------------
//  Type to configure the category of the interface (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2) as read/write, read-only or write-only by setting the value of <config_interface_type> configuration parameter.
// 
// AXI4_READWRITE - Read and Write
// AXI4_READONLY - Read Only
// AXI4_WRITEONLY - Write Only
// 
typedef enum bit [1:0]
{
    AXI4_READWRITE = 2'h0,
    AXI4_READONLY  = 2'h1,
    AXI4_WRITEONLY = 2'h2
} axi4_interface_type_e;



//------------------------------------------------------------------------------
//
// Enum: axi4_assertion_e
//
//------------------------------------------------------------------------------
//  Type defining the error messages which can be produced by the <mgc_axi4> Questa Verification IP.
// 
// Individual error messages can be disabled using the <config_enable_assertion> array of configuration bits.
// 
// AXI4_ADDRESS_WIDTH_EXCEEDS_64                               -  60000 -  AXI4 supports up to 64-bit addressing (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_ARADDR_CHANGED_BEFORE_ARREADY                          -  60001 -  The value of <ARADDR> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARADDR_UNKN                                            -  60002 -  <ARADDR> has an X value/<ARADDR> has an Z value
// AXI4_ARPROT_CHANGED_BEFORE_ARREADY                          -  60003 -  The value of <ARPROT> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARPROT_UNKN                                            -  60004 -  <ARPROT> has an X value/<ARPROT> has an Z value
// AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID                     -  60005 -  Once <ARVALID> has been asserted, <ARREADY> should be asserted within <config_max_latency_ARVALID_assertion_to_ARREADY> clock periods
// AXI4_ARREADY_UNKN                                           -  60006 -  <ARREADY> has an X value/<ARREADY> has a Z value
// AXI4_ARVALID_DEASSERTED_BEFORE_ARREADY                      -  60007 -  <ARVALID> has been de-asserted before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARVALID_HIGH_ON_FIRST_CLOCK                            -  60008 -  A master interface must begin driving <ARVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_ARVALID_UNKN                                           -  60009 -  <ARVALID> has an X value/<ARVALID> has an Z value
// AXI4_AWADDR_CHANGED_BEFORE_AWREADY                          -  60010 -  The value of <AWADDR> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWADDR_UNKN                                            -  60011 -  <AWADDR> has an X value/<AWADDR> has an Z value
// AXI4_AWPROT_CHANGED_BEFORE_AWREADY                          -  60012 -  The value of <AWPROT> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWPROT_UNKN                                            -  60013 -  <AWPROT> has an X value/<AWPROT> has an Z value
// AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID                     -  60014 -  Once <AWVALID> has been asserted, <AWREADY> should be asserted within <config_max_latency_AWVALID_assertion_to_AWREADY> clock periods
// AXI4_AWREADY_UNKN                                           -  60015 -  <AWREADY> has an X value/<AWREADY> has an Z value
// AXI4_AWVALID_DEASSERTED_BEFORE_AWREADY                      -  60016 -  <AWVALID> has been de-asserted before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWVALID_HIGH_ON_FIRST_CLOCK                            -  60017 -  A master interface must begin driving <AWVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_AWVALID_UNKN                                           -  60018 -  <AWVALID> has an X value/<AWVALID> has an Z value
// AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID                       -  60019 -  Once <BVALID> has been asserted, <BREADY> should be asserted within <config_max_latency_BVALID_assertion_to_BREADY> clock periods
// AXI4_BREADY_UNKN                                            -  60020 -  <BREADY> has an X value/<BREADY> has an Z value
// AXI4_BRESP_CHANGED_BEFORE_BREADY                            -  60021 -  The value of <BRESP> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BRESP_UNKN                                             -  60022 -  <BRESP> has an X value/<BRESP> has a Z value
// AXI4_BVALID_DEASSERTED_BEFORE_BREADY                        -  60023 -  <BVALID> has been de-asserted before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BVALID_HIGH_EXITING_RESET                              -  60024 -  <BVALID> should have been driven low when exiting reset (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_BVALID_UNKN                                            -  60025 -  <BVALID> has an X value/<BVALID> has a Z value
// AXI4_DEC_ERR_RESP_FOR_READ                                  -  60026 -  DEPRECATED - This assertion is now DEPRECATED, as DECERR response for a read transaction is not a protocol violation.
// AXI4_DEC_ERR_RESP_FOR_WRITE                                 -  60027 -  DEPRECATED - This assertion is now DEPRECATED, as DECERR response for a write transaction is not a protocol violation.
// AXI4_EXOKAY_RESPONSE_NORMAL_READ                            -  60028 -  Slave has responded <AXI4_EXOKAY> to a non exclusive read transfer
// AXI4_EXOKAY_RESPONSE_NORMAL_WRITE                           -  60029 -  Slave has responded <AXI4_EXOKAY> to a non exclusive write transfer 
// AXI4_INVALID_WRITE_STROBES_ON_UNALIGNED_WRITE_TRANSFER      -  60030 -  Write strobe(s) incorrect for address/size of an unaligned transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.3)
// AXI4_MINIMUM_SLAVE_ADDRESS_SPACE_VIOLATION                  -  60031 -  The minimum address space occupied by a single slave device is 4 kilobytes (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.2)
// AXI4_PARAM_READ_DATA_BUS_WIDTH                              -  60032 -  The value of <AXI4_RDATA_WIDTH> must be either 32 or 64 (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.2)
// AXI4_PARAM_WRITE_DATA_BUS_WIDTH                             -  60033 -  The value of <AXI4_WDATA_WIDTH> must be either 32 or 64 (see AMBA AXI and ACE Protocol Specification IHI0022D section B1.1.2)
// AXI4_READ_DATA_BEFORE_ADDRESS                               -  60034 -  An unexpected read response has occurred (there are no outstanding read transactions with this id) (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.3)
// AXI4_READ_DATA_CHANGED_BEFORE_RREADY                        -  60035 -  The value of <RDATA> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_READ_DATA_UNKN                                         -  60036 -  <RDATA> has an X value/<RDATA> has a Z value
// AXI4_READ_RESP_CHANGED_BEFORE_RREADY                        -  60037 -  The value of <RRESP> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID                       -  60038 -  Once <RVALID> has been asserted, <RREADY> should be asserted within <config_max_latency_RVALID_assertion_to_RREADY> clock periods
// AXI4_RREADY_UNKN                                            -  60039 -  <RREADY> has an X value/<RREADY> has a Z value
// AXI4_RRESP_UNKN                                             -  60040 -  <RRESP> has an X value/<RRESP> has a Z value
// AXI4_RVALID_DEASSERTED_BEFORE_RREADY                        -  60041 -  <RVALID> has been de-asserted before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RVALID_HIGH_EXITING_RESET                              -  60042 -  <RVALID> should have been driven low when exiting reset (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_RVALID_UNKN                                            -  60043 -  <RVALID> has an X value/<RVALID> has a Z value
// AXI4_SLV_ERR_RESP_FOR_READ                                  -  60044 -  DEPRECATED - This assertion is now DEPRECATED, as SLVERR response for a read transaction is not a protocol violation.
// AXI4_SLV_ERR_RESP_FOR_WRITE                                 -  60045 -  DEPRECATED - This assertion is now DEPRECATED, as SLVERR response for a write transaction is not a protocol violation.
// AXI4_TIMEOUT_WAITING_FOR_READ_RESPONSE                      -  60046 -  Timed-out waiting for a read response (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_TIMEOUT_WAITING_FOR_WRITE_RESPONSE                     -  60047 -  Timed-out waiting for a write response (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_INVALID_LANE            -  60048 -  On a lane whose strobe is 0, the value of <WDATA> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_VALID_LANE              -  60049 -  On a lane whose strobe is 1, the value of <WDATA> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID                       -  60050 -  Once <WVALID> has been asserted, <WREADY> should be asserted within <config_max_latency_WVALID_assertion_to_WREADY> clock periods
// AXI4_WREADY_UNKN                                            -  60051 -  <WREADY> has an X value/<WREADY> has a Z value
// AXI4_WRITE_DATA_BEFORE_ADDRESS                              -  60052 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol write data beat can occurred before the corresponding address phase
// AXI4_WRITE_DATA_UNKN_ON_INVALID_LANE                        -  60053 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol Byte lanes for which strobe is 0 could take unknown value.
// AXI4_WRITE_DATA_UNKN_ON_VALID_LANE                          -  60054 -  On a lane whose strobe is 1, <WDATA> has an X value/<WDATA> has a Z value
// AXI4_WRITE_RESPONSE_WITHOUT_ADDR_DATA                       -  60055 -  An unexpected write response has occurred (there are no outstanding write transactions with this id) 
// AXI4_WSTRB_CHANGED_BEFORE_WREADY                            -  60056 -  The value of <WSTRB> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WSTRB_UNKN                                             -  60057 -  <WSTRB> has an X value/<WSTRB> has an Z value
// AXI4_WVALID_DEASSERTED_BEFORE_WREADY                        -  60058 -  <WVALID> has been de-asserted before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WVALID_HIGH_ON_FIRST_CLOCK                             -  60059 -  A master interface must begin driving <WVALID> high only at a rising clock edge after <ARESETn> is HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)
// AXI4_WVALID_UNKN                                            -  60060 -  <WVALID> has an X value/<WVALID> has an Z value
// MVC_FAILED_POSTCONDITION                                    -  60061 -  A postcondition failed
// MVC_FAILED_RECOGNITION                                      -  60062 -  An item failed to be recognized  
// AXI4_DEC_ERR_ILLEGAL_FOR_MAPPED_SLAVE_ADDR                  -  60063 -  Slave receives a burst to a mapped address but responds with DECERR (signaled by <AXI4_DECERR>) (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.4)
// AXI4_AWVALID_HIGH_DURING_RESET                              -  60064 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol AWVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                     
// AXI4_WVALID_HIGH_DURING_RESET                               -  60065 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol WVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                       
// AXI4_BVALID_HIGH_DURING_RESET                               -  60066 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol BVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                        
// AXI4_ARVALID_HIGH_DURING_RESET                              -  60067 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol ARVALID timing during the reset state is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                          
// AXI4_RVALID_HIGH_DURING_RESET                               -  60068 -  DEPRECATED - This assertion is now DEPRECATED, as per protocol RVALID timing during the reset state  is not defined (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.1.2)                                                                                                                         
// AXI4_TIMEOUT_WAITING_FOR_WRITE_DATA                         -  60069 -  Timed-out waiting for a data phase in write data burst (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.6)
// AXI4_TIMEOUT_WAITING_FOR_WRITE_ADDR_AFTER_DATA              -  60070 -  Timed-out waiting for a write address phase to be coming after data see AMBA AXI and ACE Protocol Specification IHI0022D section A2.2
// AXI4_ARESETn_SIGNAL_Z                                       -  60071 -  AXI4 Reset signal has z value
// AXI4_ARESETn_SIGNAL_X                                       -  60072 -  AXI4 Reset signal has x value
// AXI4_ADDR_FOR_READ_BURST_ACROSS_4K_BOUNDARY                 -  60073 -  This read transaction has crossed a 4KB boundary (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1) 
// AXI4_ADDR_FOR_WRITE_BURST_ACROSS_4K_BOUNDARY                -  60074 -  This write transaction has crossed a 4KB boundary (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ARADDR_FALLS_IN_REGION_HOLE                            -  60075 -  The <ARADDR> value cannot be decoded to a region in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_ARBURST_CHANGED_BEFORE_ARREADY                         -  60076 -  The value of <ARBURST> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARBURST_UNKN                                           -  60077 -  <ARBURST> has an X value/<ARBURST> has an Z value
// AXI4_ARCACHE_CHANGED_BEFORE_ARREADY                         -  60078 -  The value of <ARCACHE> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARCACHE_UNKN                                           -  60079 -  <ARCACHE> has an X value/<ARCACHE> has an Z value
// AXI4_ARID_CHANGED_BEFORE_ARREADY                            -  60080 -  The value of <ARID> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARID_UNKN                                              -  60081 -  <ARID> has an X value/<ARID> has an Z value
// AXI4_ARLEN_CHANGED_BEFORE_ARREADY                           -  60082 -   The value of <ARLEN> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARLEN_UNKN                                             -  60083 -  <ARLEN> has an X value/<ARLEN> has an Z value
// AXI4_ARLOCK_CHANGED_BEFORE_ARREADY                          -  60084 -  The value of <ARLOCK> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARLOCK_UNKN                                            -  60085 -  <ARLOCK> has an X value/<ARLOCK> has an Z value
// AXI4_ARQOS_CHANGED_BEFORE_ARREADY                           -  60086 -  The value of <ARQOS>  has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARQOS_UNKN                                             -  60087 -  <ARQOS>  has an X value/<ARQOS> has an Z value
// AXI4_ARREGION_CHANGED_BEFORE_ARREADY                        -  60088 -  The value of <ARREGION> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARREGION_MISMATCH                                      -  60089 -  The <ARREGION> value does not match the value defined in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_ARREGION_UNKN                                          -  60090 -  <ARREGION> has an X value/<ARREGION> has an Z value
// AXI4_ARSIZE_CHANGED_BEFORE_ARREADY                          -  60091 -  The value of <ARSIZE> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARSIZE_UNKN                                            -  60092 -  <ARSIZE> has an X value/<ARSIZE> has an Z value
// AXI4_ARUSER_CHANGED_BEFORE_ARREADY                          -  60093 -  The value of <ARUSER> has changed from its initial value between the time <ARVALID> was asserted, and before <ARREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_ARUSER_UNKN                                            -  60094 -  <ARUSER> has an X value/<ARUSER> has an Z value
// AXI4_AWADDR_FALLS_IN_REGION_HOLE                            -  60095 -  The addr value cannot be decoded to a region in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_AWBURST_CHANGED_BEFORE_AWREADY                         -  60096 -  The value of <AWBURST> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWBURST_UNKN                                           -  60097 -  <AWBURST> has an X value/<AWBURST> has an Z value
// AXI4_AWCACHE_CHANGED_BEFORE_AWREADY                         -  60098 -  The value of <AWCACHE> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWCACHE_UNKN                                           -  60099 -  <AWCACHE> has an X value/AWCACHE has an Z value
// AXI4_AWID_CHANGED_BEFORE_AWREADY                            -  60100 -  The value of <AWID> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWID_UNKN                                              -  60101 -  <AWID> has an X value/<AWID> has an Z value
// AXI4_AWLEN_CHANGED_BEFORE_AWREADY                           -  60102 -  The value of <AWLEN> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWLEN_UNKN                                             -  60103 -  <AWLEN> has an X value/<AWLEN> has an Z value
// AXI4_AWLOCK_CHANGED_BEFORE_AWREADY                          -  60104 -  The value of <AWLOCK> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWLOCK_UNKN                                            -  60105 -  <AWLOCK> has an X value/<AWLOCK> has an Z value
// AXI4_AWQOS_CHANGED_BEFORE_AWREADY                           -  60106 -  The value of <AWQOS> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWQOS_UNKN                                             -  60107 -  <AWQOS> has an X value/<AWQOS> has an Z value
// AXI4_AWREGION_CHANGED_BEFORE_AWREADY                        -  60108 -  The value of <AWREGION> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWREGION_MISMATCH                                      -  60109 -  The <AWREGION> value does not match the value defined in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_AWREGION_UNKN                                          -  60110 -  <AWREGION> has an X value/<AWREGION> has an Z value
// AXI4_AWSIZE_CHANGED_BEFORE_AWREADY                          -  60111 -  The value of <AWSIZE> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWSIZE_UNKN                                            -  60112 -  <AWSIZE> has an X value/<AWSIZE> has an Z value
// AXI4_AWUSER_CHANGED_BEFORE_AWREADY                          -  60113 -  The value of <AWUSER> has changed from its initial value between the time <AWVALID> was asserted, and before <AWREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_AWUSER_UNKN                                            -  60114 -  <AWUSER> has an X value/<AWUSER> has an Z value
// AXI4_BID_CHANGED_BEFORE_BREADY                              -  60115 -  The value of <BID> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BID_UNKN                                               -  60116 -  <BID> has an X value/<BID> has a Z value
// AXI4_BUSER_CHANGED_BEFORE_BREADY                            -  60117 -  The value of <BUSER> has changed from its initial value between the time <BVALID> was asserted, and before <BREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_BUSER_UNKN                                             -  60118 -  <BUSER> has an X value/<BUSER> has a Z value
// AXI4_EXCLUSIVE_READ_ACCESS_MODIFIABLE                       -  60119 -  Exclusive read access must not have AxCACHE value that indicates that the transaction is cacheable (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_128              -  60120 -  Number of bytes in an exclusive read transaction must be less than or equal to 128 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_128             -  60121 -  Number of bytes in an exclusive write transaction must be less than or equal to 128 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_NOT_POWER_OF_2           -  60122 -  Number of bytes of an exclusive read transaction is not a power of 2 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_NOT_POWER_OF_2          -  60123 -  Number of bytes of an exclusive write transaction is not a power of 2 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_READ_LENGTH_EXCEEDS_16                       -  60124 -  Exclusive read accesses are not permitted to use a burst length greater than 16 (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_ADDRESS_NOT_SAME_AS_RD                    -  60125 -  Exclusive write does not match the address of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_BURST_NOT_SAME_AS_RD                      -  60126 -  Exclusive write does not match the burst setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_CACHE_NOT_SAME_AS_RD                      -  60127 -  Exclusive write does not match the cache setting of the previous exclusive read to this id (see the ARM AXI4 compliance-checker AXI4_RECM_EXCL_MATCH assertion code)
// AXI4_EXCLUSIVE_WRITE_ACCESS_MODIFIABLE                      -  60128 -  Exclusive write access must not have AxCACHE value that indicates that the transaction is cacheable (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_LENGTH_NOT_SAME_AS_RD                     -  60129 -  Exclusive write does not match the length of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_PROT_NOT_SAME_AS_RD                       -  60130 -  Exclusive write does not match the prot setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_REGION_NOT_SAME_AS_RD                     -  60131 -  Exclusive write does not match the region setting of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCLUSIVE_WR_SIZE_NOT_SAME_AS_RD                       -  60132 -  Exclusive write does not match the size of the previous exclusive read to this id (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_EXOKAY_RESP_EXPECTED_OKAY                        -  60133 -  Expected <AXI4_OKAY> response to this exclusive read (because the parameters did not meet the the restrictions), but got <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS       -  60134 -  Response for an exclusive read to a slave which does not support exclusive access should be <AXI4_OKAY>, but it returned <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.5)
// AXI4_EX_RD_OKAY_RESP_EXPECTED_EXOKAY                        -  60135 -  Expected <AXI4_EXOKAY> response to this exclusive read (because the parameters met the restrictions), but got <AXI4_OKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EX_RD_WHEN_EX_NOT_ENABLED                              -  60136 -  An exclusive read should not be issued when exclusive transactions are not enabled
// AXI4_EX_WRITE_BEFORE_EX_READ_RESPONSE                       -  60137 -  Exclusive write has occurred, with no previous exclusive read
// AXI4_EX_WRITE_EXOKAY_RESP_EXPECTED_OKAY                     -  60138 -  Exclusive write has not been successful, yet slave has responded with <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.2)
// AXI4_EX_WRITE_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS    -  60139 -  Response for an exclusive write to a slave which does not support exclusive access should be <AXI4_OKAY>, but it returned <AXI4_EXOKAY> (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.5)
// AXI4_EX_WRITE_OKAY_RESP_EXPECTED_EXOKAY                     -  60140 -  An <AXI4_OKAY> response to an exclusive write occurred, but an <AXI4_EXOKAY> response had been expected (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.2). If the slave has multiple interfaces to the system, this check should be disabled, as it is possible for this response to occur as a result of activity on another port.
// AXI4_EX_WR_WHEN_EX_NOT_ENABLED                              -  60141 -  An exclusive write should not be issued when exclusive transactions are not enabled
// AXI4_ILLEGAL_ARCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     -  60142 -  For a read from a cacheable address region, one of bits 2 or 3 of the cache parameter must be HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_ARCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION -  60143 -  For a read from a non-cacheable address region, bits 2 and 3 of the cache parameter must be LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_AWCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     -  60144 -  For a write to a cacheable address region, one of bits 2 or 3 of the cache parameter must be HIGH (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_AWCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION -  60145 -  For a write to a non-cacheable address region, bits 2 and 3 of the cache parameter must be LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.5)
// AXI4_ILLEGAL_LENGTH_FIXED_READ_BURST                        -  60146 -  In the last read address phase burst_length has an illegal value for a burst of type AXI4_FIXED (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_LENGTH_FIXED_WRITE_BURST                       -  60147 -  In the last write address phase burst_length has an illegal value for a burst of type AXI4_FIXED (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1) 
// AXI4_ILLEGAL_LENGTH_WRAPPING_READ_BURST                     -  60148 -  In the last read address phase burst_length has an illegal value for a burst of type AXI4_WRAP (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_LENGTH_WRAPPING_WRITE_BURST                    -  60149 -  In the last write address phase burst_length has an illegal value for a burst of type AXI4_WRAP (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_READ                        -  60150 -  Error response (<AXI4_SLVERR>/<AXI4_DECERR>) received for an exclusive read
// AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_WRITE                       -  60151 -  Error response (<AXI4_SLVERR>/<AXI4_DECERR>) received for an exclusive write
// AXI4_INVALID_REGION_CARDINALITY                             -  60152 -  The configuration parameter <config_slave_regions> does not lie in the range 1-16, inclusive (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1).
// AXI4_INVALID_WRITE_STROBES_ON_ALIGNED_WRITE_TRANSFER        -  60153 -  Write strobe(s) incorrect for address/size of an aligned transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.3)
// AXI4_NON_INCREASING_REGION_SPECIFICATION                    -  60154 -  A region address-range has an upper bound smaller than the lower bound.
// AXI4_NON_ZERO_ARQOS                                         -  60155 -  The master is configured to not participate in the Quality-of-Service scheme, but <ARQOS>  is not 4'b0000, as it should be (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.1.2)
// AXI4_NON_ZERO_AWQOS                                         -  60156 -  The master is configured to not participate in the Quality-of-Service scheme, but <AWQOS> is not 4'b0000, as it should be (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.1.2)
// AXI4_OVERLAPPING_REGION                                     -  60157 -  An address-range in the region map overlaps with another address in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1). 
// AXI4_PARAM_READ_REORDERING_DEPTH_EQUALS_ZERO                -  60158 -  The user-supplied config_read_data_reordering_depth should be greater than zero (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_PARAM_READ_REORDERING_DEPTH_EXCEEDS_MAX_ID             -  60159 -  The user-supplied config_read_data_reordering_depth exceeds the maximum possible value, as defined by the AXI4_ID_WIDTH parameter (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_12                   -  60160 -  The RA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_13                   -  60161 -  The RA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_4                    -  60162 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_5                    -  60163 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_8                    -  60164 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_9                    -  60165 -  The RA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_READ_BURST_LENGTH_VIOLATION                            -  60166 -  The burst_length implied by the number of beats actually read does not match the burst_length defined by the <axi4_master_read_addr_channel_phase>
// AXI4_READ_BURST_MAXIMUM_LENGTH_VIOLATION                    -  60167 -  256 read data beats were seen without <RLAST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_READ_BURST_SIZE_VIOLATION                              -  60168 -  In this read transaction, size has been set too high for the defined data buswidth
// AXI4_READ_EXCLUSIVE_ENCODING_VIOLATION                      -  60169 -  A read-only interface does not support exclusive accesses (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2.2)
// AXI4_READ_REORDERING_VIOLATION                              -  60170 -  The arrival of a read response has exceeded the read reordering depth (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_TRANSFER_EXCEEDS_ADDRESS_SPACE                    -  60171 -  This read transfer runs off the edge of the address space defined by AXI4_ADDRESS_WIDTH (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_REGION_SMALLER_THAN_4KB                                -  60172 -  An address-range in the region map is smaller than 4kB (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_RESERVED_ARBURST_ENCODING                              -  60173 -  The reserved encoding of 2'b11 should not be used for <ARBURST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_RESERVED_AWBURST_ENCODING                              -  60174 -  The reserved encoding of 2'b11 should not be used for <AWBURST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_RID_CHANGED_BEFORE_RREADY                              -  60175 -  The value of <RID> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RID_UNKN                                               -  60176 -  <RID> has an X value/<RID> has a Z value
// AXI4_RLAST_CHANGED_BEFORE_RREADY                            -  60177 -  The value of <RLAST> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RLAST_UNKN                                             -  60178 -  <RLAST> has an X value/<RLAST> has a Z value
// AXI4_RUSER_CHANGED_BEFORE_RREADY                            -  60179 -  The value of <RUSER> has changed from its initial value between the time <RVALID> was asserted, and before <RREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_RUSER_UNKN                                             -  60180 -  <RUSER> has an X value/<RUSER> has a Z value
// AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_READ                   -  60181 -  Exclusive read accesses must have address aligned to the total number of bytes in the transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_WRITE                  -  60182 -  Exclusive write accesses must have address aligned to the total number of bytes in the transaction (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_UNALIGNED_ADDR_FOR_WRAPPING_READ_BURST                 -  60183 -  Wrapping bursts must have address aligned to the start of the read transfer (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_UNALIGNED_ADDR_FOR_WRAPPING_WRITE_BURST                -  60184 -  Wrapping bursts must have address aligned to the start of the write transfet (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)   
// AXI4_WLAST_CHANGED_BEFORE_WREADY                            -  60185 -  The value of <WLAST> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WLAST_UNKN                                             -  60186 -  <WLAST> has an X value/<WLAST> has an Z value
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_12                  -  60187 -  The WA bit of the cache parameter should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_13                  -  60188 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_4                   -  60189 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_5                   -  60190 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_8                   -  60191 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_9                   -  60192 -  The WA of the cache parameter bit should not be HIGH when the Modifiable bit is LOW (see AMBA AXI and ACE Protocol Specification IHI0022D section A4.4)
// AXI4_WRITE_BURST_LENGTH_VIOLATION                           -  60193 -  The number of data beats in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_STROBES_LENGTH_VIOLATION                         -  60194 -  The size of the write_strobes array in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_USER_DATA_LENGTH_VIOLATION                       -  60195 -  The size of the wdata_user_data array in a write transfer should match the value given by <AWLEN>
// AXI4_WRITE_BURST_MAXIMUM_LENGTH_VIOLATION                   -  60196 -  256 write data beats were seen without <WLAST> (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1 )
// AXI4_WRITE_BURST_SIZE_VIOLATION                             -  60197 -  In this write transaction, size has been set too high for the defined data buswidth
// AXI4_WRITE_EXCLUSIVE_ENCODING_VIOLATION                     -  60198 -  A write-only interface does not support exclusive accesses (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.2.3)
// AXI4_WRITE_STROBE_FIXED_BURST_VIOLATION                     -  60199 -  Write strobe(s) incorrect for the address/size of a fixed transfer 
// AXI4_WRITE_TRANSFER_EXCEEDS_ADDRESS_SPACE                   -  60200 -  This write transfer runs off the edge of the address space defined by <AXI4_ADDRESS_WIDTH> (see AMBA AXI and ACE Protocol Specification IHI0022D section A10.3.1)
// AXI4_WRONG_ARREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    -  60201 -  The region value should be 4'b0000 for a read from a slave with a single address decode in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1)
// AXI4_WRONG_AWREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    -  60202 -  The region value should be 4'b0000 for a write to a slave with a single address decode in the region map (see AMBA AXI and ACE Protocol Specification IHI0022D section A8.2.1) 
// AXI4_WUSER_CHANGED_BEFORE_WREADY                            -  60203 -  The value of <WUSER> has changed from its initial value between the time <WVALID> was asserted, and before <WREADY> was asserted (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.2.1)
// AXI4_WUSER_UNKN                                             -  60204 -  <WUSER> has an X value/<WUSER> has an Z value
// AXI4_EXCL_RD_WHILE_EXCL_WR_IN_PROGRESS_SAME_ID              -  60205 -  Master starts an exclusive read burst while exclusive write burst with same ID tag is in progress (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_EXCL_WR_WHILE_EXCL_RD_IN_PROGRESS_SAME_ID              -  60206 -  Master starts an exclusive write burst while exclusive read burst with same ID tag is in progress (see AMBA AXI and ACE Protocol Specification IHI0022D section A7.2.4)
// AXI4_RLAST_VIOLATION                                        -  60207 -  RLAST signal should be asserted along with the final transfer of the read data burst (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_WLAST_ASSERTED_DURING_DATA_PHASE_OTHER_THAN_LAST       -  60208 -  Wlast must only be asserted during the last data phase (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_READ_ISSUING_CAPABILITY_VIOLATION                      -  60209 -  Number of outstanding Read transactions exceeded maximum Read issuing capability 
// AXI4_WRITE_ISSUING_CAPABILITY_VIOLATION                     -  60210 -  Number of outstanding Write transactions exceeded maximum Write issuing capability
// AXI4_COMBINED_ISSUING_CAPABILITY_VIOLATION                  -  60211 -  Number of outstanding Read and Write transactions exceeded maximum combined issuing capability
// AXI4_READ_ACCEPTANCE_CAPABILITY_VIOLATION                   -  60212 -  Number of outstanding Read transactions exceeded maximum Read acceptance capability
// AXI4_WRITE_ACCEPTANCE_CAPABILITY_VIOLATION                  -  60213 -  Number of outstanding Write transactions exceeded maximum Write acceptance capability
// AXI4_COMBINED_ACCEPTANCE_CAPABILITY_VIOLATION               -  60214 -  Number of outstanding Read and Write transactions exceeded maximum combined acceptance capability
// AXI4_READ_INTERLEAVING_VIOLATION                            -  60215 -  The number of read transactions being interleaved has crossed the interleaving depth (see AMBA AXI and ACE Protocol Specification IHI0022D section A5.3.1)
// AXI4_READ_BURST_SIZE_ILL                                    -  60216 -  In this read transaction, ARSIZE has been set to an illegal value (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_WRITE_BURST_SIZE_ILL                                   -  60217 -  In this write transaction, AWSIZE has been set to an illegal value (see AMBA AXI and ACE Protocol Specification IHI0022D section A3.4.1)
// AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_MAX              -  60218 -  Number of bytes in an exclusive read transaction must be less than or equal to maximum allowed count
// AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_MAX             -  60219 -  Number of bytes in an exclusive write transaction must be less than or equal to maximum allowed count
typedef enum bit [7:0]
{
    AXI4_ADDRESS_WIDTH_EXCEEDS_64                               = 8'h00,
    AXI4_ARADDR_CHANGED_BEFORE_ARREADY                          = 8'h01,
    AXI4_ARADDR_UNKN                                            = 8'h02,
    AXI4_ARPROT_CHANGED_BEFORE_ARREADY                          = 8'h03,
    AXI4_ARPROT_UNKN                                            = 8'h04,
    AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID                     = 8'h05,
    AXI4_ARREADY_UNKN                                           = 8'h06,
    AXI4_ARVALID_DEASSERTED_BEFORE_ARREADY                      = 8'h07,
    AXI4_ARVALID_HIGH_ON_FIRST_CLOCK                            = 8'h08,
    AXI4_ARVALID_UNKN                                           = 8'h09,
    AXI4_AWADDR_CHANGED_BEFORE_AWREADY                          = 8'h0a,
    AXI4_AWADDR_UNKN                                            = 8'h0b,
    AXI4_AWPROT_CHANGED_BEFORE_AWREADY                          = 8'h0c,
    AXI4_AWPROT_UNKN                                            = 8'h0d,
    AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID                     = 8'h0e,
    AXI4_AWREADY_UNKN                                           = 8'h0f,
    AXI4_AWVALID_DEASSERTED_BEFORE_AWREADY                      = 8'h10,
    AXI4_AWVALID_HIGH_ON_FIRST_CLOCK                            = 8'h11,
    AXI4_AWVALID_UNKN                                           = 8'h12,
    AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID                       = 8'h13,
    AXI4_BREADY_UNKN                                            = 8'h14,
    AXI4_BRESP_CHANGED_BEFORE_BREADY                            = 8'h15,
    AXI4_BRESP_UNKN                                             = 8'h16,
    AXI4_BVALID_DEASSERTED_BEFORE_BREADY                        = 8'h17,
    AXI4_BVALID_HIGH_EXITING_RESET                              = 8'h18,
    AXI4_BVALID_UNKN                                            = 8'h19,
    AXI4_DEC_ERR_RESP_FOR_READ                                  = 8'h1a,
    AXI4_DEC_ERR_RESP_FOR_WRITE                                 = 8'h1b,
    AXI4_EXOKAY_RESPONSE_NORMAL_READ                            = 8'h1c,
    AXI4_EXOKAY_RESPONSE_NORMAL_WRITE                           = 8'h1d,
    AXI4_INVALID_WRITE_STROBES_ON_UNALIGNED_WRITE_TRANSFER      = 8'h1e,
    AXI4_MINIMUM_SLAVE_ADDRESS_SPACE_VIOLATION                  = 8'h1f,
    AXI4_PARAM_READ_DATA_BUS_WIDTH                              = 8'h20,
    AXI4_PARAM_WRITE_DATA_BUS_WIDTH                             = 8'h21,
    AXI4_READ_DATA_BEFORE_ADDRESS                               = 8'h22,
    AXI4_READ_DATA_CHANGED_BEFORE_RREADY                        = 8'h23,
    AXI4_READ_DATA_UNKN                                         = 8'h24,
    AXI4_READ_RESP_CHANGED_BEFORE_RREADY                        = 8'h25,
    AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID                       = 8'h26,
    AXI4_RREADY_UNKN                                            = 8'h27,
    AXI4_RRESP_UNKN                                             = 8'h28,
    AXI4_RVALID_DEASSERTED_BEFORE_RREADY                        = 8'h29,
    AXI4_RVALID_HIGH_EXITING_RESET                              = 8'h2a,
    AXI4_RVALID_UNKN                                            = 8'h2b,
    AXI4_SLV_ERR_RESP_FOR_READ                                  = 8'h2c,
    AXI4_SLV_ERR_RESP_FOR_WRITE                                 = 8'h2d,
    AXI4_TIMEOUT_WAITING_FOR_READ_RESPONSE                      = 8'h2e,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_RESPONSE                     = 8'h2f,
    AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_INVALID_LANE            = 8'h30,
    AXI4_WDATA_CHANGED_BEFORE_WREADY_ON_VALID_LANE              = 8'h31,
    AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID                       = 8'h32,
    AXI4_WREADY_UNKN                                            = 8'h33,
    AXI4_WRITE_DATA_BEFORE_ADDRESS                              = 8'h34,
    AXI4_WRITE_DATA_UNKN_ON_INVALID_LANE                        = 8'h35,
    AXI4_WRITE_DATA_UNKN_ON_VALID_LANE                          = 8'h36,
    AXI4_WRITE_RESPONSE_WITHOUT_ADDR_DATA                       = 8'h37,
    AXI4_WSTRB_CHANGED_BEFORE_WREADY                            = 8'h38,
    AXI4_WSTRB_UNKN                                             = 8'h39,
    AXI4_WVALID_DEASSERTED_BEFORE_WREADY                        = 8'h3a,
    AXI4_WVALID_HIGH_ON_FIRST_CLOCK                             = 8'h3b,
    AXI4_WVALID_UNKN                                            = 8'h3c,
    MVC_FAILED_POSTCONDITION                                    = 8'h3d,
    MVC_FAILED_RECOGNITION                                      = 8'h3e,
    AXI4_DEC_ERR_ILLEGAL_FOR_MAPPED_SLAVE_ADDR                  = 8'h3f,
    AXI4_AWVALID_HIGH_DURING_RESET                              = 8'h40,
    AXI4_WVALID_HIGH_DURING_RESET                               = 8'h41,
    AXI4_BVALID_HIGH_DURING_RESET                               = 8'h42,
    AXI4_ARVALID_HIGH_DURING_RESET                              = 8'h43,
    AXI4_RVALID_HIGH_DURING_RESET                               = 8'h44,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_DATA                         = 8'h45,
    AXI4_TIMEOUT_WAITING_FOR_WRITE_ADDR_AFTER_DATA              = 8'h46,
    AXI4_ARESETn_SIGNAL_Z                                       = 8'h47,
    AXI4_ARESETn_SIGNAL_X                                       = 8'h48,
    AXI4_ADDR_FOR_READ_BURST_ACROSS_4K_BOUNDARY                 = 8'h49,
    AXI4_ADDR_FOR_WRITE_BURST_ACROSS_4K_BOUNDARY                = 8'h4a,
    AXI4_ARADDR_FALLS_IN_REGION_HOLE                            = 8'h4b,
    AXI4_ARBURST_CHANGED_BEFORE_ARREADY                         = 8'h4c,
    AXI4_ARBURST_UNKN                                           = 8'h4d,
    AXI4_ARCACHE_CHANGED_BEFORE_ARREADY                         = 8'h4e,
    AXI4_ARCACHE_UNKN                                           = 8'h4f,
    AXI4_ARID_CHANGED_BEFORE_ARREADY                            = 8'h50,
    AXI4_ARID_UNKN                                              = 8'h51,
    AXI4_ARLEN_CHANGED_BEFORE_ARREADY                           = 8'h52,
    AXI4_ARLEN_UNKN                                             = 8'h53,
    AXI4_ARLOCK_CHANGED_BEFORE_ARREADY                          = 8'h54,
    AXI4_ARLOCK_UNKN                                            = 8'h55,
    AXI4_ARQOS_CHANGED_BEFORE_ARREADY                           = 8'h56,
    AXI4_ARQOS_UNKN                                             = 8'h57,
    AXI4_ARREGION_CHANGED_BEFORE_ARREADY                        = 8'h58,
    AXI4_ARREGION_MISMATCH                                      = 8'h59,
    AXI4_ARREGION_UNKN                                          = 8'h5a,
    AXI4_ARSIZE_CHANGED_BEFORE_ARREADY                          = 8'h5b,
    AXI4_ARSIZE_UNKN                                            = 8'h5c,
    AXI4_ARUSER_CHANGED_BEFORE_ARREADY                          = 8'h5d,
    AXI4_ARUSER_UNKN                                            = 8'h5e,
    AXI4_AWADDR_FALLS_IN_REGION_HOLE                            = 8'h5f,
    AXI4_AWBURST_CHANGED_BEFORE_AWREADY                         = 8'h60,
    AXI4_AWBURST_UNKN                                           = 8'h61,
    AXI4_AWCACHE_CHANGED_BEFORE_AWREADY                         = 8'h62,
    AXI4_AWCACHE_UNKN                                           = 8'h63,
    AXI4_AWID_CHANGED_BEFORE_AWREADY                            = 8'h64,
    AXI4_AWID_UNKN                                              = 8'h65,
    AXI4_AWLEN_CHANGED_BEFORE_AWREADY                           = 8'h66,
    AXI4_AWLEN_UNKN                                             = 8'h67,
    AXI4_AWLOCK_CHANGED_BEFORE_AWREADY                          = 8'h68,
    AXI4_AWLOCK_UNKN                                            = 8'h69,
    AXI4_AWQOS_CHANGED_BEFORE_AWREADY                           = 8'h6a,
    AXI4_AWQOS_UNKN                                             = 8'h6b,
    AXI4_AWREGION_CHANGED_BEFORE_AWREADY                        = 8'h6c,
    AXI4_AWREGION_MISMATCH                                      = 8'h6d,
    AXI4_AWREGION_UNKN                                          = 8'h6e,
    AXI4_AWSIZE_CHANGED_BEFORE_AWREADY                          = 8'h6f,
    AXI4_AWSIZE_UNKN                                            = 8'h70,
    AXI4_AWUSER_CHANGED_BEFORE_AWREADY                          = 8'h71,
    AXI4_AWUSER_UNKN                                            = 8'h72,
    AXI4_BID_CHANGED_BEFORE_BREADY                              = 8'h73,
    AXI4_BID_UNKN                                               = 8'h74,
    AXI4_BUSER_CHANGED_BEFORE_BREADY                            = 8'h75,
    AXI4_BUSER_UNKN                                             = 8'h76,
    AXI4_EXCLUSIVE_READ_ACCESS_MODIFIABLE                       = 8'h77,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_128              = 8'h78,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_128             = 8'h79,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_NOT_POWER_OF_2           = 8'h7a,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_NOT_POWER_OF_2          = 8'h7b,
    AXI4_EXCLUSIVE_READ_LENGTH_EXCEEDS_16                       = 8'h7c,
    AXI4_EXCLUSIVE_WR_ADDRESS_NOT_SAME_AS_RD                    = 8'h7d,
    AXI4_EXCLUSIVE_WR_BURST_NOT_SAME_AS_RD                      = 8'h7e,
    AXI4_EXCLUSIVE_WR_CACHE_NOT_SAME_AS_RD                      = 8'h7f,
    AXI4_EXCLUSIVE_WRITE_ACCESS_MODIFIABLE                      = 8'h80,
    AXI4_EXCLUSIVE_WR_LENGTH_NOT_SAME_AS_RD                     = 8'h81,
    AXI4_EXCLUSIVE_WR_PROT_NOT_SAME_AS_RD                       = 8'h82,
    AXI4_EXCLUSIVE_WR_REGION_NOT_SAME_AS_RD                     = 8'h83,
    AXI4_EXCLUSIVE_WR_SIZE_NOT_SAME_AS_RD                       = 8'h84,
    AXI4_EX_RD_EXOKAY_RESP_EXPECTED_OKAY                        = 8'h85,
    AXI4_EX_RD_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS       = 8'h86,
    AXI4_EX_RD_OKAY_RESP_EXPECTED_EXOKAY                        = 8'h87,
    AXI4_EX_RD_WHEN_EX_NOT_ENABLED                              = 8'h88,
    AXI4_EX_WRITE_BEFORE_EX_READ_RESPONSE                       = 8'h89,
    AXI4_EX_WRITE_EXOKAY_RESP_EXPECTED_OKAY                     = 8'h8a,
    AXI4_EX_WRITE_EXOKAY_RESP_SLAVE_WITHOUT_EXCLUSIVE_ACCESS    = 8'h8b,
    AXI4_EX_WRITE_OKAY_RESP_EXPECTED_EXOKAY                     = 8'h8c,
    AXI4_EX_WR_WHEN_EX_NOT_ENABLED                              = 8'h8d,
    AXI4_ILLEGAL_ARCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     = 8'h8e,
    AXI4_ILLEGAL_ARCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION = 8'h8f,
    AXI4_ILLEGAL_AWCACHE_VALUE_FOR_CACHEABLE_ADDRESS_REGION     = 8'h90,
    AXI4_ILLEGAL_AWCACHE_VALUE_FOR_NON_CACHEABLE_ADDRESS_REGION = 8'h91,
    AXI4_ILLEGAL_LENGTH_FIXED_READ_BURST                        = 8'h92,
    AXI4_ILLEGAL_LENGTH_FIXED_WRITE_BURST                       = 8'h93,
    AXI4_ILLEGAL_LENGTH_WRAPPING_READ_BURST                     = 8'h94,
    AXI4_ILLEGAL_LENGTH_WRAPPING_WRITE_BURST                    = 8'h95,
    AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_READ                        = 8'h96,
    AXI4_ILLEGAL_RESPONSE_EXCLUSIVE_WRITE                       = 8'h97,
    AXI4_INVALID_REGION_CARDINALITY                             = 8'h98,
    AXI4_INVALID_WRITE_STROBES_ON_ALIGNED_WRITE_TRANSFER        = 8'h99,
    AXI4_NON_INCREASING_REGION_SPECIFICATION                    = 8'h9a,
    AXI4_NON_ZERO_ARQOS                                         = 8'h9b,
    AXI4_NON_ZERO_AWQOS                                         = 8'h9c,
    AXI4_OVERLAPPING_REGION                                     = 8'h9d,
    AXI4_PARAM_READ_REORDERING_DEPTH_EQUALS_ZERO                = 8'h9e,
    AXI4_PARAM_READ_REORDERING_DEPTH_EXCEEDS_MAX_ID             = 8'h9f,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_12                   = 8'ha0,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_13                   = 8'ha1,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_4                    = 8'ha2,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_5                    = 8'ha3,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_8                    = 8'ha4,
    AXI4_READ_ALLOCATE_WHEN_NON_MODIFIABLE_9                    = 8'ha5,
    AXI4_READ_BURST_LENGTH_VIOLATION                            = 8'ha6,
    AXI4_READ_BURST_MAXIMUM_LENGTH_VIOLATION                    = 8'ha7,
    AXI4_READ_BURST_SIZE_VIOLATION                              = 8'ha8,
    AXI4_READ_EXCLUSIVE_ENCODING_VIOLATION                      = 8'ha9,
    AXI4_READ_REORDERING_VIOLATION                              = 8'haa,
    AXI4_READ_TRANSFER_EXCEEDS_ADDRESS_SPACE                    = 8'hab,
    AXI4_REGION_SMALLER_THAN_4KB                                = 8'hac,
    AXI4_RESERVED_ARBURST_ENCODING                              = 8'had,
    AXI4_RESERVED_AWBURST_ENCODING                              = 8'hae,
    AXI4_RID_CHANGED_BEFORE_RREADY                              = 8'haf,
    AXI4_RID_UNKN                                               = 8'hb0,
    AXI4_RLAST_CHANGED_BEFORE_RREADY                            = 8'hb1,
    AXI4_RLAST_UNKN                                             = 8'hb2,
    AXI4_RUSER_CHANGED_BEFORE_RREADY                            = 8'hb3,
    AXI4_RUSER_UNKN                                             = 8'hb4,
    AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_READ                   = 8'hb5,
    AXI4_UNALIGNED_ADDRESS_FOR_EXCLUSIVE_WRITE                  = 8'hb6,
    AXI4_UNALIGNED_ADDR_FOR_WRAPPING_READ_BURST                 = 8'hb7,
    AXI4_UNALIGNED_ADDR_FOR_WRAPPING_WRITE_BURST                = 8'hb8,
    AXI4_WLAST_CHANGED_BEFORE_WREADY                            = 8'hb9,
    AXI4_WLAST_UNKN                                             = 8'hba,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_12                  = 8'hbb,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_13                  = 8'hbc,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_4                   = 8'hbd,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_5                   = 8'hbe,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_8                   = 8'hbf,
    AXI4_WRITE_ALLOCATE_WHEN_NON_MODIFIABLE_9                   = 8'hc0,
    AXI4_WRITE_BURST_LENGTH_VIOLATION                           = 8'hc1,
    AXI4_WRITE_STROBES_LENGTH_VIOLATION                         = 8'hc2,
    AXI4_WRITE_USER_DATA_LENGTH_VIOLATION                       = 8'hc3,
    AXI4_WRITE_BURST_MAXIMUM_LENGTH_VIOLATION                   = 8'hc4,
    AXI4_WRITE_BURST_SIZE_VIOLATION                             = 8'hc5,
    AXI4_WRITE_EXCLUSIVE_ENCODING_VIOLATION                     = 8'hc6,
    AXI4_WRITE_STROBE_FIXED_BURST_VIOLATION                     = 8'hc7,
    AXI4_WRITE_TRANSFER_EXCEEDS_ADDRESS_SPACE                   = 8'hc8,
    AXI4_WRONG_ARREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    = 8'hc9,
    AXI4_WRONG_AWREGION_FOR_SLAVE_WITH_SINGLE_ADDRESS_DECODE    = 8'hca,
    AXI4_WUSER_CHANGED_BEFORE_WREADY                            = 8'hcb,
    AXI4_WUSER_UNKN                                             = 8'hcc,
    AXI4_EXCL_RD_WHILE_EXCL_WR_IN_PROGRESS_SAME_ID              = 8'hcd,
    AXI4_EXCL_WR_WHILE_EXCL_RD_IN_PROGRESS_SAME_ID              = 8'hce,
    AXI4_RLAST_VIOLATION                                        = 8'hcf,
    AXI4_WLAST_ASSERTED_DURING_DATA_PHASE_OTHER_THAN_LAST       = 8'hd0,
    AXI4_READ_ISSUING_CAPABILITY_VIOLATION                      = 8'hd1,
    AXI4_WRITE_ISSUING_CAPABILITY_VIOLATION                     = 8'hd2,
    AXI4_COMBINED_ISSUING_CAPABILITY_VIOLATION                  = 8'hd3,
    AXI4_READ_ACCEPTANCE_CAPABILITY_VIOLATION                   = 8'hd4,
    AXI4_WRITE_ACCEPTANCE_CAPABILITY_VIOLATION                  = 8'hd5,
    AXI4_COMBINED_ACCEPTANCE_CAPABILITY_VIOLATION               = 8'hd6,
    AXI4_READ_INTERLEAVING_VIOLATION                            = 8'hd7,
    AXI4_READ_BURST_SIZE_ILL                                    = 8'hd8,
    AXI4_WRITE_BURST_SIZE_ILL                                   = 8'hd9,
    AXI4_EXCLUSIVE_READ_BYTES_TRANSFER_EXCEEDS_MAX              = 8'hda,
    AXI4_EXCLUSIVE_WRITE_BYTES_TRANSFER_EXCEEDS_MAX             = 8'hdb
} axi4_assertion_e;


//------------------------------------------------------------------------------
//
// Struct: axi4_en_perf_stats_s
//
//------------------------------------------------------------------------------
// This struct specifies the control over axi4 performance statistics
// 
// enable - Enables performance statistics collection
// 
// Default: 0 (perf stats collection is disabled)
// 
// step - Sets the step size of the sliding window for performance monitoring.
// The step size within the sliding window is measured in terms of clock cycle
// 
// Default: 100 (clock cycles)
// 
// multiple - Sets the number of steps within the sliding window. 
// The minimum value can be 1 and the maximum 10.
// 
// Default: 5
//------------------------------------------------------------------------------

typedef struct
{
    bit enable;
    int step;
    int multiple;
} axi4_en_perf_stats_s;


//------------------------------------------------------------------------------
//
// Struct: axi4_debug_type_s
//
//------------------------------------------------------------------------------
// The field of this struct if set to 1 enables printing of
//  debug statements.
//  master_seq - Enables master_seq debug
//  slave_seq - Enables slave_seq debug
// 
// Default: 0 (Debug statements printing is disabled)
//------------------------------------------------------------------------------

typedef struct packed
{
    bit master_seq;
    bit slave_seq;
    bit vip_config;
} axi4_debug_type_s;


//-------------------------------------------------------------------------
// Private Types
//-------------------------------------------------------------------------

typedef bit [1023:0] axi4_max_bits_t;

// enum: axi4_config_e
//
// An enum which fields corresponding to each configuration parameter of the VIP
//    AXI4_CONFIG_AXI4LITE_axi4 - 
//          This configuration can be enabled to use the AXI4 QVIP as AXI4-Lite.
//         Setting this variable to 1 enables the user to use the AXI4 BFM and all as AXI4-Lite
//         Setting this variable to 0 ensures usual usage as AXI4.
//          
//         Default: false(behaves as AXI4 interface)
//         
//    AXI4_CONFIG_INTERFACE_TYPE - 
//          
//         A configuration parameter (of type <axi4_interface_type_e>) defining the
//         interface category of the interface: read/write, read only or write only.
//         
//         Default: <AXI4_READWRITE>.
//         
//    AXI4_CONFIG_ENABLE_RLAST - 
//          
//         Enables use of optional RLAST signal.
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ID - 
//          
//         Enables use of optional AWID, BID, ARID, RID signals.
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ALL_ASSERTIONS - 
//         
//         Enables all protocol assertions. 
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_ASSERTION - 
//         
//         Enables individual protocol assertion.
//         This variable controls whether specific assertion within QVIP (of type <axi4_assertion_e>) is enabled or disabled.
//         Individual assertion can be disabled as follows:-
//         //-----------------------------------------------------------------------
//         // < BFM interface>.config_enable_assertion[<name of assertion>] = 1'b0;
//         //-----------------------------------------------------------------------
//         
//         For example, the assertion AXI4_READ_DATA_UNKN is disabled as follows:
//         <bfm>.config_enable_assertion[AXI4_READ_DATA_UNKN] =  1'b0; 
//         
//         Here bfm is the AXI4 interface instance name for which the assertion to be disabled. 
//         
//         Default: All assertions are enabled
//           
//         
//    AXI4_CONFIG_ENABLE_USER_SIDEBAND - 
//         
//         Enables user sideband signals. 
//         
//         Default: 1
//         
//    AXI4_CONFIG_ENABLE_SLAVE_EXCLUSIVE - 
//         
//         Enables exclusive transactions support for slave.
//         If disabled, every exclusive read/write returns an OKAY response,
//         and exclusive write updates memory. 
//         
//         Default: 1  
//         
//    AXI4_CONFIG_READ_DATA_REORDERING_DEPTH - 
//         
//         Sets the maximum number of different read transaction addresses for which read 
//         data(response) can be sent in any order from slave. 
//         
//         Default: 2 ** AXI4_ID_WIDTH
//         
//    AXI4_CONFIG_MAX_TRANSACTION_TIME_FACTOR - 
//          
//         Sets maximum timeout period (in terms of ACLK) for any complete read or write transaction, which
//         includes time period for all individual phases of transaction. 
//         
//         Default: 100000 clock cycles
//         
//    AXI4_CONFIG_BURST_TIMEOUT_FACTOR - 
//          
//         Sets maximum timeout period (in terms of ACLK) between individual phases of a transaction. 
//         
//         Default: 10000 clock cycles 
//         
//    AXI4_CONFIG_MAX_LATENCY_AWVALID_ASSERTION_TO_AWREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of AWVALID to assertion of AWREADY.
//         An error message AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID is generated if AWREADY is not asserted
//         after assertion of AWVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_ARVALID_ASSERTION_TO_ARREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of ARVALID to assertion of ARREADY.
//         An error message AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID is generated if ARREADY is not asserted
//         after assertion of ARVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_RVALID_ASSERTION_TO_RREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of RVALID to assertion of RREADY.
//         An error message AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID is generated if RREADY is not asserted
//         after assertion of RVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_BVALID_ASSERTION_TO_BREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of BVALID to assertion of BREADY.
//         An error message AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID is generated if BREADY is not asserted
//         after assertion of BVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_MAX_LATENCY_WVALID_ASSERTION_TO_WREADY - 
//          
//         Sets maximum timeout period (in terms of ACLK) from assertion of WVALID to assertion of WREADY.
//         An error message AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID is generated if WREADY is not asserted
//         after assertion of WVALID within this period. 
//         
//         Default: 10000 clock cycles
//         
//    AXI4_CONFIG_ENABLE_QOS - 
//          
//         Enables quality-of-service support. 
//         
//         Default: 1 
//         
//    AXI4_CONFIG_SETUP_TIME - 
//         
//         Sets number of simulation time units from the setup time to the active 
//         clock edge of ACLK. The setup time will always be less than the time period
//         of the clock. 
//         
//         Default: 0
//         
//    AXI4_CONFIG_HOLD_TIME - 
//         
//         Sets number of simulation time units from the hold time to the active 
//         clock edge of ACLK. 
//         
//         Default: 0
//         
//    AXI4_CONFIG_NUM_MAX_OUTSTANDING_READS - 
//         
//         Configures maximum number of read outstanding transfers allowed on the bus.
//         
//         Default: -1
//         
//    AXI4_CONFIG_NUM_MAX_OUTSTANDING_WRITES - 
//                                                                                   
//         Configures maximum number of write outstanding transfers allowed on the bus. 
//                                                                                      
//         Default: -1                                                                 
//         
//    AXI4_CONFIG_MAX_OUTSTANDING_WR -  Configures maximum possible outstanding Write transactions
//    AXI4_CONFIG_MAX_OUTSTANDING_RD -  Configures maximum possible outstanding Read transactions
//    AXI4_CONFIG_MAX_OUTSTANDING_RW -  Configures maximum possible outstanding Combined (Read and Write) transactions
//    AXI4_CONFIG_NUM_OUTSTANDING_WR_PHASE -  Provides current number of outstanding Write address phases 
//    AXI4_CONFIG_NUM_OUTSTANDING_RD_PHASE -  Provides current number of outstanding Read address phases
//    AXI4_CONFIG_IS_ISSUING -  Enables Master component to use "config_max_outstanding_wr/config_max_outstanding_rd/config_max_outstanding_rw" variables for transaction issuing capability when set to true

typedef enum bit [7:0]
{
    AXI4_CONFIG_AXI4LITE_axi4           = 8'd0,
    AXI4_CONFIG_INTERFACE_TYPE               = 8'd1,
    AXI4_CONFIG_ENABLE_RLAST                 = 8'd2,
    AXI4_CONFIG_ENABLE_ID                    = 8'd3,
    AXI4_CONFIG_ENABLE_ALL_ASSERTIONS        = 8'd4,
    AXI4_CONFIG_ENABLE_ASSERTION             = 8'd5,
    AXI4_CONFIG_ENABLE_USER_SIDEBAND         = 8'd6,
    AXI4_CONFIG_AXI4LITE_TR_ID               = 8'd7,
    AXI4_CONFIG_ENABLE_SLAVE_EXCLUSIVE       = 8'd8,
    AXI4_CONFIG_READ_DATA_REORDERING_DEPTH   = 8'd9,
    AXI4_CONFIG_SLAVE_START_ADDR             = 8'd10,
    AXI4_CONFIG_SLAVE_END_ADDR               = 8'd11,
    AXI4_CONFIG_MAX_TRANSACTION_TIME_FACTOR  = 8'd12,
    AXI4_CONFIG_BURST_TIMEOUT_FACTOR         = 8'd13,
    AXI4_CONFIG_MAX_LATENCY_AWVALID_ASSERTION_TO_AWREADY = 8'd14,
    AXI4_CONFIG_MAX_LATENCY_ARVALID_ASSERTION_TO_ARREADY = 8'd15,
    AXI4_CONFIG_MAX_LATENCY_RVALID_ASSERTION_TO_RREADY = 8'd16,
    AXI4_CONFIG_MAX_LATENCY_BVALID_ASSERTION_TO_BREADY = 8'd17,
    AXI4_CONFIG_MAX_LATENCY_WVALID_ASSERTION_TO_WREADY = 8'd18,
    AXI4_CONFIG_ENABLE_QOS                   = 8'd19,
    AXI4_CONFIG_SETUP_TIME                   = 8'd20,
    AXI4_CONFIG_HOLD_TIME                    = 8'd21,
    AXI4_CONFIG_NUM_MAX_OUTSTANDING_READS    = 8'd22,
    AXI4_CONFIG_NUM_MAX_OUTSTANDING_WRITES   = 8'd23,
    AXI4_CONFIG_MAX_OUTSTANDING_WR           = 8'd24,
    AXI4_CONFIG_MAX_OUTSTANDING_RD           = 8'd25,
    AXI4_CONFIG_MAX_OUTSTANDING_RW           = 8'd26,
    AXI4_CONFIG_NUM_OUTSTANDING_WR_PHASE     = 8'd27,
    AXI4_CONFIG_NUM_OUTSTANDING_RD_PHASE     = 8'd28,
    AXI4_CONFIG_IS_ISSUING                   = 8'd29
} axi4_config_e;

// enum: axi4_vhd_if_e
//
// For VHDL use only
typedef enum int
{
    AXI4_VHD_SET_CONFIG                         = 32'd0,
    AXI4_VHD_GET_CONFIG                         = 32'd1,
    AXI4_VHD_CREATE_WRITE_TRANSACTION           = 32'd2,
    AXI4_VHD_CREATE_READ_TRANSACTION            = 32'd3,
    AXI4_VHD_SET_READ_OR_WRITE                  = 32'd4,
    AXI4_VHD_GET_READ_OR_WRITE                  = 32'd5,
    AXI4_VHD_SET_ADDR                           = 32'd6,
    AXI4_VHD_GET_ADDR                           = 32'd7,
    AXI4_VHD_SET_PROT                           = 32'd8,
    AXI4_VHD_GET_PROT                           = 32'd9,
    AXI4_VHD_SET_REGION                         = 32'd10,
    AXI4_VHD_GET_REGION                         = 32'd11,
    AXI4_VHD_SET_SIZE                           = 32'd12,
    AXI4_VHD_GET_SIZE                           = 32'd13,
    AXI4_VHD_SET_BURST                          = 32'd14,
    AXI4_VHD_GET_BURST                          = 32'd15,
    AXI4_VHD_SET_LOCK                           = 32'd16,
    AXI4_VHD_GET_LOCK                           = 32'd17,
    AXI4_VHD_SET_CACHE                          = 32'd18,
    AXI4_VHD_GET_CACHE                          = 32'd19,
    AXI4_VHD_SET_QOS                            = 32'd20,
    AXI4_VHD_GET_QOS                            = 32'd21,
    AXI4_VHD_SET_ID                             = 32'd22,
    AXI4_VHD_GET_ID                             = 32'd23,
    AXI4_VHD_SET_BURST_LENGTH                   = 32'd24,
    AXI4_VHD_GET_BURST_LENGTH                   = 32'd25,
    AXI4_VHD_SET_ADDR_USER                      = 32'd26,
    AXI4_VHD_GET_ADDR_USER                      = 32'd27,
    AXI4_VHD_SET_DATA_WORDS                     = 32'd28,
    AXI4_VHD_GET_DATA_WORDS                     = 32'd29,
    AXI4_VHD_SET_WRITE_STROBES                  = 32'd30,
    AXI4_VHD_GET_WRITE_STROBES                  = 32'd31,
    AXI4_VHD_SET_DATA_USER                      = 32'd32,
    AXI4_VHD_GET_DATA_USER                      = 32'd33,
    AXI4_VHD_SET_RESP                           = 32'd34,
    AXI4_VHD_GET_RESP                           = 32'd35,
    AXI4_VHD_SET_RESP_USER                      = 32'd36,
    AXI4_VHD_GET_RESP_USER                      = 32'd37,
    AXI4_VHD_SET_ADDRESS_VALID_DELAY            = 32'd38,
    AXI4_VHD_GET_ADDRESS_VALID_DELAY            = 32'd39,
    AXI4_VHD_SET_DATA_VALID_DELAY               = 32'd40,
    AXI4_VHD_GET_DATA_VALID_DELAY               = 32'd41,
    AXI4_VHD_SET_WRITE_RESPONSE_VALID_DELAY     = 32'd42,
    AXI4_VHD_GET_WRITE_RESPONSE_VALID_DELAY     = 32'd43,
    AXI4_VHD_SET_ADDRESS_READY_DELAY            = 32'd44,
    AXI4_VHD_GET_ADDRESS_READY_DELAY            = 32'd45,
    AXI4_VHD_SET_DATA_READY_DELAY               = 32'd46,
    AXI4_VHD_GET_DATA_READY_DELAY               = 32'd47,
    AXI4_VHD_SET_WRITE_RESPONSE_READY_DELAY     = 32'd48,
    AXI4_VHD_GET_WRITE_RESPONSE_READY_DELAY     = 32'd49,
    AXI4_VHD_SET_GEN_WRITE_STROBES              = 32'd50,
    AXI4_VHD_GET_GEN_WRITE_STROBES              = 32'd51,
    AXI4_VHD_SET_OPERATION_MODE                 = 32'd52,
    AXI4_VHD_GET_OPERATION_MODE                 = 32'd53,
    AXI4_VHD_SET_WRITE_DATA_MODE                = 32'd54,
    AXI4_VHD_GET_WRITE_DATA_MODE                = 32'd55,
    AXI4_VHD_SET_DATA_BEAT_DONE                 = 32'd56,
    AXI4_VHD_GET_DATA_BEAT_DONE                 = 32'd57,
    AXI4_VHD_SET_TRANSACTION_DONE               = 32'd58,
    AXI4_VHD_GET_TRANSACTION_DONE               = 32'd59,
    AXI4_VHD_EXECUTE_TRANSACTION                = 32'd60,
    AXI4_VHD_GET_RW_TRANSACTION                 = 32'd61,
    AXI4_VHD_EXECUTE_WRITE_DATA_BURST           = 32'd62,
    AXI4_VHD_GET_WRITE_DATA_BURST               = 32'd63,
    AXI4_VHD_EXECUTE_READ_ADDR_PHASE            = 32'd64,
    AXI4_VHD_GET_READ_ADDR_PHASE                = 32'd65,
    AXI4_VHD_EXECUTE_READ_DATA_BURST            = 32'd66,
    AXI4_VHD_GET_READ_DATA_BURST                = 32'd67,
    AXI4_VHD_EXECUTE_READ_DATA_PHASE            = 32'd68,
    AXI4_VHD_GET_READ_DATA_PHASE                = 32'd69,
    AXI4_VHD_EXECUTE_WRITE_ADDR_PHASE           = 32'd70,
    AXI4_VHD_GET_WRITE_ADDR_PHASE               = 32'd71,
    AXI4_VHD_EXECUTE_WRITE_DATA_PHASE           = 32'd72,
    AXI4_VHD_GET_WRITE_DATA_PHASE               = 32'd73,
    AXI4_VHD_EXECUTE_WRITE_RESPONSE_PHASE       = 32'd74,
    AXI4_VHD_GET_WRITE_RESPONSE_PHASE           = 32'd75,
    AXI4_VHD_GET_READ_ADDR_CYCLE                = 32'd76,
    AXI4_VHD_EXECUTE_READ_ADDR_READY            = 32'd77,
    AXI4_VHD_GET_READ_ADDR_READY                = 32'd78,
    AXI4_VHD_GET_READ_DATA_CYCLE                = 32'd79,
    AXI4_VHD_EXECUTE_READ_DATA_READY            = 32'd80,
    AXI4_VHD_GET_READ_DATA_READY                = 32'd81,
    AXI4_VHD_GET_WRITE_ADDR_CYCLE               = 32'd82,
    AXI4_VHD_EXECUTE_WRITE_ADDR_READY           = 32'd83,
    AXI4_VHD_GET_WRITE_ADDR_READY               = 32'd84,
    AXI4_VHD_GET_WRITE_DATA_CYCLE               = 32'd85,
    AXI4_VHD_EXECUTE_WRITE_DATA_READY           = 32'd86,
    AXI4_VHD_GET_WRITE_DATA_READY               = 32'd87,
    AXI4_VHD_GET_WRITE_RESPONSE_CYCLE           = 32'd88,
    AXI4_VHD_EXECUTE_WRITE_RESP_READY           = 32'd89,
    AXI4_VHD_GET_WRITE_RESP_READY               = 32'd90,
    AXI4_VHD_CREATE_MONITOR_TRANSACTION         = 32'd91,
    AXI4_VHD_CREATE_SLAVE_TRANSACTION           = 32'd92,
    AXI4_VHD_PUSH_TRANSACTION_ID                = 32'd93,
    AXI4_VHD_POP_TRANSACTION_ID                 = 32'd94,
    AXI4_VHD_GET_WRITE_ADDR_DATA                = 32'd95,
    AXI4_VHD_GET_READ_ADDR                      = 32'd96,
    AXI4_VHD_SET_READ_DATA                      = 32'd97,
    AXI4_VHD_PRINT                              = 32'd98,
    AXI4_VHD_DESTRUCT_TRANSACTION               = 32'd99,
    AXI4_VHD_WAIT_ON                            = 32'd100
} axi4_vhd_if_e;


typedef enum bit [7:0]
{
    AXI4_CLOCK_POSEDGE = 8'd0,
    AXI4_CLOCK_NEGEDGE = 8'd1,
    AXI4_CLOCK_ANYEDGE = 8'd2,
    AXI4_CLOCK_0_TO_1  = 8'd3,
    AXI4_CLOCK_1_TO_0  = 8'd4,
    AXI4_RESET_POSEDGE = 8'd5,
    AXI4_RESET_NEGEDGE = 8'd6,
    AXI4_RESET_ANYEDGE = 8'd7,
    AXI4_RESET_0_TO_1  = 8'd8,
    AXI4_RESET_1_TO_0  = 8'd9
} axi4_wait_e;

`ifndef MAX_AXI4_ADDRESS_WIDTH
  `define MAX_AXI4_ADDRESS_WIDTH 64
`endif

`ifndef MAX_AXI4_RDATA_WIDTH
  `define MAX_AXI4_RDATA_WIDTH 1024
`endif

`ifndef MAX_AXI4_WDATA_WIDTH
  `define MAX_AXI4_WDATA_WIDTH 1024
`endif

`ifndef MAX_AXI4_ID_WIDTH
  `define MAX_AXI4_ID_WIDTH 18
`endif

`ifndef MAX_AXI4_USER_WIDTH
  `define MAX_AXI4_USER_WIDTH 8
`endif

`ifndef MAX_AXI4_REGION_MAP_SIZE
  `define MAX_AXI4_REGION_MAP_SIZE 16
`endif

// enum: axi4_operation_mode_e
//
typedef enum int
{
    AXI4_TRANSACTION_NON_BLOCKING = 32'd0,
    AXI4_TRANSACTION_BLOCKING     = 32'd1
} axi4_operation_mode_e;

// enum: axi4_write_data_mode_e
//
typedef enum int
{
    AXI4_DATA_AFTER_ADDRESS = 32'd0,
    AXI4_DATA_WITH_ADDRESS  = 32'd1
} axi4_write_data_mode_e;

// Global Transaction Class
class axi4_transaction;
    // Protocol 
    axi4_rw_e read_or_write;
    bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0] addr;
    axi4_prot_e prot;
    bit [3:0] region;
    axi4_size_e size;
    axi4_burst_e burst;
    axi4_lock_e lock;
    axi4_cache_e cache;
    bit [3:0] qos;
    bit [((`MAX_AXI4_ID_WIDTH) - 1):0] id;
    bit [7:0] burst_length;
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] addr_user;
    bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0] data_words [];
    bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [];
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] data_user [];
    axi4_response_e resp[];
    bit [((`MAX_AXI4_USER_WIDTH) - 1):0] resp_user [];
    int address_valid_delay;
    int data_valid_delay[];
    int write_response_valid_delay;
    int address_ready_delay;
    int data_ready_delay[];
    int write_response_ready_delay;

    // Housekeeping
    bit gen_write_strobes = 1'b1;
    axi4_operation_mode_e  operation_mode  = AXI4_TRANSACTION_BLOCKING;
    axi4_write_data_mode_e write_data_mode = AXI4_DATA_WITH_ADDRESS;
    bit data_beat_done[];
    bit transaction_done;

    // This varaible is for printing component name and should not be visible/documented
    string driver_name;

    function void set_read_or_write( input axi4_rw_e lread_or_write );
      read_or_write = lread_or_write;
    endfunction

    function axi4_rw_e get_read_or_write();
      return read_or_write;
    endfunction

    function void set_addr( input bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0] laddr );
      addr = laddr;
    endfunction

    function bit [((`MAX_AXI4_ADDRESS_WIDTH) - 1):0]  get_addr();
      return addr;
    endfunction

    function void set_prot( input axi4_prot_e lprot );
      prot = lprot;
    endfunction

    function axi4_prot_e get_prot();
      return prot;
    endfunction

    function void set_region( input bit [3:0] lregion );
      region = lregion;
    endfunction

    function bit [3:0]  get_region();
      return region;
    endfunction

    function void set_size( input axi4_size_e lsize );
      size = lsize;
    endfunction

    function axi4_size_e get_size();
      return size;
    endfunction

    function void set_burst( input axi4_burst_e lburst );
      burst = lburst;
    endfunction

    function axi4_burst_e get_burst();
      return burst;
    endfunction

    function void set_lock( input axi4_lock_e llock );
      lock = llock;
    endfunction

    function axi4_lock_e get_lock();
      return lock;
    endfunction

    function void set_cache( input axi4_cache_e lcache );
      cache = lcache;
    endfunction

    function axi4_cache_e get_cache();
      return cache;
    endfunction

    function void set_qos( input bit [3:0] lqos );
      qos = lqos;
    endfunction

    function bit [3:0]  get_qos();
      return qos;
    endfunction

    function void set_id( input bit [((`MAX_AXI4_ID_WIDTH) - 1):0] lid );
      id = lid;
    endfunction

    function bit [((`MAX_AXI4_ID_WIDTH) - 1):0]  get_id();
      return id;
    endfunction

    function void set_burst_length( input bit [7:0] lburst_length );
      burst_length = lburst_length;
      data_words           = new[(lburst_length + 1)];
      write_strobes        = new[(lburst_length + 1)];
      data_user            = new[(lburst_length + 1)];
      resp                 = new[(lburst_length + 1)];
      resp_user            = new[(lburst_length + 1)];
      data_valid_delay     = new[(lburst_length + 1)];
      data_ready_delay     = new[(lburst_length + 1)];
      data_beat_done       = new[(lburst_length + 1)];
    endfunction

    function bit [7:0]  get_burst_length();
      return burst_length;
    endfunction

    function void set_addr_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] laddr_user );
      addr_user = laddr_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_addr_user();
      return addr_user;
    endfunction

    function void set_data_words( input bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0] ldata_words, input int index = 0 );
      data_words[index] = ldata_words;
    endfunction

    function bit [(((((`MAX_AXI4_RDATA_WIDTH > `MAX_AXI4_WDATA_WIDTH) ? `MAX_AXI4_RDATA_WIDTH : `MAX_AXI4_WDATA_WIDTH))) - 1):0]  get_data_words( input int index = 0 );
      return data_words[index];
    endfunction

    function void set_write_strobes( input bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0] lwrite_strobes, input int index = 0 );
      write_strobes[index] = lwrite_strobes;
    endfunction

    function bit [(((`MAX_AXI4_WDATA_WIDTH / 8)) - 1):0]  get_write_strobes( input int index = 0 );
      return write_strobes[index];
    endfunction

    function void set_data_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] ldata_user, input int index = 0 );
      data_user[index] = ldata_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_data_user( input int index = 0 );
      return data_user[index];
    endfunction

    function void set_resp( input axi4_response_e lresp, input int index = 0 );
      resp[index] = lresp;
    endfunction

    function axi4_response_e get_resp( input int index = 0 );
      return resp[index];
    endfunction

    function void set_resp_user( input bit [((`MAX_AXI4_USER_WIDTH) - 1):0] lresp_user, input int index = 0 );
      resp_user[index] = lresp_user;
    endfunction

    function bit [((`MAX_AXI4_USER_WIDTH) - 1):0]  get_resp_user( input int index = 0 );
      return resp_user[index];
    endfunction

    function void set_address_valid_delay( input int laddress_valid_delay );
      address_valid_delay = laddress_valid_delay;
    endfunction

    function int get_address_valid_delay();
      return address_valid_delay;
    endfunction

    function void set_data_valid_delay( input int ldata_valid_delay, input int index = 0 );
      data_valid_delay[index] = ldata_valid_delay;
    endfunction

    function int get_data_valid_delay( input int index = 0 );
      return data_valid_delay[index];
    endfunction

    function void set_write_response_valid_delay( input int lwrite_response_valid_delay );
      write_response_valid_delay = lwrite_response_valid_delay;
    endfunction

    function int get_write_response_valid_delay();
      return write_response_valid_delay;
    endfunction

    function void set_address_ready_delay( input int laddress_ready_delay );
      address_ready_delay = laddress_ready_delay;
    endfunction

    function int get_address_ready_delay();
      return address_ready_delay;
    endfunction

    function void set_data_ready_delay( input int ldata_ready_delay, input int index = 0 );
      data_ready_delay[index] = ldata_ready_delay;
    endfunction

    function int get_data_ready_delay( input int index = 0 );
      return data_ready_delay[index];
    endfunction

    function void set_write_response_ready_delay( input int lwrite_response_ready_delay );
      write_response_ready_delay = lwrite_response_ready_delay;
    endfunction

    function int get_write_response_ready_delay();
      return write_response_ready_delay;
    endfunction

    function void set_gen_write_strobes( input bit lgen_write_strobes);
      gen_write_strobes = lgen_write_strobes;
    endfunction

    function bit get_gen_write_strobes();
      return gen_write_strobes;
    endfunction

    function void set_operation_mode( input axi4_operation_mode_e loperation_mode );
      operation_mode = loperation_mode;
    endfunction

    function axi4_operation_mode_e get_operation_mode();
      return operation_mode;
    endfunction

    function void set_write_data_mode( input axi4_write_data_mode_e lwrite_data_mode );
      write_data_mode = lwrite_data_mode;
    endfunction

    function axi4_write_data_mode_e get_write_data_mode();
      return write_data_mode;
    endfunction

    function void set_data_beat_done( input int ldata_beat_done, input int index = 0 );
      data_beat_done[index] = ldata_beat_done;
    endfunction

    function int get_data_beat_done( input int index = 0 );
      return data_beat_done[index];
    endfunction

    function void set_transaction_done( input int ltransaction_done );
      transaction_done = ltransaction_done;
    endfunction

    function int get_transaction_done();
      return transaction_done;
    endfunction

    // Function: do_print
    //
    // Prints axi4_transaction transaction attributes
    function void print (bit print_delays = 1'b0);
      $display("------------------------------------------------------------------------");
      $display("%0t: %s axi4_transaction", $time, driver_name);
      $display("------------------------------------------------------------------------");
      $display("read_or_write : %s", read_or_write.name());
      $display("addr : 'h%h", addr);
      $display("prot : %s", prot.name());
      $display("region : 'h%h", region);
      $display("size : %s", size.name());
      $display("burst : %s", burst.name());
      $display("lock : %s", lock.name());
      $display("cache : %s", cache.name());
      $display("qos : 'h%h", qos);
      $display("id : 'h%h", id);
      $display("burst_length : 'h%h", burst_length);
      $display("addr_user : 'h%h", addr_user);
      foreach( data_words[i0_1] )
        $display("data_words[%0d] : 'h%h", i0_1, data_words[i0_1]);
      foreach( write_strobes[i0_1] )
        $display("write_strobes[%0d] : 'h%h", i0_1, write_strobes[i0_1]);
      foreach( data_user[i0_1] )
        $display("data_user[%0d] : 'h%h", i0_1, data_user[i0_1]);
      foreach( resp[i0_1] )
        $display("resp[%0d] : %s", i0_1, resp[i0_1].name());
      foreach( resp_user[i0_1] )
        $display("resp_user[%0d] : 'h%h", i0_1, resp_user[i0_1]);
      $display("gen_write_strobes : 'b%b", gen_write_strobes );
      $display("operation_mode   : %s", operation_mode.name() );
      $display("write_data_mode  : %s", write_data_mode.name() );
      foreach( data_beat_done[i0_1] )
        $display("data_beat_done[%0d] : 'b%b", i0_1, data_beat_done[i0_1] );
      $display("transaction_done : 'b%b", transaction_done );
      if ( print_delays == 1'b1 )
      begin
        $display("address_valid_delay : %0d", address_valid_delay);
        foreach( data_valid_delay[i0_1] )
          $display("data_valid_delay[%0d] : %0d", i0_1, data_valid_delay[i0_1]);
        $display("write_response_valid_delay : %0d", write_response_valid_delay);
        $display("address_ready_delay : %0d", address_ready_delay);
        foreach( data_ready_delay[i0_1] )
          $display("data_ready_delay[%0d] : %0d", i0_1, data_ready_delay[i0_1]);
        $display("write_response_ready_delay : %0d", write_response_ready_delay);
      end
    endfunction
endclass


//------------------------------------------------------------------------------
//
// Enum: axi4_call_back_e
//
//------------------------------------------------------------------------------
//
// The types of callback that can be registered with this <dvc_axi4> interface.
//
// AXI4_REPORTER_CB                         - Callback used to redirect BFM assertion and debug messages to the UVM messaging system
//
typedef enum
{
    AXI4_REPORTER_CB                         = 0
} axi4_call_back_e;
`endif
endpackage

import mgc_axi4_pkg::*;
`ifdef QUESTA
// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

`ifndef QVIP_MIX_AND_MATCH
(* cy_so="libaxi4_IN_SystemVerilog_MTI_full_DVC" *)
(* on_lib_load="axi4_IN_SystemVerilog_load" *)
`endif

interface mgc_common_axi4 #( int AXI4_ADDRESS_WIDTH = 64, int AXI4_RDATA_WIDTH = 1024, int AXI4_WDATA_WIDTH = 1024, int AXI4_ID_WIDTH = 18, int AXI4_USER_WIDTH = 8, int AXI4_REGION_MAP_SIZE = 16 )
    (input wire iACLK, input wire iARESETn);

import QUESTA_MVC::questa_mvc_reporter;
import QUESTA_MVC::questa_mvc_item_comms_semantic;
import QUESTA_MVC::questa_mvc_edge;
import QUESTA_MVC::QUESTA_MVC_POSEDGE;
import QUESTA_MVC::QUESTA_MVC_NEGEDGE;
import QUESTA_MVC::QUESTA_MVC_ANYEDGE;
import QUESTA_MVC::QUESTA_MVC_0_TO_1_EDGE;
import QUESTA_MVC::QUESTA_MVC_1_TO_0_EDGE;




    //-------------------------------------------------------------------------
    // Private wires
    //-------------------------------------------------------------------------
    wire ACLK;
    wire ARESETn;
    wire AWVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR;
    wire [2:0] AWPROT;
    wire [3:0] AWREGION;
    wire [7:0] AWLEN;
    wire [3:0] AWSIZE;
    wire [1:0] AWBURST;
    wire AWLOCK;
    wire [3:0] AWCACHE;
    wire [3:0] AWQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] AWID;
    wire [((AXI4_USER_WIDTH) - 1):0] AWUSER;
    wire AWREADY;
    wire ARVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR;
    wire [2:0] ARPROT;
    wire [3:0] ARREGION;
    wire [7:0] ARLEN;
    wire [3:0] ARSIZE;
    wire [1:0] ARBURST;
    wire ARLOCK;
    wire [3:0] ARCACHE;
    wire [3:0] ARQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] ARID;
    wire [((AXI4_USER_WIDTH) - 1):0] ARUSER;
    wire ARREADY;
    wire RVALID;
    wire [((AXI4_RDATA_WIDTH) - 1):0] RDATA;
    wire [1:0] RRESP;
    wire RLAST;
    wire [((AXI4_ID_WIDTH) - 1):0] RID;
    wire [((AXI4_USER_WIDTH) - 1):0] RUSER;
    wire RREADY;
    wire WVALID;
    wire [((AXI4_WDATA_WIDTH) - 1):0] WDATA;
    wire [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB;
    wire WLAST;
    wire [((AXI4_USER_WIDTH) - 1):0] WUSER;
    wire WREADY;
    wire BVALID;
    wire [1:0] BRESP;
    wire [((AXI4_ID_WIDTH) - 1):0] BID;
    wire [((AXI4_USER_WIDTH) - 1):0] BUSER;
    wire BREADY;



    // Propagate global signals onto interface wires
    assign ACLK = iACLK;
    assign ARESETn = iARESETn;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Enables
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_rlast
    //
    //  
    // Enables use of optional RLAST signal.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable optional RLAST signal"
    bit config_enable_rlast;

    // Variable: config_enable_id
    //
    //  
    // Enables use of optional AWID, BID, ARID, RID signals.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable ID signals"
    bit config_enable_id;

    // Variable: config_enable_all_assertions
    //
    // 
    // Enables all protocol assertions. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable all protocol assertions"
    bit config_enable_all_assertions;

    // Variable: config_enable_assertion
    //
    // 
    // Enables individual protocol assertion.
    // This variable controls whether specific assertion within QVIP (of type <axi4_assertion_e>) is enabled or disabled.
    // Individual assertion can be disabled as follows:-
    // //-----------------------------------------------------------------------
    // // < BFM interface>.config_enable_assertion[<name of assertion>] = 1'b0;
    // //-----------------------------------------------------------------------
    // 
    // For example, the assertion AXI4_READ_DATA_UNKN is disabled as follows:
    // <bfm>.config_enable_assertion[AXI4_READ_DATA_UNKN] =  1'b0; 
    // 
    // Here bfm is the AXI4 interface instance name for which the assertion to be disabled. 
    // 
    // Default: All assertions are enabled
    //   
    // 
    //
    // mentor configurator specification name "Enable individual protocol assertion" 
    bit [255:0] config_enable_assertion;

    // Variable: config_enable_user_sideband
    //
    // 
    // Enables user sideband signals. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable user sideband signals"
    bit config_enable_user_sideband;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Slave behavior control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_slave_exclusive
    //
    // 
    // Enables exclusive transactions support for slave.
    // If disabled, every exclusive read/write returns an OKAY response,
    // and exclusive write updates memory. 
    // 
    // Default: 1  
    // 
    //
    // mentor configurator specification name "Enable exclusive transaction support"
    bit config_enable_slave_exclusive;

    // Variable: config_read_data_reordering_depth
    //
    // 
    // Sets the maximum number of different read transaction addresses for which read 
    // data(response) can be sent in any order from slave. 
    // 
    // Default: 2 ** AXI4_ID_WIDTH
    // 
    //
    // mentor configurator specification name "Read data reordering depth"
    int unsigned config_read_data_reordering_depth;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Timeout control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_max_transaction_time_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) for any complete read or write transaction, which
    // includes time period for all individual phases of transaction. 
    // 
    // Default: 100000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout for complete Read/Write transaction"
    int unsigned config_max_transaction_time_factor;

    // Variable: config_burst_timeout_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) between individual phases of a transaction. 
    // 
    // Default: 10000 clock cycles 
    // 
    //
    // mentor configurator specification name "Burst timeout between individual phases of a transaction"
    int unsigned config_burst_timeout_factor;

    // Variable: config_max_latency_AWVALID_assertion_to_AWREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of AWVALID to assertion of AWREADY.
    // An error message AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID is generated if AWREADY is not asserted
    // after assertion of AWVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from AWVALID to AWREADY assertion"
    int unsigned config_max_latency_AWVALID_assertion_to_AWREADY;

    // Variable: config_max_latency_ARVALID_assertion_to_ARREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of ARVALID to assertion of ARREADY.
    // An error message AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID is generated if ARREADY is not asserted
    // after assertion of ARVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from ARVALID to ARREADY assertion"
    int unsigned config_max_latency_ARVALID_assertion_to_ARREADY;

    // Variable: config_max_latency_RVALID_assertion_to_RREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of RVALID to assertion of RREADY.
    // An error message AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID is generated if RREADY is not asserted
    // after assertion of RVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from RVALID to RREADY assertion"
    int unsigned config_max_latency_RVALID_assertion_to_RREADY;

    // Variable: config_max_latency_BVALID_assertion_to_BREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of BVALID to assertion of BREADY.
    // An error message AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID is generated if BREADY is not asserted
    // after assertion of BVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from BVALID to BREADY assertion"
    int unsigned config_max_latency_BVALID_assertion_to_BREADY;

    // Variable: config_max_latency_WVALID_assertion_to_WREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of WVALID to assertion of WREADY.
    // An error message AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID is generated if WREADY is not asserted
    // after assertion of WVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from WVALID to WREADY assertion"
    int unsigned config_max_latency_WVALID_assertion_to_WREADY;

    // 
    // //-----------------------------------------------------------------------------
    // Group: QOS and Region control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_qos
    //
    //  
    // Enables quality-of-service support. 
    // 
    // Default: 1 
    // 
    //
    // mentor configurator specification name "Enable quality of service support"
    bit config_enable_qos;

    // Variable: config_setup_time
    //
    // 
    // Sets number of simulation time units from the setup time to the active 
    // clock edge of ACLK. The setup time will always be less than the time period
    // of the clock. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_setup_time;

    // Variable: config_hold_time
    //
    // 
    // Sets number of simulation time units from the hold time to the active 
    // clock edge of ACLK. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_hold_time;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Master Outstanding Control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_num_max_outstanding_reads
    //
    // 
    // Configures maximum number of read outstanding transfers allowed on the bus.
    // 
    // Default: -1
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding reads"
    int config_num_max_outstanding_reads;

    // Variable: config_num_max_outstanding_writes
    //
    //                                                                           
    // Configures maximum number of write outstanding transfers allowed on the bus. 
    //                                                                              
    // Default: -1                                                                 
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding writes"
    int config_num_max_outstanding_writes;

    // Variable: config_max_outstanding_wr
    //
    // Configures maximum possible outstanding Write transactions
    //
    int config_max_outstanding_wr;

    // Variable: config_max_outstanding_rd
    //
    // Configures maximum possible outstanding Read transactions
    //
    int config_max_outstanding_rd;

    // Variable: config_max_outstanding_rw
    //
    // Configures maximum possible outstanding Combined (Read and Write) transactions
    //
    int config_max_outstanding_rw;

    // Variable: config_num_outstanding_wr_phase
    //
    // Provides current number of outstanding Write address phases 
    //
    int config_num_outstanding_wr_phase;

    // Variable: config_num_outstanding_rd_phase
    //
    // Provides current number of outstanding Read address phases
    //
    int config_num_outstanding_rd_phase;

    // Variable: config_is_issuing
    //
    // Enables Master component to use "config_max_outstanding_wr/config_max_outstanding_rd/config_max_outstanding_rw" variables for transaction issuing capability when set to true
    //
    bit config_is_issuing;


    //-------------------------------------------------------------------------
    // Deprecated variables - writing to these variables will cause a warning to be issued.
    //-------------------------------------------------------------------------
    bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr;


    //-------------------------------------------------------------------------
    // Private variables
    //-------------------------------------------------------------------------
    bit config_axi4lite_interface;
    axi4_interface_type_e config_interface_type;


    //-------------------------------------------------------------------------
    // Initialise local variables
    //-------------------------------------------------------------------------
    initial
    begin
         config_axi4lite_tr_id = 0;
    end

    //------------------------------------------------------------------------------
    // Group:- Interface ends
    //------------------------------------------------------------------------------
    //
    longint axi4_master_end;

    // Function:- get_axi4_master_end
    //
    // Returns a handle to the <master> end of this instance of the <axi4> interface.

    function longint get_axi4_master_end();
        return axi4_master_end;
    endfunction

    longint axi4_slave_end;

    // Function:- get_axi4_slave_end
    //
    // Returns a handle to the <slave> end of this instance of the <axi4> interface.

    function longint get_axi4_slave_end();
        return axi4_slave_end;
    endfunction

    longint axi4__monitor_end;

    // Function:- get_axi4__monitor_end
    //
    // Returns a handle to the <_monitor> end of this instance of the <axi4> interface.

    function longint get_axi4__monitor_end();
        return axi4__monitor_end;
    endfunction


    // Group:- Abstraction Levels
    // 
    // These functions are used set or get the abstraction levels of an interface end.
    // See <Abstraction Levels of Interface Ends> for more details on the meaning of
    // TLM or WLM connected and the valid combinations.


    //-------------------------------------------------------------------------
    // Function:- axi4_set_master_abstraction_level
    //
    //     Function to set whether the <master> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_master_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_master_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <master> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_master_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_set_slave_abstraction_level
    //
    //     Function to set whether the <slave> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_slave_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_slave_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <slave> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_slave_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    import "DPI-C" context function longint dvc_axi4_initialise_SystemVerilog
    (
        int     usage_code,
        string  iface_version,
        longint generate_ver,
        int     qvip_mix_and_match,
        output longint master_end,
        output longint slave_end,
        output longint _monitor_end,
        input int AXI4_ADDRESS_WIDTH,
        input int AXI4_RDATA_WIDTH,
        input int AXI4_WDATA_WIDTH,
        input int AXI4_ID_WIDTH,
        input int AXI4_USER_WIDTH,
        input int AXI4_REGION_MAP_SIZE
    );

    `ifndef MVC_axi4_VERSION
    `define MVC_axi4_VERSION ""
    `endif

    // Handle to the linkage
    (* elab_init *) longint _interface_ref =
                                dvc_axi4_initialise_SystemVerilog
                                (
                                    18102076,
                                    `MVC_axi4_VERSION,
                                    20220701,
                                    `ifdef QVIP_MIX_AND_MATCH
                                    1
                                    `else
                                    0
                                    `endif
                                    ,
                                    axi4_master_end,
                                    axi4_slave_end,
                                    axi4__monitor_end,
                                    AXI4_ADDRESS_WIDTH,
                                    AXI4_RDATA_WIDTH,
                                    AXI4_WDATA_WIDTH,
                                    AXI4_ID_WIDTH,
                                    AXI4_USER_WIDTH,
                                    AXI4_REGION_MAP_SIZE
                                ); // DPI call to create transactor (called at
                                     // elaboration time as initialiser)

    questa_mvc_reporter endPoint[longint];
    export "DPI-C" dvc_axi4_process_reports = function process_reports;
    function void process_reports( input longint recipient, input string category, input string objectName, input string instanceName, input string error_no, input string severity, input string mess );
        if( endPoint.exists(recipient) )
            endPoint[recipient].report_message( category, "dvc_axi4", 0, objectName, instanceName, error_no, severity, mess );
        else
            $error("Invalid recipient (%d) when processing report", recipient);
    endfunction

    import "DPI-C" context dvc_axi4_register_end_point = function void axi4_register_end_point( input longint iface_ref, input longint as_end, input string name );

    // A function for registering a reporter to capture any reports coming from as_end
    function automatic void register_end_point( input longint as_end, input questa_mvc_reporter rep = null );
        if ( rep != null )
        begin
            if ( ( rep.name == "" ) || ( rep.name == "NULL" ) )
            begin
                $display("Error: %m: Reporter passed to register_end_point has a reserved name. Neither an empty string nor the string 'NULL' can be used.");
            end
            else
            begin
                axi4_register_end_point( _interface_ref, as_end, rep.name );
                endPoint[as_end] = rep;
            end
        end
        else
        begin
            axi4_register_end_point( _interface_ref, as_end, "NULL" );
            endPoint.delete( as_end );
        end
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group:- Registering Reports
    //
    //
    // The following methods are used to register a custom reporting object as
    // described in the Questa Verification IP base library section, <Customizing Error-Reporting>.
    // 
    //-------------------------------------------------------------------------

    function void register_interface_reporter( input questa_mvc_reporter _rep = null );
        register_end_point( _interface_ref, _rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_master_reporter
    //
    // Function used to register a reporter for the <master> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the master end.
    //
    function void register_master_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_master_end, rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_slave_reporter
    //
    // Function used to register a reporter for the <slave> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the slave end.
    //
    function void register_slave_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_slave_end, rep );
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_mvc_reporter
    //
    // Function used to get the handle for an already registered reporter.
    // By default returns the reporter associated with this interface. If an end handle is passed,
    // then the reporter for that end.
    //
    // Arguments:
    //    as_end - Optional, a handle for an end of this interface.
    //
    function questa_mvc_reporter get_mvc_reporter
    (
        input longint as_end = 0
    );
        if ( as_end == 0 )
            as_end = _interface_ref;
        if ( endPoint.exists( as_end ) )
            return endPoint[ as_end ];
        else
            return null;
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group: BFM Utility/Convenience Methods
    //
    // This is the group of utility functions provided by the QVIP BFM to
    // communicate from the SV world to the QVIP BFM.
    // This set of APIs can be used to either get status/statistics
    // information from the BFM or to set values in a particular database 
    // in the BFM. Please refer to individual functions for more information.
    //
    //-------------------------------------------------------------------------

    function automatic void fn_get_raddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output int num_beats_rcvd
    );
         fn_get_raddr_phase_info_C(valid,id,addr,num_beats_rcvd);
    endfunction

    function automatic void fn_get_waddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr
    );
         fn_get_waddr_phase_info_C(valid,id,addr);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_add_addr_map_entry
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets address map fields inside internal BFM. 
    // 
    // Inputs:
    // region - Region name 
    // addr - Start address of region
    // size - Size of address region
    function automatic void fn_add_addr_map_entry
    (
        input string region,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input longint unsigned size
    );
         fn_add_addr_map_entry_C(region,addr,size);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_add_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets delay values for write address-data pair and 
    // between write data beats, initiated by master end. 
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_add_wr_delay
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        const ref int unsigned data2data[]
    );
         fn_add_wr_delay_C(region,id,addr2data,data2data);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_delete_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function deletes delay values for a region-id pair
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_delete_wr_delay
    (
        input string region,
        input bit [17:0] id
    );
         fn_delete_wr_delay_C(region,id);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_drive_signals
    //-------------------------------------------------------------------------
    // A function to drive the required value by the user on the desired signal when VALID is 0. 
    //   It takes the name of the signal and value to be driven on that signal as an input:
    // 
    //   Note - In case, the user wants to retain the same value on the signal (irrespective of Valid signal), 
    //   the value needs to be passed as -1.
    function automatic void fn_drive_signals
    (
        input string signal,
        input longint value
    );
         fn_drive_signals_C(signal,value);
    endfunction

    function automatic void fn_set_address_map_entry
    (
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] start_addr,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] end_addr
    );
         fn_set_address_map_entry_C(start_addr,end_addr);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_set_wr_def_delays
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets default delay values for write address-data pair and data beats 
    // initiated by master end between input max. and min. values. In case non-randomized 
    // default value is desired, user can set max. and min. to same value.
    // 
    // Inputs:
    // min_addr2data - Minimum value of default delays to be inserted between write address and data phase.
    // min_data2data - Minimum value of default delays to be inserted between data beats
    // max_addr2data - Maximum value of default delays to be inserted between write address and data phase.
    // max_data2data - Maximum value of default delays to be inserted between data beats
    function automatic void fn_set_wr_def_delays
    (
        input int unsigned min_addr2data,
        const ref int unsigned min_data2data[],
        input int unsigned max_addr2data,
        const ref int unsigned max_data2data[]
    );
         fn_set_wr_def_delays_C(min_addr2data,min_data2data,max_addr2data,max_data2data);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_print_perf_stats
    //-------------------------------------------------------------------------
    // API used to display the performance statistics of the current simulation. This
    // API shall be called only if axi4_vip_config variable 'cfg.agent_cfg.en_perf_stats.enable' has been configured to '1'.
    // 
    // Every statistics has 4 basic units:
    // minimum - Tells about the minimum metric within sliding window  
    // maximum - Tells about the maximum metric within sliding window 
    // mean    - Tells about the mean metric within a sliding window
    // sw      - Stores the calculation of the metrics over sliding window. 
    //           Statistics are collected for each of the sliding window segments 
    //           individually and then the actual statistic is calculated based 
    //           on all of the segments
    // 
    // The statistics are calculated for basic categories -
    // 
    // bus_usage: Stores the bus occupancy(as a percentage out of 100) for
    //   - Read transactions
    //   - Write transactions
    //   - Read Data burst
    //   - Write Data burst
    // See also, <axi4_bus_usage_s>
    // 
    // bandwidth: Stores the bandwidth value for
    //   - Read transactions
    //   - Write transactions
    // See also, <axi4_bandwidth_s>
    // 
    // latency: Stores the interphase delays as per the following categories
    //   - Write address to next write address delay
    //   - Write address to write data delay
    //   - Write data to write response delay
    //   - Read address to next read address delay
    //   - Read address to read data delay
    // See also, <axi4_latency_s>
    // 
    // waits: Stores the VALID-READY waits that are involved in each of the phases
    //   - Write address
    //   - Write data
    //   - Write response
    //   - Read address
    //   - Read data
    // See also, <axi4_waits_s>
    function automatic void fn_print_perf_stats
    (
    );
         fn_print_perf_stats_C();
    endfunction

    // Declare user visible wires variables, for non-continuous assignments.
    logic m_ACLK = 'z;
    logic m_ARESETn = 'z;
    logic m_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_AWADDR = 'z;
    logic [2:0] m_AWPROT = 'z;
    logic [3:0] m_AWREGION = 'z;
    logic [7:0] m_AWLEN = 'z;
    logic [3:0] m_AWSIZE = 'z;
    logic [1:0] m_AWBURST = 'z;
    logic m_AWLOCK = 'z;
    logic [3:0] m_AWCACHE = 'z;
    logic [3:0] m_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_AWUSER = 'z;
    logic m_AWREADY = 'z;
    logic m_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_ARADDR = 'z;
    logic [2:0] m_ARPROT = 'z;
    logic [3:0] m_ARREGION = 'z;
    logic [7:0] m_ARLEN = 'z;
    logic [3:0] m_ARSIZE = 'z;
    logic [1:0] m_ARBURST = 'z;
    logic m_ARLOCK = 'z;
    logic [3:0] m_ARCACHE = 'z;
    logic [3:0] m_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_ARUSER = 'z;
    logic m_ARREADY = 'z;
    logic m_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] m_RDATA = 'z;
    logic [1:0] m_RRESP = 'z;
    logic m_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_RUSER = 'z;
    logic m_RREADY = 'z;
    logic m_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] m_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] m_WSTRB = 'z;
    logic m_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_WUSER = 'z;
    logic m_WREADY = 'z;
    logic m_BVALID = 'z;
    logic [1:0] m_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_BUSER = 'z;
    logic m_BREADY = 'z;

    // Forces a sweep through the wire change checkers at time 0 to get around
    // process kick-off order unknowns
    bit _check_t0_values;
    always_comb _check_t0_values = 1;

    // handle control
    longint last_start_time = 0;

    longint last_end_time = 0;

    export "DPI-C" dvc_axi4_set_start_end_times = function set_start_end_times;

    function void set_start_end_times(longint _start, longint _end);
        last_start_time = _start;
        last_end_time = _end;
    endfunction


    function longint get_last_handle();
        return -1;
    endfunction


    function longint get_last_start_time();
        return last_start_time;
    endfunction


    function longint get_last_end_time();
        return last_end_time;
    endfunction


    //-------------------------------------------------------------------------
    // Tasks to wait for a number of specified edges on a wire
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_ACLK
    //     Wait for the specified change on wire <axi4::ACLK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ACLK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ACLK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ACLK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ACLK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ACLK === 0 );
                    @( ACLK );
                end
                while ( ACLK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ACLK === 1 );
                    @( ACLK );
                end
                while ( ACLK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARESETn
    //     Wait for the specified change on wire <axi4::ARESETn>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARESETn( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARESETn);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARESETn);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARESETn);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 0 );
                    @( ARESETn );
                end
                while ( ARESETn !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 1 );
                    @( ARESETn );
                end
                while ( ARESETn !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWVALID
    //     Wait for the specified change on wire <axi4::AWVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 0 );
                    @( AWVALID );
                end
                while ( AWVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 1 );
                    @( AWVALID );
                end
                while ( AWVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 0 );
                    @( AWADDR );
                end
                while ( AWADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 1 );
                    @( AWADDR );
                end
                while ( AWADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR_index1
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 0 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 1 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 0 );
                    @( AWPROT );
                end
                while ( AWPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 1 );
                    @( AWPROT );
                end
                while ( AWPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT_index1
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 0 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 1 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 0 );
                    @( AWREGION );
                end
                while ( AWREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 1 );
                    @( AWREGION );
                end
                while ( AWREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION_index1
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 0 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 1 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 0 );
                    @( AWLEN );
                end
                while ( AWLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 1 );
                    @( AWLEN );
                end
                while ( AWLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN_index1
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 0 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 1 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 0 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 1 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE_index1
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 0 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 1 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 0 );
                    @( AWBURST );
                end
                while ( AWBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 1 );
                    @( AWBURST );
                end
                while ( AWBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST_index1
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 0 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 1 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLOCK
    //     Wait for the specified change on wire <axi4::AWLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 0 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 1 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 0 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 1 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE_index1
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 0 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 1 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 0 );
                    @( AWQOS );
                end
                while ( AWQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 1 );
                    @( AWQOS );
                end
                while ( AWQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS_index1
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 0 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 1 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID === 0 );
                    @( AWID );
                end
                while ( AWID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID === 1 );
                    @( AWID );
                end
                while ( AWID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID_index1
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 0 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 1 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 0 );
                    @( AWUSER );
                end
                while ( AWUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 1 );
                    @( AWUSER );
                end
                while ( AWUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER_index1
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 0 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 1 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREADY
    //     Wait for the specified change on wire <axi4::AWREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 0 );
                    @( AWREADY );
                end
                while ( AWREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 1 );
                    @( AWREADY );
                end
                while ( AWREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARVALID
    //     Wait for the specified change on wire <axi4::ARVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 0 );
                    @( ARVALID );
                end
                while ( ARVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 1 );
                    @( ARVALID );
                end
                while ( ARVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 0 );
                    @( ARADDR );
                end
                while ( ARADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 1 );
                    @( ARADDR );
                end
                while ( ARADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR_index1
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 0 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 1 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 0 );
                    @( ARPROT );
                end
                while ( ARPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 1 );
                    @( ARPROT );
                end
                while ( ARPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT_index1
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 0 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 1 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 0 );
                    @( ARREGION );
                end
                while ( ARREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 1 );
                    @( ARREGION );
                end
                while ( ARREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION_index1
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 0 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 1 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 0 );
                    @( ARLEN );
                end
                while ( ARLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 1 );
                    @( ARLEN );
                end
                while ( ARLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN_index1
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 0 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 1 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 0 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 1 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE_index1
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 0 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 1 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 0 );
                    @( ARBURST );
                end
                while ( ARBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 1 );
                    @( ARBURST );
                end
                while ( ARBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST_index1
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 0 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 1 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLOCK
    //     Wait for the specified change on wire <axi4::ARLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 0 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 1 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 0 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 1 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE_index1
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 0 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 1 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 0 );
                    @( ARQOS );
                end
                while ( ARQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 1 );
                    @( ARQOS );
                end
                while ( ARQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS_index1
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 0 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 1 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID === 0 );
                    @( ARID );
                end
                while ( ARID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID === 1 );
                    @( ARID );
                end
                while ( ARID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID_index1
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 0 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 1 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 0 );
                    @( ARUSER );
                end
                while ( ARUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 1 );
                    @( ARUSER );
                end
                while ( ARUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER_index1
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 0 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 1 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREADY
    //     Wait for the specified change on wire <axi4::ARREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 0 );
                    @( ARREADY );
                end
                while ( ARREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 1 );
                    @( ARREADY );
                end
                while ( ARREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RVALID
    //     Wait for the specified change on wire <axi4::RVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RVALID === 0 );
                    @( RVALID );
                end
                while ( RVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RVALID === 1 );
                    @( RVALID );
                end
                while ( RVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA === 0 );
                    @( RDATA );
                end
                while ( RDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA === 1 );
                    @( RDATA );
                end
                while ( RDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA_index1
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 0 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 1 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP === 0 );
                    @( RRESP );
                end
                while ( RRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP === 1 );
                    @( RRESP );
                end
                while ( RRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP_index1
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 0 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 1 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RLAST
    //     Wait for the specified change on wire <axi4::RLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RLAST === 0 );
                    @( RLAST );
                end
                while ( RLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RLAST === 1 );
                    @( RLAST );
                end
                while ( RLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID === 0 );
                    @( RID );
                end
                while ( RID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID === 1 );
                    @( RID );
                end
                while ( RID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID_index1
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 0 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 1 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER === 0 );
                    @( RUSER );
                end
                while ( RUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER === 1 );
                    @( RUSER );
                end
                while ( RUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER_index1
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 0 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 1 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RREADY
    //     Wait for the specified change on wire <axi4::RREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RREADY === 0 );
                    @( RREADY );
                end
                while ( RREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RREADY === 1 );
                    @( RREADY );
                end
                while ( RREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WVALID
    //     Wait for the specified change on wire <axi4::WVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WVALID === 0 );
                    @( WVALID );
                end
                while ( WVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WVALID === 1 );
                    @( WVALID );
                end
                while ( WVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA === 0 );
                    @( WDATA );
                end
                while ( WDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA === 1 );
                    @( WDATA );
                end
                while ( WDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA_index1
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 0 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 1 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 0 );
                    @( WSTRB );
                end
                while ( WSTRB !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 1 );
                    @( WSTRB );
                end
                while ( WSTRB !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB_index1
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 0 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 1 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WLAST
    //     Wait for the specified change on wire <axi4::WLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WLAST === 0 );
                    @( WLAST );
                end
                while ( WLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WLAST === 1 );
                    @( WLAST );
                end
                while ( WLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER === 0 );
                    @( WUSER );
                end
                while ( WUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER === 1 );
                    @( WUSER );
                end
                while ( WUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER_index1
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 0 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 1 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WREADY
    //     Wait for the specified change on wire <axi4::WREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WREADY === 0 );
                    @( WREADY );
                end
                while ( WREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WREADY === 1 );
                    @( WREADY );
                end
                while ( WREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BVALID
    //     Wait for the specified change on wire <axi4::BVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BVALID === 0 );
                    @( BVALID );
                end
                while ( BVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BVALID === 1 );
                    @( BVALID );
                end
                while ( BVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP === 0 );
                    @( BRESP );
                end
                while ( BRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP === 1 );
                    @( BRESP );
                end
                while ( BRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP_index1
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 0 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 1 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID === 0 );
                    @( BID );
                end
                while ( BID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID === 1 );
                    @( BID );
                end
                while ( BID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID_index1
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 0 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 1 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER === 0 );
                    @( BUSER );
                end
                while ( BUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER === 1 );
                    @( BUSER );
                end
                while ( BUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER_index1
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 0 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 1 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BREADY
    //     Wait for the specified change on wire <axi4::BREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BREADY === 0 );
                    @( BREADY );
                end
                while ( BREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BREADY === 1 );
                    @( BREADY );
                end
                while ( BREADY !== 0 );
            end
        end
    endtask

    //-------------------------------------------------------------------------
    // Tasks/functions to set/get wires
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_ACLK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ACLK>.
    //
    // Parameters:
    //     ACLK_param - The value to set onto wire <ACLK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ACLK( logic ACLK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ACLK = ACLK_param;
        else
            m_ACLK <= ACLK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ACLK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ACLK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ACLK>.
    //
    function automatic logic get_ACLK(  );
        return ACLK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARESETn
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARESETn>.
    //
    // Parameters:
    //     ARESETn_param - The value to set onto wire <ARESETn>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARESETn( logic ARESETn_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARESETn = ARESETn_param;
        else
            m_ARESETn <= ARESETn_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARESETn
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARESETn>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARESETn>.
    //
    function automatic logic get_ARESETn(  );
        return ARESETn;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWVALID>.
    //
    // Parameters:
    //     AWVALID_param - The value to set onto wire <AWVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWVALID( logic AWVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWVALID = AWVALID_param;
        else
            m_AWVALID <= AWVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWVALID>.
    //
    function automatic logic get_AWVALID(  );
        return AWVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWADDR>.
    //
    // Parameters:
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR = AWADDR_param;
        else
            m_AWADDR <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR_index1( int _this_dot_1, logic  AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR[_this_dot_1] = AWADDR_param;
        else
            m_AWADDR[_this_dot_1] <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_AWADDR(  );
        return AWADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic   get_AWADDR_index1( int _this_dot_1 );
        return AWADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWPROT>.
    //
    // Parameters:
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT( logic [2:0] AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT = AWPROT_param;
        else
            m_AWPROT <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT_index1( int _this_dot_1, logic  AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT[_this_dot_1] = AWPROT_param;
        else
            m_AWPROT[_this_dot_1] <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic [2:0]  get_AWPROT(  );
        return AWPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic   get_AWPROT_index1( int _this_dot_1 );
        return AWPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREGION>.
    //
    // Parameters:
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION( logic [3:0] AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION = AWREGION_param;
        else
            m_AWREGION <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION_index1( int _this_dot_1, logic  AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION[_this_dot_1] = AWREGION_param;
        else
            m_AWREGION[_this_dot_1] <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic [3:0]  get_AWREGION(  );
        return AWREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic   get_AWREGION_index1( int _this_dot_1 );
        return AWREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLEN>.
    //
    // Parameters:
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN( logic [7:0] AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN = AWLEN_param;
        else
            m_AWLEN <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN_index1( int _this_dot_1, logic  AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN[_this_dot_1] = AWLEN_param;
        else
            m_AWLEN[_this_dot_1] <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic [7:0]  get_AWLEN(  );
        return AWLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic   get_AWLEN_index1( int _this_dot_1 );
        return AWLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWSIZE>.
    //
    // Parameters:
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE( logic [3:0] AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE = AWSIZE_param;
        else
            m_AWSIZE <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE_index1( int _this_dot_1, logic  AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE[_this_dot_1] = AWSIZE_param;
        else
            m_AWSIZE[_this_dot_1] <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic [3:0]  get_AWSIZE(  );
        return AWSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic   get_AWSIZE_index1( int _this_dot_1 );
        return AWSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWBURST>.
    //
    // Parameters:
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST( logic [1:0] AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST = AWBURST_param;
        else
            m_AWBURST <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST_index1( int _this_dot_1, logic  AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST[_this_dot_1] = AWBURST_param;
        else
            m_AWBURST[_this_dot_1] <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic [1:0]  get_AWBURST(  );
        return AWBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic   get_AWBURST_index1( int _this_dot_1 );
        return AWBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLOCK>.
    //
    // Parameters:
    //     AWLOCK_param - The value to set onto wire <AWLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLOCK( logic AWLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLOCK = AWLOCK_param;
        else
            m_AWLOCK <= AWLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLOCK>.
    //
    function automatic logic get_AWLOCK(  );
        return AWLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWCACHE>.
    //
    // Parameters:
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE( logic [3:0] AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE = AWCACHE_param;
        else
            m_AWCACHE <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE_index1( int _this_dot_1, logic  AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE[_this_dot_1] = AWCACHE_param;
        else
            m_AWCACHE[_this_dot_1] <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic [3:0]  get_AWCACHE(  );
        return AWCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic   get_AWCACHE_index1( int _this_dot_1 );
        return AWCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWQOS>.
    //
    // Parameters:
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS( logic [3:0] AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS = AWQOS_param;
        else
            m_AWQOS <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS_index1( int _this_dot_1, logic  AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS[_this_dot_1] = AWQOS_param;
        else
            m_AWQOS[_this_dot_1] <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic [3:0]  get_AWQOS(  );
        return AWQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic   get_AWQOS_index1( int _this_dot_1 );
        return AWQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWID>.
    //
    // Parameters:
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID( logic [((AXI4_ID_WIDTH) - 1):0] AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID = AWID_param;
        else
            m_AWID <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID_index1( int _this_dot_1, logic  AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID[_this_dot_1] = AWID_param;
        else
            m_AWID[_this_dot_1] <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_AWID(  );
        return AWID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic   get_AWID_index1( int _this_dot_1 );
        return AWID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWUSER>.
    //
    // Parameters:
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER( logic [((AXI4_USER_WIDTH) - 1):0] AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER = AWUSER_param;
        else
            m_AWUSER <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER_index1( int _this_dot_1, logic  AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER[_this_dot_1] = AWUSER_param;
        else
            m_AWUSER[_this_dot_1] <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_AWUSER(  );
        return AWUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic   get_AWUSER_index1( int _this_dot_1 );
        return AWUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREADY>.
    //
    // Parameters:
    //     AWREADY_param - The value to set onto wire <AWREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREADY( logic AWREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREADY = AWREADY_param;
        else
            m_AWREADY <= AWREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREADY>.
    //
    function automatic logic get_AWREADY(  );
        return AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARVALID>.
    //
    // Parameters:
    //     ARVALID_param - The value to set onto wire <ARVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARVALID( logic ARVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARVALID = ARVALID_param;
        else
            m_ARVALID <= ARVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARVALID>.
    //
    function automatic logic get_ARVALID(  );
        return ARVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARADDR>.
    //
    // Parameters:
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR = ARADDR_param;
        else
            m_ARADDR <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR_index1( int _this_dot_1, logic  ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR[_this_dot_1] = ARADDR_param;
        else
            m_ARADDR[_this_dot_1] <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_ARADDR(  );
        return ARADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic   get_ARADDR_index1( int _this_dot_1 );
        return ARADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARPROT>.
    //
    // Parameters:
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT( logic [2:0] ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT = ARPROT_param;
        else
            m_ARPROT <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT_index1( int _this_dot_1, logic  ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT[_this_dot_1] = ARPROT_param;
        else
            m_ARPROT[_this_dot_1] <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic [2:0]  get_ARPROT(  );
        return ARPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic   get_ARPROT_index1( int _this_dot_1 );
        return ARPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREGION>.
    //
    // Parameters:
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION( logic [3:0] ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION = ARREGION_param;
        else
            m_ARREGION <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION_index1( int _this_dot_1, logic  ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION[_this_dot_1] = ARREGION_param;
        else
            m_ARREGION[_this_dot_1] <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic [3:0]  get_ARREGION(  );
        return ARREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic   get_ARREGION_index1( int _this_dot_1 );
        return ARREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLEN>.
    //
    // Parameters:
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN( logic [7:0] ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN = ARLEN_param;
        else
            m_ARLEN <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN_index1( int _this_dot_1, logic  ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN[_this_dot_1] = ARLEN_param;
        else
            m_ARLEN[_this_dot_1] <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic [7:0]  get_ARLEN(  );
        return ARLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic   get_ARLEN_index1( int _this_dot_1 );
        return ARLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARSIZE>.
    //
    // Parameters:
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE( logic [3:0] ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE = ARSIZE_param;
        else
            m_ARSIZE <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE_index1( int _this_dot_1, logic  ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE[_this_dot_1] = ARSIZE_param;
        else
            m_ARSIZE[_this_dot_1] <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic [3:0]  get_ARSIZE(  );
        return ARSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic   get_ARSIZE_index1( int _this_dot_1 );
        return ARSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARBURST>.
    //
    // Parameters:
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST( logic [1:0] ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST = ARBURST_param;
        else
            m_ARBURST <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST_index1( int _this_dot_1, logic  ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST[_this_dot_1] = ARBURST_param;
        else
            m_ARBURST[_this_dot_1] <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic [1:0]  get_ARBURST(  );
        return ARBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic   get_ARBURST_index1( int _this_dot_1 );
        return ARBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLOCK>.
    //
    // Parameters:
    //     ARLOCK_param - The value to set onto wire <ARLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLOCK( logic ARLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLOCK = ARLOCK_param;
        else
            m_ARLOCK <= ARLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLOCK>.
    //
    function automatic logic get_ARLOCK(  );
        return ARLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARCACHE>.
    //
    // Parameters:
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE( logic [3:0] ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE = ARCACHE_param;
        else
            m_ARCACHE <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE_index1( int _this_dot_1, logic  ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE[_this_dot_1] = ARCACHE_param;
        else
            m_ARCACHE[_this_dot_1] <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic [3:0]  get_ARCACHE(  );
        return ARCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic   get_ARCACHE_index1( int _this_dot_1 );
        return ARCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARQOS>.
    //
    // Parameters:
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS( logic [3:0] ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS = ARQOS_param;
        else
            m_ARQOS <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS_index1( int _this_dot_1, logic  ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS[_this_dot_1] = ARQOS_param;
        else
            m_ARQOS[_this_dot_1] <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic [3:0]  get_ARQOS(  );
        return ARQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic   get_ARQOS_index1( int _this_dot_1 );
        return ARQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARID>.
    //
    // Parameters:
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID( logic [((AXI4_ID_WIDTH) - 1):0] ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID = ARID_param;
        else
            m_ARID <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID_index1( int _this_dot_1, logic  ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID[_this_dot_1] = ARID_param;
        else
            m_ARID[_this_dot_1] <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_ARID(  );
        return ARID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic   get_ARID_index1( int _this_dot_1 );
        return ARID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARUSER>.
    //
    // Parameters:
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER( logic [((AXI4_USER_WIDTH) - 1):0] ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER = ARUSER_param;
        else
            m_ARUSER <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER_index1( int _this_dot_1, logic  ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER[_this_dot_1] = ARUSER_param;
        else
            m_ARUSER[_this_dot_1] <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_ARUSER(  );
        return ARUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic   get_ARUSER_index1( int _this_dot_1 );
        return ARUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREADY>.
    //
    // Parameters:
    //     ARREADY_param - The value to set onto wire <ARREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREADY( logic ARREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREADY = ARREADY_param;
        else
            m_ARREADY <= ARREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREADY>.
    //
    function automatic logic get_ARREADY(  );
        return ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RVALID>.
    //
    // Parameters:
    //     RVALID_param - The value to set onto wire <RVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RVALID( logic RVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RVALID = RVALID_param;
        else
            m_RVALID <= RVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RVALID>.
    //
    function automatic logic get_RVALID(  );
        return RVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <RDATA>.
    //
    // Parameters:
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA( logic [((AXI4_RDATA_WIDTH) - 1):0] RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA = RDATA_param;
        else
            m_RDATA <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA_index1( int _this_dot_1, logic  RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA[_this_dot_1] = RDATA_param;
        else
            m_RDATA[_this_dot_1] <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <RDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic [((AXI4_RDATA_WIDTH) - 1):0]  get_RDATA(  );
        return RDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic   get_RDATA_index1( int _this_dot_1 );
        return RDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <RRESP>.
    //
    // Parameters:
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP( logic [1:0] RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP = RRESP_param;
        else
            m_RRESP <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP_index1( int _this_dot_1, logic  RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP[_this_dot_1] = RRESP_param;
        else
            m_RRESP[_this_dot_1] <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <RRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic [1:0]  get_RRESP(  );
        return RRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic   get_RRESP_index1( int _this_dot_1 );
        return RRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <RLAST>.
    //
    // Parameters:
    //     RLAST_param - The value to set onto wire <RLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RLAST( logic RLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RLAST = RLAST_param;
        else
            m_RLAST <= RLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <RLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RLAST>.
    //
    function automatic logic get_RLAST(  );
        return RLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RID>.
    //
    // Parameters:
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID( logic [((AXI4_ID_WIDTH) - 1):0] RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID = RID_param;
        else
            m_RID <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID_index1( int _this_dot_1, logic  RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID[_this_dot_1] = RID_param;
        else
            m_RID[_this_dot_1] <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_RID(  );
        return RID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic   get_RID_index1( int _this_dot_1 );
        return RID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <RUSER>.
    //
    // Parameters:
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER( logic [((AXI4_USER_WIDTH) - 1):0] RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER = RUSER_param;
        else
            m_RUSER <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER_index1( int _this_dot_1, logic  RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER[_this_dot_1] = RUSER_param;
        else
            m_RUSER[_this_dot_1] <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <RUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_RUSER(  );
        return RUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic   get_RUSER_index1( int _this_dot_1 );
        return RUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <RREADY>.
    //
    // Parameters:
    //     RREADY_param - The value to set onto wire <RREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RREADY( logic RREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RREADY = RREADY_param;
        else
            m_RREADY <= RREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RREADY>.
    //
    function automatic logic get_RREADY(  );
        return RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <WVALID>.
    //
    // Parameters:
    //     WVALID_param - The value to set onto wire <WVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WVALID( logic WVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WVALID = WVALID_param;
        else
            m_WVALID <= WVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <WVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WVALID>.
    //
    function automatic logic get_WVALID(  );
        return WVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <WDATA>.
    //
    // Parameters:
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA( logic [((AXI4_WDATA_WIDTH) - 1):0] WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA = WDATA_param;
        else
            m_WDATA <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA_index1( int _this_dot_1, logic  WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA[_this_dot_1] = WDATA_param;
        else
            m_WDATA[_this_dot_1] <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <WDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic [((AXI4_WDATA_WIDTH) - 1):0]  get_WDATA(  );
        return WDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic   get_WDATA_index1( int _this_dot_1 );
        return WDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB
    //-------------------------------------------------------------------------
    //     Set the value of wire <WSTRB>.
    //
    // Parameters:
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB( logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB = WSTRB_param;
        else
            m_WSTRB <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB_index1( int _this_dot_1, logic  WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB[_this_dot_1] = WSTRB_param;
        else
            m_WSTRB[_this_dot_1] <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB
    //-------------------------------------------------------------------------
    //     Get the value of wire <WSTRB>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0]  get_WSTRB(  );
        return WSTRB;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic   get_WSTRB_index1( int _this_dot_1 );
        return WSTRB[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <WLAST>.
    //
    // Parameters:
    //     WLAST_param - The value to set onto wire <WLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WLAST( logic WLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WLAST = WLAST_param;
        else
            m_WLAST <= WLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <WLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WLAST>.
    //
    function automatic logic get_WLAST(  );
        return WLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <WUSER>.
    //
    // Parameters:
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER( logic [((AXI4_USER_WIDTH) - 1):0] WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER = WUSER_param;
        else
            m_WUSER <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER_index1( int _this_dot_1, logic  WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER[_this_dot_1] = WUSER_param;
        else
            m_WUSER[_this_dot_1] <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <WUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_WUSER(  );
        return WUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic   get_WUSER_index1( int _this_dot_1 );
        return WUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <WREADY>.
    //
    // Parameters:
    //     WREADY_param - The value to set onto wire <WREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WREADY( logic WREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WREADY = WREADY_param;
        else
            m_WREADY <= WREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WREADY>.
    //
    function automatic logic get_WREADY(  );
        return WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BVALID>.
    //
    // Parameters:
    //     BVALID_param - The value to set onto wire <BVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BVALID( logic BVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BVALID = BVALID_param;
        else
            m_BVALID <= BVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BVALID>.
    //
    function automatic logic get_BVALID(  );
        return BVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <BRESP>.
    //
    // Parameters:
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP( logic [1:0] BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP = BRESP_param;
        else
            m_BRESP <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP_index1( int _this_dot_1, logic  BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP[_this_dot_1] = BRESP_param;
        else
            m_BRESP[_this_dot_1] <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <BRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic [1:0]  get_BRESP(  );
        return BRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic   get_BRESP_index1( int _this_dot_1 );
        return BRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BID>.
    //
    // Parameters:
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID( logic [((AXI4_ID_WIDTH) - 1):0] BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID = BID_param;
        else
            m_BID <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID_index1( int _this_dot_1, logic  BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID[_this_dot_1] = BID_param;
        else
            m_BID[_this_dot_1] <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_BID(  );
        return BID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic   get_BID_index1( int _this_dot_1 );
        return BID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <BUSER>.
    //
    // Parameters:
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER( logic [((AXI4_USER_WIDTH) - 1):0] BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER = BUSER_param;
        else
            m_BUSER <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER_index1( int _this_dot_1, logic  BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER[_this_dot_1] = BUSER_param;
        else
            m_BUSER[_this_dot_1] <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <BUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_BUSER(  );
        return BUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic   get_BUSER_index1( int _this_dot_1 );
        return BUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <BREADY>.
    //
    // Parameters:
    //     BREADY_param - The value to set onto wire <BREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BREADY( logic BREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BREADY = BREADY_param;
        else
            m_BREADY <= BREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BREADY>.
    //
    function automatic logic get_BREADY(  );
        return BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id( bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id_param );
        config_axi4lite_tr_id = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id_index1( int _this_dot_1, bit  config_axi4lite_tr_id_param );
        config_axi4lite_tr_id[_this_dot_1] = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit [((AXI4_ID_WIDTH) - 1):0]  get_config_axi4lite_tr_id(  );
        return config_axi4lite_tr_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit   get_config_axi4lite_tr_id_index1( int _this_dot_1 );
        return config_axi4lite_tr_id[_this_dot_1];
    endfunction

    //-------------------------------------------------------------------------
    // Tasks to wait for a change to a global variable with read access
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_interface
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_interface>.
    //
    task automatic wait_for_config_axi4lite_interface(  );
        begin
            @( config_axi4lite_interface );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_interface_type
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_interface_type>.
    //
    task automatic wait_for_config_interface_type(  );
        begin
            @( config_interface_type );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_rlast
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_rlast>.
    //
    task automatic wait_for_config_enable_rlast(  );
        begin
            @( config_enable_rlast );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_id>.
    //
    task automatic wait_for_config_enable_id(  );
        begin
            @( config_enable_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_all_assertions
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_all_assertions>.
    //
    task automatic wait_for_config_enable_all_assertions(  );
        begin
            @( config_enable_all_assertions );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    task automatic wait_for_config_enable_assertion(  );
        begin
            @( config_enable_assertion );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_enable_assertion_index1( input int _this_dot_1 );
        begin
            @( config_enable_assertion[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_user_sideband
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_user_sideband>.
    //
    task automatic wait_for_config_enable_user_sideband(  );
        begin
            @( config_enable_user_sideband );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_slave_exclusive
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_slave_exclusive>.
    //
    task automatic wait_for_config_enable_slave_exclusive(  );
        begin
            @( config_enable_slave_exclusive );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_read_data_reordering_depth
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_read_data_reordering_depth>.
    //
    task automatic wait_for_config_read_data_reordering_depth(  );
        begin
            @( config_read_data_reordering_depth );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    task automatic wait_for_config_slave_start_addr(  );
        begin
            @( config_slave_start_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_start_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_start_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    task automatic wait_for_config_slave_end_addr(  );
        begin
            @( config_slave_end_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_end_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_end_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_transaction_time_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_transaction_time_factor>.
    //
    task automatic wait_for_config_max_transaction_time_factor(  );
        begin
            @( config_max_transaction_time_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_burst_timeout_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_burst_timeout_factor>.
    //
    task automatic wait_for_config_burst_timeout_factor(  );
        begin
            @( config_burst_timeout_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_AWVALID_assertion_to_AWREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    task automatic wait_for_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        begin
            @( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_ARVALID_assertion_to_ARREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    task automatic wait_for_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        begin
            @( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_RVALID_assertion_to_RREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_RVALID_assertion_to_RREADY>.
    //
    task automatic wait_for_config_max_latency_RVALID_assertion_to_RREADY(  );
        begin
            @( config_max_latency_RVALID_assertion_to_RREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_BVALID_assertion_to_BREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_BVALID_assertion_to_BREADY>.
    //
    task automatic wait_for_config_max_latency_BVALID_assertion_to_BREADY(  );
        begin
            @( config_max_latency_BVALID_assertion_to_BREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_WVALID_assertion_to_WREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_WVALID_assertion_to_WREADY>.
    //
    task automatic wait_for_config_max_latency_WVALID_assertion_to_WREADY(  );
        begin
            @( config_max_latency_WVALID_assertion_to_WREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_qos
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_qos>.
    //
    task automatic wait_for_config_enable_qos(  );
        begin
            @( config_enable_qos );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_setup_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_setup_time>.
    //
    task automatic wait_for_config_setup_time(  );
        begin
            @( config_setup_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_hold_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_hold_time>.
    //
    task automatic wait_for_config_hold_time(  );
        begin
            @( config_hold_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_reads
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_reads>.
    //
    task automatic wait_for_config_num_max_outstanding_reads(  );
        begin
            @( config_num_max_outstanding_reads );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_writes
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_writes>.
    //
    task automatic wait_for_config_num_max_outstanding_writes(  );
        begin
            @( config_num_max_outstanding_writes );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_wr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_wr>.
    //
    task automatic wait_for_config_max_outstanding_wr(  );
        begin
            @( config_max_outstanding_wr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rd
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rd>.
    //
    task automatic wait_for_config_max_outstanding_rd(  );
        begin
            @( config_max_outstanding_rd );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rw
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rw>.
    //
    task automatic wait_for_config_max_outstanding_rw(  );
        begin
            @( config_max_outstanding_rw );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_wr_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_wr_phase>.
    //
    task automatic wait_for_config_num_outstanding_wr_phase(  );
        begin
            @( config_num_outstanding_wr_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_rd_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_rd_phase>.
    //
    task automatic wait_for_config_num_outstanding_rd_phase(  );
        begin
            @( config_num_outstanding_rd_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_is_issuing
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_is_issuing>.
    //
    task automatic wait_for_config_is_issuing(  );
        begin
            @( config_is_issuing );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    task automatic wait_for_config_axi4lite_tr_id(  );
        begin
            @( config_axi4lite_tr_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_axi4lite_tr_id_index1( input int _this_dot_1 );
        begin
            @( config_axi4lite_tr_id[_this_dot_1] );
        end
    endtask


    //-------------------------------------------------------------------------
    // Functions to set global variables with write access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //     config_axi4lite_interface_param - The value to assign to variable <config_axi4lite_interface>.
    //
    function automatic void set_config_axi4lite_interface( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_interface_type
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_interface_type>.
    //
    // Parameters:
    //     config_interface_type_param - The value to assign to variable <config_interface_type>.
    //
    function automatic void set_config_interface_type( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //     config_enable_rlast_param - The value to assign to variable <config_enable_rlast>.
    //
    function automatic void set_config_enable_rlast( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_id>.
    //
    // Parameters:
    //     config_enable_id_param - The value to assign to variable <config_enable_id>.
    //
    function automatic void set_config_enable_id( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //     config_enable_all_assertions_param - The value to assign to variable <config_enable_all_assertions>.
    //
    function automatic void set_config_enable_all_assertions( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion_index1( int _this_dot_1, bit  config_enable_assertion_param );
        config_enable_assertion[_this_dot_1] = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //     config_enable_user_sideband_param - The value to assign to variable <config_enable_user_sideband>.
    //
    function automatic void set_config_enable_user_sideband( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //     config_enable_slave_exclusive_param - The value to assign to variable <config_enable_slave_exclusive>.
    //
    function automatic void set_config_enable_slave_exclusive( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //     config_read_data_reordering_depth_param - The value to assign to variable <config_read_data_reordering_depth>.
    //
    function automatic void set_config_read_data_reordering_depth( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        config_slave_start_addr = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr_index1( int _this_dot_1, bit  config_slave_start_addr_param );
        config_slave_start_addr[_this_dot_1] = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        config_slave_end_addr = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr_index1( int _this_dot_1, bit  config_slave_end_addr_param );
        config_slave_end_addr[_this_dot_1] = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //     config_max_transaction_time_factor_param - The value to assign to variable <config_max_transaction_time_factor>.
    //
    function automatic void set_config_max_transaction_time_factor( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //     config_burst_timeout_factor_param - The value to assign to variable <config_burst_timeout_factor>.
    //
    function automatic void set_config_burst_timeout_factor( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //     config_max_latency_AWVALID_assertion_to_AWREADY_param - The value to assign to variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic void set_config_max_latency_AWVALID_assertion_to_AWREADY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //     config_max_latency_ARVALID_assertion_to_ARREADY_param - The value to assign to variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic void set_config_max_latency_ARVALID_assertion_to_ARREADY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //     config_max_latency_RVALID_assertion_to_RREADY_param - The value to assign to variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic void set_config_max_latency_RVALID_assertion_to_RREADY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //     config_max_latency_BVALID_assertion_to_BREADY_param - The value to assign to variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic void set_config_max_latency_BVALID_assertion_to_BREADY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //     config_max_latency_WVALID_assertion_to_WREADY_param - The value to assign to variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic void set_config_max_latency_WVALID_assertion_to_WREADY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_qos
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_qos>.
    //
    // Parameters:
    //     config_enable_qos_param - The value to assign to variable <config_enable_qos>.
    //
    function automatic void set_config_enable_qos( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_setup_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_setup_time>.
    //
    // Parameters:
    //     config_setup_time_param - The value to assign to variable <config_setup_time>.
    //
    function automatic void set_config_setup_time( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_hold_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_hold_time>.
    //
    // Parameters:
    //     config_hold_time_param - The value to assign to variable <config_hold_time>.
    //
    function automatic void set_config_hold_time( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //     config_num_max_outstanding_reads_param - The value to assign to variable <config_num_max_outstanding_reads>.
    //
    function automatic void set_config_num_max_outstanding_reads( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //     config_num_max_outstanding_writes_param - The value to assign to variable <config_num_max_outstanding_writes>.
    //
    function automatic void set_config_num_max_outstanding_writes( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //     config_max_outstanding_wr_param - The value to assign to variable <config_max_outstanding_wr>.
    //
    function automatic void set_config_max_outstanding_wr( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //     config_max_outstanding_rd_param - The value to assign to variable <config_max_outstanding_rd>.
    //
    function automatic void set_config_max_outstanding_rd( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //     config_max_outstanding_rw_param - The value to assign to variable <config_max_outstanding_rw>.
    //
    function automatic void set_config_max_outstanding_rw( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //     config_num_outstanding_wr_phase_param - The value to assign to variable <config_num_outstanding_wr_phase>.
    //
    function automatic void set_config_num_outstanding_wr_phase( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //     config_num_outstanding_rd_phase_param - The value to assign to variable <config_num_outstanding_rd_phase>.
    //
    function automatic void set_config_num_outstanding_rd_phase( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_is_issuing
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_is_issuing>.
    //
    // Parameters:
    //     config_is_issuing_param - The value to assign to variable <config_is_issuing>.
    //
    function automatic void set_config_is_issuing( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to get global variables with read access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_interface>.
    //
    function automatic bit get_config_axi4lite_interface(  );
        return config_axi4lite_interface;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_interface_type
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_interface_type>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_interface_type>.
    //
    function automatic axi4_interface_type_e get_config_interface_type(  );
        return config_interface_type;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_rlast>.
    //
    function automatic bit get_config_enable_rlast(  );
        return config_enable_rlast;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_id>.
    //
    function automatic bit get_config_enable_id(  );
        return config_enable_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_all_assertions>.
    //
    function automatic bit get_config_enable_all_assertions(  );
        return config_enable_all_assertions;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit [255:0]  get_config_enable_assertion(  );
        return config_enable_assertion;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit   get_config_enable_assertion_index1( int _this_dot_1 );
        return config_enable_assertion[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_user_sideband>.
    //
    function automatic bit get_config_enable_user_sideband(  );
        return config_enable_user_sideband;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_slave_exclusive>.
    //
    function automatic bit get_config_enable_slave_exclusive(  );
        return config_enable_slave_exclusive;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_read_data_reordering_depth>.
    //
    function automatic int unsigned get_config_read_data_reordering_depth(  );
        return config_read_data_reordering_depth;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_start_addr(  );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit   get_config_slave_start_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_end_addr(  );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit   get_config_slave_end_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_transaction_time_factor>.
    //
    function automatic int unsigned get_config_max_transaction_time_factor(  );
        return config_max_transaction_time_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_burst_timeout_factor>.
    //
    function automatic int unsigned get_config_burst_timeout_factor(  );
        return config_burst_timeout_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic int unsigned get_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        return config_max_latency_AWVALID_assertion_to_AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic int unsigned get_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        return config_max_latency_ARVALID_assertion_to_ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic int unsigned get_config_max_latency_RVALID_assertion_to_RREADY(  );
        return config_max_latency_RVALID_assertion_to_RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic int unsigned get_config_max_latency_BVALID_assertion_to_BREADY(  );
        return config_max_latency_BVALID_assertion_to_BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic int unsigned get_config_max_latency_WVALID_assertion_to_WREADY(  );
        return config_max_latency_WVALID_assertion_to_WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_qos
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_qos>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_qos>.
    //
    function automatic bit get_config_enable_qos(  );
        return config_enable_qos;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_setup_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_setup_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_setup_time>.
    //
    function automatic longint unsigned get_config_setup_time(  );
        return config_setup_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_hold_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_hold_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_hold_time>.
    //
    function automatic longint unsigned get_config_hold_time(  );
        return config_hold_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_reads>.
    //
    function automatic int get_config_num_max_outstanding_reads(  );
        return config_num_max_outstanding_reads;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_writes>.
    //
    function automatic int get_config_num_max_outstanding_writes(  );
        return config_num_max_outstanding_writes;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_wr>.
    //
    function automatic int get_config_max_outstanding_wr(  );
        return config_max_outstanding_wr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rd>.
    //
    function automatic int get_config_max_outstanding_rd(  );
        return config_max_outstanding_rd;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rw>.
    //
    function automatic int get_config_max_outstanding_rw(  );
        return config_max_outstanding_rw;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_wr_phase>.
    //
    function automatic int get_config_num_outstanding_wr_phase(  );
        return config_num_outstanding_wr_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_rd_phase>.
    //
    function automatic int get_config_num_outstanding_rd_phase(  );
        return config_num_outstanding_rd_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_is_issuing
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_is_issuing>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_is_issuing>.
    //
    function automatic bit get_config_is_issuing(  );
        return config_is_issuing;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to set/get generic interface configuration
    //-------------------------------------------------------------------------

    function void set_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0,
        input int arg10 = 0
    );
        axi4_set_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10 );
    endfunction

    function int get_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0
    );
        return axi4_get_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 );
    endfunction

    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //-------------------------------------------------------------------------
    function string get_full_name();
        return axi4_get_full_name();
    endfunction

    //--------------------------------------------------------------------------
    //
    // Group:- Monitor Value Change on Variable
    //
    //--------------------------------------------------------------------------

    function automatic void axi4_local_set_config_axi4lite_interface_from_SystemVerilog( ref bit config_axi4lite_interface_param );
        dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog( _interface_ref, config_axi4lite_interface );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_axi4lite_interface_from_SystemVerilog( config_axi4lite_interface );
        end
    end

    function automatic void axi4_local_set_config_interface_type_from_SystemVerilog( ref axi4_interface_type_e config_interface_type_param );
        dvc_axi4_set_config_interface_type_from_SystemVerilog( _interface_ref, config_interface_type );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_interface_type_from_SystemVerilog( config_interface_type );
        end
    end

    function automatic void axi4_local_set_config_enable_rlast_from_SystemVerilog( ref bit config_enable_rlast_param );
        dvc_axi4_set_config_enable_rlast_from_SystemVerilog( _interface_ref, config_enable_rlast );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_rlast_from_SystemVerilog( config_enable_rlast );
        end
    end

    function automatic void axi4_local_set_config_enable_id_from_SystemVerilog( ref bit config_enable_id_param );
        dvc_axi4_set_config_enable_id_from_SystemVerilog( _interface_ref, config_enable_id );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_id_from_SystemVerilog( config_enable_id );
        end
    end

    function automatic void axi4_local_set_config_enable_all_assertions_from_SystemVerilog( ref bit config_enable_all_assertions_param );
        dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog( _interface_ref, config_enable_all_assertions );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_all_assertions_from_SystemVerilog( config_enable_all_assertions );
        end
    end

    function automatic void axi4_local_set_config_enable_assertion_from_SystemVerilog( ref bit [255:0] config_enable_assertion_param );
        dvc_axi4_set_config_enable_assertion_from_SystemVerilog( _interface_ref, config_enable_assertion );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_assertion_from_SystemVerilog( config_enable_assertion );
        end
    end

    function automatic void axi4_local_set_config_enable_user_sideband_from_SystemVerilog( ref bit config_enable_user_sideband_param );
        dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog( _interface_ref, config_enable_user_sideband );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_user_sideband_from_SystemVerilog( config_enable_user_sideband );
        end
    end

    function automatic void axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( ref bit config_enable_slave_exclusive_param );
        dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog( _interface_ref, config_enable_slave_exclusive );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( config_enable_slave_exclusive );
        end
    end

    function automatic void axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( ref int unsigned config_read_data_reordering_depth_param );
        dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog( _interface_ref, config_read_data_reordering_depth );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( config_read_data_reordering_depth );
        end
    end

    function automatic void axi4_local_set_config_slave_start_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        dvc_axi4_set_config_slave_start_addr_from_SystemVerilog( _interface_ref, config_slave_start_addr );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_start_addr_from_SystemVerilog( config_slave_start_addr );
        end
    end

    function automatic void axi4_local_set_config_slave_end_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        dvc_axi4_set_config_slave_end_addr_from_SystemVerilog( _interface_ref, config_slave_end_addr );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_end_addr_from_SystemVerilog( config_slave_end_addr );
        end
    end

    function automatic void axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( ref int unsigned config_max_transaction_time_factor_param );
        dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog( _interface_ref, config_max_transaction_time_factor );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( config_max_transaction_time_factor );
        end
    end

    function automatic void axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( ref int unsigned config_burst_timeout_factor_param );
        dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog( _interface_ref, config_burst_timeout_factor );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( config_burst_timeout_factor );
        end
    end

    function automatic void axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( ref int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( _interface_ref, config_max_latency_AWVALID_assertion_to_AWREADY );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( ref int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( _interface_ref, config_max_latency_ARVALID_assertion_to_ARREADY );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( ref int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( _interface_ref, config_max_latency_RVALID_assertion_to_RREADY );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( config_max_latency_RVALID_assertion_to_RREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( ref int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( _interface_ref, config_max_latency_BVALID_assertion_to_BREADY );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( config_max_latency_BVALID_assertion_to_BREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( ref int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( _interface_ref, config_max_latency_WVALID_assertion_to_WREADY );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( config_max_latency_WVALID_assertion_to_WREADY );
        end
    end

    function automatic void axi4_local_set_config_enable_qos_from_SystemVerilog( ref bit config_enable_qos_param );
        dvc_axi4_set_config_enable_qos_from_SystemVerilog( _interface_ref, config_enable_qos );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_qos_from_SystemVerilog( config_enable_qos );
        end
    end

    function automatic void axi4_local_set_config_setup_time_from_SystemVerilog( ref longint unsigned config_setup_time_param );
        dvc_axi4_set_config_setup_time_from_SystemVerilog( _interface_ref, config_setup_time );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_setup_time_from_SystemVerilog( config_setup_time );
        end
    end

    function automatic void axi4_local_set_config_hold_time_from_SystemVerilog( ref longint unsigned config_hold_time_param );
        dvc_axi4_set_config_hold_time_from_SystemVerilog( _interface_ref, config_hold_time );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_hold_time_from_SystemVerilog( config_hold_time );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( ref int config_num_max_outstanding_reads_param );
        dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog( _interface_ref, config_num_max_outstanding_reads );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( config_num_max_outstanding_reads );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( ref int config_num_max_outstanding_writes_param );
        dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog( _interface_ref, config_num_max_outstanding_writes );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( config_num_max_outstanding_writes );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( ref int config_max_outstanding_wr_param );
        dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog( _interface_ref, config_max_outstanding_wr );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( config_max_outstanding_wr );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( ref int config_max_outstanding_rd_param );
        dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog( _interface_ref, config_max_outstanding_rd );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( config_max_outstanding_rd );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( ref int config_max_outstanding_rw_param );
        dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog( _interface_ref, config_max_outstanding_rw );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( config_max_outstanding_rw );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( ref int config_num_outstanding_wr_phase_param );
        dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog( _interface_ref, config_num_outstanding_wr_phase );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( config_num_outstanding_wr_phase );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( ref int config_num_outstanding_rd_phase_param );
        dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog( _interface_ref, config_num_outstanding_rd_phase );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( config_num_outstanding_rd_phase );
        end
    end

    function automatic void axi4_local_set_config_is_issuing_from_SystemVerilog( ref bit config_is_issuing_param );
        dvc_axi4_set_config_is_issuing_from_SystemVerilog( _interface_ref, config_is_issuing );
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_is_issuing_from_SystemVerilog( config_is_issuing );
        end
    end

    //-------------------------------------------------------------------------
    // Transaction interface
    //-------------------------------------------------------------------------

    task automatic dvc_activate_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref axi4_rw_e read_or_write,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref int address_valid_delay,
        ref int data_valid_delay[],
        ref int write_response_valid_delay,
        ref int address_ready_delay,
        ref int data_ready_delay[],
        ref int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
                // Call function to provide sized and unsized params.
                // In addition gets back updated sizes of unsized params.
                axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, addr, prot, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, data_words_DIMS0, write_strobes, write_strobes_DIMS0, wdata_user_data, wdata_user_data_DIMS0, resp, resp_DIMS0, resp_user_data, resp_user_data_DIMS0, address_valid_delay, data_valid_delay, data_valid_delay_DIMS0, write_response_valid_delay, address_ready_delay, data_ready_delay, data_ready_delay_DIMS0, write_response_ready_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        data_words = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        write_strobes = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        wdata_user_data = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        resp = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        resp_user_data = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_valid_delay_DIMS0 != 0)
                    begin
                        data_valid_delay = new [data_valid_delay_DIMS0];
                    end
                    else
                    begin
                        data_valid_delay = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_ready_delay_DIMS0 != 0)
                    begin
                        data_ready_delay = new [data_ready_delay_DIMS0];
                    end
                    else
                    begin
                        data_ready_delay = new [1];  // Create dummy instead of a zero sized array
                    end
                    // Call function to get the sized and unsized params
                    axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, addr, prot, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, write_strobes, wdata_user_data, resp, resp_user_data, address_valid_delay, data_valid_delay, write_response_valid_delay, address_ready_delay, data_ready_delay, write_response_ready_delay, _unit_id); // DPI call to imported task
                    if (data_words_DIMS0 == 0)
                        data_words.delete;  // Delete each zero sized param
                    if (write_strobes_DIMS0 == 0)
                        write_strobes.delete;  // Delete each zero sized param
                    if (wdata_user_data_DIMS0 == 0)
                        wdata_user_data.delete;  // Delete each zero sized param
                    if (resp_DIMS0 == 0)
                        resp.delete;  // Delete each zero sized param
                    if (resp_user_data_DIMS0 == 0)
                        resp_user_data.delete;  // Delete each zero sized param
                    if (data_valid_delay_DIMS0 == 0)
                        data_valid_delay.delete;  // Delete each zero sized param
                    if (data_ready_delay_DIMS0 == 0)
                        data_ready_delay.delete;  // Delete each zero sized param
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_rw_e read_or_write,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output int address_valid_delay,
        ref int data_valid_delay[],
        output int write_response_valid_delay,
        output int address_ready_delay,
        ref int data_ready_delay[],
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
                // Call function to get unsized params sizes.
                axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_valid_delay_DIMS0, data_ready_delay_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    data_words = new [data_words_DIMS0];
                end
                else
                begin
                    data_words = new [1];  // Create dummy instead of a zero sized array
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    write_strobes = new [1];  // Create dummy instead of a zero sized array
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    wdata_user_data = new [1];  // Create dummy instead of a zero sized array
                end
                if (resp_DIMS0 != 0)
                begin
                    resp = new [resp_DIMS0];
                end
                else
                begin
                    resp = new [1];  // Create dummy instead of a zero sized array
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    resp_user_data = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_valid_delay_DIMS0 != 0)
                begin
                    data_valid_delay = new [data_valid_delay_DIMS0];
                end
                else
                begin
                    data_valid_delay = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_ready_delay_DIMS0 != 0)
                begin
                    data_ready_delay = new [data_ready_delay_DIMS0];
                end
                else
                begin
                    data_ready_delay = new [1];  // Create dummy instead of a zero sized array
                end
                // Call function to get the sized and unsized params
                axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, addr, prot, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, write_strobes, wdata_user_data, resp, resp_user_data, address_valid_delay, data_valid_delay, write_response_valid_delay, address_ready_delay, data_ready_delay, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
                if (data_words_DIMS0 == 0)
                    data_words.delete;  // Delete each zero sized param
                if (write_strobes_DIMS0 == 0)
                    write_strobes.delete;  // Delete each zero sized param
                if (wdata_user_data_DIMS0 == 0)
                    wdata_user_data.delete;  // Delete each zero sized param
                if (resp_DIMS0 == 0)
                    resp.delete;  // Delete each zero sized param
                if (resp_user_data_DIMS0 == 0)
                    resp_user_data.delete;  // Delete each zero sized param
                if (data_valid_delay_DIMS0 == 0)
                    data_valid_delay.delete;  // Delete each zero sized param
                if (data_ready_delay_DIMS0 == 0)
                    data_ready_delay.delete;  // Delete each zero sized param
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int rresp_delays[255:0],
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to provide sized and unsized params.
                // In addition gets back updated sizes of unsized params.
                axi4_read_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, rresp_delays, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, data_words_DIMS0, resp, resp_DIMS0, resp_user_data, resp_user_data_DIMS0, addr_start_time, addr_end_time, data_start_time, data_start_time_DIMS0, data_end_time, data_end_time_DIMS0, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        data_words = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        resp = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        resp_user_data = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        data_start_time = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        data_end_time = new [1];  // Create dummy instead of a zero sized array
                    end
                    // Call function to get the sized and unsized params
                    axi4_read_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, rresp_delays, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, resp, resp_user_data, addr_start_time, addr_end_time, data_start_time, data_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    if (data_words_DIMS0 == 0)
                        data_words.delete;  // Delete each zero sized param
                    if (resp_DIMS0 == 0)
                        resp.delete;  // Delete each zero sized param
                    if (resp_user_data_DIMS0 == 0)
                        resp_user_data.delete;  // Delete each zero sized param
                    if (data_start_time_DIMS0 == 0)
                        data_start_time.delete;  // Delete each zero sized param
                    if (data_end_time_DIMS0 == 0)
                        data_end_time.delete;  // Delete each zero sized param
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        ref int rresp_delays[255:0],
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_read_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    data_words = new [data_words_DIMS0];
                end
                else
                begin
                    data_words = new [1];  // Create dummy instead of a zero sized array
                end
                if (resp_DIMS0 != 0)
                begin
                    resp = new [resp_DIMS0];
                end
                else
                begin
                    resp = new [1];  // Create dummy instead of a zero sized array
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    resp_user_data = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    data_start_time = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    data_end_time = new [1];  // Create dummy instead of a zero sized array
                end
                // Call function to get the sized and unsized params
                axi4_read_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, rresp_delays, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, resp, resp_user_data, addr_start_time, addr_end_time, data_start_time, data_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                if (data_words_DIMS0 == 0)
                    data_words.delete;  // Delete each zero sized param
                if (resp_DIMS0 == 0)
                    resp.delete;  // Delete each zero sized param
                if (resp_user_data_DIMS0 == 0)
                    resp_user_data.delete;  // Delete each zero sized param
                if (data_start_time_DIMS0 == 0)
                    data_start_time.delete;  // Delete each zero sized param
                if (data_end_time_DIMS0 == 0)
                    data_end_time.delete;  // Delete each zero sized param
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int bresp_delay,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp,
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref longint wr_resp_start_time,
        ref longint wr_resp_end_time,
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to provide sized and unsized params.
                // In addition gets back updated sizes of unsized params.
                axi4_write_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, bresp_delay, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, data_words_DIMS0, write_strobes, write_strobes_DIMS0, wdata_user_data, wdata_user_data_DIMS0, resp, resp_user_data, addr_start_time, addr_end_time, data_start_time, data_start_time_DIMS0, data_end_time, data_end_time_DIMS0, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        data_words = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        write_strobes = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        wdata_user_data = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        data_start_time = new [1];  // Create dummy instead of a zero sized array
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        data_end_time = new [1];  // Create dummy instead of a zero sized array
                    end
                    // Call function to get the sized and unsized params
                    axi4_write_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, bresp_delay, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, write_strobes, wdata_user_data, resp, resp_user_data, addr_start_time, addr_end_time, data_start_time, data_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    if (data_words_DIMS0 == 0)
                        data_words.delete;  // Delete each zero sized param
                    if (write_strobes_DIMS0 == 0)
                        write_strobes.delete;  // Delete each zero sized param
                    if (wdata_user_data_DIMS0 == 0)
                        wdata_user_data.delete;  // Delete each zero sized param
                    if (data_start_time_DIMS0 == 0)
                        data_start_time.delete;  // Delete each zero sized param
                    if (data_end_time_DIMS0 == 0)
                        data_end_time.delete;  // Delete each zero sized param
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        output axi4_response_e resp,
        output bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    data_words = new [data_words_DIMS0];
                end
                else
                begin
                    data_words = new [1];  // Create dummy instead of a zero sized array
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    write_strobes = new [1];  // Create dummy instead of a zero sized array
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    wdata_user_data = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    data_start_time = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    data_end_time = new [1];  // Create dummy instead of a zero sized array
                end
                // Call function to get the sized and unsized params
                axi4_write_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, addr, prot, bresp_delay, region, size, burst, lock, cache, qos, id, burst_length, addr_user_data, data_words, write_strobes, wdata_user_data, resp, resp_user_data, addr_start_time, addr_end_time, data_start_time, data_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                if (data_words_DIMS0 == 0)
                    data_words.delete;  // Delete each zero sized param
                if (write_strobes_DIMS0 == 0)
                    write_strobes.delete;  // Delete each zero sized param
                if (wdata_user_data_DIMS0 == 0)
                    wdata_user_data.delete;  // Delete each zero sized param
                if (data_start_time_DIMS0 == 0)
                    data_start_time.delete;  // Delete each zero sized param
                if (data_end_time_DIMS0 == 0)
                    data_end_time.delete;  // Delete each zero sized param
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        input int wdata_burst_start_clock,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to provide sized and unsized params.
            axi4_write_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, burst_length, data_words, write_strobes, user_data, data_start_time, data_end_time, data_valid2ready, data2data_clock, wdata_burst_start_clock, _unit_id);
        end
    endtask

    task automatic dvc_get_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        output int wdata_burst_start_clock,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                automatic int data_valid2ready_DIMS0;
                automatic int data2data_clock_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, data_valid2ready_DIMS0, data2data_clock_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    data_words = new [data_words_DIMS0];
                end
                else
                begin
                    data_words = new [1];  // Create dummy instead of a zero sized array
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    write_strobes = new [1];  // Create dummy instead of a zero sized array
                end
                if (user_data_DIMS0 != 0)
                begin
                    user_data = new [user_data_DIMS0];
                end
                else
                begin
                    user_data = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    data_start_time = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    data_end_time = new [1];  // Create dummy instead of a zero sized array
                end
                if (data_valid2ready_DIMS0 != 0)
                begin
                    data_valid2ready = new [data_valid2ready_DIMS0];
                end
                else
                begin
                    data_valid2ready = new [1];  // Create dummy instead of a zero sized array
                end
                if (data2data_clock_DIMS0 != 0)
                begin
                    data2data_clock = new [data2data_clock_DIMS0];
                end
                else
                begin
                    data2data_clock = new [1];  // Create dummy instead of a zero sized array
                end
                // Call function to get the sized and unsized params
                axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, burst_length, data_words, write_strobes, user_data, data_start_time, data_end_time, data_valid2ready, data2data_clock, wdata_burst_start_clock, _unit_id, _using); // DPI call to imported task
                if (data_words_DIMS0 == 0)
                    data_words.delete;  // Delete each zero sized param
                if (write_strobes_DIMS0 == 0)
                    write_strobes.delete;  // Delete each zero sized param
                if (user_data_DIMS0 == 0)
                    user_data.delete;  // Delete each zero sized param
                if (data_start_time_DIMS0 == 0)
                    data_start_time.delete;  // Delete each zero sized param
                if (data_end_time_DIMS0 == 0)
                    data_end_time.delete;  // Delete each zero sized param
                if (data_valid2ready_DIMS0 == 0)
                    data_valid2ready.delete;  // Delete each zero sized param
                if (data2data_clock_DIMS0 == 0)
                    data2data_clock.delete;  // Delete each zero sized param
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, addr, prot, region, burst_length, size, burst, lock, cache, qos, id, addr_user, address_valid_delay, address_ready_delay, _unit_id);
        end
    endtask

    task automatic dvc_get_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, addr, prot, region, burst_length, size, burst, lock, cache, qos, id, addr_user, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, _unit_id);
        end
    endtask

    task automatic dvc_get_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, data, resp, last, id, user_data, data_ready_delay, _unit_id);
        end
    endtask

    task automatic dvc_get_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, data, resp, last, id, user_data, data_ready_delay, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, addr, prot, region, burst_length, size, burst, lock, cache, qos, id, addr_user, address_valid_delay, address_ready_delay, _unit_id);
        end
    endtask

    task automatic dvc_get_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, addr, prot, region, burst_length, size, burst, lock, cache, qos, id, addr_user, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, data, write_strobes, last, user_data, data_ready_delay, _unit_id);
        end
    endtask

    task automatic dvc_get_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, data, write_strobes, last, user_data, data_ready_delay, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, id, user_data, write_response_ready_delay, _unit_id);
        end
    endtask

    task automatic dvc_get_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, resp, id, user_data, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, addr, region, cache, burst_length, size, burst, lock, qos, id, user_data, prot, _unit_id);
        end
    endtask

    task automatic dvc_get_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, addr, region, cache, burst_length, size, burst, lock, qos, id, user_data, prot, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, data, resp, last, id, user_data, _unit_id);
        end
    endtask

    task automatic dvc_get_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, data, resp, last, id, user_data, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, addr, region, cache, burst_length, size, burst, lock, qos, id, user_data, prot, _unit_id);
        end
    endtask

    task automatic dvc_get_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, addr, region, cache, burst_length, size, burst, lock, qos, id, user_data, prot, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, data, strb, last, user_data, _unit_id);
        end
    endtask

    task automatic dvc_get_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, data, strb, last, user_data, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, id, user_data, _unit_id);
        end
    endtask

    task automatic dvc_get_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, resp, id, user_data, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask


    //-------------------------------------------------------------------------
    // Generic Interface Configuration Support
    //

    import "DPI-C" context dvc_axi4_set_interface = function void axi4_set_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9,
        input int arg10
    );
    import "DPI-C" context dvc_axi4_get_interface = function int axi4_get_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9
    );


    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //
    import "DPI-C" context dvc_axi4_get_full_name = function string axi4_get_full_name();



    //-------------------------------------------------------------------------
    // Abstraction level Support
    //

    import "DPI-C" context dvc_axi4_set_master_end_abstraction_level =
    function void axi4_set_master_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_master_end_abstraction_level =
    function void axi4_get_master_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_set_slave_end_abstraction_level =
    function void axi4_set_slave_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_slave_end_abstraction_level =
    function void axi4_get_slave_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );

    //-------------------------------------------------------------------------
    // Wire Level Interface Support
    //
    logic internal_ACLK = 'z;
    logic internal_ARESETn = 'z;
    logic internal_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_AWADDR = 'z;
    logic [2:0] internal_AWPROT = 'z;
    logic [3:0] internal_AWREGION = 'z;
    logic [7:0] internal_AWLEN = 'z;
    logic [3:0] internal_AWSIZE = 'z;
    logic [1:0] internal_AWBURST = 'z;
    logic internal_AWLOCK = 'z;
    logic [3:0] internal_AWCACHE = 'z;
    logic [3:0] internal_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_AWUSER = 'z;
    logic internal_AWREADY = 'z;
    logic internal_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_ARADDR = 'z;
    logic [2:0] internal_ARPROT = 'z;
    logic [3:0] internal_ARREGION = 'z;
    logic [7:0] internal_ARLEN = 'z;
    logic [3:0] internal_ARSIZE = 'z;
    logic [1:0] internal_ARBURST = 'z;
    logic internal_ARLOCK = 'z;
    logic [3:0] internal_ARCACHE = 'z;
    logic [3:0] internal_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_ARUSER = 'z;
    logic internal_ARREADY = 'z;
    logic internal_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] internal_RDATA = 'z;
    logic [1:0] internal_RRESP = 'z;
    logic internal_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_RUSER = 'z;
    logic internal_RREADY = 'z;
    logic internal_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] internal_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] internal_WSTRB = 'z;
    logic internal_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_WUSER = 'z;
    logic internal_WREADY = 'z;
    logic internal_BVALID = 'z;
    logic [1:0] internal_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_BUSER = 'z;
    logic internal_BREADY = 'z;
    import "DPI-C" context dvc_axi4_set_ACLK_from_SystemVerilog = function void dvc_axi4_set_ACLK_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit ACLK_param
    );
    import "DPI-C" context dvc_axi4_get_ACLK_into_SystemVerilog = function void dvc_axi4_get_ACLK_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit ACLK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ACLK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARESETn_from_SystemVerilog = function void dvc_axi4_set_ARESETn_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARESETn_param
    );
    import "DPI-C" context dvc_axi4_get_ARESETn_into_SystemVerilog = function void dvc_axi4_get_ARESETn_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARESETn_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARESETn_from_CY;

    import "DPI-C" context dvc_axi4_set_AWVALID_from_SystemVerilog = function void dvc_axi4_set_AWVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWVALID_param
    );
    import "DPI-C" context dvc_axi4_get_AWVALID_into_SystemVerilog = function void dvc_axi4_get_AWVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWADDR_from_SystemVerilog = function void dvc_axi4_set_AWADDR_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR_param
    );
    import "DPI-C" context dvc_axi4_get_AWADDR_into_SystemVerilog = function void dvc_axi4_get_AWADDR_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_AWPROT_from_SystemVerilog = function void dvc_axi4_set_AWPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] AWPROT_param
    );
    import "DPI-C" context dvc_axi4_get_AWPROT_into_SystemVerilog = function void dvc_axi4_get_AWPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] AWPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREGION_from_SystemVerilog = function void dvc_axi4_set_AWREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWREGION_param
    );
    import "DPI-C" context dvc_axi4_get_AWREGION_into_SystemVerilog = function void dvc_axi4_get_AWREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLEN_from_SystemVerilog = function void dvc_axi4_set_AWLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] AWLEN_param
    );
    import "DPI-C" context dvc_axi4_get_AWLEN_into_SystemVerilog = function void dvc_axi4_get_AWLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] AWLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_AWSIZE_from_SystemVerilog = function void dvc_axi4_set_AWSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_AWSIZE_into_SystemVerilog = function void dvc_axi4_get_AWSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWBURST_from_SystemVerilog = function void dvc_axi4_set_AWBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] AWBURST_param
    );
    import "DPI-C" context dvc_axi4_get_AWBURST_into_SystemVerilog = function void dvc_axi4_get_AWBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] AWBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLOCK_from_SystemVerilog = function void dvc_axi4_set_AWLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_AWLOCK_into_SystemVerilog = function void dvc_axi4_get_AWLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_AWCACHE_from_SystemVerilog = function void dvc_axi4_set_AWCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_AWCACHE_into_SystemVerilog = function void dvc_axi4_get_AWCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWQOS_from_SystemVerilog = function void dvc_axi4_set_AWQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWQOS_param
    );
    import "DPI-C" context dvc_axi4_get_AWQOS_into_SystemVerilog = function void dvc_axi4_get_AWQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_AWID_from_SystemVerilog = function void dvc_axi4_set_AWID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ID_WIDTH) - 1):0] AWID_param
    );
    import "DPI-C" context dvc_axi4_get_AWID_into_SystemVerilog = function void dvc_axi4_get_AWID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ID_WIDTH) - 1):0] AWID_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWUSER_from_SystemVerilog = function void dvc_axi4_set_AWUSER_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_USER_WIDTH) - 1):0] AWUSER_param
    );
    import "DPI-C" context dvc_axi4_get_AWUSER_into_SystemVerilog = function void dvc_axi4_get_AWUSER_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_USER_WIDTH) - 1):0] AWUSER_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREADY_from_SystemVerilog = function void dvc_axi4_set_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_AWREADY_into_SystemVerilog = function void dvc_axi4_get_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_ARVALID_from_SystemVerilog = function void dvc_axi4_set_ARVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARVALID_param
    );
    import "DPI-C" context dvc_axi4_get_ARVALID_into_SystemVerilog = function void dvc_axi4_get_ARVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARADDR_from_SystemVerilog = function void dvc_axi4_set_ARADDR_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR_param
    );
    import "DPI-C" context dvc_axi4_get_ARADDR_into_SystemVerilog = function void dvc_axi4_get_ARADDR_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_ARPROT_from_SystemVerilog = function void dvc_axi4_set_ARPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] ARPROT_param
    );
    import "DPI-C" context dvc_axi4_get_ARPROT_into_SystemVerilog = function void dvc_axi4_get_ARPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] ARPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREGION_from_SystemVerilog = function void dvc_axi4_set_ARREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARREGION_param
    );
    import "DPI-C" context dvc_axi4_get_ARREGION_into_SystemVerilog = function void dvc_axi4_get_ARREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLEN_from_SystemVerilog = function void dvc_axi4_set_ARLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] ARLEN_param
    );
    import "DPI-C" context dvc_axi4_get_ARLEN_into_SystemVerilog = function void dvc_axi4_get_ARLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] ARLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_ARSIZE_from_SystemVerilog = function void dvc_axi4_set_ARSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_ARSIZE_into_SystemVerilog = function void dvc_axi4_get_ARSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARBURST_from_SystemVerilog = function void dvc_axi4_set_ARBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] ARBURST_param
    );
    import "DPI-C" context dvc_axi4_get_ARBURST_into_SystemVerilog = function void dvc_axi4_get_ARBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] ARBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLOCK_from_SystemVerilog = function void dvc_axi4_set_ARLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_ARLOCK_into_SystemVerilog = function void dvc_axi4_get_ARLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARCACHE_from_SystemVerilog = function void dvc_axi4_set_ARCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_ARCACHE_into_SystemVerilog = function void dvc_axi4_get_ARCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARQOS_from_SystemVerilog = function void dvc_axi4_set_ARQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARQOS_param
    );
    import "DPI-C" context dvc_axi4_get_ARQOS_into_SystemVerilog = function void dvc_axi4_get_ARQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_ARID_from_SystemVerilog = function void dvc_axi4_set_ARID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ID_WIDTH) - 1):0] ARID_param
    );
    import "DPI-C" context dvc_axi4_get_ARID_into_SystemVerilog = function void dvc_axi4_get_ARID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ID_WIDTH) - 1):0] ARID_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARUSER_from_SystemVerilog = function void dvc_axi4_set_ARUSER_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_USER_WIDTH) - 1):0] ARUSER_param
    );
    import "DPI-C" context dvc_axi4_get_ARUSER_into_SystemVerilog = function void dvc_axi4_get_ARUSER_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_USER_WIDTH) - 1):0] ARUSER_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREADY_from_SystemVerilog = function void dvc_axi4_set_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_ARREADY_into_SystemVerilog = function void dvc_axi4_get_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_RVALID_from_SystemVerilog = function void dvc_axi4_set_RVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RVALID_param
    );
    import "DPI-C" context dvc_axi4_get_RVALID_into_SystemVerilog = function void dvc_axi4_get_RVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_RVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_RDATA_from_SystemVerilog = function void dvc_axi4_set_RDATA_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_RDATA_WIDTH) - 1):0] RDATA_param
    );
    import "DPI-C" context dvc_axi4_get_RDATA_into_SystemVerilog = function void dvc_axi4_get_RDATA_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_RDATA_WIDTH) - 1):0] RDATA_param

    );
    export "DPI-C" function dvc_axi4_initialise_RDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_RRESP_from_SystemVerilog = function void dvc_axi4_set_RRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] RRESP_param
    );
    import "DPI-C" context dvc_axi4_get_RRESP_into_SystemVerilog = function void dvc_axi4_get_RRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] RRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_RRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_RLAST_from_SystemVerilog = function void dvc_axi4_set_RLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RLAST_param
    );
    import "DPI-C" context dvc_axi4_get_RLAST_into_SystemVerilog = function void dvc_axi4_get_RLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_RLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_RID_from_SystemVerilog = function void dvc_axi4_set_RID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ID_WIDTH) - 1):0] RID_param
    );
    import "DPI-C" context dvc_axi4_get_RID_into_SystemVerilog = function void dvc_axi4_get_RID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ID_WIDTH) - 1):0] RID_param

    );
    export "DPI-C" function dvc_axi4_initialise_RID_from_CY;

    import "DPI-C" context dvc_axi4_set_RUSER_from_SystemVerilog = function void dvc_axi4_set_RUSER_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_USER_WIDTH) - 1):0] RUSER_param
    );
    import "DPI-C" context dvc_axi4_get_RUSER_into_SystemVerilog = function void dvc_axi4_get_RUSER_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_USER_WIDTH) - 1):0] RUSER_param

    );
    export "DPI-C" function dvc_axi4_initialise_RUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_RREADY_from_SystemVerilog = function void dvc_axi4_set_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_RREADY_into_SystemVerilog = function void dvc_axi4_get_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_WVALID_from_SystemVerilog = function void dvc_axi4_set_WVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WVALID_param
    );
    import "DPI-C" context dvc_axi4_get_WVALID_into_SystemVerilog = function void dvc_axi4_get_WVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_WVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_WDATA_from_SystemVerilog = function void dvc_axi4_set_WDATA_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_WDATA_WIDTH) - 1):0] WDATA_param
    );
    import "DPI-C" context dvc_axi4_get_WDATA_into_SystemVerilog = function void dvc_axi4_get_WDATA_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_WDATA_WIDTH) - 1):0] WDATA_param

    );
    export "DPI-C" function dvc_axi4_initialise_WDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_WSTRB_from_SystemVerilog = function void dvc_axi4_set_WSTRB_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB_param
    );
    import "DPI-C" context dvc_axi4_get_WSTRB_into_SystemVerilog = function void dvc_axi4_get_WSTRB_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB_param

    );
    export "DPI-C" function dvc_axi4_initialise_WSTRB_from_CY;

    import "DPI-C" context dvc_axi4_set_WLAST_from_SystemVerilog = function void dvc_axi4_set_WLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WLAST_param
    );
    import "DPI-C" context dvc_axi4_get_WLAST_into_SystemVerilog = function void dvc_axi4_get_WLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_WLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_WUSER_from_SystemVerilog = function void dvc_axi4_set_WUSER_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_USER_WIDTH) - 1):0] WUSER_param
    );
    import "DPI-C" context dvc_axi4_get_WUSER_into_SystemVerilog = function void dvc_axi4_get_WUSER_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_USER_WIDTH) - 1):0] WUSER_param

    );
    export "DPI-C" function dvc_axi4_initialise_WUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_WREADY_from_SystemVerilog = function void dvc_axi4_set_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_WREADY_into_SystemVerilog = function void dvc_axi4_get_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_BVALID_from_SystemVerilog = function void dvc_axi4_set_BVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BVALID_param
    );
    import "DPI-C" context dvc_axi4_get_BVALID_into_SystemVerilog = function void dvc_axi4_get_BVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_BVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_BRESP_from_SystemVerilog = function void dvc_axi4_set_BRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] BRESP_param
    );
    import "DPI-C" context dvc_axi4_get_BRESP_into_SystemVerilog = function void dvc_axi4_get_BRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] BRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_BRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_BID_from_SystemVerilog = function void dvc_axi4_set_BID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_ID_WIDTH) - 1):0] BID_param
    );
    import "DPI-C" context dvc_axi4_get_BID_into_SystemVerilog = function void dvc_axi4_get_BID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_ID_WIDTH) - 1):0] BID_param

    );
    export "DPI-C" function dvc_axi4_initialise_BID_from_CY;

    import "DPI-C" context dvc_axi4_set_BUSER_from_SystemVerilog = function void dvc_axi4_set_BUSER_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [((AXI4_USER_WIDTH) - 1):0] BUSER_param
    );
    import "DPI-C" context dvc_axi4_get_BUSER_into_SystemVerilog = function void dvc_axi4_get_BUSER_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [((AXI4_USER_WIDTH) - 1):0] BUSER_param

    );
    export "DPI-C" function dvc_axi4_initialise_BUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_BREADY_from_SystemVerilog = function void dvc_axi4_set_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_BREADY_into_SystemVerilog = function void dvc_axi4_get_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog = function void dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_axi4lite_interface_param
    );
    import "DPI-C" context dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog = function void dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_axi4lite_interface_param

    );
    export "DPI-C" function dvc_axi4_set_config_axi4lite_interface_from_CY;

    import "DPI-C" context dvc_axi4_set_config_interface_type_from_SystemVerilog = function void dvc_axi4_set_config_interface_type_from_SystemVerilog
    (
        input longint _iface_ref,
        input axi4_interface_type_e config_interface_type_param
    );
    import "DPI-C" context dvc_axi4_get_config_interface_type_into_SystemVerilog = function void dvc_axi4_get_config_interface_type_into_SystemVerilog
    (
        input longint _iface_ref,
        output axi4_interface_type_e config_interface_type_param

    );
    export "DPI-C" function dvc_axi4_set_config_interface_type_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_rlast_from_SystemVerilog = function void dvc_axi4_set_config_enable_rlast_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_rlast_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_rlast_into_SystemVerilog = function void dvc_axi4_get_config_enable_rlast_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_rlast_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_rlast_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_id_from_SystemVerilog = function void dvc_axi4_set_config_enable_id_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_id_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_id_into_SystemVerilog = function void dvc_axi4_get_config_enable_id_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_id_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_id_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog = function void dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_all_assertions_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog = function void dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_all_assertions_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_all_assertions_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_assertion_from_SystemVerilog = function void dvc_axi4_set_config_enable_assertion_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit [255:0] config_enable_assertion_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_assertion_into_SystemVerilog = function void dvc_axi4_get_config_enable_assertion_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit [255:0] config_enable_assertion_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_assertion_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog = function void dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_user_sideband_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog = function void dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_user_sideband_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_user_sideband_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog = function void dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_slave_exclusive_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog = function void dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_slave_exclusive_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_slave_exclusive_from_CY;

    import "DPI-C" context dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog = function void dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_read_data_reordering_depth_param
    );
    import "DPI-C" context dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog = function void dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_read_data_reordering_depth_param

    );
    export "DPI-C" function dvc_axi4_set_config_read_data_reordering_depth_from_CY;

    import "DPI-C" context dvc_axi4_set_config_slave_start_addr_from_SystemVerilog = function void dvc_axi4_set_config_slave_start_addr_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param
    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param

    );
    export "DPI-C" function dvc_axi4_set_config_slave_start_addr_from_CY;

    import "DPI-C" context dvc_axi4_set_config_slave_end_addr_from_SystemVerilog = function void dvc_axi4_set_config_slave_end_addr_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param
    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param

    );
    export "DPI-C" function dvc_axi4_set_config_slave_end_addr_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog = function void dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_transaction_time_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog = function void dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_transaction_time_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_transaction_time_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog = function void dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_burst_timeout_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog = function void dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_burst_timeout_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_burst_timeout_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_RVALID_assertion_to_RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_RVALID_assertion_to_RREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_BVALID_assertion_to_BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_BVALID_assertion_to_BREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_WVALID_assertion_to_WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_WVALID_assertion_to_WREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_qos_from_SystemVerilog = function void dvc_axi4_set_config_enable_qos_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_qos_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_qos_into_SystemVerilog = function void dvc_axi4_get_config_enable_qos_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_qos_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_qos_from_CY;

    import "DPI-C" context dvc_axi4_set_config_setup_time_from_SystemVerilog = function void dvc_axi4_set_config_setup_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_setup_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_setup_time_into_SystemVerilog = function void dvc_axi4_get_config_setup_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_setup_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_setup_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_hold_time_from_SystemVerilog = function void dvc_axi4_set_config_hold_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_hold_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_hold_time_into_SystemVerilog = function void dvc_axi4_get_config_hold_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_hold_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_hold_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_reads_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_reads_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_reads_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_writes_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_writes_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_writes_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_wr_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_wr_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_wr_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rd_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rd_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rd_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rw_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rw_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rw_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_wr_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_wr_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_wr_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_rd_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_rd_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_rd_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_is_issuing_from_SystemVerilog = function void dvc_axi4_set_config_is_issuing_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_is_issuing_param
    );
    import "DPI-C" context dvc_axi4_get_config_is_issuing_into_SystemVerilog = function void dvc_axi4_get_config_is_issuing_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_is_issuing_param

    );
    export "DPI-C" function dvc_axi4_set_config_is_issuing_from_CY;

    function void dvc_axi4_initialise_ACLK_from_CY();
        internal_ACLK = 'z;
        m_ACLK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARESETn_from_CY();
        internal_ARESETn = 'z;
        m_ARESETn = 'z;
    endfunction

    function void dvc_axi4_initialise_AWVALID_from_CY();
        internal_AWVALID = 'z;
        m_AWVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_AWADDR_from_CY();
        internal_AWADDR = 'z;
        m_AWADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_AWPROT_from_CY();
        internal_AWPROT = 'z;
        m_AWPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREGION_from_CY();
        internal_AWREGION = 'z;
        m_AWREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLEN_from_CY();
        internal_AWLEN = 'z;
        m_AWLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_AWSIZE_from_CY();
        internal_AWSIZE = 'z;
        m_AWSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWBURST_from_CY();
        internal_AWBURST = 'z;
        m_AWBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLOCK_from_CY();
        internal_AWLOCK = 'z;
        m_AWLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_AWCACHE_from_CY();
        internal_AWCACHE = 'z;
        m_AWCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWQOS_from_CY();
        internal_AWQOS = 'z;
        m_AWQOS = 'z;
    endfunction

    function void dvc_axi4_initialise_AWID_from_CY();
        internal_AWID = 'z;
        m_AWID = 'z;
    endfunction

    function void dvc_axi4_initialise_AWUSER_from_CY();
        internal_AWUSER = 'z;
        m_AWUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREADY_from_CY();
        internal_AWREADY = 'z;
        m_AWREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_ARVALID_from_CY();
        internal_ARVALID = 'z;
        m_ARVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_ARADDR_from_CY();
        internal_ARADDR = 'z;
        m_ARADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_ARPROT_from_CY();
        internal_ARPROT = 'z;
        m_ARPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREGION_from_CY();
        internal_ARREGION = 'z;
        m_ARREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLEN_from_CY();
        internal_ARLEN = 'z;
        m_ARLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_ARSIZE_from_CY();
        internal_ARSIZE = 'z;
        m_ARSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARBURST_from_CY();
        internal_ARBURST = 'z;
        m_ARBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLOCK_from_CY();
        internal_ARLOCK = 'z;
        m_ARLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARCACHE_from_CY();
        internal_ARCACHE = 'z;
        m_ARCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARQOS_from_CY();
        internal_ARQOS = 'z;
        m_ARQOS = 'z;
    endfunction

    function void dvc_axi4_initialise_ARID_from_CY();
        internal_ARID = 'z;
        m_ARID = 'z;
    endfunction

    function void dvc_axi4_initialise_ARUSER_from_CY();
        internal_ARUSER = 'z;
        m_ARUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREADY_from_CY();
        internal_ARREADY = 'z;
        m_ARREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_RVALID_from_CY();
        internal_RVALID = 'z;
        m_RVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_RDATA_from_CY();
        internal_RDATA = 'z;
        m_RDATA = 'z;
    endfunction

    function void dvc_axi4_initialise_RRESP_from_CY();
        internal_RRESP = 'z;
        m_RRESP = 'z;
    endfunction

    function void dvc_axi4_initialise_RLAST_from_CY();
        internal_RLAST = 'z;
        m_RLAST = 'z;
    endfunction

    function void dvc_axi4_initialise_RID_from_CY();
        internal_RID = 'z;
        m_RID = 'z;
    endfunction

    function void dvc_axi4_initialise_RUSER_from_CY();
        internal_RUSER = 'z;
        m_RUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_RREADY_from_CY();
        internal_RREADY = 'z;
        m_RREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_WVALID_from_CY();
        internal_WVALID = 'z;
        m_WVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_WDATA_from_CY();
        internal_WDATA = 'z;
        m_WDATA = 'z;
    endfunction

    function void dvc_axi4_initialise_WSTRB_from_CY();
        internal_WSTRB = 'z;
        m_WSTRB = 'z;
    endfunction

    function void dvc_axi4_initialise_WLAST_from_CY();
        internal_WLAST = 'z;
        m_WLAST = 'z;
    endfunction

    function void dvc_axi4_initialise_WUSER_from_CY();
        internal_WUSER = 'z;
        m_WUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_WREADY_from_CY();
        internal_WREADY = 'z;
        m_WREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_BVALID_from_CY();
        internal_BVALID = 'z;
        m_BVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_BRESP_from_CY();
        internal_BRESP = 'z;
        m_BRESP = 'z;
    endfunction

    function void dvc_axi4_initialise_BID_from_CY();
        internal_BID = 'z;
        m_BID = 'z;
    endfunction

    function void dvc_axi4_initialise_BUSER_from_CY();
        internal_BUSER = 'z;
        m_BUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_BREADY_from_CY();
        internal_BREADY = 'z;
        m_BREADY = 'z;
    endfunction

    function void dvc_axi4_set_config_axi4lite_interface_from_CY( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction

    function void dvc_axi4_set_config_interface_type_from_CY( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction

    function void dvc_axi4_set_config_enable_rlast_from_CY( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction

    function void dvc_axi4_set_config_enable_id_from_CY( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction

    function void dvc_axi4_set_config_enable_all_assertions_from_CY( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction

    function void dvc_axi4_set_config_enable_assertion_from_CY( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction

    function void dvc_axi4_set_config_enable_user_sideband_from_CY( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction

    function void dvc_axi4_set_config_enable_slave_exclusive_from_CY( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction

    function void dvc_axi4_set_config_read_data_reordering_depth_from_CY( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction

    function void dvc_axi4_set_config_slave_start_addr_from_CY( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        config_slave_start_addr = config_slave_start_addr_param;
    endfunction

    function void dvc_axi4_set_config_slave_end_addr_from_CY( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        config_slave_end_addr = config_slave_end_addr_param;
    endfunction

    function void dvc_axi4_set_config_max_transaction_time_factor_from_CY( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction

    function void dvc_axi4_set_config_burst_timeout_factor_from_CY( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction

    function void dvc_axi4_set_config_enable_qos_from_CY( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction

    function void dvc_axi4_set_config_setup_time_from_CY( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction

    function void dvc_axi4_set_config_hold_time_from_CY( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_reads_from_CY( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_writes_from_CY( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_wr_from_CY( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rd_from_CY( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rw_from_CY( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_wr_phase_from_CY( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_rd_phase_from_CY( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction

    function void dvc_axi4_set_config_is_issuing_from_CY( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //--------------------------------------------------------------------------
    //
    // Group:- TLM Interface Support
    //
    //--------------------------------------------------------------------------
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_rw_e read_or_write,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        input bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        input axi4_response_e resp[],
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        input int data_valid_delay[],
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        input int data_ready_delay[],
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_rw_e read_or_write,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        inout bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        inout axi4_response_e resp[],
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        inout int address_valid_delay,
        inout int data_valid_delay[],
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        inout int data_ready_delay[],
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_rw_e read_or_write,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        inout bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        inout axi4_response_e resp[],
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output int address_valid_delay,
        inout int data_valid_delay[],
        output int write_response_valid_delay,
        output int address_ready_delay,
        inout int data_ready_delay[],
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout int rresp_delays[255:0],
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        input axi4_response_e resp[],
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint addr_start_time,
        inout longint addr_end_time,
        input longint data_start_time[],
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input longint data_end_time[],
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout int rresp_delays[255:0],
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        inout axi4_response_e resp[],
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout longint data_start_time[],
        inout longint data_end_time[],
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output int rresp_delays[255:0],
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        inout axi4_response_e resp[],
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output longint addr_start_time,
        output longint addr_end_time,
        inout longint data_start_time[],
        inout longint data_end_time[],
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        input bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout axi4_response_e resp,
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        inout longint addr_start_time,
        inout longint addr_end_time,
        input longint data_start_time[],
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input longint data_end_time[],
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [((AXI4_ID_WIDTH) - 1):0] id,
        inout bit [7:0] burst_length,
        inout bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        inout bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        inout axi4_response_e resp,
        inout bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout longint data_start_time[],
        inout longint data_end_time[],
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        inout bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        inout bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        output axi4_response_e resp,
        output bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        output longint addr_start_time,
        output longint addr_end_time,
        inout longint data_start_time[],
        inout longint data_end_time[],
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_SendSendingSent_SystemVerilog =
    task axi4_write_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int burst_length,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        input longint data_start_time[],
        input longint data_end_time[],
        input int data_valid2ready[],
        input int data2data_clock[],
        input int wdata_burst_start_clock,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid2ready_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data2data_clock_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output int burst_length,
        inout bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        inout bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        inout bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        inout longint data_start_time[],
        inout longint data_end_time[],
        inout int data_valid2ready[],
        inout int data2data_clock[],
        output int wdata_burst_start_clock,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_data_burst_SendSendingSent_SystemVerilog =
    task axi4_read_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_fn_get_raddr_phase_info = function void fn_get_raddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output int num_beats_rcvd
    );

    import "DPI-C" context dvc_axi4_fn_get_waddr_phase_info = function void fn_get_waddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr
    );

    import "DPI-C" context dvc_axi4_fn_add_addr_map_entry = function void fn_add_addr_map_entry_C
    (
        input string region,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input longint unsigned size
    );

    import "DPI-C" context dvc_axi4_fn_add_wr_delay = function void fn_add_wr_delay_C
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        input int unsigned data2data[]
    );

    import "DPI-C" context dvc_axi4_fn_delete_wr_delay = function void fn_delete_wr_delay_C
    (
        input string region,
        input bit [17:0] id
    );

    import "DPI-C" context dvc_axi4_fn_drive_signals = function void fn_drive_signals_C
    (
        input string signal,
        input longint value
    );

    import "DPI-C" context dvc_axi4_fn_set_address_map_entry = function void fn_set_address_map_entry_C
    (
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] start_addr,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] end_addr
    );

    import "DPI-C" context dvc_axi4_fn_set_wr_def_delays = function void fn_set_wr_def_delays_C
    (
        input int unsigned min_addr2data,
        input int unsigned min_data2data[],
        input int unsigned max_addr2data,
        input int unsigned max_data2data[]
    );

    import "DPI-C" context dvc_axi4_fn_print_perf_stats = function void fn_print_perf_stats_C
    (
    );

    // Waiter task and control
    reg sim_wait_for_control = 0;

    always @(posedge sim_wait_for_control)
    begin
        disable wait_for;
        sim_wait_for_control = 0;
    end

    export "DPI-C" dvc_axi4_wait_for = task wait_for;

    task wait_for();
        begin
            wait(0 == 1);
        end
    endtask

    // Drive wires (from Cohesive) 
    assign ACLK = internal_ACLK;
    assign ARESETn = internal_ARESETn;
    assign AWVALID = internal_AWVALID;
    assign AWADDR = internal_AWADDR;
    assign AWPROT = internal_AWPROT;
    assign AWREGION = internal_AWREGION;
    assign AWLEN = internal_AWLEN;
    assign AWSIZE = internal_AWSIZE;
    assign AWBURST = internal_AWBURST;
    assign AWLOCK = internal_AWLOCK;
    assign AWCACHE = internal_AWCACHE;
    assign AWQOS = internal_AWQOS;
    assign AWID = internal_AWID;
    assign AWUSER = internal_AWUSER;
    assign AWREADY = internal_AWREADY;
    assign ARVALID = internal_ARVALID;
    assign ARADDR = internal_ARADDR;
    assign ARPROT = internal_ARPROT;
    assign ARREGION = internal_ARREGION;
    assign ARLEN = internal_ARLEN;
    assign ARSIZE = internal_ARSIZE;
    assign ARBURST = internal_ARBURST;
    assign ARLOCK = internal_ARLOCK;
    assign ARCACHE = internal_ARCACHE;
    assign ARQOS = internal_ARQOS;
    assign ARID = internal_ARID;
    assign ARUSER = internal_ARUSER;
    assign ARREADY = internal_ARREADY;
    assign RVALID = internal_RVALID;
    assign RDATA = internal_RDATA;
    assign RRESP = internal_RRESP;
    assign RLAST = internal_RLAST;
    assign RID = internal_RID;
    assign RUSER = internal_RUSER;
    assign RREADY = internal_RREADY;
    assign WVALID = internal_WVALID;
    assign WDATA = internal_WDATA;
    assign WSTRB = internal_WSTRB;
    assign WLAST = internal_WLAST;
    assign WUSER = internal_WUSER;
    assign WREADY = internal_WREADY;
    assign BVALID = internal_BVALID;
    assign BRESP = internal_BRESP;
    assign BID = internal_BID;
    assign BUSER = internal_BUSER;
    assign BREADY = internal_BREADY;
    // Drive wires (from User) 
    assign ACLK = m_ACLK;
    assign ARESETn = m_ARESETn;
    assign AWVALID = m_AWVALID;
    assign AWADDR = m_AWADDR;
    assign AWPROT = m_AWPROT;
    assign AWREGION = m_AWREGION;
    assign AWLEN = m_AWLEN;
    assign AWSIZE = m_AWSIZE;
    assign AWBURST = m_AWBURST;
    assign AWLOCK = m_AWLOCK;
    assign AWCACHE = m_AWCACHE;
    assign AWQOS = m_AWQOS;
    assign AWID = m_AWID;
    assign AWUSER = m_AWUSER;
    assign AWREADY = m_AWREADY;
    assign ARVALID = m_ARVALID;
    assign ARADDR = m_ARADDR;
    assign ARPROT = m_ARPROT;
    assign ARREGION = m_ARREGION;
    assign ARLEN = m_ARLEN;
    assign ARSIZE = m_ARSIZE;
    assign ARBURST = m_ARBURST;
    assign ARLOCK = m_ARLOCK;
    assign ARCACHE = m_ARCACHE;
    assign ARQOS = m_ARQOS;
    assign ARID = m_ARID;
    assign ARUSER = m_ARUSER;
    assign ARREADY = m_ARREADY;
    assign RVALID = m_RVALID;
    assign RDATA = m_RDATA;
    assign RRESP = m_RRESP;
    assign RLAST = m_RLAST;
    assign RID = m_RID;
    assign RUSER = m_RUSER;
    assign RREADY = m_RREADY;
    assign WVALID = m_WVALID;
    assign WDATA = m_WDATA;
    assign WSTRB = m_WSTRB;
    assign WLAST = m_WLAST;
    assign WUSER = m_WUSER;
    assign WREADY = m_WREADY;
    assign BVALID = m_BVALID;
    assign BRESP = m_BRESP;
    assign BID = m_BID;
    assign BUSER = m_BUSER;
    assign BREADY = m_BREADY;

    reg ACLK_changed = 0;
    reg ARESETn_changed = 0;
    reg AWVALID_changed = 0;
    reg AWADDR_changed = 0;
    reg AWPROT_changed = 0;
    reg AWREGION_changed = 0;
    reg AWLEN_changed = 0;
    reg AWSIZE_changed = 0;
    reg AWBURST_changed = 0;
    reg AWLOCK_changed = 0;
    reg AWCACHE_changed = 0;
    reg AWQOS_changed = 0;
    reg AWID_changed = 0;
    reg AWUSER_changed = 0;
    reg AWREADY_changed = 0;
    reg ARVALID_changed = 0;
    reg ARADDR_changed = 0;
    reg ARPROT_changed = 0;
    reg ARREGION_changed = 0;
    reg ARLEN_changed = 0;
    reg ARSIZE_changed = 0;
    reg ARBURST_changed = 0;
    reg ARLOCK_changed = 0;
    reg ARCACHE_changed = 0;
    reg ARQOS_changed = 0;
    reg ARID_changed = 0;
    reg ARUSER_changed = 0;
    reg ARREADY_changed = 0;
    reg RVALID_changed = 0;
    reg RDATA_changed = 0;
    reg RRESP_changed = 0;
    reg RLAST_changed = 0;
    reg RID_changed = 0;
    reg RUSER_changed = 0;
    reg RREADY_changed = 0;
    reg WVALID_changed = 0;
    reg WDATA_changed = 0;
    reg WSTRB_changed = 0;
    reg WLAST_changed = 0;
    reg WUSER_changed = 0;
    reg WREADY_changed = 0;
    reg BVALID_changed = 0;
    reg BRESP_changed = 0;
    reg BID_changed = 0;
    reg BUSER_changed = 0;
    reg BREADY_changed = 0;
    reg config_axi4lite_interface_changed = 0;
    reg config_interface_type_changed = 0;
    reg config_enable_rlast_changed = 0;
    reg config_enable_id_changed = 0;
    reg config_enable_all_assertions_changed = 0;
    reg config_enable_assertion_changed = 0;
    reg config_enable_user_sideband_changed = 0;
    reg config_enable_slave_exclusive_changed = 0;
    reg config_read_data_reordering_depth_changed = 0;
    reg config_slave_start_addr_changed = 0;
    reg config_slave_end_addr_changed = 0;
    reg config_max_transaction_time_factor_changed = 0;
    reg config_burst_timeout_factor_changed = 0;
    reg config_max_latency_AWVALID_assertion_to_AWREADY_changed = 0;
    reg config_max_latency_ARVALID_assertion_to_ARREADY_changed = 0;
    reg config_max_latency_RVALID_assertion_to_RREADY_changed = 0;
    reg config_max_latency_BVALID_assertion_to_BREADY_changed = 0;
    reg config_max_latency_WVALID_assertion_to_WREADY_changed = 0;
    reg config_enable_qos_changed = 0;
    reg config_setup_time_changed = 0;
    reg config_hold_time_changed = 0;
    reg config_num_max_outstanding_reads_changed = 0;
    reg config_num_max_outstanding_writes_changed = 0;
    reg config_max_outstanding_wr_changed = 0;
    reg config_max_outstanding_rd_changed = 0;
    reg config_max_outstanding_rw_changed = 0;
    reg config_num_outstanding_wr_phase_changed = 0;
    reg config_num_outstanding_rd_phase_changed = 0;
    reg config_is_issuing_changed = 0;

    // SV wire change monitors

    always @( ACLK or posedge _check_t0_values )
    begin
        dvc_axi4_set_ACLK_from_SystemVerilog(_interface_ref, ACLK); // DPI call to imported task
    end

    always @( ARESETn or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARESETn_from_SystemVerilog(_interface_ref, ARESETn); // DPI call to imported task
    end

    always @( AWVALID or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWVALID_from_SystemVerilog(_interface_ref, AWVALID); // DPI call to imported task
    end

    always @( AWADDR or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWADDR_from_SystemVerilog(_interface_ref, AWADDR); // DPI call to imported task
    end

    always @( AWPROT or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWPROT_from_SystemVerilog(_interface_ref, AWPROT); // DPI call to imported task
    end

    always @( AWREGION or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWREGION_from_SystemVerilog(_interface_ref, AWREGION); // DPI call to imported task
    end

    always @( AWLEN or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWLEN_from_SystemVerilog(_interface_ref, AWLEN); // DPI call to imported task
    end

    always @( AWSIZE or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWSIZE_from_SystemVerilog(_interface_ref, AWSIZE); // DPI call to imported task
    end

    always @( AWBURST or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWBURST_from_SystemVerilog(_interface_ref, AWBURST); // DPI call to imported task
    end

    always @( AWLOCK or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWLOCK_from_SystemVerilog(_interface_ref, AWLOCK); // DPI call to imported task
    end

    always @( AWCACHE or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWCACHE_from_SystemVerilog(_interface_ref, AWCACHE); // DPI call to imported task
    end

    always @( AWQOS or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWQOS_from_SystemVerilog(_interface_ref, AWQOS); // DPI call to imported task
    end

    always @( AWID or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWID_from_SystemVerilog(_interface_ref, AWID); // DPI call to imported task
    end

    always @( AWUSER or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWUSER_from_SystemVerilog(_interface_ref, AWUSER); // DPI call to imported task
    end

    always @( AWREADY or posedge _check_t0_values )
    begin
        dvc_axi4_set_AWREADY_from_SystemVerilog(_interface_ref, AWREADY); // DPI call to imported task
    end

    always @( ARVALID or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARVALID_from_SystemVerilog(_interface_ref, ARVALID); // DPI call to imported task
    end

    always @( ARADDR or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARADDR_from_SystemVerilog(_interface_ref, ARADDR); // DPI call to imported task
    end

    always @( ARPROT or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARPROT_from_SystemVerilog(_interface_ref, ARPROT); // DPI call to imported task
    end

    always @( ARREGION or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARREGION_from_SystemVerilog(_interface_ref, ARREGION); // DPI call to imported task
    end

    always @( ARLEN or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARLEN_from_SystemVerilog(_interface_ref, ARLEN); // DPI call to imported task
    end

    always @( ARSIZE or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARSIZE_from_SystemVerilog(_interface_ref, ARSIZE); // DPI call to imported task
    end

    always @( ARBURST or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARBURST_from_SystemVerilog(_interface_ref, ARBURST); // DPI call to imported task
    end

    always @( ARLOCK or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARLOCK_from_SystemVerilog(_interface_ref, ARLOCK); // DPI call to imported task
    end

    always @( ARCACHE or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARCACHE_from_SystemVerilog(_interface_ref, ARCACHE); // DPI call to imported task
    end

    always @( ARQOS or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARQOS_from_SystemVerilog(_interface_ref, ARQOS); // DPI call to imported task
    end

    always @( ARID or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARID_from_SystemVerilog(_interface_ref, ARID); // DPI call to imported task
    end

    always @( ARUSER or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARUSER_from_SystemVerilog(_interface_ref, ARUSER); // DPI call to imported task
    end

    always @( ARREADY or posedge _check_t0_values )
    begin
        dvc_axi4_set_ARREADY_from_SystemVerilog(_interface_ref, ARREADY); // DPI call to imported task
    end

    always @( RVALID or posedge _check_t0_values )
    begin
        dvc_axi4_set_RVALID_from_SystemVerilog(_interface_ref, RVALID); // DPI call to imported task
    end

    always @( RDATA or posedge _check_t0_values )
    begin
        dvc_axi4_set_RDATA_from_SystemVerilog(_interface_ref, RDATA); // DPI call to imported task
    end

    always @( RRESP or posedge _check_t0_values )
    begin
        dvc_axi4_set_RRESP_from_SystemVerilog(_interface_ref, RRESP); // DPI call to imported task
    end

    always @( RLAST or posedge _check_t0_values )
    begin
        dvc_axi4_set_RLAST_from_SystemVerilog(_interface_ref, RLAST); // DPI call to imported task
    end

    always @( RID or posedge _check_t0_values )
    begin
        dvc_axi4_set_RID_from_SystemVerilog(_interface_ref, RID); // DPI call to imported task
    end

    always @( RUSER or posedge _check_t0_values )
    begin
        dvc_axi4_set_RUSER_from_SystemVerilog(_interface_ref, RUSER); // DPI call to imported task
    end

    always @( RREADY or posedge _check_t0_values )
    begin
        dvc_axi4_set_RREADY_from_SystemVerilog(_interface_ref, RREADY); // DPI call to imported task
    end

    always @( WVALID or posedge _check_t0_values )
    begin
        dvc_axi4_set_WVALID_from_SystemVerilog(_interface_ref, WVALID); // DPI call to imported task
    end

    always @( WDATA or posedge _check_t0_values )
    begin
        dvc_axi4_set_WDATA_from_SystemVerilog(_interface_ref, WDATA); // DPI call to imported task
    end

    always @( WSTRB or posedge _check_t0_values )
    begin
        dvc_axi4_set_WSTRB_from_SystemVerilog(_interface_ref, WSTRB); // DPI call to imported task
    end

    always @( WLAST or posedge _check_t0_values )
    begin
        dvc_axi4_set_WLAST_from_SystemVerilog(_interface_ref, WLAST); // DPI call to imported task
    end

    always @( WUSER or posedge _check_t0_values )
    begin
        dvc_axi4_set_WUSER_from_SystemVerilog(_interface_ref, WUSER); // DPI call to imported task
    end

    always @( WREADY or posedge _check_t0_values )
    begin
        dvc_axi4_set_WREADY_from_SystemVerilog(_interface_ref, WREADY); // DPI call to imported task
    end

    always @( BVALID or posedge _check_t0_values )
    begin
        dvc_axi4_set_BVALID_from_SystemVerilog(_interface_ref, BVALID); // DPI call to imported task
    end

    always @( BRESP or posedge _check_t0_values )
    begin
        dvc_axi4_set_BRESP_from_SystemVerilog(_interface_ref, BRESP); // DPI call to imported task
    end

    always @( BID or posedge _check_t0_values )
    begin
        dvc_axi4_set_BID_from_SystemVerilog(_interface_ref, BID); // DPI call to imported task
    end

    always @( BUSER or posedge _check_t0_values )
    begin
        dvc_axi4_set_BUSER_from_SystemVerilog(_interface_ref, BUSER); // DPI call to imported task
    end

    always @( BREADY or posedge _check_t0_values )
    begin
        dvc_axi4_set_BREADY_from_SystemVerilog(_interface_ref, BREADY); // DPI call to imported task
    end


    // CY wire and variable changed flag monitors

    always @(posedge ACLK_changed or posedge _check_t0_values )
    begin
        while (ACLK_changed == 1'b1)
        begin
            dvc_axi4_get_ACLK_into_SystemVerilog( _interface_ref, internal_ACLK ); // DPI call to imported task
            ACLK_changed = 1'b0;
            #0 if ( ACLK !== internal_ACLK )
            begin
                dvc_axi4_set_ACLK_from_SystemVerilog( _interface_ref, ACLK );
            end
        end
    end

    always @(posedge ARESETn_changed or posedge _check_t0_values )
    begin
        while (ARESETn_changed == 1'b1)
        begin
            dvc_axi4_get_ARESETn_into_SystemVerilog( _interface_ref, internal_ARESETn ); // DPI call to imported task
            ARESETn_changed = 1'b0;
            #0 if ( ARESETn !== internal_ARESETn )
            begin
                dvc_axi4_set_ARESETn_from_SystemVerilog( _interface_ref, ARESETn );
            end
        end
    end

    always @(posedge AWVALID_changed or posedge _check_t0_values )
    begin
        while (AWVALID_changed == 1'b1)
        begin
            dvc_axi4_get_AWVALID_into_SystemVerilog( _interface_ref, internal_AWVALID ); // DPI call to imported task
            AWVALID_changed = 1'b0;
            #0 if ( AWVALID !== internal_AWVALID )
            begin
                dvc_axi4_set_AWVALID_from_SystemVerilog( _interface_ref, AWVALID );
            end
        end
    end

    always @(posedge AWADDR_changed or posedge _check_t0_values )
    begin
        while (AWADDR_changed == 1'b1)
        begin
            dvc_axi4_get_AWADDR_into_SystemVerilog( _interface_ref, internal_AWADDR ); // DPI call to imported task
            AWADDR_changed = 1'b0;
            #0 if ( AWADDR !== internal_AWADDR )
            begin
                dvc_axi4_set_AWADDR_from_SystemVerilog( _interface_ref, AWADDR );
            end
        end
    end

    always @(posedge AWPROT_changed or posedge _check_t0_values )
    begin
        while (AWPROT_changed == 1'b1)
        begin
            dvc_axi4_get_AWPROT_into_SystemVerilog( _interface_ref, internal_AWPROT ); // DPI call to imported task
            AWPROT_changed = 1'b0;
            #0 if ( AWPROT !== internal_AWPROT )
            begin
                dvc_axi4_set_AWPROT_from_SystemVerilog( _interface_ref, AWPROT );
            end
        end
    end

    always @(posedge AWREGION_changed or posedge _check_t0_values )
    begin
        while (AWREGION_changed == 1'b1)
        begin
            dvc_axi4_get_AWREGION_into_SystemVerilog( _interface_ref, internal_AWREGION ); // DPI call to imported task
            AWREGION_changed = 1'b0;
            #0 if ( AWREGION !== internal_AWREGION )
            begin
                dvc_axi4_set_AWREGION_from_SystemVerilog( _interface_ref, AWREGION );
            end
        end
    end

    always @(posedge AWLEN_changed or posedge _check_t0_values )
    begin
        while (AWLEN_changed == 1'b1)
        begin
            dvc_axi4_get_AWLEN_into_SystemVerilog( _interface_ref, internal_AWLEN ); // DPI call to imported task
            AWLEN_changed = 1'b0;
            #0 if ( AWLEN !== internal_AWLEN )
            begin
                dvc_axi4_set_AWLEN_from_SystemVerilog( _interface_ref, AWLEN );
            end
        end
    end

    always @(posedge AWSIZE_changed or posedge _check_t0_values )
    begin
        while (AWSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_AWSIZE_into_SystemVerilog( _interface_ref, internal_AWSIZE ); // DPI call to imported task
            AWSIZE_changed = 1'b0;
            #0 if ( AWSIZE !== internal_AWSIZE )
            begin
                dvc_axi4_set_AWSIZE_from_SystemVerilog( _interface_ref, AWSIZE );
            end
        end
    end

    always @(posedge AWBURST_changed or posedge _check_t0_values )
    begin
        while (AWBURST_changed == 1'b1)
        begin
            dvc_axi4_get_AWBURST_into_SystemVerilog( _interface_ref, internal_AWBURST ); // DPI call to imported task
            AWBURST_changed = 1'b0;
            #0 if ( AWBURST !== internal_AWBURST )
            begin
                dvc_axi4_set_AWBURST_from_SystemVerilog( _interface_ref, AWBURST );
            end
        end
    end

    always @(posedge AWLOCK_changed or posedge _check_t0_values )
    begin
        while (AWLOCK_changed == 1'b1)
        begin
            dvc_axi4_get_AWLOCK_into_SystemVerilog( _interface_ref, internal_AWLOCK ); // DPI call to imported task
            AWLOCK_changed = 1'b0;
            #0 if ( AWLOCK !== internal_AWLOCK )
            begin
                dvc_axi4_set_AWLOCK_from_SystemVerilog( _interface_ref, AWLOCK );
            end
        end
    end

    always @(posedge AWCACHE_changed or posedge _check_t0_values )
    begin
        while (AWCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_AWCACHE_into_SystemVerilog( _interface_ref, internal_AWCACHE ); // DPI call to imported task
            AWCACHE_changed = 1'b0;
            #0 if ( AWCACHE !== internal_AWCACHE )
            begin
                dvc_axi4_set_AWCACHE_from_SystemVerilog( _interface_ref, AWCACHE );
            end
        end
    end

    always @(posedge AWQOS_changed or posedge _check_t0_values )
    begin
        while (AWQOS_changed == 1'b1)
        begin
            dvc_axi4_get_AWQOS_into_SystemVerilog( _interface_ref, internal_AWQOS ); // DPI call to imported task
            AWQOS_changed = 1'b0;
            #0 if ( AWQOS !== internal_AWQOS )
            begin
                dvc_axi4_set_AWQOS_from_SystemVerilog( _interface_ref, AWQOS );
            end
        end
    end

    always @(posedge AWID_changed or posedge _check_t0_values )
    begin
        while (AWID_changed == 1'b1)
        begin
            dvc_axi4_get_AWID_into_SystemVerilog( _interface_ref, internal_AWID ); // DPI call to imported task
            AWID_changed = 1'b0;
            #0 if ( AWID !== internal_AWID )
            begin
                dvc_axi4_set_AWID_from_SystemVerilog( _interface_ref, AWID );
            end
        end
    end

    always @(posedge AWUSER_changed or posedge _check_t0_values )
    begin
        while (AWUSER_changed == 1'b1)
        begin
            dvc_axi4_get_AWUSER_into_SystemVerilog( _interface_ref, internal_AWUSER ); // DPI call to imported task
            AWUSER_changed = 1'b0;
            #0 if ( AWUSER !== internal_AWUSER )
            begin
                dvc_axi4_set_AWUSER_from_SystemVerilog( _interface_ref, AWUSER );
            end
        end
    end

    always @(posedge AWREADY_changed or posedge _check_t0_values )
    begin
        while (AWREADY_changed == 1'b1)
        begin
            dvc_axi4_get_AWREADY_into_SystemVerilog( _interface_ref, internal_AWREADY ); // DPI call to imported task
            AWREADY_changed = 1'b0;
            #0 if ( AWREADY !== internal_AWREADY )
            begin
                dvc_axi4_set_AWREADY_from_SystemVerilog( _interface_ref, AWREADY );
            end
        end
    end

    always @(posedge ARVALID_changed or posedge _check_t0_values )
    begin
        while (ARVALID_changed == 1'b1)
        begin
            dvc_axi4_get_ARVALID_into_SystemVerilog( _interface_ref, internal_ARVALID ); // DPI call to imported task
            ARVALID_changed = 1'b0;
            #0 if ( ARVALID !== internal_ARVALID )
            begin
                dvc_axi4_set_ARVALID_from_SystemVerilog( _interface_ref, ARVALID );
            end
        end
    end

    always @(posedge ARADDR_changed or posedge _check_t0_values )
    begin
        while (ARADDR_changed == 1'b1)
        begin
            dvc_axi4_get_ARADDR_into_SystemVerilog( _interface_ref, internal_ARADDR ); // DPI call to imported task
            ARADDR_changed = 1'b0;
            #0 if ( ARADDR !== internal_ARADDR )
            begin
                dvc_axi4_set_ARADDR_from_SystemVerilog( _interface_ref, ARADDR );
            end
        end
    end

    always @(posedge ARPROT_changed or posedge _check_t0_values )
    begin
        while (ARPROT_changed == 1'b1)
        begin
            dvc_axi4_get_ARPROT_into_SystemVerilog( _interface_ref, internal_ARPROT ); // DPI call to imported task
            ARPROT_changed = 1'b0;
            #0 if ( ARPROT !== internal_ARPROT )
            begin
                dvc_axi4_set_ARPROT_from_SystemVerilog( _interface_ref, ARPROT );
            end
        end
    end

    always @(posedge ARREGION_changed or posedge _check_t0_values )
    begin
        while (ARREGION_changed == 1'b1)
        begin
            dvc_axi4_get_ARREGION_into_SystemVerilog( _interface_ref, internal_ARREGION ); // DPI call to imported task
            ARREGION_changed = 1'b0;
            #0 if ( ARREGION !== internal_ARREGION )
            begin
                dvc_axi4_set_ARREGION_from_SystemVerilog( _interface_ref, ARREGION );
            end
        end
    end

    always @(posedge ARLEN_changed or posedge _check_t0_values )
    begin
        while (ARLEN_changed == 1'b1)
        begin
            dvc_axi4_get_ARLEN_into_SystemVerilog( _interface_ref, internal_ARLEN ); // DPI call to imported task
            ARLEN_changed = 1'b0;
            #0 if ( ARLEN !== internal_ARLEN )
            begin
                dvc_axi4_set_ARLEN_from_SystemVerilog( _interface_ref, ARLEN );
            end
        end
    end

    always @(posedge ARSIZE_changed or posedge _check_t0_values )
    begin
        while (ARSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_ARSIZE_into_SystemVerilog( _interface_ref, internal_ARSIZE ); // DPI call to imported task
            ARSIZE_changed = 1'b0;
            #0 if ( ARSIZE !== internal_ARSIZE )
            begin
                dvc_axi4_set_ARSIZE_from_SystemVerilog( _interface_ref, ARSIZE );
            end
        end
    end

    always @(posedge ARBURST_changed or posedge _check_t0_values )
    begin
        while (ARBURST_changed == 1'b1)
        begin
            dvc_axi4_get_ARBURST_into_SystemVerilog( _interface_ref, internal_ARBURST ); // DPI call to imported task
            ARBURST_changed = 1'b0;
            #0 if ( ARBURST !== internal_ARBURST )
            begin
                dvc_axi4_set_ARBURST_from_SystemVerilog( _interface_ref, ARBURST );
            end
        end
    end

    always @(posedge ARLOCK_changed or posedge _check_t0_values )
    begin
        while (ARLOCK_changed == 1'b1)
        begin
            dvc_axi4_get_ARLOCK_into_SystemVerilog( _interface_ref, internal_ARLOCK ); // DPI call to imported task
            ARLOCK_changed = 1'b0;
            #0 if ( ARLOCK !== internal_ARLOCK )
            begin
                dvc_axi4_set_ARLOCK_from_SystemVerilog( _interface_ref, ARLOCK );
            end
        end
    end

    always @(posedge ARCACHE_changed or posedge _check_t0_values )
    begin
        while (ARCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_ARCACHE_into_SystemVerilog( _interface_ref, internal_ARCACHE ); // DPI call to imported task
            ARCACHE_changed = 1'b0;
            #0 if ( ARCACHE !== internal_ARCACHE )
            begin
                dvc_axi4_set_ARCACHE_from_SystemVerilog( _interface_ref, ARCACHE );
            end
        end
    end

    always @(posedge ARQOS_changed or posedge _check_t0_values )
    begin
        while (ARQOS_changed == 1'b1)
        begin
            dvc_axi4_get_ARQOS_into_SystemVerilog( _interface_ref, internal_ARQOS ); // DPI call to imported task
            ARQOS_changed = 1'b0;
            #0 if ( ARQOS !== internal_ARQOS )
            begin
                dvc_axi4_set_ARQOS_from_SystemVerilog( _interface_ref, ARQOS );
            end
        end
    end

    always @(posedge ARID_changed or posedge _check_t0_values )
    begin
        while (ARID_changed == 1'b1)
        begin
            dvc_axi4_get_ARID_into_SystemVerilog( _interface_ref, internal_ARID ); // DPI call to imported task
            ARID_changed = 1'b0;
            #0 if ( ARID !== internal_ARID )
            begin
                dvc_axi4_set_ARID_from_SystemVerilog( _interface_ref, ARID );
            end
        end
    end

    always @(posedge ARUSER_changed or posedge _check_t0_values )
    begin
        while (ARUSER_changed == 1'b1)
        begin
            dvc_axi4_get_ARUSER_into_SystemVerilog( _interface_ref, internal_ARUSER ); // DPI call to imported task
            ARUSER_changed = 1'b0;
            #0 if ( ARUSER !== internal_ARUSER )
            begin
                dvc_axi4_set_ARUSER_from_SystemVerilog( _interface_ref, ARUSER );
            end
        end
    end

    always @(posedge ARREADY_changed or posedge _check_t0_values )
    begin
        while (ARREADY_changed == 1'b1)
        begin
            dvc_axi4_get_ARREADY_into_SystemVerilog( _interface_ref, internal_ARREADY ); // DPI call to imported task
            ARREADY_changed = 1'b0;
            #0 if ( ARREADY !== internal_ARREADY )
            begin
                dvc_axi4_set_ARREADY_from_SystemVerilog( _interface_ref, ARREADY );
            end
        end
    end

    always @(posedge RVALID_changed or posedge _check_t0_values )
    begin
        while (RVALID_changed == 1'b1)
        begin
            dvc_axi4_get_RVALID_into_SystemVerilog( _interface_ref, internal_RVALID ); // DPI call to imported task
            RVALID_changed = 1'b0;
            #0 if ( RVALID !== internal_RVALID )
            begin
                dvc_axi4_set_RVALID_from_SystemVerilog( _interface_ref, RVALID );
            end
        end
    end

    always @(posedge RDATA_changed or posedge _check_t0_values )
    begin
        while (RDATA_changed == 1'b1)
        begin
            dvc_axi4_get_RDATA_into_SystemVerilog( _interface_ref, internal_RDATA ); // DPI call to imported task
            RDATA_changed = 1'b0;
            #0 if ( RDATA !== internal_RDATA )
            begin
                dvc_axi4_set_RDATA_from_SystemVerilog( _interface_ref, RDATA );
            end
        end
    end

    always @(posedge RRESP_changed or posedge _check_t0_values )
    begin
        while (RRESP_changed == 1'b1)
        begin
            dvc_axi4_get_RRESP_into_SystemVerilog( _interface_ref, internal_RRESP ); // DPI call to imported task
            RRESP_changed = 1'b0;
            #0 if ( RRESP !== internal_RRESP )
            begin
                dvc_axi4_set_RRESP_from_SystemVerilog( _interface_ref, RRESP );
            end
        end
    end

    always @(posedge RLAST_changed or posedge _check_t0_values )
    begin
        while (RLAST_changed == 1'b1)
        begin
            dvc_axi4_get_RLAST_into_SystemVerilog( _interface_ref, internal_RLAST ); // DPI call to imported task
            RLAST_changed = 1'b0;
            #0 if ( RLAST !== internal_RLAST )
            begin
                dvc_axi4_set_RLAST_from_SystemVerilog( _interface_ref, RLAST );
            end
        end
    end

    always @(posedge RID_changed or posedge _check_t0_values )
    begin
        while (RID_changed == 1'b1)
        begin
            dvc_axi4_get_RID_into_SystemVerilog( _interface_ref, internal_RID ); // DPI call to imported task
            RID_changed = 1'b0;
            #0 if ( RID !== internal_RID )
            begin
                dvc_axi4_set_RID_from_SystemVerilog( _interface_ref, RID );
            end
        end
    end

    always @(posedge RUSER_changed or posedge _check_t0_values )
    begin
        while (RUSER_changed == 1'b1)
        begin
            dvc_axi4_get_RUSER_into_SystemVerilog( _interface_ref, internal_RUSER ); // DPI call to imported task
            RUSER_changed = 1'b0;
            #0 if ( RUSER !== internal_RUSER )
            begin
                dvc_axi4_set_RUSER_from_SystemVerilog( _interface_ref, RUSER );
            end
        end
    end

    always @(posedge RREADY_changed or posedge _check_t0_values )
    begin
        while (RREADY_changed == 1'b1)
        begin
            dvc_axi4_get_RREADY_into_SystemVerilog( _interface_ref, internal_RREADY ); // DPI call to imported task
            RREADY_changed = 1'b0;
            #0 if ( RREADY !== internal_RREADY )
            begin
                dvc_axi4_set_RREADY_from_SystemVerilog( _interface_ref, RREADY );
            end
        end
    end

    always @(posedge WVALID_changed or posedge _check_t0_values )
    begin
        while (WVALID_changed == 1'b1)
        begin
            dvc_axi4_get_WVALID_into_SystemVerilog( _interface_ref, internal_WVALID ); // DPI call to imported task
            WVALID_changed = 1'b0;
            #0 if ( WVALID !== internal_WVALID )
            begin
                dvc_axi4_set_WVALID_from_SystemVerilog( _interface_ref, WVALID );
            end
        end
    end

    always @(posedge WDATA_changed or posedge _check_t0_values )
    begin
        while (WDATA_changed == 1'b1)
        begin
            dvc_axi4_get_WDATA_into_SystemVerilog( _interface_ref, internal_WDATA ); // DPI call to imported task
            WDATA_changed = 1'b0;
            #0 if ( WDATA !== internal_WDATA )
            begin
                dvc_axi4_set_WDATA_from_SystemVerilog( _interface_ref, WDATA );
            end
        end
    end

    always @(posedge WSTRB_changed or posedge _check_t0_values )
    begin
        while (WSTRB_changed == 1'b1)
        begin
            dvc_axi4_get_WSTRB_into_SystemVerilog( _interface_ref, internal_WSTRB ); // DPI call to imported task
            WSTRB_changed = 1'b0;
            #0 if ( WSTRB !== internal_WSTRB )
            begin
                dvc_axi4_set_WSTRB_from_SystemVerilog( _interface_ref, WSTRB );
            end
        end
    end

    always @(posedge WLAST_changed or posedge _check_t0_values )
    begin
        while (WLAST_changed == 1'b1)
        begin
            dvc_axi4_get_WLAST_into_SystemVerilog( _interface_ref, internal_WLAST ); // DPI call to imported task
            WLAST_changed = 1'b0;
            #0 if ( WLAST !== internal_WLAST )
            begin
                dvc_axi4_set_WLAST_from_SystemVerilog( _interface_ref, WLAST );
            end
        end
    end

    always @(posedge WUSER_changed or posedge _check_t0_values )
    begin
        while (WUSER_changed == 1'b1)
        begin
            dvc_axi4_get_WUSER_into_SystemVerilog( _interface_ref, internal_WUSER ); // DPI call to imported task
            WUSER_changed = 1'b0;
            #0 if ( WUSER !== internal_WUSER )
            begin
                dvc_axi4_set_WUSER_from_SystemVerilog( _interface_ref, WUSER );
            end
        end
    end

    always @(posedge WREADY_changed or posedge _check_t0_values )
    begin
        while (WREADY_changed == 1'b1)
        begin
            dvc_axi4_get_WREADY_into_SystemVerilog( _interface_ref, internal_WREADY ); // DPI call to imported task
            WREADY_changed = 1'b0;
            #0 if ( WREADY !== internal_WREADY )
            begin
                dvc_axi4_set_WREADY_from_SystemVerilog( _interface_ref, WREADY );
            end
        end
    end

    always @(posedge BVALID_changed or posedge _check_t0_values )
    begin
        while (BVALID_changed == 1'b1)
        begin
            dvc_axi4_get_BVALID_into_SystemVerilog( _interface_ref, internal_BVALID ); // DPI call to imported task
            BVALID_changed = 1'b0;
            #0 if ( BVALID !== internal_BVALID )
            begin
                dvc_axi4_set_BVALID_from_SystemVerilog( _interface_ref, BVALID );
            end
        end
    end

    always @(posedge BRESP_changed or posedge _check_t0_values )
    begin
        while (BRESP_changed == 1'b1)
        begin
            dvc_axi4_get_BRESP_into_SystemVerilog( _interface_ref, internal_BRESP ); // DPI call to imported task
            BRESP_changed = 1'b0;
            #0 if ( BRESP !== internal_BRESP )
            begin
                dvc_axi4_set_BRESP_from_SystemVerilog( _interface_ref, BRESP );
            end
        end
    end

    always @(posedge BID_changed or posedge _check_t0_values )
    begin
        while (BID_changed == 1'b1)
        begin
            dvc_axi4_get_BID_into_SystemVerilog( _interface_ref, internal_BID ); // DPI call to imported task
            BID_changed = 1'b0;
            #0 if ( BID !== internal_BID )
            begin
                dvc_axi4_set_BID_from_SystemVerilog( _interface_ref, BID );
            end
        end
    end

    always @(posedge BUSER_changed or posedge _check_t0_values )
    begin
        while (BUSER_changed == 1'b1)
        begin
            dvc_axi4_get_BUSER_into_SystemVerilog( _interface_ref, internal_BUSER ); // DPI call to imported task
            BUSER_changed = 1'b0;
            #0 if ( BUSER !== internal_BUSER )
            begin
                dvc_axi4_set_BUSER_from_SystemVerilog( _interface_ref, BUSER );
            end
        end
    end

    always @(posedge BREADY_changed or posedge _check_t0_values )
    begin
        while (BREADY_changed == 1'b1)
        begin
            dvc_axi4_get_BREADY_into_SystemVerilog( _interface_ref, internal_BREADY ); // DPI call to imported task
            BREADY_changed = 1'b0;
            #0 if ( BREADY !== internal_BREADY )
            begin
                dvc_axi4_set_BREADY_from_SystemVerilog( _interface_ref, BREADY );
            end
        end
    end

    always @(posedge config_axi4lite_interface_changed or posedge _check_t0_values )
    begin
        if (config_axi4lite_interface_changed == 1'b1)
        begin
            dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog( _interface_ref, config_axi4lite_interface ); // DPI call to imported task
            config_axi4lite_interface_changed = 1'b0;
        end
    end

    always @(posedge config_interface_type_changed or posedge _check_t0_values )
    begin
        if (config_interface_type_changed == 1'b1)
        begin
            dvc_axi4_get_config_interface_type_into_SystemVerilog( _interface_ref, config_interface_type ); // DPI call to imported task
            config_interface_type_changed = 1'b0;
        end
    end

    always @(posedge config_enable_rlast_changed or posedge _check_t0_values )
    begin
        if (config_enable_rlast_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_rlast_into_SystemVerilog( _interface_ref, config_enable_rlast ); // DPI call to imported task
            config_enable_rlast_changed = 1'b0;
        end
    end

    always @(posedge config_enable_id_changed or posedge _check_t0_values )
    begin
        if (config_enable_id_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_id_into_SystemVerilog( _interface_ref, config_enable_id ); // DPI call to imported task
            config_enable_id_changed = 1'b0;
        end
    end

    always @(posedge config_enable_all_assertions_changed or posedge _check_t0_values )
    begin
        if (config_enable_all_assertions_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog( _interface_ref, config_enable_all_assertions ); // DPI call to imported task
            config_enable_all_assertions_changed = 1'b0;
        end
    end

    always @(posedge config_enable_assertion_changed or posedge _check_t0_values )
    begin
        if (config_enable_assertion_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_assertion_into_SystemVerilog( _interface_ref, config_enable_assertion ); // DPI call to imported task
            config_enable_assertion_changed = 1'b0;
        end
    end

    always @(posedge config_enable_user_sideband_changed or posedge _check_t0_values )
    begin
        if (config_enable_user_sideband_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog( _interface_ref, config_enable_user_sideband ); // DPI call to imported task
            config_enable_user_sideband_changed = 1'b0;
        end
    end

    always @(posedge config_enable_slave_exclusive_changed or posedge _check_t0_values )
    begin
        if (config_enable_slave_exclusive_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog( _interface_ref, config_enable_slave_exclusive ); // DPI call to imported task
            config_enable_slave_exclusive_changed = 1'b0;
        end
    end

    always @(posedge config_read_data_reordering_depth_changed or posedge _check_t0_values )
    begin
        if (config_read_data_reordering_depth_changed == 1'b1)
        begin
            dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog( _interface_ref, config_read_data_reordering_depth ); // DPI call to imported task
            config_read_data_reordering_depth_changed = 1'b0;
        end
    end

    always @(posedge config_slave_start_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_start_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_start_addr_into_SystemVerilog( _interface_ref, config_slave_start_addr ); // DPI call to imported task
            config_slave_start_addr_changed = 1'b0;
        end
    end

    always @(posedge config_slave_end_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_end_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_end_addr_into_SystemVerilog( _interface_ref, config_slave_end_addr ); // DPI call to imported task
            config_slave_end_addr_changed = 1'b0;
        end
    end

    always @(posedge config_max_transaction_time_factor_changed or posedge _check_t0_values )
    begin
        if (config_max_transaction_time_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog( _interface_ref, config_max_transaction_time_factor ); // DPI call to imported task
            config_max_transaction_time_factor_changed = 1'b0;
        end
    end

    always @(posedge config_burst_timeout_factor_changed or posedge _check_t0_values )
    begin
        if (config_burst_timeout_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog( _interface_ref, config_burst_timeout_factor ); // DPI call to imported task
            config_burst_timeout_factor_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_AWVALID_assertion_to_AWREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_AWVALID_assertion_to_AWREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog( _interface_ref, config_max_latency_AWVALID_assertion_to_AWREADY ); // DPI call to imported task
            config_max_latency_AWVALID_assertion_to_AWREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_ARVALID_assertion_to_ARREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_ARVALID_assertion_to_ARREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog( _interface_ref, config_max_latency_ARVALID_assertion_to_ARREADY ); // DPI call to imported task
            config_max_latency_ARVALID_assertion_to_ARREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_RVALID_assertion_to_RREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_RVALID_assertion_to_RREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog( _interface_ref, config_max_latency_RVALID_assertion_to_RREADY ); // DPI call to imported task
            config_max_latency_RVALID_assertion_to_RREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_BVALID_assertion_to_BREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_BVALID_assertion_to_BREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog( _interface_ref, config_max_latency_BVALID_assertion_to_BREADY ); // DPI call to imported task
            config_max_latency_BVALID_assertion_to_BREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_WVALID_assertion_to_WREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_WVALID_assertion_to_WREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog( _interface_ref, config_max_latency_WVALID_assertion_to_WREADY ); // DPI call to imported task
            config_max_latency_WVALID_assertion_to_WREADY_changed = 1'b0;
        end
    end

    always @(posedge config_enable_qos_changed or posedge _check_t0_values )
    begin
        if (config_enable_qos_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_qos_into_SystemVerilog( _interface_ref, config_enable_qos ); // DPI call to imported task
            config_enable_qos_changed = 1'b0;
        end
    end

    always @(posedge config_setup_time_changed or posedge _check_t0_values )
    begin
        if (config_setup_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_setup_time_into_SystemVerilog( _interface_ref, config_setup_time ); // DPI call to imported task
            config_setup_time_changed = 1'b0;
        end
    end

    always @(posedge config_hold_time_changed or posedge _check_t0_values )
    begin
        if (config_hold_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_hold_time_into_SystemVerilog( _interface_ref, config_hold_time ); // DPI call to imported task
            config_hold_time_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_reads_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_reads_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog( _interface_ref, config_num_max_outstanding_reads ); // DPI call to imported task
            config_num_max_outstanding_reads_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_writes_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_writes_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog( _interface_ref, config_num_max_outstanding_writes ); // DPI call to imported task
            config_num_max_outstanding_writes_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_wr_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_wr_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog( _interface_ref, config_max_outstanding_wr ); // DPI call to imported task
            config_max_outstanding_wr_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rd_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rd_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog( _interface_ref, config_max_outstanding_rd ); // DPI call to imported task
            config_max_outstanding_rd_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rw_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rw_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog( _interface_ref, config_max_outstanding_rw ); // DPI call to imported task
            config_max_outstanding_rw_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_wr_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_wr_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_wr_phase ); // DPI call to imported task
            config_num_outstanding_wr_phase_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_rd_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_rd_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_rd_phase ); // DPI call to imported task
            config_num_outstanding_rd_phase_changed = 1'b0;
        end
    end

    always @(posedge config_is_issuing_changed or posedge _check_t0_values )
    begin
        if (config_is_issuing_changed == 1'b1)
        begin
            dvc_axi4_get_config_is_issuing_into_SystemVerilog( _interface_ref, config_is_issuing ); // DPI call to imported task
            config_is_issuing_changed = 1'b0;
        end
    end



    // Sparse array of blocking control events
    event block_control[] = new[100];

    // Unblocks a blocked clock control thread by id
    function void unblock( int unsigned id );
    begin
        -> block_control[id];
    end
    endfunction
    export "DPI-C" dvc_axi4_unblock_SystemVerilog = function unblock;

    // Blocks a blocked clock control thread by id
    task automatic block( int unsigned id );
    begin
        if (id >= block_control.size())
        begin
            int newsize  = ( (  id / 100 ) + 1 ) * 100;
            block_control = new[newsize](block_control);
        end
        @ block_control[id];
    end
    endtask
    export "DPI-C" dvc_axi4_block_SystemVerilog = task block;


    function int is_call_back_registered(int cb_name);
        case( axi4_call_back_e'(cb_name) )
          AXI4_REPORTER_CB:
          begin
              return ( endPoint.size() > 0 ) ? 1 : 0;
          end
        endcase
    endfunction

    //--------------------------------------------------------------------------------
    // Task which blocks and outputs an error if the interface has not initialized properly
    //--------------------------------------------------------------------------------

    task _initialized();
        if (_interface_ref == 0)
        begin
            $display("Error: %m - Questa Verification IP failed to initialise. Please check questa_mvc.log for details");
            wait(_interface_ref!=0);
        end
    endtask

endinterface

`elsif XCELIUM
// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

`ifndef QVIP_MIX_AND_MATCH
(* cy_so="libaxi4_IN_SystemVerilog_MTI_full_DVC" *)
(* on_lib_load="axi4_IN_SystemVerilog_load" *)
`endif

interface mgc_common_axi4 #( int AXI4_ADDRESS_WIDTH = 64, int AXI4_RDATA_WIDTH = 1024, int AXI4_WDATA_WIDTH = 1024, int AXI4_ID_WIDTH = 18, int AXI4_USER_WIDTH = 8, int AXI4_REGION_MAP_SIZE = 16 )
    (input wire iACLK, input wire iARESETn);

import QUESTA_MVC::questa_mvc_reporter;
import QUESTA_MVC::questa_mvc_item_comms_semantic;
import QUESTA_MVC::questa_mvc_edge;
import QUESTA_MVC::QUESTA_MVC_POSEDGE;
import QUESTA_MVC::QUESTA_MVC_NEGEDGE;
import QUESTA_MVC::QUESTA_MVC_ANYEDGE;
import QUESTA_MVC::QUESTA_MVC_0_TO_1_EDGE;
import QUESTA_MVC::QUESTA_MVC_1_TO_0_EDGE;




    //-------------------------------------------------------------------------
    // Private wires
    //-------------------------------------------------------------------------
    wire ACLK;
    wire ARESETn;
    wire AWVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR;
    wire [2:0] AWPROT;
    wire [3:0] AWREGION;
    wire [7:0] AWLEN;
    wire [3:0] AWSIZE;
    wire [1:0] AWBURST;
    wire AWLOCK;
    wire [3:0] AWCACHE;
    wire [3:0] AWQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] AWID;
    wire [((AXI4_USER_WIDTH) - 1):0] AWUSER;
    wire AWREADY;
    wire ARVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR;
    wire [2:0] ARPROT;
    wire [3:0] ARREGION;
    wire [7:0] ARLEN;
    wire [3:0] ARSIZE;
    wire [1:0] ARBURST;
    wire ARLOCK;
    wire [3:0] ARCACHE;
    wire [3:0] ARQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] ARID;
    wire [((AXI4_USER_WIDTH) - 1):0] ARUSER;
    wire ARREADY;
    wire RVALID;
    wire [((AXI4_RDATA_WIDTH) - 1):0] RDATA;
    wire [1:0] RRESP;
    wire RLAST;
    wire [((AXI4_ID_WIDTH) - 1):0] RID;
    wire [((AXI4_USER_WIDTH) - 1):0] RUSER;
    wire RREADY;
    wire WVALID;
    wire [((AXI4_WDATA_WIDTH) - 1):0] WDATA;
    wire [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB;
    wire WLAST;
    wire [((AXI4_USER_WIDTH) - 1):0] WUSER;
    wire WREADY;
    wire BVALID;
    wire [1:0] BRESP;
    wire [((AXI4_ID_WIDTH) - 1):0] BID;
    wire [((AXI4_USER_WIDTH) - 1):0] BUSER;
    wire BREADY;



    // Propagate global signals onto interface wires
    assign ACLK = iACLK;
    assign ARESETn = iARESETn;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Enables
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_rlast
    //
    //  
    // Enables use of optional RLAST signal.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable optional RLAST signal"
    bit config_enable_rlast;

    // Variable: config_enable_id
    //
    //  
    // Enables use of optional AWID, BID, ARID, RID signals.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable ID signals"
    bit config_enable_id;

    // Variable: config_enable_all_assertions
    //
    // 
    // Enables all protocol assertions. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable all protocol assertions"
    bit config_enable_all_assertions;

    // Variable: config_enable_assertion
    //
    // 
    // Enables individual protocol assertion.
    // This variable controls whether specific assertion within QVIP (of type <axi4_assertion_e>) is enabled or disabled.
    // Individual assertion can be disabled as follows:-
    // //-----------------------------------------------------------------------
    // // < BFM interface>.config_enable_assertion[<name of assertion>] = 1'b0;
    // //-----------------------------------------------------------------------
    // 
    // For example, the assertion AXI4_READ_DATA_UNKN is disabled as follows:
    // <bfm>.config_enable_assertion[AXI4_READ_DATA_UNKN] =  1'b0; 
    // 
    // Here bfm is the AXI4 interface instance name for which the assertion to be disabled. 
    // 
    // Default: All assertions are enabled
    //   
    // 
    //
    // mentor configurator specification name "Enable individual protocol assertion" 
    bit [255:0] config_enable_assertion;

    // Variable: config_enable_user_sideband
    //
    // 
    // Enables user sideband signals. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable user sideband signals"
    bit config_enable_user_sideband;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Slave behavior control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_slave_exclusive
    //
    // 
    // Enables exclusive transactions support for slave.
    // If disabled, every exclusive read/write returns an OKAY response,
    // and exclusive write updates memory. 
    // 
    // Default: 1  
    // 
    //
    // mentor configurator specification name "Enable exclusive transaction support"
    bit config_enable_slave_exclusive;

    // Variable: config_read_data_reordering_depth
    //
    // 
    // Sets the maximum number of different read transaction addresses for which read 
    // data(response) can be sent in any order from slave. 
    // 
    // Default: 2 ** AXI4_ID_WIDTH
    // 
    //
    // mentor configurator specification name "Read data reordering depth"
    int unsigned config_read_data_reordering_depth;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Timeout control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_max_transaction_time_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) for any complete read or write transaction, which
    // includes time period for all individual phases of transaction. 
    // 
    // Default: 100000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout for complete Read/Write transaction"
    int unsigned config_max_transaction_time_factor;

    // Variable: config_burst_timeout_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) between individual phases of a transaction. 
    // 
    // Default: 10000 clock cycles 
    // 
    //
    // mentor configurator specification name "Burst timeout between individual phases of a transaction"
    int unsigned config_burst_timeout_factor;

    // Variable: config_max_latency_AWVALID_assertion_to_AWREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of AWVALID to assertion of AWREADY.
    // An error message AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID is generated if AWREADY is not asserted
    // after assertion of AWVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from AWVALID to AWREADY assertion"
    int unsigned config_max_latency_AWVALID_assertion_to_AWREADY;

    // Variable: config_max_latency_ARVALID_assertion_to_ARREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of ARVALID to assertion of ARREADY.
    // An error message AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID is generated if ARREADY is not asserted
    // after assertion of ARVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from ARVALID to ARREADY assertion"
    int unsigned config_max_latency_ARVALID_assertion_to_ARREADY;

    // Variable: config_max_latency_RVALID_assertion_to_RREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of RVALID to assertion of RREADY.
    // An error message AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID is generated if RREADY is not asserted
    // after assertion of RVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from RVALID to RREADY assertion"
    int unsigned config_max_latency_RVALID_assertion_to_RREADY;

    // Variable: config_max_latency_BVALID_assertion_to_BREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of BVALID to assertion of BREADY.
    // An error message AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID is generated if BREADY is not asserted
    // after assertion of BVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from BVALID to BREADY assertion"
    int unsigned config_max_latency_BVALID_assertion_to_BREADY;

    // Variable: config_max_latency_WVALID_assertion_to_WREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of WVALID to assertion of WREADY.
    // An error message AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID is generated if WREADY is not asserted
    // after assertion of WVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from WVALID to WREADY assertion"
    int unsigned config_max_latency_WVALID_assertion_to_WREADY;

    // 
    // //-----------------------------------------------------------------------------
    // Group: QOS and Region control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_qos
    //
    //  
    // Enables quality-of-service support. 
    // 
    // Default: 1 
    // 
    //
    // mentor configurator specification name "Enable quality of service support"
    bit config_enable_qos;

    // Variable: config_setup_time
    //
    // 
    // Sets number of simulation time units from the setup time to the active 
    // clock edge of ACLK. The setup time will always be less than the time period
    // of the clock. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_setup_time;

    // Variable: config_hold_time
    //
    // 
    // Sets number of simulation time units from the hold time to the active 
    // clock edge of ACLK. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_hold_time;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Master Outstanding Control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_num_max_outstanding_reads
    //
    // 
    // Configures maximum number of read outstanding transfers allowed on the bus.
    // 
    // Default: -1
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding reads"
    int config_num_max_outstanding_reads;

    // Variable: config_num_max_outstanding_writes
    //
    //                                                                           
    // Configures maximum number of write outstanding transfers allowed on the bus. 
    //                                                                              
    // Default: -1                                                                 
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding writes"
    int config_num_max_outstanding_writes;

    // Variable: config_max_outstanding_wr
    //
    // Configures maximum possible outstanding Write transactions
    //
    int config_max_outstanding_wr;

    // Variable: config_max_outstanding_rd
    //
    // Configures maximum possible outstanding Read transactions
    //
    int config_max_outstanding_rd;

    // Variable: config_max_outstanding_rw
    //
    // Configures maximum possible outstanding Combined (Read and Write) transactions
    //
    int config_max_outstanding_rw;

    // Variable: config_num_outstanding_wr_phase
    //
    // Provides current number of outstanding Write address phases 
    //
    int config_num_outstanding_wr_phase;

    // Variable: config_num_outstanding_rd_phase
    //
    // Provides current number of outstanding Read address phases
    //
    int config_num_outstanding_rd_phase;

    // Variable: config_is_issuing
    //
    // Enables Master component to use "config_max_outstanding_wr/config_max_outstanding_rd/config_max_outstanding_rw" variables for transaction issuing capability when set to true
    //
    bit config_is_issuing;


    //-------------------------------------------------------------------------
    // Deprecated variables - writing to these variables will cause a warning to be issued.
    //-------------------------------------------------------------------------
    bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr;


    //-------------------------------------------------------------------------
    // Private variables
    //-------------------------------------------------------------------------
    bit config_axi4lite_interface;
    axi4_interface_type_e config_interface_type;


    //-------------------------------------------------------------------------
    // Initialise local variables
    //-------------------------------------------------------------------------
    initial
    begin
         config_axi4lite_tr_id = 0;
    end

    //------------------------------------------------------------------------------
    // Group:- Interface ends
    //------------------------------------------------------------------------------
    //
    longint axi4_master_end;

    // Function:- get_axi4_master_end
    //
    // Returns a handle to the <master> end of this instance of the <axi4> interface.

    function longint get_axi4_master_end();
        return axi4_master_end;
    endfunction

    longint axi4_slave_end;

    // Function:- get_axi4_slave_end
    //
    // Returns a handle to the <slave> end of this instance of the <axi4> interface.

    function longint get_axi4_slave_end();
        return axi4_slave_end;
    endfunction

    longint axi4__monitor_end;

    // Function:- get_axi4__monitor_end
    //
    // Returns a handle to the <_monitor> end of this instance of the <axi4> interface.

    function longint get_axi4__monitor_end();
        return axi4__monitor_end;
    endfunction


    // Group:- Abstraction Levels
    // 
    // These functions are used set or get the abstraction levels of an interface end.
    // See <Abstraction Levels of Interface Ends> for more details on the meaning of
    // TLM or WLM connected and the valid combinations.


    //-------------------------------------------------------------------------
    // Function:- axi4_set_master_abstraction_level
    //
    //     Function to set whether the <master> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_master_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_master_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <master> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_master_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_set_slave_abstraction_level
    //
    //     Function to set whether the <slave> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_slave_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_slave_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <slave> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_slave_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    import "DPI-C" context function longint dvc_axi4_initialise_SystemVerilog
    (
        int     usage_code,
        string  iface_version,
        longint generate_ver,
        int     qvip_mix_and_match,
        output longint master_end,
        output longint slave_end,
        output longint _monitor_end,
        input int AXI4_ADDRESS_WIDTH,
        input int AXI4_RDATA_WIDTH,
        input int AXI4_WDATA_WIDTH,
        input int AXI4_ID_WIDTH,
        input int AXI4_USER_WIDTH,
        input int AXI4_REGION_MAP_SIZE
    );

    `ifndef MVC_axi4_VERSION
    `define MVC_axi4_VERSION ""
    `endif

    // Handle to the linkage
    (* elab_init *) longint _interface_ref =
                                dvc_axi4_initialise_SystemVerilog
                                (
                                    18102076,
                                    `MVC_axi4_VERSION,
                                    20220701,
                                    `ifdef QVIP_MIX_AND_MATCH
                                    1
                                    `else
                                    0
                                    `endif
                                    ,
                                    axi4_master_end,
                                    axi4_slave_end,
                                    axi4__monitor_end,
                                    AXI4_ADDRESS_WIDTH,
                                    AXI4_RDATA_WIDTH,
                                    AXI4_WDATA_WIDTH,
                                    AXI4_ID_WIDTH,
                                    AXI4_USER_WIDTH,
                                    AXI4_REGION_MAP_SIZE
                                ); // DPI call to create transactor (called at
                                     // elaboration time as initialiser)

    questa_mvc_reporter endPoint[longint];
    export "DPI-C" dvc_axi4_process_reports = function process_reports;
    function void process_reports( input longint recipient, input string category, input string objectName, input string instanceName, input string error_no, input string severity, input string mess );
        if( endPoint.exists(recipient) )
            endPoint[recipient].report_message( category, "dvc_axi4", 0, objectName, instanceName, error_no, severity, mess );
        else
            $error("Invalid recipient (%d) when processing report", recipient);
    endfunction

    import "DPI-C" context dvc_axi4_register_end_point = function void axi4_register_end_point( input longint iface_ref, input longint as_end, input string name );

    // A function for registering a reporter to capture any reports coming from as_end
    function automatic void register_end_point( input longint as_end, input questa_mvc_reporter rep = null );
        if ( rep != null )
        begin
            if ( ( rep.name == "" ) || ( rep.name == "NULL" ) )
            begin
                $display("Error: %m: Reporter passed to register_end_point has a reserved name. Neither an empty string nor the string 'NULL' can be used.");
            end
            else
            begin
                axi4_register_end_point( _interface_ref, as_end, rep.name );
                endPoint[as_end] = rep;
            end
        end
        else
        begin
            axi4_register_end_point( _interface_ref, as_end, "NULL" );
            endPoint.delete( as_end );
        end
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group:- Registering Reports
    //
    //
    // The following methods are used to register a custom reporting object as
    // described in the Questa Verification IP base library section, <Customizing Error-Reporting>.
    // 
    //-------------------------------------------------------------------------

    function void register_interface_reporter( input questa_mvc_reporter _rep = null );
        register_end_point( _interface_ref, _rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_master_reporter
    //
    // Function used to register a reporter for the <master> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the master end.
    //
    function void register_master_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_master_end, rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_slave_reporter
    //
    // Function used to register a reporter for the <slave> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the slave end.
    //
    function void register_slave_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_slave_end, rep );
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_mvc_reporter
    //
    // Function used to get the handle for an already registered reporter.
    // By default returns the reporter associated with this interface. If an end handle is passed,
    // then the reporter for that end.
    //
    // Arguments:
    //    as_end - Optional, a handle for an end of this interface.
    //
    function questa_mvc_reporter get_mvc_reporter
    (
        input longint as_end = 0
    );
        if ( as_end == 0 )
            as_end = _interface_ref;
        if ( endPoint.exists( as_end ) )
            return endPoint[ as_end ];
        else
            return null;
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group:- BFM Utility/Convenience Methods
    //
    // This is the group of utility functions provided by the QVIP BFM to
    // communicate from the SV world to the QVIP BFM.
    // This set of APIs can be used to either get status/statistics
    // information from the BFM or to set values in a particular database 
    // in the BFM. Please refer to individual functions for more information.
    //
    //-------------------------------------------------------------------------

    export "DPI-C" dvc_axi4_get_temp_static_fn_get_raddr_phase_info_addr = function axi4_get_temp_static_fn_get_raddr_phase_info_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_get_raddr_phase_info_addr = function axi4_set_temp_static_fn_get_raddr_phase_info_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_get_raddr_phase_info_addr;
    function void axi4_get_temp_static_fn_get_raddr_phase_info_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_get_raddr_phase_info_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_get_raddr_phase_info_addr( input int _d1, input bit  _value );
        temp_static_fn_get_raddr_phase_info_addr[_d1] = _value;
    endfunction
    function automatic void fn_get_raddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output int num_beats_rcvd
    );
         fn_get_raddr_phase_info_C(valid,id,num_beats_rcvd);
        addr = temp_static_fn_get_raddr_phase_info_addr;
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_get_waddr_phase_info_addr = function axi4_get_temp_static_fn_get_waddr_phase_info_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_get_waddr_phase_info_addr = function axi4_set_temp_static_fn_get_waddr_phase_info_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_get_waddr_phase_info_addr;
    function void axi4_get_temp_static_fn_get_waddr_phase_info_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_get_waddr_phase_info_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_get_waddr_phase_info_addr( input int _d1, input bit  _value );
        temp_static_fn_get_waddr_phase_info_addr[_d1] = _value;
    endfunction
    function automatic void fn_get_waddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr
    );
         fn_get_waddr_phase_info_C(valid,id);
        addr = temp_static_fn_get_waddr_phase_info_addr;
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_add_addr_map_entry_addr = function axi4_get_temp_static_fn_add_addr_map_entry_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_add_addr_map_entry_addr = function axi4_set_temp_static_fn_add_addr_map_entry_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_add_addr_map_entry_addr;
    function void axi4_get_temp_static_fn_add_addr_map_entry_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_add_addr_map_entry_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_add_addr_map_entry_addr( input int _d1, input bit  _value );
        temp_static_fn_add_addr_map_entry_addr[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_add_addr_map_entry
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets address map fields inside internal BFM. 
    // 
    // Inputs:
    // region - Region name 
    // addr - Start address of region
    // size - Size of address region
    function automatic void fn_add_addr_map_entry
    (
        input string region,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input longint unsigned size
    );
        temp_static_fn_add_addr_map_entry_addr = addr;
         fn_add_addr_map_entry_C(region,size);
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_add_wr_delay_data2data = function axi4_get_temp_static_fn_add_wr_delay_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_add_wr_delay_data2data = function axi4_set_temp_static_fn_add_wr_delay_data2data;
    int unsigned temp_static_fn_add_wr_delay_data2data[];
    function void axi4_get_temp_static_fn_add_wr_delay_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_add_wr_delay_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_add_wr_delay_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_add_wr_delay_data2data[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_add_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets delay values for write address-data pair and 
    // between write data beats, initiated by master end. 
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_add_wr_delay
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        const ref int unsigned data2data[]
    );
        int tmp_data2data_DIMS0;
        tmp_data2data_DIMS0 = data2data.size();
        temp_static_fn_add_wr_delay_data2data = data2data;
         fn_add_wr_delay_C(region,id,addr2data,tmp_data2data_DIMS0);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_delete_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function deletes delay values for a region-id pair
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_delete_wr_delay
    (
        input string region,
        input bit [17:0] id
    );
         fn_delete_wr_delay_C(region,id);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_drive_signals
    //-------------------------------------------------------------------------
    // A function to drive the required value by the user on the desired signal when VALID is 0. 
    //   It takes the name of the signal and value to be driven on that signal as an input:
    // 
    //   Note - In case, the user wants to retain the same value on the signal (irrespective of Valid signal), 
    //   the value needs to be passed as -1.
    function automatic void fn_drive_signals
    (
        input string signal,
        input longint value
    );
         fn_drive_signals_C(signal,value);
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_set_address_map_entry_start_addr = function axi4_get_temp_static_fn_set_address_map_entry_start_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_address_map_entry_start_addr = function axi4_set_temp_static_fn_set_address_map_entry_start_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_set_address_map_entry_start_addr;
    function void axi4_get_temp_static_fn_set_address_map_entry_start_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_set_address_map_entry_start_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_address_map_entry_start_addr( input int _d1, input bit  _value );
        temp_static_fn_set_address_map_entry_start_addr[_d1] = _value;
    endfunction
    export "DPI-C" dvc_axi4_get_temp_static_fn_set_address_map_entry_end_addr = function axi4_get_temp_static_fn_set_address_map_entry_end_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_address_map_entry_end_addr = function axi4_set_temp_static_fn_set_address_map_entry_end_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_set_address_map_entry_end_addr;
    function void axi4_get_temp_static_fn_set_address_map_entry_end_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_set_address_map_entry_end_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_address_map_entry_end_addr( input int _d1, input bit  _value );
        temp_static_fn_set_address_map_entry_end_addr[_d1] = _value;
    endfunction
    function automatic void fn_set_address_map_entry
    (
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] start_addr,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] end_addr
    );
        temp_static_fn_set_address_map_entry_start_addr = start_addr;
        temp_static_fn_set_address_map_entry_end_addr = end_addr;
         fn_set_address_map_entry_C();
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_set_wr_def_delays_min_data2data = function axi4_get_temp_static_fn_set_wr_def_delays_min_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_wr_def_delays_min_data2data = function axi4_set_temp_static_fn_set_wr_def_delays_min_data2data;
    int unsigned temp_static_fn_set_wr_def_delays_min_data2data[];
    function void axi4_get_temp_static_fn_set_wr_def_delays_min_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_set_wr_def_delays_min_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_wr_def_delays_min_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_set_wr_def_delays_min_data2data[_d1] = _value;
    endfunction
    export "DPI-C" dvc_axi4_get_temp_static_fn_set_wr_def_delays_max_data2data = function axi4_get_temp_static_fn_set_wr_def_delays_max_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_wr_def_delays_max_data2data = function axi4_set_temp_static_fn_set_wr_def_delays_max_data2data;
    int unsigned temp_static_fn_set_wr_def_delays_max_data2data[];
    function void axi4_get_temp_static_fn_set_wr_def_delays_max_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_set_wr_def_delays_max_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_wr_def_delays_max_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_set_wr_def_delays_max_data2data[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_set_wr_def_delays
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets default delay values for write address-data pair and data beats 
    // initiated by master end between input max. and min. values. In case non-randomized 
    // default value is desired, user can set max. and min. to same value.
    // 
    // Inputs:
    // min_addr2data - Minimum value of default delays to be inserted between write address and data phase.
    // min_data2data - Minimum value of default delays to be inserted between data beats
    // max_addr2data - Maximum value of default delays to be inserted between write address and data phase.
    // max_data2data - Maximum value of default delays to be inserted between data beats
    function automatic void fn_set_wr_def_delays
    (
        input int unsigned min_addr2data,
        const ref int unsigned min_data2data[],
        input int unsigned max_addr2data,
        const ref int unsigned max_data2data[]
    );
        int tmp_min_data2data_DIMS0;
        int tmp_max_data2data_DIMS0;
        tmp_min_data2data_DIMS0 = min_data2data.size();
        tmp_max_data2data_DIMS0 = max_data2data.size();
        temp_static_fn_set_wr_def_delays_min_data2data = min_data2data;
        temp_static_fn_set_wr_def_delays_max_data2data = max_data2data;
         fn_set_wr_def_delays_C(min_addr2data,tmp_min_data2data_DIMS0,max_addr2data,tmp_max_data2data_DIMS0);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_print_perf_stats
    //-------------------------------------------------------------------------
    // API used to display the performance statistics of the current simulation. This
    // API shall be called only if axi4_vip_config variable 'cfg.agent_cfg.en_perf_stats.enable' has been configured to '1'.
    // 
    // Every statistics has 4 basic units:
    // minimum - Tells about the minimum metric within sliding window  
    // maximum - Tells about the maximum metric within sliding window 
    // mean    - Tells about the mean metric within a sliding window
    // sw      - Stores the calculation of the metrics over sliding window. 
    //           Statistics are collected for each of the sliding window segments 
    //           individually and then the actual statistic is calculated based 
    //           on all of the segments
    // 
    // The statistics are calculated for basic categories -
    // 
    // bus_usage: Stores the bus occupancy(as a percentage out of 100) for
    //   - Read transactions
    //   - Write transactions
    //   - Read Data burst
    //   - Write Data burst
    // See also, <axi4_bus_usage_s>
    // 
    // bandwidth: Stores the bandwidth value for
    //   - Read transactions
    //   - Write transactions
    // See also, <axi4_bandwidth_s>
    // 
    // latency: Stores the interphase delays as per the following categories
    //   - Write address to next write address delay
    //   - Write address to write data delay
    //   - Write data to write response delay
    //   - Read address to next read address delay
    //   - Read address to read data delay
    // See also, <axi4_latency_s>
    // 
    // waits: Stores the VALID-READY waits that are involved in each of the phases
    //   - Write address
    //   - Write data
    //   - Write response
    //   - Read address
    //   - Read data
    // See also, <axi4_waits_s>
    function automatic void fn_print_perf_stats
    (
    );
         fn_print_perf_stats_C();
    endfunction

    // Declare user visible wires variables, for non-continuous assignments.
    logic m_ACLK = 'z;
    logic m_ARESETn = 'z;
    logic m_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_AWADDR = 'z;
    logic [2:0] m_AWPROT = 'z;
    logic [3:0] m_AWREGION = 'z;
    logic [7:0] m_AWLEN = 'z;
    logic [3:0] m_AWSIZE = 'z;
    logic [1:0] m_AWBURST = 'z;
    logic m_AWLOCK = 'z;
    logic [3:0] m_AWCACHE = 'z;
    logic [3:0] m_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_AWUSER = 'z;
    logic m_AWREADY = 'z;
    logic m_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_ARADDR = 'z;
    logic [2:0] m_ARPROT = 'z;
    logic [3:0] m_ARREGION = 'z;
    logic [7:0] m_ARLEN = 'z;
    logic [3:0] m_ARSIZE = 'z;
    logic [1:0] m_ARBURST = 'z;
    logic m_ARLOCK = 'z;
    logic [3:0] m_ARCACHE = 'z;
    logic [3:0] m_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_ARUSER = 'z;
    logic m_ARREADY = 'z;
    logic m_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] m_RDATA = 'z;
    logic [1:0] m_RRESP = 'z;
    logic m_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_RUSER = 'z;
    logic m_RREADY = 'z;
    logic m_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] m_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] m_WSTRB = 'z;
    logic m_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_WUSER = 'z;
    logic m_WREADY = 'z;
    logic m_BVALID = 'z;
    logic [1:0] m_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_BUSER = 'z;
    logic m_BREADY = 'z;

    // Forces a sweep through the wire change checkers at time 0 to get around
    // process kick-off order unknowns
    bit _check_t0_values;
    always_comb _check_t0_values = 1;

    // handle control
    longint last_start_time = 0;

    longint last_end_time = 0;

    export "DPI-C" dvc_axi4_set_start_end_times = function set_start_end_times;

    function void set_start_end_times(longint _start, longint _end);
        last_start_time = _start;
        last_end_time = _end;
    endfunction


    function longint get_last_handle();
        return -1;
    endfunction


    function longint get_last_start_time();
        return last_start_time;
    endfunction


    function longint get_last_end_time();
        return last_end_time;
    endfunction


    //-------------------------------------------------------------------------
    // Tasks to wait for a number of specified edges on a wire
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_ACLK
    //     Wait for the specified change on wire <axi4::ACLK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ACLK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ACLK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ACLK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ACLK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ACLK === 0 );
                    @( ACLK );
                end
                while ( ACLK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ACLK === 1 );
                    @( ACLK );
                end
                while ( ACLK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARESETn
    //     Wait for the specified change on wire <axi4::ARESETn>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARESETn( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARESETn);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARESETn);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARESETn);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 0 );
                    @( ARESETn );
                end
                while ( ARESETn !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 1 );
                    @( ARESETn );
                end
                while ( ARESETn !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWVALID
    //     Wait for the specified change on wire <axi4::AWVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 0 );
                    @( AWVALID );
                end
                while ( AWVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 1 );
                    @( AWVALID );
                end
                while ( AWVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 0 );
                    @( AWADDR );
                end
                while ( AWADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 1 );
                    @( AWADDR );
                end
                while ( AWADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR_index1
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 0 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 1 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 0 );
                    @( AWPROT );
                end
                while ( AWPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 1 );
                    @( AWPROT );
                end
                while ( AWPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT_index1
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 0 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 1 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 0 );
                    @( AWREGION );
                end
                while ( AWREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 1 );
                    @( AWREGION );
                end
                while ( AWREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION_index1
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 0 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 1 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 0 );
                    @( AWLEN );
                end
                while ( AWLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 1 );
                    @( AWLEN );
                end
                while ( AWLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN_index1
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 0 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 1 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 0 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 1 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE_index1
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 0 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 1 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 0 );
                    @( AWBURST );
                end
                while ( AWBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 1 );
                    @( AWBURST );
                end
                while ( AWBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST_index1
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 0 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 1 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLOCK
    //     Wait for the specified change on wire <axi4::AWLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 0 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 1 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 0 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 1 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE_index1
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 0 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 1 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 0 );
                    @( AWQOS );
                end
                while ( AWQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 1 );
                    @( AWQOS );
                end
                while ( AWQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS_index1
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 0 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 1 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID === 0 );
                    @( AWID );
                end
                while ( AWID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID === 1 );
                    @( AWID );
                end
                while ( AWID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID_index1
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 0 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 1 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 0 );
                    @( AWUSER );
                end
                while ( AWUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 1 );
                    @( AWUSER );
                end
                while ( AWUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER_index1
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 0 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 1 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREADY
    //     Wait for the specified change on wire <axi4::AWREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 0 );
                    @( AWREADY );
                end
                while ( AWREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 1 );
                    @( AWREADY );
                end
                while ( AWREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARVALID
    //     Wait for the specified change on wire <axi4::ARVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 0 );
                    @( ARVALID );
                end
                while ( ARVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 1 );
                    @( ARVALID );
                end
                while ( ARVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 0 );
                    @( ARADDR );
                end
                while ( ARADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 1 );
                    @( ARADDR );
                end
                while ( ARADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR_index1
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 0 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 1 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 0 );
                    @( ARPROT );
                end
                while ( ARPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 1 );
                    @( ARPROT );
                end
                while ( ARPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT_index1
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 0 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 1 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 0 );
                    @( ARREGION );
                end
                while ( ARREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 1 );
                    @( ARREGION );
                end
                while ( ARREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION_index1
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 0 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 1 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 0 );
                    @( ARLEN );
                end
                while ( ARLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 1 );
                    @( ARLEN );
                end
                while ( ARLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN_index1
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 0 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 1 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 0 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 1 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE_index1
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 0 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 1 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 0 );
                    @( ARBURST );
                end
                while ( ARBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 1 );
                    @( ARBURST );
                end
                while ( ARBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST_index1
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 0 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 1 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLOCK
    //     Wait for the specified change on wire <axi4::ARLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 0 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 1 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 0 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 1 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE_index1
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 0 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 1 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 0 );
                    @( ARQOS );
                end
                while ( ARQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 1 );
                    @( ARQOS );
                end
                while ( ARQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS_index1
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 0 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 1 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID === 0 );
                    @( ARID );
                end
                while ( ARID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID === 1 );
                    @( ARID );
                end
                while ( ARID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID_index1
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 0 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 1 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 0 );
                    @( ARUSER );
                end
                while ( ARUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 1 );
                    @( ARUSER );
                end
                while ( ARUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER_index1
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 0 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 1 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREADY
    //     Wait for the specified change on wire <axi4::ARREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 0 );
                    @( ARREADY );
                end
                while ( ARREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 1 );
                    @( ARREADY );
                end
                while ( ARREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RVALID
    //     Wait for the specified change on wire <axi4::RVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RVALID === 0 );
                    @( RVALID );
                end
                while ( RVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RVALID === 1 );
                    @( RVALID );
                end
                while ( RVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA === 0 );
                    @( RDATA );
                end
                while ( RDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA === 1 );
                    @( RDATA );
                end
                while ( RDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA_index1
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 0 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 1 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP === 0 );
                    @( RRESP );
                end
                while ( RRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP === 1 );
                    @( RRESP );
                end
                while ( RRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP_index1
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 0 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 1 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RLAST
    //     Wait for the specified change on wire <axi4::RLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RLAST === 0 );
                    @( RLAST );
                end
                while ( RLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RLAST === 1 );
                    @( RLAST );
                end
                while ( RLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID === 0 );
                    @( RID );
                end
                while ( RID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID === 1 );
                    @( RID );
                end
                while ( RID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID_index1
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 0 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 1 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER === 0 );
                    @( RUSER );
                end
                while ( RUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER === 1 );
                    @( RUSER );
                end
                while ( RUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER_index1
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 0 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 1 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RREADY
    //     Wait for the specified change on wire <axi4::RREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RREADY === 0 );
                    @( RREADY );
                end
                while ( RREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RREADY === 1 );
                    @( RREADY );
                end
                while ( RREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WVALID
    //     Wait for the specified change on wire <axi4::WVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WVALID === 0 );
                    @( WVALID );
                end
                while ( WVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WVALID === 1 );
                    @( WVALID );
                end
                while ( WVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA === 0 );
                    @( WDATA );
                end
                while ( WDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA === 1 );
                    @( WDATA );
                end
                while ( WDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA_index1
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 0 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 1 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 0 );
                    @( WSTRB );
                end
                while ( WSTRB !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 1 );
                    @( WSTRB );
                end
                while ( WSTRB !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB_index1
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 0 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 1 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WLAST
    //     Wait for the specified change on wire <axi4::WLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WLAST === 0 );
                    @( WLAST );
                end
                while ( WLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WLAST === 1 );
                    @( WLAST );
                end
                while ( WLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER === 0 );
                    @( WUSER );
                end
                while ( WUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER === 1 );
                    @( WUSER );
                end
                while ( WUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER_index1
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 0 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 1 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WREADY
    //     Wait for the specified change on wire <axi4::WREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WREADY === 0 );
                    @( WREADY );
                end
                while ( WREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WREADY === 1 );
                    @( WREADY );
                end
                while ( WREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BVALID
    //     Wait for the specified change on wire <axi4::BVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BVALID === 0 );
                    @( BVALID );
                end
                while ( BVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BVALID === 1 );
                    @( BVALID );
                end
                while ( BVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP === 0 );
                    @( BRESP );
                end
                while ( BRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP === 1 );
                    @( BRESP );
                end
                while ( BRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP_index1
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 0 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 1 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID === 0 );
                    @( BID );
                end
                while ( BID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID === 1 );
                    @( BID );
                end
                while ( BID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID_index1
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 0 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 1 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER === 0 );
                    @( BUSER );
                end
                while ( BUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER === 1 );
                    @( BUSER );
                end
                while ( BUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER_index1
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 0 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 1 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BREADY
    //     Wait for the specified change on wire <axi4::BREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BREADY === 0 );
                    @( BREADY );
                end
                while ( BREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BREADY === 1 );
                    @( BREADY );
                end
                while ( BREADY !== 0 );
            end
        end
    endtask

    //-------------------------------------------------------------------------
    // Tasks/functions to set/get wires
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_ACLK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ACLK>.
    //
    // Parameters:
    //     ACLK_param - The value to set onto wire <ACLK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ACLK( logic ACLK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ACLK = ACLK_param;
        else
            m_ACLK <= ACLK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ACLK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ACLK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ACLK>.
    //
    function automatic logic get_ACLK(  );
        return ACLK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARESETn
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARESETn>.
    //
    // Parameters:
    //     ARESETn_param - The value to set onto wire <ARESETn>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARESETn( logic ARESETn_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARESETn = ARESETn_param;
        else
            m_ARESETn <= ARESETn_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARESETn
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARESETn>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARESETn>.
    //
    function automatic logic get_ARESETn(  );
        return ARESETn;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWVALID>.
    //
    // Parameters:
    //     AWVALID_param - The value to set onto wire <AWVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWVALID( logic AWVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWVALID = AWVALID_param;
        else
            m_AWVALID <= AWVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWVALID>.
    //
    function automatic logic get_AWVALID(  );
        return AWVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWADDR>.
    //
    // Parameters:
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR = AWADDR_param;
        else
            m_AWADDR <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR_index1( int _this_dot_1, logic  AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR[_this_dot_1] = AWADDR_param;
        else
            m_AWADDR[_this_dot_1] <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_AWADDR(  );
        return AWADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic   get_AWADDR_index1( int _this_dot_1 );
        return AWADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWPROT>.
    //
    // Parameters:
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT( logic [2:0] AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT = AWPROT_param;
        else
            m_AWPROT <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT_index1( int _this_dot_1, logic  AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT[_this_dot_1] = AWPROT_param;
        else
            m_AWPROT[_this_dot_1] <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic [2:0]  get_AWPROT(  );
        return AWPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic   get_AWPROT_index1( int _this_dot_1 );
        return AWPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREGION>.
    //
    // Parameters:
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION( logic [3:0] AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION = AWREGION_param;
        else
            m_AWREGION <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION_index1( int _this_dot_1, logic  AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION[_this_dot_1] = AWREGION_param;
        else
            m_AWREGION[_this_dot_1] <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic [3:0]  get_AWREGION(  );
        return AWREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic   get_AWREGION_index1( int _this_dot_1 );
        return AWREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLEN>.
    //
    // Parameters:
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN( logic [7:0] AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN = AWLEN_param;
        else
            m_AWLEN <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN_index1( int _this_dot_1, logic  AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN[_this_dot_1] = AWLEN_param;
        else
            m_AWLEN[_this_dot_1] <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic [7:0]  get_AWLEN(  );
        return AWLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic   get_AWLEN_index1( int _this_dot_1 );
        return AWLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWSIZE>.
    //
    // Parameters:
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE( logic [3:0] AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE = AWSIZE_param;
        else
            m_AWSIZE <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE_index1( int _this_dot_1, logic  AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE[_this_dot_1] = AWSIZE_param;
        else
            m_AWSIZE[_this_dot_1] <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic [3:0]  get_AWSIZE(  );
        return AWSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic   get_AWSIZE_index1( int _this_dot_1 );
        return AWSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWBURST>.
    //
    // Parameters:
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST( logic [1:0] AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST = AWBURST_param;
        else
            m_AWBURST <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST_index1( int _this_dot_1, logic  AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST[_this_dot_1] = AWBURST_param;
        else
            m_AWBURST[_this_dot_1] <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic [1:0]  get_AWBURST(  );
        return AWBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic   get_AWBURST_index1( int _this_dot_1 );
        return AWBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLOCK>.
    //
    // Parameters:
    //     AWLOCK_param - The value to set onto wire <AWLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLOCK( logic AWLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLOCK = AWLOCK_param;
        else
            m_AWLOCK <= AWLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLOCK>.
    //
    function automatic logic get_AWLOCK(  );
        return AWLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWCACHE>.
    //
    // Parameters:
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE( logic [3:0] AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE = AWCACHE_param;
        else
            m_AWCACHE <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE_index1( int _this_dot_1, logic  AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE[_this_dot_1] = AWCACHE_param;
        else
            m_AWCACHE[_this_dot_1] <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic [3:0]  get_AWCACHE(  );
        return AWCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic   get_AWCACHE_index1( int _this_dot_1 );
        return AWCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWQOS>.
    //
    // Parameters:
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS( logic [3:0] AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS = AWQOS_param;
        else
            m_AWQOS <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS_index1( int _this_dot_1, logic  AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS[_this_dot_1] = AWQOS_param;
        else
            m_AWQOS[_this_dot_1] <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic [3:0]  get_AWQOS(  );
        return AWQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic   get_AWQOS_index1( int _this_dot_1 );
        return AWQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWID>.
    //
    // Parameters:
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID( logic [((AXI4_ID_WIDTH) - 1):0] AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID = AWID_param;
        else
            m_AWID <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID_index1( int _this_dot_1, logic  AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID[_this_dot_1] = AWID_param;
        else
            m_AWID[_this_dot_1] <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_AWID(  );
        return AWID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic   get_AWID_index1( int _this_dot_1 );
        return AWID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWUSER>.
    //
    // Parameters:
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER( logic [((AXI4_USER_WIDTH) - 1):0] AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER = AWUSER_param;
        else
            m_AWUSER <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER_index1( int _this_dot_1, logic  AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER[_this_dot_1] = AWUSER_param;
        else
            m_AWUSER[_this_dot_1] <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_AWUSER(  );
        return AWUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic   get_AWUSER_index1( int _this_dot_1 );
        return AWUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREADY>.
    //
    // Parameters:
    //     AWREADY_param - The value to set onto wire <AWREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREADY( logic AWREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREADY = AWREADY_param;
        else
            m_AWREADY <= AWREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREADY>.
    //
    function automatic logic get_AWREADY(  );
        return AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARVALID>.
    //
    // Parameters:
    //     ARVALID_param - The value to set onto wire <ARVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARVALID( logic ARVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARVALID = ARVALID_param;
        else
            m_ARVALID <= ARVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARVALID>.
    //
    function automatic logic get_ARVALID(  );
        return ARVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARADDR>.
    //
    // Parameters:
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR = ARADDR_param;
        else
            m_ARADDR <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR_index1( int _this_dot_1, logic  ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR[_this_dot_1] = ARADDR_param;
        else
            m_ARADDR[_this_dot_1] <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_ARADDR(  );
        return ARADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic   get_ARADDR_index1( int _this_dot_1 );
        return ARADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARPROT>.
    //
    // Parameters:
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT( logic [2:0] ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT = ARPROT_param;
        else
            m_ARPROT <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT_index1( int _this_dot_1, logic  ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT[_this_dot_1] = ARPROT_param;
        else
            m_ARPROT[_this_dot_1] <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic [2:0]  get_ARPROT(  );
        return ARPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic   get_ARPROT_index1( int _this_dot_1 );
        return ARPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREGION>.
    //
    // Parameters:
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION( logic [3:0] ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION = ARREGION_param;
        else
            m_ARREGION <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION_index1( int _this_dot_1, logic  ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION[_this_dot_1] = ARREGION_param;
        else
            m_ARREGION[_this_dot_1] <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic [3:0]  get_ARREGION(  );
        return ARREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic   get_ARREGION_index1( int _this_dot_1 );
        return ARREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLEN>.
    //
    // Parameters:
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN( logic [7:0] ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN = ARLEN_param;
        else
            m_ARLEN <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN_index1( int _this_dot_1, logic  ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN[_this_dot_1] = ARLEN_param;
        else
            m_ARLEN[_this_dot_1] <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic [7:0]  get_ARLEN(  );
        return ARLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic   get_ARLEN_index1( int _this_dot_1 );
        return ARLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARSIZE>.
    //
    // Parameters:
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE( logic [3:0] ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE = ARSIZE_param;
        else
            m_ARSIZE <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE_index1( int _this_dot_1, logic  ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE[_this_dot_1] = ARSIZE_param;
        else
            m_ARSIZE[_this_dot_1] <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic [3:0]  get_ARSIZE(  );
        return ARSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic   get_ARSIZE_index1( int _this_dot_1 );
        return ARSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARBURST>.
    //
    // Parameters:
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST( logic [1:0] ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST = ARBURST_param;
        else
            m_ARBURST <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST_index1( int _this_dot_1, logic  ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST[_this_dot_1] = ARBURST_param;
        else
            m_ARBURST[_this_dot_1] <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic [1:0]  get_ARBURST(  );
        return ARBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic   get_ARBURST_index1( int _this_dot_1 );
        return ARBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLOCK>.
    //
    // Parameters:
    //     ARLOCK_param - The value to set onto wire <ARLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLOCK( logic ARLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLOCK = ARLOCK_param;
        else
            m_ARLOCK <= ARLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLOCK>.
    //
    function automatic logic get_ARLOCK(  );
        return ARLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARCACHE>.
    //
    // Parameters:
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE( logic [3:0] ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE = ARCACHE_param;
        else
            m_ARCACHE <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE_index1( int _this_dot_1, logic  ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE[_this_dot_1] = ARCACHE_param;
        else
            m_ARCACHE[_this_dot_1] <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic [3:0]  get_ARCACHE(  );
        return ARCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic   get_ARCACHE_index1( int _this_dot_1 );
        return ARCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARQOS>.
    //
    // Parameters:
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS( logic [3:0] ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS = ARQOS_param;
        else
            m_ARQOS <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS_index1( int _this_dot_1, logic  ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS[_this_dot_1] = ARQOS_param;
        else
            m_ARQOS[_this_dot_1] <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic [3:0]  get_ARQOS(  );
        return ARQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic   get_ARQOS_index1( int _this_dot_1 );
        return ARQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARID>.
    //
    // Parameters:
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID( logic [((AXI4_ID_WIDTH) - 1):0] ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID = ARID_param;
        else
            m_ARID <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID_index1( int _this_dot_1, logic  ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID[_this_dot_1] = ARID_param;
        else
            m_ARID[_this_dot_1] <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_ARID(  );
        return ARID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic   get_ARID_index1( int _this_dot_1 );
        return ARID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARUSER>.
    //
    // Parameters:
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER( logic [((AXI4_USER_WIDTH) - 1):0] ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER = ARUSER_param;
        else
            m_ARUSER <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER_index1( int _this_dot_1, logic  ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER[_this_dot_1] = ARUSER_param;
        else
            m_ARUSER[_this_dot_1] <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_ARUSER(  );
        return ARUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic   get_ARUSER_index1( int _this_dot_1 );
        return ARUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREADY>.
    //
    // Parameters:
    //     ARREADY_param - The value to set onto wire <ARREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREADY( logic ARREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREADY = ARREADY_param;
        else
            m_ARREADY <= ARREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREADY>.
    //
    function automatic logic get_ARREADY(  );
        return ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RVALID>.
    //
    // Parameters:
    //     RVALID_param - The value to set onto wire <RVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RVALID( logic RVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RVALID = RVALID_param;
        else
            m_RVALID <= RVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RVALID>.
    //
    function automatic logic get_RVALID(  );
        return RVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <RDATA>.
    //
    // Parameters:
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA( logic [((AXI4_RDATA_WIDTH) - 1):0] RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA = RDATA_param;
        else
            m_RDATA <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA_index1( int _this_dot_1, logic  RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA[_this_dot_1] = RDATA_param;
        else
            m_RDATA[_this_dot_1] <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <RDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic [((AXI4_RDATA_WIDTH) - 1):0]  get_RDATA(  );
        return RDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic   get_RDATA_index1( int _this_dot_1 );
        return RDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <RRESP>.
    //
    // Parameters:
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP( logic [1:0] RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP = RRESP_param;
        else
            m_RRESP <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP_index1( int _this_dot_1, logic  RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP[_this_dot_1] = RRESP_param;
        else
            m_RRESP[_this_dot_1] <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <RRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic [1:0]  get_RRESP(  );
        return RRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic   get_RRESP_index1( int _this_dot_1 );
        return RRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <RLAST>.
    //
    // Parameters:
    //     RLAST_param - The value to set onto wire <RLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RLAST( logic RLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RLAST = RLAST_param;
        else
            m_RLAST <= RLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <RLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RLAST>.
    //
    function automatic logic get_RLAST(  );
        return RLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RID>.
    //
    // Parameters:
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID( logic [((AXI4_ID_WIDTH) - 1):0] RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID = RID_param;
        else
            m_RID <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID_index1( int _this_dot_1, logic  RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID[_this_dot_1] = RID_param;
        else
            m_RID[_this_dot_1] <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_RID(  );
        return RID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic   get_RID_index1( int _this_dot_1 );
        return RID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <RUSER>.
    //
    // Parameters:
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER( logic [((AXI4_USER_WIDTH) - 1):0] RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER = RUSER_param;
        else
            m_RUSER <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER_index1( int _this_dot_1, logic  RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER[_this_dot_1] = RUSER_param;
        else
            m_RUSER[_this_dot_1] <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <RUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_RUSER(  );
        return RUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic   get_RUSER_index1( int _this_dot_1 );
        return RUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <RREADY>.
    //
    // Parameters:
    //     RREADY_param - The value to set onto wire <RREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RREADY( logic RREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RREADY = RREADY_param;
        else
            m_RREADY <= RREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RREADY>.
    //
    function automatic logic get_RREADY(  );
        return RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <WVALID>.
    //
    // Parameters:
    //     WVALID_param - The value to set onto wire <WVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WVALID( logic WVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WVALID = WVALID_param;
        else
            m_WVALID <= WVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <WVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WVALID>.
    //
    function automatic logic get_WVALID(  );
        return WVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <WDATA>.
    //
    // Parameters:
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA( logic [((AXI4_WDATA_WIDTH) - 1):0] WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA = WDATA_param;
        else
            m_WDATA <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA_index1( int _this_dot_1, logic  WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA[_this_dot_1] = WDATA_param;
        else
            m_WDATA[_this_dot_1] <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <WDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic [((AXI4_WDATA_WIDTH) - 1):0]  get_WDATA(  );
        return WDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic   get_WDATA_index1( int _this_dot_1 );
        return WDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB
    //-------------------------------------------------------------------------
    //     Set the value of wire <WSTRB>.
    //
    // Parameters:
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB( logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB = WSTRB_param;
        else
            m_WSTRB <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB_index1( int _this_dot_1, logic  WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB[_this_dot_1] = WSTRB_param;
        else
            m_WSTRB[_this_dot_1] <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB
    //-------------------------------------------------------------------------
    //     Get the value of wire <WSTRB>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0]  get_WSTRB(  );
        return WSTRB;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic   get_WSTRB_index1( int _this_dot_1 );
        return WSTRB[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <WLAST>.
    //
    // Parameters:
    //     WLAST_param - The value to set onto wire <WLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WLAST( logic WLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WLAST = WLAST_param;
        else
            m_WLAST <= WLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <WLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WLAST>.
    //
    function automatic logic get_WLAST(  );
        return WLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <WUSER>.
    //
    // Parameters:
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER( logic [((AXI4_USER_WIDTH) - 1):0] WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER = WUSER_param;
        else
            m_WUSER <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER_index1( int _this_dot_1, logic  WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER[_this_dot_1] = WUSER_param;
        else
            m_WUSER[_this_dot_1] <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <WUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_WUSER(  );
        return WUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic   get_WUSER_index1( int _this_dot_1 );
        return WUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <WREADY>.
    //
    // Parameters:
    //     WREADY_param - The value to set onto wire <WREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WREADY( logic WREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WREADY = WREADY_param;
        else
            m_WREADY <= WREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WREADY>.
    //
    function automatic logic get_WREADY(  );
        return WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BVALID>.
    //
    // Parameters:
    //     BVALID_param - The value to set onto wire <BVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BVALID( logic BVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BVALID = BVALID_param;
        else
            m_BVALID <= BVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BVALID>.
    //
    function automatic logic get_BVALID(  );
        return BVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <BRESP>.
    //
    // Parameters:
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP( logic [1:0] BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP = BRESP_param;
        else
            m_BRESP <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP_index1( int _this_dot_1, logic  BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP[_this_dot_1] = BRESP_param;
        else
            m_BRESP[_this_dot_1] <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <BRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic [1:0]  get_BRESP(  );
        return BRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic   get_BRESP_index1( int _this_dot_1 );
        return BRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BID>.
    //
    // Parameters:
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID( logic [((AXI4_ID_WIDTH) - 1):0] BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID = BID_param;
        else
            m_BID <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID_index1( int _this_dot_1, logic  BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID[_this_dot_1] = BID_param;
        else
            m_BID[_this_dot_1] <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_BID(  );
        return BID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic   get_BID_index1( int _this_dot_1 );
        return BID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <BUSER>.
    //
    // Parameters:
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER( logic [((AXI4_USER_WIDTH) - 1):0] BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER = BUSER_param;
        else
            m_BUSER <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER_index1( int _this_dot_1, logic  BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER[_this_dot_1] = BUSER_param;
        else
            m_BUSER[_this_dot_1] <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <BUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_BUSER(  );
        return BUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic   get_BUSER_index1( int _this_dot_1 );
        return BUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <BREADY>.
    //
    // Parameters:
    //     BREADY_param - The value to set onto wire <BREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BREADY( logic BREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BREADY = BREADY_param;
        else
            m_BREADY <= BREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BREADY>.
    //
    function automatic logic get_BREADY(  );
        return BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id( bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id_param );
        config_axi4lite_tr_id = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id_index1( int _this_dot_1, bit  config_axi4lite_tr_id_param );
        config_axi4lite_tr_id[_this_dot_1] = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit [((AXI4_ID_WIDTH) - 1):0]  get_config_axi4lite_tr_id(  );
        return config_axi4lite_tr_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit   get_config_axi4lite_tr_id_index1( int _this_dot_1 );
        return config_axi4lite_tr_id[_this_dot_1];
    endfunction

    //-------------------------------------------------------------------------
    // Tasks to wait for a change to a global variable with read access
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_interface
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_interface>.
    //
    task automatic wait_for_config_axi4lite_interface(  );
        begin
            @( config_axi4lite_interface );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_interface_type
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_interface_type>.
    //
    task automatic wait_for_config_interface_type(  );
        begin
            @( config_interface_type );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_rlast
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_rlast>.
    //
    task automatic wait_for_config_enable_rlast(  );
        begin
            @( config_enable_rlast );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_id>.
    //
    task automatic wait_for_config_enable_id(  );
        begin
            @( config_enable_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_all_assertions
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_all_assertions>.
    //
    task automatic wait_for_config_enable_all_assertions(  );
        begin
            @( config_enable_all_assertions );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    task automatic wait_for_config_enable_assertion(  );
        begin
            @( config_enable_assertion );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_enable_assertion_index1( input int _this_dot_1 );
        begin
            @( config_enable_assertion[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_user_sideband
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_user_sideband>.
    //
    task automatic wait_for_config_enable_user_sideband(  );
        begin
            @( config_enable_user_sideband );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_slave_exclusive
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_slave_exclusive>.
    //
    task automatic wait_for_config_enable_slave_exclusive(  );
        begin
            @( config_enable_slave_exclusive );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_read_data_reordering_depth
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_read_data_reordering_depth>.
    //
    task automatic wait_for_config_read_data_reordering_depth(  );
        begin
            @( config_read_data_reordering_depth );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    task automatic wait_for_config_slave_start_addr(  );
        begin
            @( config_slave_start_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_start_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_start_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    task automatic wait_for_config_slave_end_addr(  );
        begin
            @( config_slave_end_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_end_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_end_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_transaction_time_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_transaction_time_factor>.
    //
    task automatic wait_for_config_max_transaction_time_factor(  );
        begin
            @( config_max_transaction_time_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_burst_timeout_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_burst_timeout_factor>.
    //
    task automatic wait_for_config_burst_timeout_factor(  );
        begin
            @( config_burst_timeout_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_AWVALID_assertion_to_AWREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    task automatic wait_for_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        begin
            @( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_ARVALID_assertion_to_ARREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    task automatic wait_for_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        begin
            @( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_RVALID_assertion_to_RREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_RVALID_assertion_to_RREADY>.
    //
    task automatic wait_for_config_max_latency_RVALID_assertion_to_RREADY(  );
        begin
            @( config_max_latency_RVALID_assertion_to_RREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_BVALID_assertion_to_BREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_BVALID_assertion_to_BREADY>.
    //
    task automatic wait_for_config_max_latency_BVALID_assertion_to_BREADY(  );
        begin
            @( config_max_latency_BVALID_assertion_to_BREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_WVALID_assertion_to_WREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_WVALID_assertion_to_WREADY>.
    //
    task automatic wait_for_config_max_latency_WVALID_assertion_to_WREADY(  );
        begin
            @( config_max_latency_WVALID_assertion_to_WREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_qos
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_qos>.
    //
    task automatic wait_for_config_enable_qos(  );
        begin
            @( config_enable_qos );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_setup_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_setup_time>.
    //
    task automatic wait_for_config_setup_time(  );
        begin
            @( config_setup_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_hold_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_hold_time>.
    //
    task automatic wait_for_config_hold_time(  );
        begin
            @( config_hold_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_reads
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_reads>.
    //
    task automatic wait_for_config_num_max_outstanding_reads(  );
        begin
            @( config_num_max_outstanding_reads );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_writes
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_writes>.
    //
    task automatic wait_for_config_num_max_outstanding_writes(  );
        begin
            @( config_num_max_outstanding_writes );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_wr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_wr>.
    //
    task automatic wait_for_config_max_outstanding_wr(  );
        begin
            @( config_max_outstanding_wr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rd
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rd>.
    //
    task automatic wait_for_config_max_outstanding_rd(  );
        begin
            @( config_max_outstanding_rd );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rw
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rw>.
    //
    task automatic wait_for_config_max_outstanding_rw(  );
        begin
            @( config_max_outstanding_rw );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_wr_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_wr_phase>.
    //
    task automatic wait_for_config_num_outstanding_wr_phase(  );
        begin
            @( config_num_outstanding_wr_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_rd_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_rd_phase>.
    //
    task automatic wait_for_config_num_outstanding_rd_phase(  );
        begin
            @( config_num_outstanding_rd_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_is_issuing
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_is_issuing>.
    //
    task automatic wait_for_config_is_issuing(  );
        begin
            @( config_is_issuing );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    task automatic wait_for_config_axi4lite_tr_id(  );
        begin
            @( config_axi4lite_tr_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_axi4lite_tr_id_index1( input int _this_dot_1 );
        begin
            @( config_axi4lite_tr_id[_this_dot_1] );
        end
    endtask


    //-------------------------------------------------------------------------
    // Functions to set global variables with write access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //     config_axi4lite_interface_param - The value to assign to variable <config_axi4lite_interface>.
    //
    function automatic void set_config_axi4lite_interface( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_interface_type
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_interface_type>.
    //
    // Parameters:
    //     config_interface_type_param - The value to assign to variable <config_interface_type>.
    //
    function automatic void set_config_interface_type( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //     config_enable_rlast_param - The value to assign to variable <config_enable_rlast>.
    //
    function automatic void set_config_enable_rlast( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_id>.
    //
    // Parameters:
    //     config_enable_id_param - The value to assign to variable <config_enable_id>.
    //
    function automatic void set_config_enable_id( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //     config_enable_all_assertions_param - The value to assign to variable <config_enable_all_assertions>.
    //
    function automatic void set_config_enable_all_assertions( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion_index1( int _this_dot_1, bit  config_enable_assertion_param );
        config_enable_assertion[_this_dot_1] = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //     config_enable_user_sideband_param - The value to assign to variable <config_enable_user_sideband>.
    //
    function automatic void set_config_enable_user_sideband( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //     config_enable_slave_exclusive_param - The value to assign to variable <config_enable_slave_exclusive>.
    //
    function automatic void set_config_enable_slave_exclusive( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //     config_read_data_reordering_depth_param - The value to assign to variable <config_read_data_reordering_depth>.
    //
    function automatic void set_config_read_data_reordering_depth( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        config_slave_start_addr = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr_index1( int _this_dot_1, bit  config_slave_start_addr_param );
        config_slave_start_addr[_this_dot_1] = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        config_slave_end_addr = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr_index1( int _this_dot_1, bit  config_slave_end_addr_param );
        config_slave_end_addr[_this_dot_1] = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //     config_max_transaction_time_factor_param - The value to assign to variable <config_max_transaction_time_factor>.
    //
    function automatic void set_config_max_transaction_time_factor( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //     config_burst_timeout_factor_param - The value to assign to variable <config_burst_timeout_factor>.
    //
    function automatic void set_config_burst_timeout_factor( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //     config_max_latency_AWVALID_assertion_to_AWREADY_param - The value to assign to variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic void set_config_max_latency_AWVALID_assertion_to_AWREADY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //     config_max_latency_ARVALID_assertion_to_ARREADY_param - The value to assign to variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic void set_config_max_latency_ARVALID_assertion_to_ARREADY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //     config_max_latency_RVALID_assertion_to_RREADY_param - The value to assign to variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic void set_config_max_latency_RVALID_assertion_to_RREADY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //     config_max_latency_BVALID_assertion_to_BREADY_param - The value to assign to variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic void set_config_max_latency_BVALID_assertion_to_BREADY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //     config_max_latency_WVALID_assertion_to_WREADY_param - The value to assign to variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic void set_config_max_latency_WVALID_assertion_to_WREADY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_qos
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_qos>.
    //
    // Parameters:
    //     config_enable_qos_param - The value to assign to variable <config_enable_qos>.
    //
    function automatic void set_config_enable_qos( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_setup_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_setup_time>.
    //
    // Parameters:
    //     config_setup_time_param - The value to assign to variable <config_setup_time>.
    //
    function automatic void set_config_setup_time( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_hold_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_hold_time>.
    //
    // Parameters:
    //     config_hold_time_param - The value to assign to variable <config_hold_time>.
    //
    function automatic void set_config_hold_time( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //     config_num_max_outstanding_reads_param - The value to assign to variable <config_num_max_outstanding_reads>.
    //
    function automatic void set_config_num_max_outstanding_reads( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //     config_num_max_outstanding_writes_param - The value to assign to variable <config_num_max_outstanding_writes>.
    //
    function automatic void set_config_num_max_outstanding_writes( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //     config_max_outstanding_wr_param - The value to assign to variable <config_max_outstanding_wr>.
    //
    function automatic void set_config_max_outstanding_wr( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //     config_max_outstanding_rd_param - The value to assign to variable <config_max_outstanding_rd>.
    //
    function automatic void set_config_max_outstanding_rd( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //     config_max_outstanding_rw_param - The value to assign to variable <config_max_outstanding_rw>.
    //
    function automatic void set_config_max_outstanding_rw( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //     config_num_outstanding_wr_phase_param - The value to assign to variable <config_num_outstanding_wr_phase>.
    //
    function automatic void set_config_num_outstanding_wr_phase( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //     config_num_outstanding_rd_phase_param - The value to assign to variable <config_num_outstanding_rd_phase>.
    //
    function automatic void set_config_num_outstanding_rd_phase( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_is_issuing
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_is_issuing>.
    //
    // Parameters:
    //     config_is_issuing_param - The value to assign to variable <config_is_issuing>.
    //
    function automatic void set_config_is_issuing( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to get global variables with read access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_interface>.
    //
    function automatic bit get_config_axi4lite_interface(  );
        return config_axi4lite_interface;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_interface_type
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_interface_type>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_interface_type>.
    //
    function automatic axi4_interface_type_e get_config_interface_type(  );
        return config_interface_type;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_rlast>.
    //
    function automatic bit get_config_enable_rlast(  );
        return config_enable_rlast;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_id>.
    //
    function automatic bit get_config_enable_id(  );
        return config_enable_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_all_assertions>.
    //
    function automatic bit get_config_enable_all_assertions(  );
        return config_enable_all_assertions;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit [255:0]  get_config_enable_assertion(  );
        return config_enable_assertion;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit   get_config_enable_assertion_index1( int _this_dot_1 );
        return config_enable_assertion[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_user_sideband>.
    //
    function automatic bit get_config_enable_user_sideband(  );
        return config_enable_user_sideband;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_slave_exclusive>.
    //
    function automatic bit get_config_enable_slave_exclusive(  );
        return config_enable_slave_exclusive;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_read_data_reordering_depth>.
    //
    function automatic int unsigned get_config_read_data_reordering_depth(  );
        return config_read_data_reordering_depth;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_start_addr(  );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit   get_config_slave_start_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_end_addr(  );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit   get_config_slave_end_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_transaction_time_factor>.
    //
    function automatic int unsigned get_config_max_transaction_time_factor(  );
        return config_max_transaction_time_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_burst_timeout_factor>.
    //
    function automatic int unsigned get_config_burst_timeout_factor(  );
        return config_burst_timeout_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic int unsigned get_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        return config_max_latency_AWVALID_assertion_to_AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic int unsigned get_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        return config_max_latency_ARVALID_assertion_to_ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic int unsigned get_config_max_latency_RVALID_assertion_to_RREADY(  );
        return config_max_latency_RVALID_assertion_to_RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic int unsigned get_config_max_latency_BVALID_assertion_to_BREADY(  );
        return config_max_latency_BVALID_assertion_to_BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic int unsigned get_config_max_latency_WVALID_assertion_to_WREADY(  );
        return config_max_latency_WVALID_assertion_to_WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_qos
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_qos>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_qos>.
    //
    function automatic bit get_config_enable_qos(  );
        return config_enable_qos;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_setup_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_setup_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_setup_time>.
    //
    function automatic longint unsigned get_config_setup_time(  );
        return config_setup_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_hold_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_hold_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_hold_time>.
    //
    function automatic longint unsigned get_config_hold_time(  );
        return config_hold_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_reads>.
    //
    function automatic int get_config_num_max_outstanding_reads(  );
        return config_num_max_outstanding_reads;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_writes>.
    //
    function automatic int get_config_num_max_outstanding_writes(  );
        return config_num_max_outstanding_writes;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_wr>.
    //
    function automatic int get_config_max_outstanding_wr(  );
        return config_max_outstanding_wr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rd>.
    //
    function automatic int get_config_max_outstanding_rd(  );
        return config_max_outstanding_rd;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rw>.
    //
    function automatic int get_config_max_outstanding_rw(  );
        return config_max_outstanding_rw;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_wr_phase>.
    //
    function automatic int get_config_num_outstanding_wr_phase(  );
        return config_num_outstanding_wr_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_rd_phase>.
    //
    function automatic int get_config_num_outstanding_rd_phase(  );
        return config_num_outstanding_rd_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_is_issuing
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_is_issuing>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_is_issuing>.
    //
    function automatic bit get_config_is_issuing(  );
        return config_is_issuing;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to set/get generic interface configuration
    //-------------------------------------------------------------------------

    function void set_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0,
        input int arg10 = 0
    );
        axi4_set_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10 );
    endfunction

    function int get_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0
    );
        return axi4_get_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 );
    endfunction

    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //-------------------------------------------------------------------------
    function string get_full_name();
        return axi4_get_full_name();
    endfunction

    //--------------------------------------------------------------------------
    //
    // Group:- Monitor Value Change on Variable
    //
    //--------------------------------------------------------------------------

    function automatic void axi4_local_set_config_axi4lite_interface_from_SystemVerilog( ref bit config_axi4lite_interface_param );
            dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog( _interface_ref,config_axi4lite_interface); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_axi4lite_interface_from_SystemVerilog( config_axi4lite_interface );
        end
    end

    function automatic void axi4_local_set_config_interface_type_from_SystemVerilog( ref axi4_interface_type_e config_interface_type_param );
            dvc_axi4_set_config_interface_type_from_SystemVerilog( _interface_ref,config_interface_type); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_interface_type_from_SystemVerilog( config_interface_type );
        end
    end

    function automatic void axi4_local_set_config_enable_rlast_from_SystemVerilog( ref bit config_enable_rlast_param );
            dvc_axi4_set_config_enable_rlast_from_SystemVerilog( _interface_ref,config_enable_rlast); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_rlast_from_SystemVerilog( config_enable_rlast );
        end
    end

    function automatic void axi4_local_set_config_enable_id_from_SystemVerilog( ref bit config_enable_id_param );
            dvc_axi4_set_config_enable_id_from_SystemVerilog( _interface_ref,config_enable_id); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_id_from_SystemVerilog( config_enable_id );
        end
    end

    function automatic void axi4_local_set_config_enable_all_assertions_from_SystemVerilog( ref bit config_enable_all_assertions_param );
            dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog( _interface_ref,config_enable_all_assertions); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_all_assertions_from_SystemVerilog( config_enable_all_assertions );
        end
    end

    function automatic void axi4_local_set_config_enable_assertion_from_SystemVerilog( ref bit [255:0] config_enable_assertion_param );
            dvc_axi4_set_config_enable_assertion_from_SystemVerilog( _interface_ref,config_enable_assertion); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_assertion_from_SystemVerilog( config_enable_assertion );
        end
    end

    function automatic void axi4_local_set_config_enable_user_sideband_from_SystemVerilog( ref bit config_enable_user_sideband_param );
            dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog( _interface_ref,config_enable_user_sideband); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_user_sideband_from_SystemVerilog( config_enable_user_sideband );
        end
    end

    function automatic void axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( ref bit config_enable_slave_exclusive_param );
            dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog( _interface_ref,config_enable_slave_exclusive); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( config_enable_slave_exclusive );
        end
    end

    function automatic void axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( ref int unsigned config_read_data_reordering_depth_param );
            dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog( _interface_ref,config_read_data_reordering_depth); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( config_read_data_reordering_depth );
        end
    end

    function automatic void axi4_local_set_config_slave_start_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1( _interface_ref,_this_dot_1,config_slave_start_addr[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
            dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog( _interface_ref ); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_start_addr_from_SystemVerilog( config_slave_start_addr );
        end
    end

    function automatic void axi4_local_set_config_slave_end_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1( _interface_ref,_this_dot_1,config_slave_end_addr[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
            dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog( _interface_ref ); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_end_addr_from_SystemVerilog( config_slave_end_addr );
        end
    end

    function automatic void axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( ref int unsigned config_max_transaction_time_factor_param );
            dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog( _interface_ref,config_max_transaction_time_factor); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( config_max_transaction_time_factor );
        end
    end

    function automatic void axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( ref int unsigned config_burst_timeout_factor_param );
            dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog( _interface_ref,config_burst_timeout_factor); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( config_burst_timeout_factor );
        end
    end

    function automatic void axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( ref int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
            dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( _interface_ref,config_max_latency_AWVALID_assertion_to_AWREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( ref int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
            dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( _interface_ref,config_max_latency_ARVALID_assertion_to_ARREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( ref int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
            dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( _interface_ref,config_max_latency_RVALID_assertion_to_RREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( config_max_latency_RVALID_assertion_to_RREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( ref int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
            dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( _interface_ref,config_max_latency_BVALID_assertion_to_BREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( config_max_latency_BVALID_assertion_to_BREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( ref int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
            dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( _interface_ref,config_max_latency_WVALID_assertion_to_WREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( config_max_latency_WVALID_assertion_to_WREADY );
        end
    end

    function automatic void axi4_local_set_config_enable_qos_from_SystemVerilog( ref bit config_enable_qos_param );
            dvc_axi4_set_config_enable_qos_from_SystemVerilog( _interface_ref,config_enable_qos); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_qos_from_SystemVerilog( config_enable_qos );
        end
    end

    function automatic void axi4_local_set_config_setup_time_from_SystemVerilog( ref longint unsigned config_setup_time_param );
            dvc_axi4_set_config_setup_time_from_SystemVerilog( _interface_ref,config_setup_time); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_setup_time_from_SystemVerilog( config_setup_time );
        end
    end

    function automatic void axi4_local_set_config_hold_time_from_SystemVerilog( ref longint unsigned config_hold_time_param );
            dvc_axi4_set_config_hold_time_from_SystemVerilog( _interface_ref,config_hold_time); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_hold_time_from_SystemVerilog( config_hold_time );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( ref int config_num_max_outstanding_reads_param );
            dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog( _interface_ref,config_num_max_outstanding_reads); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( config_num_max_outstanding_reads );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( ref int config_num_max_outstanding_writes_param );
            dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog( _interface_ref,config_num_max_outstanding_writes); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( config_num_max_outstanding_writes );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( ref int config_max_outstanding_wr_param );
            dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog( _interface_ref,config_max_outstanding_wr); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( config_max_outstanding_wr );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( ref int config_max_outstanding_rd_param );
            dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog( _interface_ref,config_max_outstanding_rd); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( config_max_outstanding_rd );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( ref int config_max_outstanding_rw_param );
            dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog( _interface_ref,config_max_outstanding_rw); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( config_max_outstanding_rw );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( ref int config_num_outstanding_wr_phase_param );
            dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog( _interface_ref,config_num_outstanding_wr_phase); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( config_num_outstanding_wr_phase );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( ref int config_num_outstanding_rd_phase_param );
            dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog( _interface_ref,config_num_outstanding_rd_phase); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( config_num_outstanding_rd_phase );
        end
    end

    function automatic void axi4_local_set_config_is_issuing_from_SystemVerilog( ref bit config_is_issuing_param );
            dvc_axi4_set_config_is_issuing_from_SystemVerilog( _interface_ref,config_is_issuing); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_is_issuing_from_SystemVerilog( config_is_issuing );
        end
    end

    //-------------------------------------------------------------------------
    // Transaction interface
    //-------------------------------------------------------------------------

    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_rw_transaction_addr;
    function void axi4_get_temp_static_rw_transaction_addr( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_addr[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_addr( input int _d1, input bit  _value );
        temp_static_rw_transaction_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_rw_transaction_id;
    function void axi4_get_temp_static_rw_transaction_id( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_id[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_id( input int _d1, input bit  _value );
        temp_static_rw_transaction_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_addr_user_data;
    function void axi4_get_temp_static_rw_transaction_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_addr_user_data( input int _d1, input bit  _value );
        temp_static_rw_transaction_addr_user_data[_d1] = _value;
    endfunction
    bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] temp_static_rw_transaction_data_words [];
    function void axi4_get_temp_static_rw_transaction_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_rw_transaction_write_strobes [];
    function void axi4_get_temp_static_rw_transaction_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_wdata_user_data [];
    function void axi4_get_temp_static_rw_transaction_wdata_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_wdata_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_wdata_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_wdata_user_data[_d1][_d2] = _value;
    endfunction
    axi4_response_e temp_static_rw_transaction_resp[];
    function void axi4_get_temp_static_rw_transaction_resp( input int _d1, output axi4_response_e _value );
        _value = temp_static_rw_transaction_resp[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_resp( input int _d1, input axi4_response_e _value );
        temp_static_rw_transaction_resp[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_resp_user_data [];
    function void axi4_get_temp_static_rw_transaction_resp_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_resp_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_resp_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_resp_user_data[_d1][_d2] = _value;
    endfunction
    int temp_static_rw_transaction_data_valid_delay[];
    function void axi4_get_temp_static_rw_transaction_data_valid_delay( input int _d1, output int _value );
        _value = temp_static_rw_transaction_data_valid_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_valid_delay( input int _d1, input int _value );
        temp_static_rw_transaction_data_valid_delay[_d1] = _value;
    endfunction
    int temp_static_rw_transaction_data_ready_delay[];
    function void axi4_get_temp_static_rw_transaction_data_ready_delay( input int _d1, output int _value );
        _value = temp_static_rw_transaction_data_ready_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_ready_delay( input int _d1, input int _value );
        temp_static_rw_transaction_data_ready_delay[_d1] = _value;
    endfunction
    int unsigned temp_static_rw_transaction_write_data_beats_delay[];
    function void axi4_get_temp_static_rw_transaction_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_rw_transaction_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_rw_transaction_write_data_beats_delay[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr;
    function void axi4_get_temp_static_read_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr( input int _d1, input bit  _value );
        temp_static_read_addr[_d1] = _value;
    endfunction
    int temp_static_read_rresp_delays[255:0];
    function void axi4_get_temp_static_read_rresp_delays( input int _d1, output int _value );
        _value = temp_static_read_rresp_delays[_d1];
    endfunction
    function void axi4_set_temp_static_read_rresp_delays( input int _d1, input int _value );
        temp_static_read_rresp_delays[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_id;
    function void axi4_get_temp_static_read_id( input int _d1, output bit  _value );
        _value = temp_static_read_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_id( input int _d1, input bit  _value );
        temp_static_read_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_user_data;
    function void axi4_get_temp_static_read_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_user_data( input int _d1, input bit  _value );
        temp_static_read_addr_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_data_words [];
    function void axi4_get_temp_static_read_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_read_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_read_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_read_data_words[_d1][_d2] = _value;
    endfunction
    axi4_response_e temp_static_read_resp[];
    function void axi4_get_temp_static_read_resp( input int _d1, output axi4_response_e _value );
        _value = temp_static_read_resp[_d1];
    endfunction
    function void axi4_set_temp_static_read_resp( input int _d1, input axi4_response_e _value );
        temp_static_read_resp[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_resp_user_data [];
    function void axi4_get_temp_static_read_resp_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_read_resp_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_read_resp_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_read_resp_user_data[_d1][_d2] = _value;
    endfunction
    longint temp_static_read_data_start_time[];
    function void axi4_get_temp_static_read_data_start_time( input int _d1, output longint _value );
        _value = temp_static_read_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_read_data_start_time( input int _d1, input longint _value );
        temp_static_read_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_read_data_end_time[];
    function void axi4_get_temp_static_read_data_end_time( input int _d1, output longint _value );
        _value = temp_static_read_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_read_data_end_time( input int _d1, input longint _value );
        temp_static_read_data_end_time[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr;
    function void axi4_get_temp_static_write_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr( input int _d1, input bit  _value );
        temp_static_write_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_id;
    function void axi4_get_temp_static_write_id( input int _d1, output bit  _value );
        _value = temp_static_write_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_id( input int _d1, input bit  _value );
        temp_static_write_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_user_data;
    function void axi4_get_temp_static_write_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_user_data( input int _d1, input bit  _value );
        temp_static_write_addr_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_data_words [];
    function void axi4_get_temp_static_write_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_write_strobes [];
    function void axi4_get_temp_static_write_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_write_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_wdata_user_data [];
    function void axi4_get_temp_static_write_wdata_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_wdata_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_wdata_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_write_wdata_user_data[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_user_data;
    function void axi4_get_temp_static_write_resp_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_user_data[_d1] = _value;
    endfunction
    longint temp_static_write_data_start_time[];
    function void axi4_get_temp_static_write_data_start_time( input int _d1, output longint _value );
        _value = temp_static_write_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_start_time( input int _d1, input longint _value );
        temp_static_write_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_write_data_end_time[];
    function void axi4_get_temp_static_write_data_end_time( input int _d1, output longint _value );
        _value = temp_static_write_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_end_time( input int _d1, input longint _value );
        temp_static_write_data_end_time[_d1] = _value;
    endfunction
    int unsigned temp_static_write_write_data_beats_delay[];
    function void axi4_get_temp_static_write_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_write_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_write_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_write_write_data_beats_delay[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_data_burst_data_words [];
    function void axi4_get_temp_static_write_data_burst_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_data_burst_write_strobes [];
    function void axi4_get_temp_static_write_data_burst_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_data_burst_user_data [];
    function void axi4_get_temp_static_write_data_burst_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_user_data[_d1][_d2] = _value;
    endfunction
    longint temp_static_write_data_burst_data_start_time[];
    function void axi4_get_temp_static_write_data_burst_data_start_time( input int _d1, output longint _value );
        _value = temp_static_write_data_burst_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_start_time( input int _d1, input longint _value );
        temp_static_write_data_burst_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_write_data_burst_data_end_time[];
    function void axi4_get_temp_static_write_data_burst_data_end_time( input int _d1, output longint _value );
        _value = temp_static_write_data_burst_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_end_time( input int _d1, input longint _value );
        temp_static_write_data_burst_data_end_time[_d1] = _value;
    endfunction
    int unsigned temp_static_write_data_burst_write_data_beats_delay[];
    function void axi4_get_temp_static_write_data_burst_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_write_data_burst_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_write_data_burst_write_data_beats_delay[_d1] = _value;
    endfunction
    int temp_static_write_data_burst_data_valid2ready[];
    function void axi4_get_temp_static_write_data_burst_data_valid2ready( input int _d1, output int _value );
        _value = temp_static_write_data_burst_data_valid2ready[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_valid2ready( input int _d1, input int _value );
        temp_static_write_data_burst_data_valid2ready[_d1] = _value;
    endfunction
    int temp_static_write_data_burst_data2data_clock[];
    function void axi4_get_temp_static_write_data_burst_data2data_clock( input int _d1, output int _value );
        _value = temp_static_write_data_burst_data2data_clock[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data2data_clock( input int _d1, input int _value );
        temp_static_write_data_burst_data2data_clock[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr_channel_phase_addr;
    function void axi4_get_temp_static_read_addr_channel_phase_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_addr( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_addr_channel_phase_id;
    function void axi4_get_temp_static_read_addr_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_id( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_channel_phase_addr_user;
    function void axi4_get_temp_static_read_addr_channel_phase_addr_user( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_addr_user[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_addr_user( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_addr_user[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_channel_phase_data;
    function void axi4_get_temp_static_read_channel_phase_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_data( input int _d1, input bit  _value );
        temp_static_read_channel_phase_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_channel_phase_id;
    function void axi4_get_temp_static_read_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_id( input int _d1, input bit  _value );
        temp_static_read_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_channel_phase_user_data;
    function void axi4_get_temp_static_read_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_read_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr_channel_phase_addr;
    function void axi4_get_temp_static_write_addr_channel_phase_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_addr( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_addr_channel_phase_id;
    function void axi4_get_temp_static_write_addr_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_id( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_channel_phase_addr_user;
    function void axi4_get_temp_static_write_addr_channel_phase_addr_user( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_addr_user[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_addr_user( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_addr_user[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_channel_phase_data;
    function void axi4_get_temp_static_write_channel_phase_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_data( input int _d1, input bit  _value );
        temp_static_write_channel_phase_data[_d1] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_channel_phase_write_strobes;
    function void axi4_get_temp_static_write_channel_phase_write_strobes( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_write_strobes[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_write_strobes( input int _d1, input bit  _value );
        temp_static_write_channel_phase_write_strobes[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_channel_phase_user_data;
    function void axi4_get_temp_static_write_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_write_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_resp_channel_phase_id;
    function void axi4_get_temp_static_write_resp_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_phase_id( input int _d1, input bit  _value );
        temp_static_write_resp_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_channel_phase_user_data;
    function void axi4_get_temp_static_write_resp_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_addr;
    function void axi4_get_temp_static_read_addr_channel_cycle_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_addr( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_id;
    function void axi4_get_temp_static_read_addr_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_user_data;
    function void axi4_get_temp_static_read_addr_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_channel_cycle_data;
    function void axi4_get_temp_static_read_channel_cycle_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_data( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_channel_cycle_id;
    function void axi4_get_temp_static_read_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_channel_cycle_user_data;
    function void axi4_get_temp_static_read_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_addr;
    function void axi4_get_temp_static_write_addr_channel_cycle_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_addr( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_id;
    function void axi4_get_temp_static_write_addr_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_user_data;
    function void axi4_get_temp_static_write_addr_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_channel_cycle_data;
    function void axi4_get_temp_static_write_channel_cycle_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_data( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_data[_d1] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_channel_cycle_strb;
    function void axi4_get_temp_static_write_channel_cycle_strb( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_strb[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_strb( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_strb[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_channel_cycle_user_data;
    function void axi4_get_temp_static_write_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_resp_channel_cycle_id;
    function void axi4_get_temp_static_write_resp_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_write_resp_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_channel_cycle_user_data;
    function void axi4_get_temp_static_write_resp_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_channel_cycle_user_data[_d1] = _value;
    endfunction
    task automatic dvc_activate_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref axi4_rw_e read_or_write,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref int address_valid_delay,
        ref int data_valid_delay[],
        ref int write_response_valid_delay,
        ref int address_ready_delay,
        ref int data_ready_delay[],
        ref int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_rw_transaction_addr = addr;
            temp_static_rw_transaction_id = id;
            temp_static_rw_transaction_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_rw_transaction_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_rw_transaction_write_strobes = write_strobes;
            wdata_user_data_DIMS0 = wdata_user_data.size();
            temp_static_rw_transaction_wdata_user_data = wdata_user_data;
            resp_DIMS0 = resp.size();
            temp_static_rw_transaction_resp = resp;
            resp_user_data_DIMS0 = resp_user_data.size();
            temp_static_rw_transaction_resp_user_data = resp_user_data;
            data_valid_delay_DIMS0 = data_valid_delay.size();
            temp_static_rw_transaction_data_valid_delay = data_valid_delay;
            data_ready_delay_DIMS0 = data_ready_delay.size();
            temp_static_rw_transaction_data_ready_delay = data_ready_delay;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp_DIMS0, resp_user_data_DIMS0, address_valid_delay, data_valid_delay_DIMS0, write_response_valid_delay, address_ready_delay, data_ready_delay_DIMS0, write_response_ready_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_words.delete();
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_write_strobes.delete();
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_wdata_user_data.delete();
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_resp.delete();
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_resp_user_data.delete();
                    end
                    if (data_valid_delay_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_valid_delay = new [data_valid_delay_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_valid_delay.delete();
                    end
                    if (data_ready_delay_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_ready_delay = new [data_ready_delay_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_ready_delay.delete();
                    end
                    // Call function to get the sized params
                    axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, address_valid_delay, write_response_valid_delay, address_ready_delay, write_response_ready_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_rw_transaction_addr;
                    id = temp_static_rw_transaction_id;
                    addr_user_data = temp_static_rw_transaction_addr_user_data;
                    data_words = temp_static_rw_transaction_data_words;
                    temp_static_rw_transaction_data_words.delete();
                    write_strobes = temp_static_rw_transaction_write_strobes;
                    temp_static_rw_transaction_write_strobes.delete();
                    wdata_user_data = temp_static_rw_transaction_wdata_user_data;
                    temp_static_rw_transaction_wdata_user_data.delete();
                    resp = temp_static_rw_transaction_resp;
                    temp_static_rw_transaction_resp.delete();
                    resp_user_data = temp_static_rw_transaction_resp_user_data;
                    temp_static_rw_transaction_resp_user_data.delete();
                    data_valid_delay = temp_static_rw_transaction_data_valid_delay;
                    temp_static_rw_transaction_data_valid_delay.delete();
                    data_ready_delay = temp_static_rw_transaction_data_ready_delay;
                    temp_static_rw_transaction_data_ready_delay.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_rw_e read_or_write,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output int address_valid_delay,
        ref int data_valid_delay[],
        output int write_response_valid_delay,
        output int address_ready_delay,
        ref int data_ready_delay[],
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
                // Call function to get unsized params sizes.
                axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_valid_delay_DIMS0, data_ready_delay_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_write_strobes.delete();
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_wdata_user_data.delete();
                end
                if (resp_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_resp = new [resp_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_resp.delete();
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_resp_user_data.delete();
                end
                if (data_valid_delay_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_valid_delay = new [data_valid_delay_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_valid_delay.delete();
                end
                if (data_ready_delay_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_ready_delay = new [data_ready_delay_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_ready_delay.delete();
                end
                // Call function to get the sized params
                axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, address_valid_delay, write_response_valid_delay, address_ready_delay, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_rw_transaction_addr;
                id = temp_static_rw_transaction_id;
                addr_user_data = temp_static_rw_transaction_addr_user_data;
                data_words = temp_static_rw_transaction_data_words;
                temp_static_rw_transaction_data_words.delete();
                write_strobes = temp_static_rw_transaction_write_strobes;
                temp_static_rw_transaction_write_strobes.delete();
                wdata_user_data = temp_static_rw_transaction_wdata_user_data;
                temp_static_rw_transaction_wdata_user_data.delete();
                resp = temp_static_rw_transaction_resp;
                temp_static_rw_transaction_resp.delete();
                resp_user_data = temp_static_rw_transaction_resp_user_data;
                temp_static_rw_transaction_resp_user_data.delete();
                data_valid_delay = temp_static_rw_transaction_data_valid_delay;
                temp_static_rw_transaction_data_valid_delay.delete();
                data_ready_delay = temp_static_rw_transaction_data_ready_delay;
                temp_static_rw_transaction_data_ready_delay.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int rresp_delays[255:0],
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr = addr;
            temp_static_read_rresp_delays = rresp_delays;
            temp_static_read_id = id;
            temp_static_read_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_read_data_words = data_words;
            resp_DIMS0 = resp.size();
            temp_static_read_resp = resp;
            resp_user_data_DIMS0 = resp_user_data.size();
            temp_static_read_resp_user_data = resp_user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_read_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_read_data_end_time = data_end_time;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_read_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, resp_DIMS0, resp_user_data_DIMS0, addr_start_time, addr_end_time, data_start_time_DIMS0, data_end_time_DIMS0, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_read_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_words.delete();
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        temp_static_read_resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_resp.delete();
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        temp_static_read_resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_resp_user_data.delete();
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        temp_static_read_data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_start_time.delete();
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        temp_static_read_data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_end_time.delete();
                    end
                    // Call function to get the sized params
                    axi4_read_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, addr_start_time, addr_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_read_addr;
                    rresp_delays = temp_static_read_rresp_delays;
                    id = temp_static_read_id;
                    addr_user_data = temp_static_read_addr_user_data;
                    data_words = temp_static_read_data_words;
                    temp_static_read_data_words.delete();
                    resp = temp_static_read_resp;
                    temp_static_read_resp.delete();
                    resp_user_data = temp_static_read_resp_user_data;
                    temp_static_read_resp_user_data.delete();
                    data_start_time = temp_static_read_data_start_time;
                    temp_static_read_data_start_time.delete();
                    data_end_time = temp_static_read_data_end_time;
                    temp_static_read_data_end_time.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        ref int rresp_delays[255:0],
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_read_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_read_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_read_data_words.delete();
                end
                if (resp_DIMS0 != 0)
                begin
                    temp_static_read_resp = new [resp_DIMS0];
                end
                else
                begin
                    temp_static_read_resp.delete();
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    temp_static_read_resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    temp_static_read_resp_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_read_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_read_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_read_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_read_data_end_time.delete();
                end
                // Call function to get the sized params
                axi4_read_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, addr_start_time, addr_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr;
                rresp_delays = temp_static_read_rresp_delays;
                id = temp_static_read_id;
                addr_user_data = temp_static_read_addr_user_data;
                data_words = temp_static_read_data_words;
                temp_static_read_data_words.delete();
                resp = temp_static_read_resp;
                temp_static_read_resp.delete();
                resp_user_data = temp_static_read_resp_user_data;
                temp_static_read_resp_user_data.delete();
                data_start_time = temp_static_read_data_start_time;
                temp_static_read_data_start_time.delete();
                data_end_time = temp_static_read_data_end_time;
                temp_static_read_data_end_time.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int bresp_delay,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp,
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref longint wr_resp_start_time,
        ref longint wr_resp_end_time,
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr = addr;
            temp_static_write_id = id;
            temp_static_write_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_write_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_write_write_strobes = write_strobes;
            wdata_user_data_DIMS0 = wdata_user_data.size();
            temp_static_write_wdata_user_data = wdata_user_data;
            temp_static_write_resp_user_data = resp_user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_write_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_write_data_end_time = data_end_time;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_write_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp, addr_start_time, addr_end_time, data_start_time_DIMS0, data_end_time_DIMS0, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_write_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_words.delete();
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        temp_static_write_write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_write_strobes.delete();
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        temp_static_write_wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_wdata_user_data.delete();
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        temp_static_write_data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_start_time.delete();
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        temp_static_write_data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_end_time.delete();
                    end
                    // Call function to get the sized params
                    axi4_write_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, resp, addr_start_time, addr_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_write_addr;
                    id = temp_static_write_id;
                    addr_user_data = temp_static_write_addr_user_data;
                    data_words = temp_static_write_data_words;
                    temp_static_write_data_words.delete();
                    write_strobes = temp_static_write_write_strobes;
                    temp_static_write_write_strobes.delete();
                    wdata_user_data = temp_static_write_wdata_user_data;
                    temp_static_write_wdata_user_data.delete();
                    resp_user_data = temp_static_write_resp_user_data;
                    data_start_time = temp_static_write_data_start_time;
                    temp_static_write_data_start_time.delete();
                    data_end_time = temp_static_write_data_end_time;
                    temp_static_write_data_end_time.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        output axi4_response_e resp,
        output bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_write_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_write_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_write_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_write_write_strobes.delete();
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    temp_static_write_wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    temp_static_write_wdata_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_write_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_write_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_end_time.delete();
                end
                // Call function to get the sized params
                axi4_write_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, resp, addr_start_time, addr_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr;
                id = temp_static_write_id;
                addr_user_data = temp_static_write_addr_user_data;
                data_words = temp_static_write_data_words;
                temp_static_write_data_words.delete();
                write_strobes = temp_static_write_write_strobes;
                temp_static_write_write_strobes.delete();
                wdata_user_data = temp_static_write_wdata_user_data;
                temp_static_write_wdata_user_data.delete();
                resp_user_data = temp_static_write_resp_user_data;
                data_start_time = temp_static_write_data_start_time;
                temp_static_write_data_start_time.delete();
                data_end_time = temp_static_write_data_end_time;
                temp_static_write_data_end_time.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        input int wdata_burst_start_clock,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                automatic int data_valid2ready_DIMS0;
                automatic int data2data_clock_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            data_words_DIMS0 = data_words.size();
            temp_static_write_data_burst_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_write_data_burst_write_strobes = write_strobes;
            user_data_DIMS0 = user_data.size();
            temp_static_write_data_burst_user_data = user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_write_data_burst_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_write_data_burst_data_end_time = data_end_time;
            data_valid2ready_DIMS0 = data_valid2ready.size();
            temp_static_write_data_burst_data_valid2ready = data_valid2ready;
            data2data_clock_DIMS0 = data2data_clock.size();
            temp_static_write_data_burst_data2data_clock = data2data_clock;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, burst_length, data_words_DIMS0, write_strobes_DIMS0, user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, data_valid2ready_DIMS0, data2data_clock_DIMS0, wdata_burst_start_clock, _unit_id);
            // Delete the storage allocated for the static variable(s)
            temp_static_write_data_burst_data_words.delete();
            temp_static_write_data_burst_write_strobes.delete();
            temp_static_write_data_burst_user_data.delete();
            temp_static_write_data_burst_data_start_time.delete();
            temp_static_write_data_burst_data_end_time.delete();
            temp_static_write_data_burst_data_valid2ready.delete();
            temp_static_write_data_burst_data2data_clock.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        output int wdata_burst_start_clock,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                automatic int data_valid2ready_DIMS0;
                automatic int data2data_clock_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, data_valid2ready_DIMS0, data2data_clock_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_write_strobes.delete();
                end
                if (user_data_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_user_data = new [user_data_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_end_time.delete();
                end
                if (data_valid2ready_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_valid2ready = new [data_valid2ready_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_valid2ready.delete();
                end
                if (data2data_clock_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data2data_clock = new [data2data_clock_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data2data_clock.delete();
                end
                // Call function to get the sized params
                axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, burst_length, wdata_burst_start_clock, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data_words = temp_static_write_data_burst_data_words;
                temp_static_write_data_burst_data_words.delete();
                write_strobes = temp_static_write_data_burst_write_strobes;
                temp_static_write_data_burst_write_strobes.delete();
                user_data = temp_static_write_data_burst_user_data;
                temp_static_write_data_burst_user_data.delete();
                data_start_time = temp_static_write_data_burst_data_start_time;
                temp_static_write_data_burst_data_start_time.delete();
                data_end_time = temp_static_write_data_burst_data_end_time;
                temp_static_write_data_burst_data_end_time.delete();
                data_valid2ready = temp_static_write_data_burst_data_valid2ready;
                temp_static_write_data_burst_data_valid2ready.delete();
                data2data_clock = temp_static_write_data_burst_data2data_clock;
                temp_static_write_data_burst_data2data_clock.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr_channel_phase_addr = addr;
            temp_static_read_addr_channel_phase_id = id;
            temp_static_read_addr_channel_phase_addr_user = addr_user;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr_channel_phase_addr;
                id = temp_static_read_addr_channel_phase_id;
                addr_user = temp_static_read_addr_channel_phase_addr_user;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, _unit_id);
        end
    endtask

    task automatic dvc_get_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_channel_phase_data = data;
            temp_static_read_channel_phase_id = id;
            temp_static_read_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, last, data_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, last, data_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_read_channel_phase_data;
                id = temp_static_read_channel_phase_id;
                user_data = temp_static_read_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr_channel_phase_addr = addr;
            temp_static_write_addr_channel_phase_id = id;
            temp_static_write_addr_channel_phase_addr_user = addr_user;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr_channel_phase_addr;
                id = temp_static_write_addr_channel_phase_id;
                addr_user = temp_static_write_addr_channel_phase_addr_user;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_channel_phase_data = data;
            temp_static_write_channel_phase_write_strobes = write_strobes;
            temp_static_write_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, last, data_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, last, data_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_write_channel_phase_data;
                write_strobes = temp_static_write_channel_phase_write_strobes;
                user_data = temp_static_write_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_resp_channel_phase_id = id;
            temp_static_write_resp_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, write_response_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                id = temp_static_write_resp_channel_phase_id;
                user_data = temp_static_write_resp_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr_channel_cycle_addr = addr;
            temp_static_read_addr_channel_cycle_id = id;
            temp_static_read_addr_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr_channel_cycle_addr;
                id = temp_static_read_addr_channel_cycle_id;
                user_data = temp_static_read_addr_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_channel_cycle_data = data;
            temp_static_read_channel_cycle_id = id;
            temp_static_read_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, last, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, last, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_read_channel_cycle_data;
                id = temp_static_read_channel_cycle_id;
                user_data = temp_static_read_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr_channel_cycle_addr = addr;
            temp_static_write_addr_channel_cycle_id = id;
            temp_static_write_addr_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr_channel_cycle_addr;
                id = temp_static_write_addr_channel_cycle_id;
                user_data = temp_static_write_addr_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_channel_cycle_data = data;
            temp_static_write_channel_cycle_strb = strb;
            temp_static_write_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, last, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, last, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_write_channel_cycle_data;
                strb = temp_static_write_channel_cycle_strb;
                user_data = temp_static_write_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_resp_channel_cycle_id = id;
            temp_static_write_resp_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                id = temp_static_write_resp_channel_cycle_id;
                user_data = temp_static_write_resp_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask


    //-------------------------------------------------------------------------
    // Generic Interface Configuration Support
    //

    import "DPI-C" context dvc_axi4_set_interface = function void axi4_set_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9,
        input int arg10
    );
    import "DPI-C" context dvc_axi4_get_interface = function int axi4_get_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9
    );


    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //
    import "DPI-C" context dvc_axi4_get_full_name = function string axi4_get_full_name();



    //-------------------------------------------------------------------------
    // Abstraction level Support
    //

    import "DPI-C" context dvc_axi4_set_master_end_abstraction_level =
    function void axi4_set_master_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_master_end_abstraction_level =
    function void axi4_get_master_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_set_slave_end_abstraction_level =
    function void axi4_set_slave_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_slave_end_abstraction_level =
    function void axi4_get_slave_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );

    //-------------------------------------------------------------------------
    // Wire Level Interface Support
    //
    logic internal_ACLK = 'z;
    logic internal_ARESETn = 'z;
    logic internal_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_AWADDR = 'z;
    logic [2:0] internal_AWPROT = 'z;
    logic [3:0] internal_AWREGION = 'z;
    logic [7:0] internal_AWLEN = 'z;
    logic [3:0] internal_AWSIZE = 'z;
    logic [1:0] internal_AWBURST = 'z;
    logic internal_AWLOCK = 'z;
    logic [3:0] internal_AWCACHE = 'z;
    logic [3:0] internal_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_AWUSER = 'z;
    logic internal_AWREADY = 'z;
    logic internal_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_ARADDR = 'z;
    logic [2:0] internal_ARPROT = 'z;
    logic [3:0] internal_ARREGION = 'z;
    logic [7:0] internal_ARLEN = 'z;
    logic [3:0] internal_ARSIZE = 'z;
    logic [1:0] internal_ARBURST = 'z;
    logic internal_ARLOCK = 'z;
    logic [3:0] internal_ARCACHE = 'z;
    logic [3:0] internal_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_ARUSER = 'z;
    logic internal_ARREADY = 'z;
    logic internal_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] internal_RDATA = 'z;
    logic [1:0] internal_RRESP = 'z;
    logic internal_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_RUSER = 'z;
    logic internal_RREADY = 'z;
    logic internal_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] internal_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] internal_WSTRB = 'z;
    logic internal_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_WUSER = 'z;
    logic internal_WREADY = 'z;
    logic internal_BVALID = 'z;
    logic [1:0] internal_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_BUSER = 'z;
    logic internal_BREADY = 'z;
    import "DPI-C" context dvc_axi4_set_ACLK_from_SystemVerilog = function void dvc_axi4_set_ACLK_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit ACLK_param
    );
    import "DPI-C" context dvc_axi4_get_ACLK_into_SystemVerilog = function void dvc_axi4_get_ACLK_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit ACLK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ACLK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARESETn_from_SystemVerilog = function void dvc_axi4_set_ARESETn_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARESETn_param
    );
    import "DPI-C" context dvc_axi4_get_ARESETn_into_SystemVerilog = function void dvc_axi4_get_ARESETn_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARESETn_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARESETn_from_CY;

    import "DPI-C" context dvc_axi4_set_AWVALID_from_SystemVerilog = function void dvc_axi4_set_AWVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWVALID_param
    );
    import "DPI-C" context dvc_axi4_get_AWVALID_into_SystemVerilog = function void dvc_axi4_get_AWVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWADDR_from_SystemVerilog_index1 = function void dvc_axi4_set_AWADDR_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWADDR_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWADDR_from_SystemVerilog = function void dvc_axi4_propagate_AWADDR_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWADDR_into_SystemVerilog = function void dvc_axi4_get_AWADDR_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWADDR_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_AWPROT_from_SystemVerilog = function void dvc_axi4_set_AWPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] AWPROT_param
    );
    import "DPI-C" context dvc_axi4_get_AWPROT_into_SystemVerilog = function void dvc_axi4_get_AWPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] AWPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREGION_from_SystemVerilog = function void dvc_axi4_set_AWREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWREGION_param
    );
    import "DPI-C" context dvc_axi4_get_AWREGION_into_SystemVerilog = function void dvc_axi4_get_AWREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLEN_from_SystemVerilog = function void dvc_axi4_set_AWLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] AWLEN_param
    );
    import "DPI-C" context dvc_axi4_get_AWLEN_into_SystemVerilog = function void dvc_axi4_get_AWLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] AWLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_AWSIZE_from_SystemVerilog = function void dvc_axi4_set_AWSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_AWSIZE_into_SystemVerilog = function void dvc_axi4_get_AWSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWBURST_from_SystemVerilog = function void dvc_axi4_set_AWBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] AWBURST_param
    );
    import "DPI-C" context dvc_axi4_get_AWBURST_into_SystemVerilog = function void dvc_axi4_get_AWBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] AWBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLOCK_from_SystemVerilog = function void dvc_axi4_set_AWLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_AWLOCK_into_SystemVerilog = function void dvc_axi4_get_AWLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_AWCACHE_from_SystemVerilog = function void dvc_axi4_set_AWCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_AWCACHE_into_SystemVerilog = function void dvc_axi4_get_AWCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWQOS_from_SystemVerilog = function void dvc_axi4_set_AWQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWQOS_param
    );
    import "DPI-C" context dvc_axi4_get_AWQOS_into_SystemVerilog = function void dvc_axi4_get_AWQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_AWID_from_SystemVerilog_index1 = function void dvc_axi4_set_AWID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWID_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWID_from_SystemVerilog = function void dvc_axi4_propagate_AWID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWID_into_SystemVerilog = function void dvc_axi4_get_AWID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_AWUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWUSER_from_SystemVerilog = function void dvc_axi4_propagate_AWUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWUSER_into_SystemVerilog = function void dvc_axi4_get_AWUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREADY_from_SystemVerilog = function void dvc_axi4_set_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_AWREADY_into_SystemVerilog = function void dvc_axi4_get_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_ARVALID_from_SystemVerilog = function void dvc_axi4_set_ARVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARVALID_param
    );
    import "DPI-C" context dvc_axi4_get_ARVALID_into_SystemVerilog = function void dvc_axi4_get_ARVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARADDR_from_SystemVerilog_index1 = function void dvc_axi4_set_ARADDR_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARADDR_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARADDR_from_SystemVerilog = function void dvc_axi4_propagate_ARADDR_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARADDR_into_SystemVerilog = function void dvc_axi4_get_ARADDR_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARADDR_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_ARPROT_from_SystemVerilog = function void dvc_axi4_set_ARPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] ARPROT_param
    );
    import "DPI-C" context dvc_axi4_get_ARPROT_into_SystemVerilog = function void dvc_axi4_get_ARPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] ARPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREGION_from_SystemVerilog = function void dvc_axi4_set_ARREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARREGION_param
    );
    import "DPI-C" context dvc_axi4_get_ARREGION_into_SystemVerilog = function void dvc_axi4_get_ARREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLEN_from_SystemVerilog = function void dvc_axi4_set_ARLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] ARLEN_param
    );
    import "DPI-C" context dvc_axi4_get_ARLEN_into_SystemVerilog = function void dvc_axi4_get_ARLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] ARLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_ARSIZE_from_SystemVerilog = function void dvc_axi4_set_ARSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_ARSIZE_into_SystemVerilog = function void dvc_axi4_get_ARSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARBURST_from_SystemVerilog = function void dvc_axi4_set_ARBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] ARBURST_param
    );
    import "DPI-C" context dvc_axi4_get_ARBURST_into_SystemVerilog = function void dvc_axi4_get_ARBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] ARBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLOCK_from_SystemVerilog = function void dvc_axi4_set_ARLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_ARLOCK_into_SystemVerilog = function void dvc_axi4_get_ARLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARCACHE_from_SystemVerilog = function void dvc_axi4_set_ARCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_ARCACHE_into_SystemVerilog = function void dvc_axi4_get_ARCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARQOS_from_SystemVerilog = function void dvc_axi4_set_ARQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARQOS_param
    );
    import "DPI-C" context dvc_axi4_get_ARQOS_into_SystemVerilog = function void dvc_axi4_get_ARQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_ARID_from_SystemVerilog_index1 = function void dvc_axi4_set_ARID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARID_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARID_from_SystemVerilog = function void dvc_axi4_propagate_ARID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARID_into_SystemVerilog = function void dvc_axi4_get_ARID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_ARUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARUSER_from_SystemVerilog = function void dvc_axi4_propagate_ARUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARUSER_into_SystemVerilog = function void dvc_axi4_get_ARUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREADY_from_SystemVerilog = function void dvc_axi4_set_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_ARREADY_into_SystemVerilog = function void dvc_axi4_get_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_RVALID_from_SystemVerilog = function void dvc_axi4_set_RVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RVALID_param
    );
    import "DPI-C" context dvc_axi4_get_RVALID_into_SystemVerilog = function void dvc_axi4_get_RVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_RVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_RDATA_from_SystemVerilog_index1 = function void dvc_axi4_set_RDATA_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RDATA_param
    );
    import "DPI-C" context dvc_axi4_propagate_RDATA_from_SystemVerilog = function void dvc_axi4_propagate_RDATA_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RDATA_into_SystemVerilog = function void dvc_axi4_get_RDATA_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RDATA_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_RRESP_from_SystemVerilog = function void dvc_axi4_set_RRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] RRESP_param
    );
    import "DPI-C" context dvc_axi4_get_RRESP_into_SystemVerilog = function void dvc_axi4_get_RRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] RRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_RRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_RLAST_from_SystemVerilog = function void dvc_axi4_set_RLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RLAST_param
    );
    import "DPI-C" context dvc_axi4_get_RLAST_into_SystemVerilog = function void dvc_axi4_get_RLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_RLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_RID_from_SystemVerilog_index1 = function void dvc_axi4_set_RID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RID_param
    );
    import "DPI-C" context dvc_axi4_propagate_RID_from_SystemVerilog = function void dvc_axi4_propagate_RID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RID_into_SystemVerilog = function void dvc_axi4_get_RID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RID_from_CY;

    import "DPI-C" context dvc_axi4_set_RUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_RUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_RUSER_from_SystemVerilog = function void dvc_axi4_propagate_RUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RUSER_into_SystemVerilog = function void dvc_axi4_get_RUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_RREADY_from_SystemVerilog = function void dvc_axi4_set_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_RREADY_into_SystemVerilog = function void dvc_axi4_get_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_WVALID_from_SystemVerilog = function void dvc_axi4_set_WVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WVALID_param
    );
    import "DPI-C" context dvc_axi4_get_WVALID_into_SystemVerilog = function void dvc_axi4_get_WVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_WVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_WDATA_from_SystemVerilog_index1 = function void dvc_axi4_set_WDATA_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WDATA_param
    );
    import "DPI-C" context dvc_axi4_propagate_WDATA_from_SystemVerilog = function void dvc_axi4_propagate_WDATA_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WDATA_into_SystemVerilog = function void dvc_axi4_get_WDATA_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WDATA_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_WSTRB_from_SystemVerilog_index1 = function void dvc_axi4_set_WSTRB_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WSTRB_param
    );
    import "DPI-C" context dvc_axi4_propagate_WSTRB_from_SystemVerilog = function void dvc_axi4_propagate_WSTRB_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WSTRB_into_SystemVerilog = function void dvc_axi4_get_WSTRB_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WSTRB_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WSTRB_from_CY;

    import "DPI-C" context dvc_axi4_set_WLAST_from_SystemVerilog = function void dvc_axi4_set_WLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WLAST_param
    );
    import "DPI-C" context dvc_axi4_get_WLAST_into_SystemVerilog = function void dvc_axi4_get_WLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_WLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_WUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_WUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_WUSER_from_SystemVerilog = function void dvc_axi4_propagate_WUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WUSER_into_SystemVerilog = function void dvc_axi4_get_WUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_WREADY_from_SystemVerilog = function void dvc_axi4_set_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_WREADY_into_SystemVerilog = function void dvc_axi4_get_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_BVALID_from_SystemVerilog = function void dvc_axi4_set_BVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BVALID_param
    );
    import "DPI-C" context dvc_axi4_get_BVALID_into_SystemVerilog = function void dvc_axi4_get_BVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_BVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_BRESP_from_SystemVerilog = function void dvc_axi4_set_BRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] BRESP_param
    );
    import "DPI-C" context dvc_axi4_get_BRESP_into_SystemVerilog = function void dvc_axi4_get_BRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] BRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_BRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_BID_from_SystemVerilog_index1 = function void dvc_axi4_set_BID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  BID_param
    );
    import "DPI-C" context dvc_axi4_propagate_BID_from_SystemVerilog = function void dvc_axi4_propagate_BID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_BID_into_SystemVerilog = function void dvc_axi4_get_BID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_BID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_BID_from_CY;

    import "DPI-C" context dvc_axi4_set_BUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_BUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  BUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_BUSER_from_SystemVerilog = function void dvc_axi4_propagate_BUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_BUSER_into_SystemVerilog = function void dvc_axi4_get_BUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_BUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_BUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_BREADY_from_SystemVerilog = function void dvc_axi4_set_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_BREADY_into_SystemVerilog = function void dvc_axi4_get_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog = function void dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_axi4lite_interface_param
    );
    import "DPI-C" context dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog = function void dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_axi4lite_interface_param

    );
    export "DPI-C" function dvc_axi4_set_config_axi4lite_interface_from_CY;

    import "DPI-C" context dvc_axi4_set_config_interface_type_from_SystemVerilog = function void dvc_axi4_set_config_interface_type_from_SystemVerilog
    (
        input longint _iface_ref,
        input axi4_interface_type_e config_interface_type_param
    );
    import "DPI-C" context dvc_axi4_get_config_interface_type_into_SystemVerilog = function void dvc_axi4_get_config_interface_type_into_SystemVerilog
    (
        input longint _iface_ref,
        output axi4_interface_type_e config_interface_type_param

    );
    export "DPI-C" function dvc_axi4_set_config_interface_type_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_rlast_from_SystemVerilog = function void dvc_axi4_set_config_enable_rlast_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_rlast_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_rlast_into_SystemVerilog = function void dvc_axi4_get_config_enable_rlast_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_rlast_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_rlast_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_id_from_SystemVerilog = function void dvc_axi4_set_config_enable_id_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_id_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_id_into_SystemVerilog = function void dvc_axi4_get_config_enable_id_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_id_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_id_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog = function void dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_all_assertions_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog = function void dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_all_assertions_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_all_assertions_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_assertion_from_SystemVerilog = function void dvc_axi4_set_config_enable_assertion_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit [255:0] config_enable_assertion_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_assertion_into_SystemVerilog = function void dvc_axi4_get_config_enable_assertion_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit [255:0] config_enable_assertion_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_assertion_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog = function void dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_user_sideband_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog = function void dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_user_sideband_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_user_sideband_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog = function void dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_slave_exclusive_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog = function void dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_slave_exclusive_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_slave_exclusive_from_CY;

    import "DPI-C" context dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog = function void dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_read_data_reordering_depth_param
    );
    import "DPI-C" context dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog = function void dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_read_data_reordering_depth_param

    );
    export "DPI-C" function dvc_axi4_set_config_read_data_reordering_depth_from_CY;

    import "DPI-C" context dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1 = function void dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input bit  config_slave_start_addr_param
    );
    import "DPI-C" context dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog = function void dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_config_slave_start_addr_from_CY_index1;

    import "DPI-C" context dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1 = function void dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input bit  config_slave_end_addr_param
    );
    import "DPI-C" context dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog = function void dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_config_slave_end_addr_from_CY_index1;

    import "DPI-C" context dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog = function void dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_transaction_time_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog = function void dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_transaction_time_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_transaction_time_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog = function void dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_burst_timeout_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog = function void dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_burst_timeout_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_burst_timeout_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_RVALID_assertion_to_RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_RVALID_assertion_to_RREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_BVALID_assertion_to_BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_BVALID_assertion_to_BREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_WVALID_assertion_to_WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_WVALID_assertion_to_WREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_qos_from_SystemVerilog = function void dvc_axi4_set_config_enable_qos_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_qos_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_qos_into_SystemVerilog = function void dvc_axi4_get_config_enable_qos_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_qos_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_qos_from_CY;

    import "DPI-C" context dvc_axi4_set_config_setup_time_from_SystemVerilog = function void dvc_axi4_set_config_setup_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_setup_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_setup_time_into_SystemVerilog = function void dvc_axi4_get_config_setup_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_setup_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_setup_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_hold_time_from_SystemVerilog = function void dvc_axi4_set_config_hold_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_hold_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_hold_time_into_SystemVerilog = function void dvc_axi4_get_config_hold_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_hold_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_hold_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_reads_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_reads_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_reads_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_writes_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_writes_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_writes_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_wr_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_wr_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_wr_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rd_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rd_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rd_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rw_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rw_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rw_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_wr_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_wr_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_wr_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_rd_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_rd_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_rd_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_is_issuing_from_SystemVerilog = function void dvc_axi4_set_config_is_issuing_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_is_issuing_param
    );
    import "DPI-C" context dvc_axi4_get_config_is_issuing_into_SystemVerilog = function void dvc_axi4_get_config_is_issuing_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_is_issuing_param

    );
    export "DPI-C" function dvc_axi4_set_config_is_issuing_from_CY;

    function void dvc_axi4_initialise_ACLK_from_CY();
        internal_ACLK = 'z;
        m_ACLK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARESETn_from_CY();
        internal_ARESETn = 'z;
        m_ARESETn = 'z;
    endfunction

    function void dvc_axi4_initialise_AWVALID_from_CY();
        internal_AWVALID = 'z;
        m_AWVALID = 'z;
    endfunction

    function void dvc_axi4_set_AWADDR_from_CY_index1( int _this_dot_1, logic  AWADDR_param );
        internal_AWADDR[_this_dot_1] = AWADDR_param;
    endfunction

    function void dvc_axi4_initialise_AWADDR_from_CY();
        internal_AWADDR = 'z;
        m_AWADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_AWPROT_from_CY();
        internal_AWPROT = 'z;
        m_AWPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREGION_from_CY();
        internal_AWREGION = 'z;
        m_AWREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLEN_from_CY();
        internal_AWLEN = 'z;
        m_AWLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_AWSIZE_from_CY();
        internal_AWSIZE = 'z;
        m_AWSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWBURST_from_CY();
        internal_AWBURST = 'z;
        m_AWBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLOCK_from_CY();
        internal_AWLOCK = 'z;
        m_AWLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_AWCACHE_from_CY();
        internal_AWCACHE = 'z;
        m_AWCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWQOS_from_CY();
        internal_AWQOS = 'z;
        m_AWQOS = 'z;
    endfunction

    function void dvc_axi4_set_AWID_from_CY_index1( int _this_dot_1, logic  AWID_param );
        internal_AWID[_this_dot_1] = AWID_param;
    endfunction

    function void dvc_axi4_initialise_AWID_from_CY();
        internal_AWID = 'z;
        m_AWID = 'z;
    endfunction

    function void dvc_axi4_set_AWUSER_from_CY_index1( int _this_dot_1, logic  AWUSER_param );
        internal_AWUSER[_this_dot_1] = AWUSER_param;
    endfunction

    function void dvc_axi4_initialise_AWUSER_from_CY();
        internal_AWUSER = 'z;
        m_AWUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREADY_from_CY();
        internal_AWREADY = 'z;
        m_AWREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_ARVALID_from_CY();
        internal_ARVALID = 'z;
        m_ARVALID = 'z;
    endfunction

    function void dvc_axi4_set_ARADDR_from_CY_index1( int _this_dot_1, logic  ARADDR_param );
        internal_ARADDR[_this_dot_1] = ARADDR_param;
    endfunction

    function void dvc_axi4_initialise_ARADDR_from_CY();
        internal_ARADDR = 'z;
        m_ARADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_ARPROT_from_CY();
        internal_ARPROT = 'z;
        m_ARPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREGION_from_CY();
        internal_ARREGION = 'z;
        m_ARREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLEN_from_CY();
        internal_ARLEN = 'z;
        m_ARLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_ARSIZE_from_CY();
        internal_ARSIZE = 'z;
        m_ARSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARBURST_from_CY();
        internal_ARBURST = 'z;
        m_ARBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLOCK_from_CY();
        internal_ARLOCK = 'z;
        m_ARLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARCACHE_from_CY();
        internal_ARCACHE = 'z;
        m_ARCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARQOS_from_CY();
        internal_ARQOS = 'z;
        m_ARQOS = 'z;
    endfunction

    function void dvc_axi4_set_ARID_from_CY_index1( int _this_dot_1, logic  ARID_param );
        internal_ARID[_this_dot_1] = ARID_param;
    endfunction

    function void dvc_axi4_initialise_ARID_from_CY();
        internal_ARID = 'z;
        m_ARID = 'z;
    endfunction

    function void dvc_axi4_set_ARUSER_from_CY_index1( int _this_dot_1, logic  ARUSER_param );
        internal_ARUSER[_this_dot_1] = ARUSER_param;
    endfunction

    function void dvc_axi4_initialise_ARUSER_from_CY();
        internal_ARUSER = 'z;
        m_ARUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREADY_from_CY();
        internal_ARREADY = 'z;
        m_ARREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_RVALID_from_CY();
        internal_RVALID = 'z;
        m_RVALID = 'z;
    endfunction

    function void dvc_axi4_set_RDATA_from_CY_index1( int _this_dot_1, logic  RDATA_param );
        internal_RDATA[_this_dot_1] = RDATA_param;
    endfunction

    function void dvc_axi4_initialise_RDATA_from_CY();
        internal_RDATA = 'z;
        m_RDATA = 'z;
    endfunction

    function void dvc_axi4_initialise_RRESP_from_CY();
        internal_RRESP = 'z;
        m_RRESP = 'z;
    endfunction

    function void dvc_axi4_initialise_RLAST_from_CY();
        internal_RLAST = 'z;
        m_RLAST = 'z;
    endfunction

    function void dvc_axi4_set_RID_from_CY_index1( int _this_dot_1, logic  RID_param );
        internal_RID[_this_dot_1] = RID_param;
    endfunction

    function void dvc_axi4_initialise_RID_from_CY();
        internal_RID = 'z;
        m_RID = 'z;
    endfunction

    function void dvc_axi4_set_RUSER_from_CY_index1( int _this_dot_1, logic  RUSER_param );
        internal_RUSER[_this_dot_1] = RUSER_param;
    endfunction

    function void dvc_axi4_initialise_RUSER_from_CY();
        internal_RUSER = 'z;
        m_RUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_RREADY_from_CY();
        internal_RREADY = 'z;
        m_RREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_WVALID_from_CY();
        internal_WVALID = 'z;
        m_WVALID = 'z;
    endfunction

    function void dvc_axi4_set_WDATA_from_CY_index1( int _this_dot_1, logic  WDATA_param );
        internal_WDATA[_this_dot_1] = WDATA_param;
    endfunction

    function void dvc_axi4_initialise_WDATA_from_CY();
        internal_WDATA = 'z;
        m_WDATA = 'z;
    endfunction

    function void dvc_axi4_set_WSTRB_from_CY_index1( int _this_dot_1, logic  WSTRB_param );
        internal_WSTRB[_this_dot_1] = WSTRB_param;
    endfunction

    function void dvc_axi4_initialise_WSTRB_from_CY();
        internal_WSTRB = 'z;
        m_WSTRB = 'z;
    endfunction

    function void dvc_axi4_initialise_WLAST_from_CY();
        internal_WLAST = 'z;
        m_WLAST = 'z;
    endfunction

    function void dvc_axi4_set_WUSER_from_CY_index1( int _this_dot_1, logic  WUSER_param );
        internal_WUSER[_this_dot_1] = WUSER_param;
    endfunction

    function void dvc_axi4_initialise_WUSER_from_CY();
        internal_WUSER = 'z;
        m_WUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_WREADY_from_CY();
        internal_WREADY = 'z;
        m_WREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_BVALID_from_CY();
        internal_BVALID = 'z;
        m_BVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_BRESP_from_CY();
        internal_BRESP = 'z;
        m_BRESP = 'z;
    endfunction

    function void dvc_axi4_set_BID_from_CY_index1( int _this_dot_1, logic  BID_param );
        internal_BID[_this_dot_1] = BID_param;
    endfunction

    function void dvc_axi4_initialise_BID_from_CY();
        internal_BID = 'z;
        m_BID = 'z;
    endfunction

    function void dvc_axi4_set_BUSER_from_CY_index1( int _this_dot_1, logic  BUSER_param );
        internal_BUSER[_this_dot_1] = BUSER_param;
    endfunction

    function void dvc_axi4_initialise_BUSER_from_CY();
        internal_BUSER = 'z;
        m_BUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_BREADY_from_CY();
        internal_BREADY = 'z;
        m_BREADY = 'z;
    endfunction

    function void dvc_axi4_set_config_axi4lite_interface_from_CY( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction

    function void dvc_axi4_set_config_interface_type_from_CY( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction

    function void dvc_axi4_set_config_enable_rlast_from_CY( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction

    function void dvc_axi4_set_config_enable_id_from_CY( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction

    function void dvc_axi4_set_config_enable_all_assertions_from_CY( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction

    function void dvc_axi4_set_config_enable_assertion_from_CY( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction

    function void dvc_axi4_set_config_enable_user_sideband_from_CY( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction

    function void dvc_axi4_set_config_enable_slave_exclusive_from_CY( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction

    function void dvc_axi4_set_config_read_data_reordering_depth_from_CY( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction

    function void dvc_axi4_set_config_slave_start_addr_from_CY_index1( int _this_dot_1, bit  config_slave_start_addr_param );
        config_slave_start_addr[_this_dot_1] = config_slave_start_addr_param;
    endfunction

    function void dvc_axi4_set_config_slave_end_addr_from_CY_index1( int _this_dot_1, bit  config_slave_end_addr_param );
        config_slave_end_addr[_this_dot_1] = config_slave_end_addr_param;
    endfunction

    function void dvc_axi4_set_config_max_transaction_time_factor_from_CY( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction

    function void dvc_axi4_set_config_burst_timeout_factor_from_CY( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction

    function void dvc_axi4_set_config_enable_qos_from_CY( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction

    function void dvc_axi4_set_config_setup_time_from_CY( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction

    function void dvc_axi4_set_config_hold_time_from_CY( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_reads_from_CY( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_writes_from_CY( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_wr_from_CY( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rd_from_CY( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rw_from_CY( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_wr_phase_from_CY( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_rd_phase_from_CY( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction

    function void dvc_axi4_set_config_is_issuing_from_CY( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //--------------------------------------------------------------------------
    //
    // Group:- TLM Interface Support
    //
    //--------------------------------------------------------------------------
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_addr = function axi4_get_temp_static_rw_transaction_addr;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_addr = function axi4_set_temp_static_rw_transaction_addr;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_id = function axi4_get_temp_static_rw_transaction_id;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_id = function axi4_set_temp_static_rw_transaction_id;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_addr_user_data = function axi4_get_temp_static_rw_transaction_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_addr_user_data = function axi4_set_temp_static_rw_transaction_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_words = function axi4_get_temp_static_rw_transaction_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_words = function axi4_set_temp_static_rw_transaction_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_write_strobes = function axi4_get_temp_static_rw_transaction_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_write_strobes = function axi4_set_temp_static_rw_transaction_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_wdata_user_data = function axi4_get_temp_static_rw_transaction_wdata_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_wdata_user_data = function axi4_set_temp_static_rw_transaction_wdata_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_resp = function axi4_get_temp_static_rw_transaction_resp;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_resp = function axi4_set_temp_static_rw_transaction_resp;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_resp_user_data = function axi4_get_temp_static_rw_transaction_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_resp_user_data = function axi4_set_temp_static_rw_transaction_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_valid_delay = function axi4_get_temp_static_rw_transaction_data_valid_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_valid_delay = function axi4_set_temp_static_rw_transaction_data_valid_delay;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_ready_delay = function axi4_get_temp_static_rw_transaction_data_ready_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_ready_delay = function axi4_set_temp_static_rw_transaction_data_ready_delay;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_write_data_beats_delay = function axi4_get_temp_static_rw_transaction_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_write_data_beats_delay = function axi4_set_temp_static_rw_transaction_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr = function axi4_get_temp_static_read_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr = function axi4_set_temp_static_read_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_rresp_delays = function axi4_get_temp_static_read_rresp_delays;
    export "DPI-C" dvc_axi4_set_temp_static_read_rresp_delays = function axi4_set_temp_static_read_rresp_delays;
    export "DPI-C" dvc_axi4_get_temp_static_read_id = function axi4_get_temp_static_read_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_id = function axi4_set_temp_static_read_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_user_data = function axi4_get_temp_static_read_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_user_data = function axi4_set_temp_static_read_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_words = function axi4_get_temp_static_read_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_words = function axi4_set_temp_static_read_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_read_resp = function axi4_get_temp_static_read_resp;
    export "DPI-C" dvc_axi4_set_temp_static_read_resp = function axi4_set_temp_static_read_resp;
    export "DPI-C" dvc_axi4_get_temp_static_read_resp_user_data = function axi4_get_temp_static_read_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_resp_user_data = function axi4_set_temp_static_read_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_start_time = function axi4_get_temp_static_read_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_start_time = function axi4_set_temp_static_read_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_end_time = function axi4_get_temp_static_read_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_end_time = function axi4_set_temp_static_read_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr = function axi4_get_temp_static_write_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr = function axi4_set_temp_static_write_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_id = function axi4_get_temp_static_write_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_id = function axi4_set_temp_static_write_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_user_data = function axi4_get_temp_static_write_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_user_data = function axi4_set_temp_static_write_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_words = function axi4_get_temp_static_write_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_words = function axi4_set_temp_static_write_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_write_write_strobes = function axi4_get_temp_static_write_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_write_strobes = function axi4_set_temp_static_write_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_wdata_user_data = function axi4_get_temp_static_write_wdata_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_wdata_user_data = function axi4_set_temp_static_write_wdata_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_user_data = function axi4_get_temp_static_write_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_user_data = function axi4_set_temp_static_write_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_start_time = function axi4_get_temp_static_write_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_start_time = function axi4_set_temp_static_write_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_end_time = function axi4_get_temp_static_write_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_end_time = function axi4_set_temp_static_write_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_write_data_beats_delay = function axi4_get_temp_static_write_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_write_write_data_beats_delay = function axi4_set_temp_static_write_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_words = function axi4_get_temp_static_write_data_burst_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_words = function axi4_set_temp_static_write_data_burst_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_write_strobes = function axi4_get_temp_static_write_data_burst_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_write_strobes = function axi4_set_temp_static_write_data_burst_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_user_data = function axi4_get_temp_static_write_data_burst_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_user_data = function axi4_set_temp_static_write_data_burst_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_start_time = function axi4_get_temp_static_write_data_burst_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_start_time = function axi4_set_temp_static_write_data_burst_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_end_time = function axi4_get_temp_static_write_data_burst_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_end_time = function axi4_set_temp_static_write_data_burst_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_write_data_beats_delay = function axi4_get_temp_static_write_data_burst_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_write_data_beats_delay = function axi4_set_temp_static_write_data_burst_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_valid2ready = function axi4_get_temp_static_write_data_burst_data_valid2ready;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_valid2ready = function axi4_set_temp_static_write_data_burst_data_valid2ready;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data2data_clock = function axi4_get_temp_static_write_data_burst_data2data_clock;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data2data_clock = function axi4_set_temp_static_write_data_burst_data2data_clock;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_addr = function axi4_get_temp_static_read_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_addr = function axi4_set_temp_static_read_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_id = function axi4_get_temp_static_read_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_id = function axi4_set_temp_static_read_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_addr_user = function axi4_get_temp_static_read_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_addr_user = function axi4_set_temp_static_read_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_data = function axi4_get_temp_static_read_channel_phase_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_data = function axi4_set_temp_static_read_channel_phase_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_id = function axi4_get_temp_static_read_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_id = function axi4_set_temp_static_read_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_user_data = function axi4_get_temp_static_read_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_user_data = function axi4_set_temp_static_read_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_addr = function axi4_get_temp_static_write_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_addr = function axi4_set_temp_static_write_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_id = function axi4_get_temp_static_write_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_id = function axi4_set_temp_static_write_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_addr_user = function axi4_get_temp_static_write_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_addr_user = function axi4_set_temp_static_write_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_data = function axi4_get_temp_static_write_channel_phase_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_data = function axi4_set_temp_static_write_channel_phase_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_write_strobes = function axi4_get_temp_static_write_channel_phase_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_write_strobes = function axi4_set_temp_static_write_channel_phase_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_user_data = function axi4_get_temp_static_write_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_user_data = function axi4_set_temp_static_write_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_phase_id = function axi4_get_temp_static_write_resp_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_phase_id = function axi4_set_temp_static_write_resp_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_phase_user_data = function axi4_get_temp_static_write_resp_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_phase_user_data = function axi4_set_temp_static_write_resp_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_addr = function axi4_get_temp_static_read_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_addr = function axi4_set_temp_static_read_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_id = function axi4_get_temp_static_read_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_id = function axi4_set_temp_static_read_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_user_data = function axi4_get_temp_static_read_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_user_data = function axi4_set_temp_static_read_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_data = function axi4_get_temp_static_read_channel_cycle_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_data = function axi4_set_temp_static_read_channel_cycle_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_id = function axi4_get_temp_static_read_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_id = function axi4_set_temp_static_read_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_user_data = function axi4_get_temp_static_read_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_user_data = function axi4_set_temp_static_read_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_addr = function axi4_get_temp_static_write_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_addr = function axi4_set_temp_static_write_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_id = function axi4_get_temp_static_write_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_id = function axi4_set_temp_static_write_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_user_data = function axi4_get_temp_static_write_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_user_data = function axi4_set_temp_static_write_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_data = function axi4_get_temp_static_write_channel_cycle_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_data = function axi4_set_temp_static_write_channel_cycle_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_strb = function axi4_get_temp_static_write_channel_cycle_strb;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_strb = function axi4_set_temp_static_write_channel_cycle_strb;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_user_data = function axi4_get_temp_static_write_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_user_data = function axi4_set_temp_static_write_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_cycle_id = function axi4_get_temp_static_write_resp_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_cycle_id = function axi4_set_temp_static_write_resp_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_cycle_user_data = function axi4_get_temp_static_write_resp_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_cycle_user_data = function axi4_set_temp_static_write_resp_channel_cycle_user_data;
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_rw_e read_or_write,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_rw_e read_or_write,
        inout axi4_prot_e prot,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout int address_valid_delay,
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_rw_e read_or_write,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output int address_valid_delay,
        output int write_response_valid_delay,
        output int address_ready_delay,
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_prot_e prot,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output longint addr_start_time,
        output longint addr_end_time,
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout axi4_response_e resp,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout axi4_response_e resp,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output axi4_response_e resp,
        output longint addr_start_time,
        output longint addr_end_time,
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_SendSendingSent_SystemVerilog =
    task axi4_write_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int burst_length,
        input int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_valid2ready_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data2data_clock_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int wdata_burst_start_clock,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid2ready_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data2data_clock_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output int burst_length,
        output int wdata_burst_start_clock,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_data_burst_SendSendingSent_SystemVerilog =
    task axi4_read_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit last,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output bit last,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit last,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit last,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((4) - 1):0] region,
        output axi4_cache_e cache,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [((4) - 1):0] qos,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit last,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output bit last,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((4) - 1):0] region,
        output axi4_cache_e cache,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [((4) - 1):0] qos,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit last,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit last,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_fn_get_raddr_phase_info = function void fn_get_raddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id,
        output int num_beats_rcvd
    );

    import "DPI-C" context dvc_axi4_fn_get_waddr_phase_info = function void fn_get_waddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id
    );

    import "DPI-C" context dvc_axi4_fn_add_addr_map_entry = function void fn_add_addr_map_entry_C
    (
        input string region,
        input longint unsigned size
    );

    import "DPI-C" context dvc_axi4_fn_add_wr_delay = function void fn_add_wr_delay_C
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        input int data2data_DIMS0
    );

    import "DPI-C" context dvc_axi4_fn_delete_wr_delay = function void fn_delete_wr_delay_C
    (
        input string region,
        input bit [17:0] id
    );

    import "DPI-C" context dvc_axi4_fn_drive_signals = function void fn_drive_signals_C
    (
        input string signal,
        input longint value
    );

    import "DPI-C" context dvc_axi4_fn_set_address_map_entry = function void fn_set_address_map_entry_C
    (
    );

    import "DPI-C" context dvc_axi4_fn_set_wr_def_delays = function void fn_set_wr_def_delays_C
    (
        input int unsigned min_addr2data,
        input int min_data2data_DIMS0,
        input int unsigned max_addr2data,
        input int max_data2data_DIMS0
    );

    import "DPI-C" context dvc_axi4_fn_print_perf_stats = function void fn_print_perf_stats_C
    (
    );

    // Waiter task and control
    reg sim_wait_for_control = 0;

    always @(posedge sim_wait_for_control)
    begin
        disable wait_for;
        sim_wait_for_control = 0;
    end

    export "DPI-C" dvc_axi4_wait_for = task wait_for;

    task wait_for();
        begin
            wait(0 == 1);
        end
    endtask

    // Drive wires (from Cohesive) 
    assign ACLK = internal_ACLK;
    assign ARESETn = internal_ARESETn;
    assign AWVALID = internal_AWVALID;
    assign AWADDR = internal_AWADDR;
    assign AWPROT = internal_AWPROT;
    assign AWREGION = internal_AWREGION;
    assign AWLEN = internal_AWLEN;
    assign AWSIZE = internal_AWSIZE;
    assign AWBURST = internal_AWBURST;
    assign AWLOCK = internal_AWLOCK;
    assign AWCACHE = internal_AWCACHE;
    assign AWQOS = internal_AWQOS;
    assign AWID = internal_AWID;
    assign AWUSER = internal_AWUSER;
    assign AWREADY = internal_AWREADY;
    assign ARVALID = internal_ARVALID;
    assign ARADDR = internal_ARADDR;
    assign ARPROT = internal_ARPROT;
    assign ARREGION = internal_ARREGION;
    assign ARLEN = internal_ARLEN;
    assign ARSIZE = internal_ARSIZE;
    assign ARBURST = internal_ARBURST;
    assign ARLOCK = internal_ARLOCK;
    assign ARCACHE = internal_ARCACHE;
    assign ARQOS = internal_ARQOS;
    assign ARID = internal_ARID;
    assign ARUSER = internal_ARUSER;
    assign ARREADY = internal_ARREADY;
    assign RVALID = internal_RVALID;
    assign RDATA = internal_RDATA;
    assign RRESP = internal_RRESP;
    assign RLAST = internal_RLAST;
    assign RID = internal_RID;
    assign RUSER = internal_RUSER;
    assign RREADY = internal_RREADY;
    assign WVALID = internal_WVALID;
    assign WDATA = internal_WDATA;
    assign WSTRB = internal_WSTRB;
    assign WLAST = internal_WLAST;
    assign WUSER = internal_WUSER;
    assign WREADY = internal_WREADY;
    assign BVALID = internal_BVALID;
    assign BRESP = internal_BRESP;
    assign BID = internal_BID;
    assign BUSER = internal_BUSER;
    assign BREADY = internal_BREADY;
    // Drive wires (from User) 
    assign ACLK = m_ACLK;
    assign ARESETn = m_ARESETn;
    assign AWVALID = m_AWVALID;
    assign AWADDR = m_AWADDR;
    assign AWPROT = m_AWPROT;
    assign AWREGION = m_AWREGION;
    assign AWLEN = m_AWLEN;
    assign AWSIZE = m_AWSIZE;
    assign AWBURST = m_AWBURST;
    assign AWLOCK = m_AWLOCK;
    assign AWCACHE = m_AWCACHE;
    assign AWQOS = m_AWQOS;
    assign AWID = m_AWID;
    assign AWUSER = m_AWUSER;
    assign AWREADY = m_AWREADY;
    assign ARVALID = m_ARVALID;
    assign ARADDR = m_ARADDR;
    assign ARPROT = m_ARPROT;
    assign ARREGION = m_ARREGION;
    assign ARLEN = m_ARLEN;
    assign ARSIZE = m_ARSIZE;
    assign ARBURST = m_ARBURST;
    assign ARLOCK = m_ARLOCK;
    assign ARCACHE = m_ARCACHE;
    assign ARQOS = m_ARQOS;
    assign ARID = m_ARID;
    assign ARUSER = m_ARUSER;
    assign ARREADY = m_ARREADY;
    assign RVALID = m_RVALID;
    assign RDATA = m_RDATA;
    assign RRESP = m_RRESP;
    assign RLAST = m_RLAST;
    assign RID = m_RID;
    assign RUSER = m_RUSER;
    assign RREADY = m_RREADY;
    assign WVALID = m_WVALID;
    assign WDATA = m_WDATA;
    assign WSTRB = m_WSTRB;
    assign WLAST = m_WLAST;
    assign WUSER = m_WUSER;
    assign WREADY = m_WREADY;
    assign BVALID = m_BVALID;
    assign BRESP = m_BRESP;
    assign BID = m_BID;
    assign BUSER = m_BUSER;
    assign BREADY = m_BREADY;

    reg ACLK_changed = 0;
    reg ARESETn_changed = 0;
    reg AWVALID_changed = 0;
    reg AWADDR_changed = 0;
    reg AWPROT_changed = 0;
    reg AWREGION_changed = 0;
    reg AWLEN_changed = 0;
    reg AWSIZE_changed = 0;
    reg AWBURST_changed = 0;
    reg AWLOCK_changed = 0;
    reg AWCACHE_changed = 0;
    reg AWQOS_changed = 0;
    reg AWID_changed = 0;
    reg AWUSER_changed = 0;
    reg AWREADY_changed = 0;
    reg ARVALID_changed = 0;
    reg ARADDR_changed = 0;
    reg ARPROT_changed = 0;
    reg ARREGION_changed = 0;
    reg ARLEN_changed = 0;
    reg ARSIZE_changed = 0;
    reg ARBURST_changed = 0;
    reg ARLOCK_changed = 0;
    reg ARCACHE_changed = 0;
    reg ARQOS_changed = 0;
    reg ARID_changed = 0;
    reg ARUSER_changed = 0;
    reg ARREADY_changed = 0;
    reg RVALID_changed = 0;
    reg RDATA_changed = 0;
    reg RRESP_changed = 0;
    reg RLAST_changed = 0;
    reg RID_changed = 0;
    reg RUSER_changed = 0;
    reg RREADY_changed = 0;
    reg WVALID_changed = 0;
    reg WDATA_changed = 0;
    reg WSTRB_changed = 0;
    reg WLAST_changed = 0;
    reg WUSER_changed = 0;
    reg WREADY_changed = 0;
    reg BVALID_changed = 0;
    reg BRESP_changed = 0;
    reg BID_changed = 0;
    reg BUSER_changed = 0;
    reg BREADY_changed = 0;
    reg config_axi4lite_interface_changed = 0;
    reg config_interface_type_changed = 0;
    reg config_enable_rlast_changed = 0;
    reg config_enable_id_changed = 0;
    reg config_enable_all_assertions_changed = 0;
    reg config_enable_assertion_changed = 0;
    reg config_enable_user_sideband_changed = 0;
    reg config_enable_slave_exclusive_changed = 0;
    reg config_read_data_reordering_depth_changed = 0;
    reg config_slave_start_addr_changed = 0;
    reg config_slave_end_addr_changed = 0;
    reg config_max_transaction_time_factor_changed = 0;
    reg config_burst_timeout_factor_changed = 0;
    reg config_max_latency_AWVALID_assertion_to_AWREADY_changed = 0;
    reg config_max_latency_ARVALID_assertion_to_ARREADY_changed = 0;
    reg config_max_latency_RVALID_assertion_to_RREADY_changed = 0;
    reg config_max_latency_BVALID_assertion_to_BREADY_changed = 0;
    reg config_max_latency_WVALID_assertion_to_WREADY_changed = 0;
    reg config_enable_qos_changed = 0;
    reg config_setup_time_changed = 0;
    reg config_hold_time_changed = 0;
    reg config_num_max_outstanding_reads_changed = 0;
    reg config_num_max_outstanding_writes_changed = 0;
    reg config_max_outstanding_wr_changed = 0;
    reg config_max_outstanding_rd_changed = 0;
    reg config_max_outstanding_rw_changed = 0;
    reg config_num_outstanding_wr_phase_changed = 0;
    reg config_num_outstanding_rd_phase_changed = 0;
    reg config_is_issuing_changed = 0;

    // SV wire change monitors

    function automatic void axi4_local_set_ACLK_from_SystemVerilog(  );
        dvc_axi4_set_ACLK_from_SystemVerilog( _interface_ref, ACLK); // DPI call to imported task
    endfunction

    always @( ACLK or posedge _check_t0_values )
    begin
        axi4_local_set_ACLK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARESETn_from_SystemVerilog(  );
        dvc_axi4_set_ARESETn_from_SystemVerilog( _interface_ref, ARESETn); // DPI call to imported task
    endfunction

    always @( ARESETn or posedge _check_t0_values )
    begin
        axi4_local_set_ARESETn_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWVALID_from_SystemVerilog(  );
        dvc_axi4_set_AWVALID_from_SystemVerilog( _interface_ref, AWVALID); // DPI call to imported task
    endfunction

    always @( AWVALID or posedge _check_t0_values )
    begin
        axi4_local_set_AWVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWADDR_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWADDR_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWADDR[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWADDR_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWADDR or posedge _check_t0_values )
    begin
        axi4_local_set_AWADDR_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWPROT_from_SystemVerilog(  );
        dvc_axi4_set_AWPROT_from_SystemVerilog( _interface_ref, AWPROT); // DPI call to imported task
    endfunction

    always @( AWPROT or posedge _check_t0_values )
    begin
        axi4_local_set_AWPROT_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWREGION_from_SystemVerilog(  );
        dvc_axi4_set_AWREGION_from_SystemVerilog( _interface_ref, AWREGION); // DPI call to imported task
    endfunction

    always @( AWREGION or posedge _check_t0_values )
    begin
        axi4_local_set_AWREGION_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWLEN_from_SystemVerilog(  );
        dvc_axi4_set_AWLEN_from_SystemVerilog( _interface_ref, AWLEN); // DPI call to imported task
    endfunction

    always @( AWLEN or posedge _check_t0_values )
    begin
        axi4_local_set_AWLEN_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWSIZE_from_SystemVerilog(  );
        dvc_axi4_set_AWSIZE_from_SystemVerilog( _interface_ref, AWSIZE); // DPI call to imported task
    endfunction

    always @( AWSIZE or posedge _check_t0_values )
    begin
        axi4_local_set_AWSIZE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWBURST_from_SystemVerilog(  );
        dvc_axi4_set_AWBURST_from_SystemVerilog( _interface_ref, AWBURST); // DPI call to imported task
    endfunction

    always @( AWBURST or posedge _check_t0_values )
    begin
        axi4_local_set_AWBURST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWLOCK_from_SystemVerilog(  );
        dvc_axi4_set_AWLOCK_from_SystemVerilog( _interface_ref, AWLOCK); // DPI call to imported task
    endfunction

    always @( AWLOCK or posedge _check_t0_values )
    begin
        axi4_local_set_AWLOCK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWCACHE_from_SystemVerilog(  );
        dvc_axi4_set_AWCACHE_from_SystemVerilog( _interface_ref, AWCACHE); // DPI call to imported task
    endfunction

    always @( AWCACHE or posedge _check_t0_values )
    begin
        axi4_local_set_AWCACHE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWQOS_from_SystemVerilog(  );
        dvc_axi4_set_AWQOS_from_SystemVerilog( _interface_ref, AWQOS); // DPI call to imported task
    endfunction

    always @( AWQOS or posedge _check_t0_values )
    begin
        axi4_local_set_AWQOS_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWID or posedge _check_t0_values )
    begin
        axi4_local_set_AWID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWUSER or posedge _check_t0_values )
    begin
        axi4_local_set_AWUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWREADY_from_SystemVerilog(  );
        dvc_axi4_set_AWREADY_from_SystemVerilog( _interface_ref, AWREADY); // DPI call to imported task
    endfunction

    always @( AWREADY or posedge _check_t0_values )
    begin
        axi4_local_set_AWREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARVALID_from_SystemVerilog(  );
        dvc_axi4_set_ARVALID_from_SystemVerilog( _interface_ref, ARVALID); // DPI call to imported task
    endfunction

    always @( ARVALID or posedge _check_t0_values )
    begin
        axi4_local_set_ARVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARADDR_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARADDR_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARADDR[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARADDR_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARADDR or posedge _check_t0_values )
    begin
        axi4_local_set_ARADDR_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARPROT_from_SystemVerilog(  );
        dvc_axi4_set_ARPROT_from_SystemVerilog( _interface_ref, ARPROT); // DPI call to imported task
    endfunction

    always @( ARPROT or posedge _check_t0_values )
    begin
        axi4_local_set_ARPROT_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARREGION_from_SystemVerilog(  );
        dvc_axi4_set_ARREGION_from_SystemVerilog( _interface_ref, ARREGION); // DPI call to imported task
    endfunction

    always @( ARREGION or posedge _check_t0_values )
    begin
        axi4_local_set_ARREGION_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARLEN_from_SystemVerilog(  );
        dvc_axi4_set_ARLEN_from_SystemVerilog( _interface_ref, ARLEN); // DPI call to imported task
    endfunction

    always @( ARLEN or posedge _check_t0_values )
    begin
        axi4_local_set_ARLEN_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARSIZE_from_SystemVerilog(  );
        dvc_axi4_set_ARSIZE_from_SystemVerilog( _interface_ref, ARSIZE); // DPI call to imported task
    endfunction

    always @( ARSIZE or posedge _check_t0_values )
    begin
        axi4_local_set_ARSIZE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARBURST_from_SystemVerilog(  );
        dvc_axi4_set_ARBURST_from_SystemVerilog( _interface_ref, ARBURST); // DPI call to imported task
    endfunction

    always @( ARBURST or posedge _check_t0_values )
    begin
        axi4_local_set_ARBURST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARLOCK_from_SystemVerilog(  );
        dvc_axi4_set_ARLOCK_from_SystemVerilog( _interface_ref, ARLOCK); // DPI call to imported task
    endfunction

    always @( ARLOCK or posedge _check_t0_values )
    begin
        axi4_local_set_ARLOCK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARCACHE_from_SystemVerilog(  );
        dvc_axi4_set_ARCACHE_from_SystemVerilog( _interface_ref, ARCACHE); // DPI call to imported task
    endfunction

    always @( ARCACHE or posedge _check_t0_values )
    begin
        axi4_local_set_ARCACHE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARQOS_from_SystemVerilog(  );
        dvc_axi4_set_ARQOS_from_SystemVerilog( _interface_ref, ARQOS); // DPI call to imported task
    endfunction

    always @( ARQOS or posedge _check_t0_values )
    begin
        axi4_local_set_ARQOS_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARID or posedge _check_t0_values )
    begin
        axi4_local_set_ARID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARUSER or posedge _check_t0_values )
    begin
        axi4_local_set_ARUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARREADY_from_SystemVerilog(  );
        dvc_axi4_set_ARREADY_from_SystemVerilog( _interface_ref, ARREADY); // DPI call to imported task
    endfunction

    always @( ARREADY or posedge _check_t0_values )
    begin
        axi4_local_set_ARREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RVALID_from_SystemVerilog(  );
        dvc_axi4_set_RVALID_from_SystemVerilog( _interface_ref, RVALID); // DPI call to imported task
    endfunction

    always @( RVALID or posedge _check_t0_values )
    begin
        axi4_local_set_RVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RDATA_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_RDATA_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RDATA_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RDATA[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RDATA_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RDATA or posedge _check_t0_values )
    begin
        axi4_local_set_RDATA_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RRESP_from_SystemVerilog(  );
        dvc_axi4_set_RRESP_from_SystemVerilog( _interface_ref, RRESP); // DPI call to imported task
    endfunction

    always @( RRESP or posedge _check_t0_values )
    begin
        axi4_local_set_RRESP_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RLAST_from_SystemVerilog(  );
        dvc_axi4_set_RLAST_from_SystemVerilog( _interface_ref, RLAST); // DPI call to imported task
    endfunction

    always @( RLAST or posedge _check_t0_values )
    begin
        axi4_local_set_RLAST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RID or posedge _check_t0_values )
    begin
        axi4_local_set_RID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RUSER or posedge _check_t0_values )
    begin
        axi4_local_set_RUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RREADY_from_SystemVerilog(  );
        dvc_axi4_set_RREADY_from_SystemVerilog( _interface_ref, RREADY); // DPI call to imported task
    endfunction

    always @( RREADY or posedge _check_t0_values )
    begin
        axi4_local_set_RREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WVALID_from_SystemVerilog(  );
        dvc_axi4_set_WVALID_from_SystemVerilog( _interface_ref, WVALID); // DPI call to imported task
    endfunction

    always @( WVALID or posedge _check_t0_values )
    begin
        axi4_local_set_WVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WDATA_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_WDATA_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_WDATA_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WDATA[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WDATA_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WDATA or posedge _check_t0_values )
    begin
        axi4_local_set_WDATA_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WSTRB_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( (AXI4_WDATA_WIDTH / 8) ); _this_dot_1++)
        begin
            dvc_axi4_set_WSTRB_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WSTRB[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WSTRB_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WSTRB or posedge _check_t0_values )
    begin
        axi4_local_set_WSTRB_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WLAST_from_SystemVerilog(  );
        dvc_axi4_set_WLAST_from_SystemVerilog( _interface_ref, WLAST); // DPI call to imported task
    endfunction

    always @( WLAST or posedge _check_t0_values )
    begin
        axi4_local_set_WLAST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_WUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WUSER or posedge _check_t0_values )
    begin
        axi4_local_set_WUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WREADY_from_SystemVerilog(  );
        dvc_axi4_set_WREADY_from_SystemVerilog( _interface_ref, WREADY); // DPI call to imported task
    endfunction

    always @( WREADY or posedge _check_t0_values )
    begin
        axi4_local_set_WREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BVALID_from_SystemVerilog(  );
        dvc_axi4_set_BVALID_from_SystemVerilog( _interface_ref, BVALID); // DPI call to imported task
    endfunction

    always @( BVALID or posedge _check_t0_values )
    begin
        axi4_local_set_BVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BRESP_from_SystemVerilog(  );
        dvc_axi4_set_BRESP_from_SystemVerilog( _interface_ref, BRESP); // DPI call to imported task
    endfunction

    always @( BRESP or posedge _check_t0_values )
    begin
        axi4_local_set_BRESP_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_BID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,BID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_BID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( BID or posedge _check_t0_values )
    begin
        axi4_local_set_BID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_BUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,BUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_BUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( BUSER or posedge _check_t0_values )
    begin
        axi4_local_set_BUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BREADY_from_SystemVerilog(  );
        dvc_axi4_set_BREADY_from_SystemVerilog( _interface_ref, BREADY); // DPI call to imported task
    endfunction

    always @( BREADY or posedge _check_t0_values )
    begin
        axi4_local_set_BREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end


    // CY wire and variable changed flag monitors

    always @(posedge ACLK_changed or posedge _check_t0_values )
    begin
        while (ACLK_changed == 1'b1)
        begin
            dvc_axi4_get_ACLK_into_SystemVerilog( _interface_ref, internal_ACLK ); // DPI call to imported task
            ACLK_changed = 1'b0;
            #0  #0 if ( ACLK !== internal_ACLK )
            begin
                axi4_local_set_ACLK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARESETn_changed or posedge _check_t0_values )
    begin
        while (ARESETn_changed == 1'b1)
        begin
            logic temp_ARESETn;

            dvc_axi4_get_ARESETn_into_SystemVerilog( _interface_ref, temp_ARESETn ); // DPI call to imported task
            internal_ARESETn = temp_ARESETn;
            ARESETn_changed = 1'b0;
            #0  #0 if ( ARESETn !== internal_ARESETn )
            begin
                axi4_local_set_ARESETn_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWVALID_changed or posedge _check_t0_values )
    begin
        while (AWVALID_changed == 1'b1)
        begin
            logic temp_AWVALID;

            dvc_axi4_get_AWVALID_into_SystemVerilog( _interface_ref, temp_AWVALID ); // DPI call to imported task
            internal_AWVALID = temp_AWVALID;
            AWVALID_changed = 1'b0;
            #0  #0 if ( AWVALID !== internal_AWVALID )
            begin
                axi4_local_set_AWVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWADDR_changed or posedge _check_t0_values )
    begin
        while (AWADDR_changed == 1'b1)
        begin
            dvc_axi4_get_AWADDR_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWADDR_changed = 1'b0;
            #0  #0 if ( AWADDR !== internal_AWADDR )
            begin
                axi4_local_set_AWADDR_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWPROT_changed or posedge _check_t0_values )
    begin
        while (AWPROT_changed == 1'b1)
        begin
            dvc_axi4_get_AWPROT_into_SystemVerilog( _interface_ref, internal_AWPROT ); // DPI call to imported task
            AWPROT_changed = 1'b0;
            #0  #0 if ( AWPROT !== internal_AWPROT )
            begin
                axi4_local_set_AWPROT_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWREGION_changed or posedge _check_t0_values )
    begin
        while (AWREGION_changed == 1'b1)
        begin
            dvc_axi4_get_AWREGION_into_SystemVerilog( _interface_ref, internal_AWREGION ); // DPI call to imported task
            AWREGION_changed = 1'b0;
            #0  #0 if ( AWREGION !== internal_AWREGION )
            begin
                axi4_local_set_AWREGION_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWLEN_changed or posedge _check_t0_values )
    begin
        while (AWLEN_changed == 1'b1)
        begin
            dvc_axi4_get_AWLEN_into_SystemVerilog( _interface_ref, internal_AWLEN ); // DPI call to imported task
            AWLEN_changed = 1'b0;
            #0  #0 if ( AWLEN !== internal_AWLEN )
            begin
                axi4_local_set_AWLEN_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWSIZE_changed or posedge _check_t0_values )
    begin
        while (AWSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_AWSIZE_into_SystemVerilog( _interface_ref, internal_AWSIZE ); // DPI call to imported task
            AWSIZE_changed = 1'b0;
            #0  #0 if ( AWSIZE !== internal_AWSIZE )
            begin
                axi4_local_set_AWSIZE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWBURST_changed or posedge _check_t0_values )
    begin
        while (AWBURST_changed == 1'b1)
        begin
            dvc_axi4_get_AWBURST_into_SystemVerilog( _interface_ref, internal_AWBURST ); // DPI call to imported task
            AWBURST_changed = 1'b0;
            #0  #0 if ( AWBURST !== internal_AWBURST )
            begin
                axi4_local_set_AWBURST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWLOCK_changed or posedge _check_t0_values )
    begin
        while (AWLOCK_changed == 1'b1)
        begin
            logic temp_AWLOCK;

            dvc_axi4_get_AWLOCK_into_SystemVerilog( _interface_ref, temp_AWLOCK ); // DPI call to imported task
            internal_AWLOCK = temp_AWLOCK;
            AWLOCK_changed = 1'b0;
            #0  #0 if ( AWLOCK !== internal_AWLOCK )
            begin
                axi4_local_set_AWLOCK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWCACHE_changed or posedge _check_t0_values )
    begin
        while (AWCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_AWCACHE_into_SystemVerilog( _interface_ref, internal_AWCACHE ); // DPI call to imported task
            AWCACHE_changed = 1'b0;
            #0  #0 if ( AWCACHE !== internal_AWCACHE )
            begin
                axi4_local_set_AWCACHE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWQOS_changed or posedge _check_t0_values )
    begin
        while (AWQOS_changed == 1'b1)
        begin
            dvc_axi4_get_AWQOS_into_SystemVerilog( _interface_ref, internal_AWQOS ); // DPI call to imported task
            AWQOS_changed = 1'b0;
            #0  #0 if ( AWQOS !== internal_AWQOS )
            begin
                axi4_local_set_AWQOS_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWID_changed or posedge _check_t0_values )
    begin
        while (AWID_changed == 1'b1)
        begin
            dvc_axi4_get_AWID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWID_changed = 1'b0;
            #0  #0 if ( AWID !== internal_AWID )
            begin
                axi4_local_set_AWID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWUSER_changed or posedge _check_t0_values )
    begin
        while (AWUSER_changed == 1'b1)
        begin
            dvc_axi4_get_AWUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWUSER_changed = 1'b0;
            #0  #0 if ( AWUSER !== internal_AWUSER )
            begin
                axi4_local_set_AWUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWREADY_changed or posedge _check_t0_values )
    begin
        while (AWREADY_changed == 1'b1)
        begin
            logic temp_AWREADY;

            dvc_axi4_get_AWREADY_into_SystemVerilog( _interface_ref, temp_AWREADY ); // DPI call to imported task
            internal_AWREADY = temp_AWREADY;
            AWREADY_changed = 1'b0;
            #0  #0 if ( AWREADY !== internal_AWREADY )
            begin
                axi4_local_set_AWREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARVALID_changed or posedge _check_t0_values )
    begin
        while (ARVALID_changed == 1'b1)
        begin
            logic temp_ARVALID;

            dvc_axi4_get_ARVALID_into_SystemVerilog( _interface_ref, temp_ARVALID ); // DPI call to imported task
            internal_ARVALID = temp_ARVALID;
            ARVALID_changed = 1'b0;
            #0  #0 if ( ARVALID !== internal_ARVALID )
            begin
                axi4_local_set_ARVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARADDR_changed or posedge _check_t0_values )
    begin
        while (ARADDR_changed == 1'b1)
        begin
            dvc_axi4_get_ARADDR_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARADDR_changed = 1'b0;
            #0  #0 if ( ARADDR !== internal_ARADDR )
            begin
                axi4_local_set_ARADDR_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARPROT_changed or posedge _check_t0_values )
    begin
        while (ARPROT_changed == 1'b1)
        begin
            dvc_axi4_get_ARPROT_into_SystemVerilog( _interface_ref, internal_ARPROT ); // DPI call to imported task
            ARPROT_changed = 1'b0;
            #0  #0 if ( ARPROT !== internal_ARPROT )
            begin
                axi4_local_set_ARPROT_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARREGION_changed or posedge _check_t0_values )
    begin
        while (ARREGION_changed == 1'b1)
        begin
            dvc_axi4_get_ARREGION_into_SystemVerilog( _interface_ref, internal_ARREGION ); // DPI call to imported task
            ARREGION_changed = 1'b0;
            #0  #0 if ( ARREGION !== internal_ARREGION )
            begin
                axi4_local_set_ARREGION_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARLEN_changed or posedge _check_t0_values )
    begin
        while (ARLEN_changed == 1'b1)
        begin
            dvc_axi4_get_ARLEN_into_SystemVerilog( _interface_ref, internal_ARLEN ); // DPI call to imported task
            ARLEN_changed = 1'b0;
            #0  #0 if ( ARLEN !== internal_ARLEN )
            begin
                axi4_local_set_ARLEN_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARSIZE_changed or posedge _check_t0_values )
    begin
        while (ARSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_ARSIZE_into_SystemVerilog( _interface_ref, internal_ARSIZE ); // DPI call to imported task
            ARSIZE_changed = 1'b0;
            #0  #0 if ( ARSIZE !== internal_ARSIZE )
            begin
                axi4_local_set_ARSIZE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARBURST_changed or posedge _check_t0_values )
    begin
        while (ARBURST_changed == 1'b1)
        begin
            dvc_axi4_get_ARBURST_into_SystemVerilog( _interface_ref, internal_ARBURST ); // DPI call to imported task
            ARBURST_changed = 1'b0;
            #0  #0 if ( ARBURST !== internal_ARBURST )
            begin
                axi4_local_set_ARBURST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARLOCK_changed or posedge _check_t0_values )
    begin
        while (ARLOCK_changed == 1'b1)
        begin
            logic temp_ARLOCK;

            dvc_axi4_get_ARLOCK_into_SystemVerilog( _interface_ref, temp_ARLOCK ); // DPI call to imported task
            internal_ARLOCK = temp_ARLOCK;
            ARLOCK_changed = 1'b0;
            #0  #0 if ( ARLOCK !== internal_ARLOCK )
            begin
                axi4_local_set_ARLOCK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARCACHE_changed or posedge _check_t0_values )
    begin
        while (ARCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_ARCACHE_into_SystemVerilog( _interface_ref, internal_ARCACHE ); // DPI call to imported task
            ARCACHE_changed = 1'b0;
            #0  #0 if ( ARCACHE !== internal_ARCACHE )
            begin
                axi4_local_set_ARCACHE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARQOS_changed or posedge _check_t0_values )
    begin
        while (ARQOS_changed == 1'b1)
        begin
            dvc_axi4_get_ARQOS_into_SystemVerilog( _interface_ref, internal_ARQOS ); // DPI call to imported task
            ARQOS_changed = 1'b0;
            #0  #0 if ( ARQOS !== internal_ARQOS )
            begin
                axi4_local_set_ARQOS_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARID_changed or posedge _check_t0_values )
    begin
        while (ARID_changed == 1'b1)
        begin
            dvc_axi4_get_ARID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARID_changed = 1'b0;
            #0  #0 if ( ARID !== internal_ARID )
            begin
                axi4_local_set_ARID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARUSER_changed or posedge _check_t0_values )
    begin
        while (ARUSER_changed == 1'b1)
        begin
            dvc_axi4_get_ARUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARUSER_changed = 1'b0;
            #0  #0 if ( ARUSER !== internal_ARUSER )
            begin
                axi4_local_set_ARUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARREADY_changed or posedge _check_t0_values )
    begin
        while (ARREADY_changed == 1'b1)
        begin
            logic temp_ARREADY;

            dvc_axi4_get_ARREADY_into_SystemVerilog( _interface_ref, temp_ARREADY ); // DPI call to imported task
            internal_ARREADY = temp_ARREADY;
            ARREADY_changed = 1'b0;
            #0  #0 if ( ARREADY !== internal_ARREADY )
            begin
                axi4_local_set_ARREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RVALID_changed or posedge _check_t0_values )
    begin
        while (RVALID_changed == 1'b1)
        begin
            logic temp_RVALID;

            dvc_axi4_get_RVALID_into_SystemVerilog( _interface_ref, temp_RVALID ); // DPI call to imported task
            internal_RVALID = temp_RVALID;
            RVALID_changed = 1'b0;
            #0  #0 if ( RVALID !== internal_RVALID )
            begin
                axi4_local_set_RVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RDATA_changed or posedge _check_t0_values )
    begin
        while (RDATA_changed == 1'b1)
        begin
            dvc_axi4_get_RDATA_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RDATA_changed = 1'b0;
            #0  #0 if ( RDATA !== internal_RDATA )
            begin
                axi4_local_set_RDATA_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RRESP_changed or posedge _check_t0_values )
    begin
        while (RRESP_changed == 1'b1)
        begin
            dvc_axi4_get_RRESP_into_SystemVerilog( _interface_ref, internal_RRESP ); // DPI call to imported task
            RRESP_changed = 1'b0;
            #0  #0 if ( RRESP !== internal_RRESP )
            begin
                axi4_local_set_RRESP_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RLAST_changed or posedge _check_t0_values )
    begin
        while (RLAST_changed == 1'b1)
        begin
            logic temp_RLAST;

            dvc_axi4_get_RLAST_into_SystemVerilog( _interface_ref, temp_RLAST ); // DPI call to imported task
            internal_RLAST = temp_RLAST;
            RLAST_changed = 1'b0;
            #0  #0 if ( RLAST !== internal_RLAST )
            begin
                axi4_local_set_RLAST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RID_changed or posedge _check_t0_values )
    begin
        while (RID_changed == 1'b1)
        begin
            dvc_axi4_get_RID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RID_changed = 1'b0;
            #0  #0 if ( RID !== internal_RID )
            begin
                axi4_local_set_RID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RUSER_changed or posedge _check_t0_values )
    begin
        while (RUSER_changed == 1'b1)
        begin
            dvc_axi4_get_RUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RUSER_changed = 1'b0;
            #0  #0 if ( RUSER !== internal_RUSER )
            begin
                axi4_local_set_RUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RREADY_changed or posedge _check_t0_values )
    begin
        while (RREADY_changed == 1'b1)
        begin
            logic temp_RREADY;

            dvc_axi4_get_RREADY_into_SystemVerilog( _interface_ref, temp_RREADY ); // DPI call to imported task
            internal_RREADY = temp_RREADY;
            RREADY_changed = 1'b0;
            #0  #0 if ( RREADY !== internal_RREADY )
            begin
                axi4_local_set_RREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WVALID_changed or posedge _check_t0_values )
    begin
        while (WVALID_changed == 1'b1)
        begin
            logic temp_WVALID;

            dvc_axi4_get_WVALID_into_SystemVerilog( _interface_ref, temp_WVALID ); // DPI call to imported task
            internal_WVALID = temp_WVALID;
            WVALID_changed = 1'b0;
            #0  #0 if ( WVALID !== internal_WVALID )
            begin
                axi4_local_set_WVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WDATA_changed or posedge _check_t0_values )
    begin
        while (WDATA_changed == 1'b1)
        begin
            dvc_axi4_get_WDATA_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WDATA_changed = 1'b0;
            #0  #0 if ( WDATA !== internal_WDATA )
            begin
                axi4_local_set_WDATA_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WSTRB_changed or posedge _check_t0_values )
    begin
        while (WSTRB_changed == 1'b1)
        begin
            dvc_axi4_get_WSTRB_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WSTRB_changed = 1'b0;
            #0  #0 if ( WSTRB !== internal_WSTRB )
            begin
                axi4_local_set_WSTRB_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WLAST_changed or posedge _check_t0_values )
    begin
        while (WLAST_changed == 1'b1)
        begin
            logic temp_WLAST;

            dvc_axi4_get_WLAST_into_SystemVerilog( _interface_ref, temp_WLAST ); // DPI call to imported task
            internal_WLAST = temp_WLAST;
            WLAST_changed = 1'b0;
            #0  #0 if ( WLAST !== internal_WLAST )
            begin
                axi4_local_set_WLAST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WUSER_changed or posedge _check_t0_values )
    begin
        while (WUSER_changed == 1'b1)
        begin
            dvc_axi4_get_WUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WUSER_changed = 1'b0;
            #0  #0 if ( WUSER !== internal_WUSER )
            begin
                axi4_local_set_WUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WREADY_changed or posedge _check_t0_values )
    begin
        while (WREADY_changed == 1'b1)
        begin
            logic temp_WREADY;

            dvc_axi4_get_WREADY_into_SystemVerilog( _interface_ref, temp_WREADY ); // DPI call to imported task
            internal_WREADY = temp_WREADY;
            WREADY_changed = 1'b0;
            #0  #0 if ( WREADY !== internal_WREADY )
            begin
                axi4_local_set_WREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BVALID_changed or posedge _check_t0_values )
    begin
        while (BVALID_changed == 1'b1)
        begin
            logic temp_BVALID;

            dvc_axi4_get_BVALID_into_SystemVerilog( _interface_ref, temp_BVALID ); // DPI call to imported task
            internal_BVALID = temp_BVALID;
            BVALID_changed = 1'b0;
            #0  #0 if ( BVALID !== internal_BVALID )
            begin
                axi4_local_set_BVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BRESP_changed or posedge _check_t0_values )
    begin
        while (BRESP_changed == 1'b1)
        begin
            dvc_axi4_get_BRESP_into_SystemVerilog( _interface_ref, internal_BRESP ); // DPI call to imported task
            BRESP_changed = 1'b0;
            #0  #0 if ( BRESP !== internal_BRESP )
            begin
                axi4_local_set_BRESP_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BID_changed or posedge _check_t0_values )
    begin
        while (BID_changed == 1'b1)
        begin
            dvc_axi4_get_BID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            BID_changed = 1'b0;
            #0  #0 if ( BID !== internal_BID )
            begin
                axi4_local_set_BID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BUSER_changed or posedge _check_t0_values )
    begin
        while (BUSER_changed == 1'b1)
        begin
            dvc_axi4_get_BUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            BUSER_changed = 1'b0;
            #0  #0 if ( BUSER !== internal_BUSER )
            begin
                axi4_local_set_BUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BREADY_changed or posedge _check_t0_values )
    begin
        while (BREADY_changed == 1'b1)
        begin
            logic temp_BREADY;

            dvc_axi4_get_BREADY_into_SystemVerilog( _interface_ref, temp_BREADY ); // DPI call to imported task
            internal_BREADY = temp_BREADY;
            BREADY_changed = 1'b0;
            #0  #0 if ( BREADY !== internal_BREADY )
            begin
                axi4_local_set_BREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge config_axi4lite_interface_changed or posedge _check_t0_values )
    begin
        if (config_axi4lite_interface_changed == 1'b1)
        begin
            dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog( _interface_ref, config_axi4lite_interface ); // DPI call to imported task
            config_axi4lite_interface_changed = 1'b0;
        end
    end

    always @(posedge config_interface_type_changed or posedge _check_t0_values )
    begin
        if (config_interface_type_changed == 1'b1)
        begin
            dvc_axi4_get_config_interface_type_into_SystemVerilog( _interface_ref, config_interface_type ); // DPI call to imported task
            config_interface_type_changed = 1'b0;
        end
    end

    always @(posedge config_enable_rlast_changed or posedge _check_t0_values )
    begin
        if (config_enable_rlast_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_rlast_into_SystemVerilog( _interface_ref, config_enable_rlast ); // DPI call to imported task
            config_enable_rlast_changed = 1'b0;
        end
    end

    always @(posedge config_enable_id_changed or posedge _check_t0_values )
    begin
        if (config_enable_id_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_id_into_SystemVerilog( _interface_ref, config_enable_id ); // DPI call to imported task
            config_enable_id_changed = 1'b0;
        end
    end

    always @(posedge config_enable_all_assertions_changed or posedge _check_t0_values )
    begin
        if (config_enable_all_assertions_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog( _interface_ref, config_enable_all_assertions ); // DPI call to imported task
            config_enable_all_assertions_changed = 1'b0;
        end
    end

    always @(posedge config_enable_assertion_changed or posedge _check_t0_values )
    begin
        if (config_enable_assertion_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_assertion_into_SystemVerilog( _interface_ref, config_enable_assertion ); // DPI call to imported task
            config_enable_assertion_changed = 1'b0;
        end
    end

    always @(posedge config_enable_user_sideband_changed or posedge _check_t0_values )
    begin
        if (config_enable_user_sideband_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog( _interface_ref, config_enable_user_sideband ); // DPI call to imported task
            config_enable_user_sideband_changed = 1'b0;
        end
    end

    always @(posedge config_enable_slave_exclusive_changed or posedge _check_t0_values )
    begin
        if (config_enable_slave_exclusive_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog( _interface_ref, config_enable_slave_exclusive ); // DPI call to imported task
            config_enable_slave_exclusive_changed = 1'b0;
        end
    end

    always @(posedge config_read_data_reordering_depth_changed or posedge _check_t0_values )
    begin
        if (config_read_data_reordering_depth_changed == 1'b1)
        begin
            dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog( _interface_ref, config_read_data_reordering_depth ); // DPI call to imported task
            config_read_data_reordering_depth_changed = 1'b0;
        end
    end

    always @(posedge config_slave_start_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_start_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            config_slave_start_addr_changed = 1'b0;
        end
    end

    always @(posedge config_slave_end_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_end_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            config_slave_end_addr_changed = 1'b0;
        end
    end

    always @(posedge config_max_transaction_time_factor_changed or posedge _check_t0_values )
    begin
        if (config_max_transaction_time_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog( _interface_ref, config_max_transaction_time_factor ); // DPI call to imported task
            config_max_transaction_time_factor_changed = 1'b0;
        end
    end

    always @(posedge config_burst_timeout_factor_changed or posedge _check_t0_values )
    begin
        if (config_burst_timeout_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog( _interface_ref, config_burst_timeout_factor ); // DPI call to imported task
            config_burst_timeout_factor_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_AWVALID_assertion_to_AWREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_AWVALID_assertion_to_AWREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog( _interface_ref, config_max_latency_AWVALID_assertion_to_AWREADY ); // DPI call to imported task
            config_max_latency_AWVALID_assertion_to_AWREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_ARVALID_assertion_to_ARREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_ARVALID_assertion_to_ARREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog( _interface_ref, config_max_latency_ARVALID_assertion_to_ARREADY ); // DPI call to imported task
            config_max_latency_ARVALID_assertion_to_ARREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_RVALID_assertion_to_RREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_RVALID_assertion_to_RREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog( _interface_ref, config_max_latency_RVALID_assertion_to_RREADY ); // DPI call to imported task
            config_max_latency_RVALID_assertion_to_RREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_BVALID_assertion_to_BREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_BVALID_assertion_to_BREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog( _interface_ref, config_max_latency_BVALID_assertion_to_BREADY ); // DPI call to imported task
            config_max_latency_BVALID_assertion_to_BREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_WVALID_assertion_to_WREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_WVALID_assertion_to_WREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog( _interface_ref, config_max_latency_WVALID_assertion_to_WREADY ); // DPI call to imported task
            config_max_latency_WVALID_assertion_to_WREADY_changed = 1'b0;
        end
    end

    always @(posedge config_enable_qos_changed or posedge _check_t0_values )
    begin
        if (config_enable_qos_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_qos_into_SystemVerilog( _interface_ref, config_enable_qos ); // DPI call to imported task
            config_enable_qos_changed = 1'b0;
        end
    end

    always @(posedge config_setup_time_changed or posedge _check_t0_values )
    begin
        if (config_setup_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_setup_time_into_SystemVerilog( _interface_ref, config_setup_time ); // DPI call to imported task
            config_setup_time_changed = 1'b0;
        end
    end

    always @(posedge config_hold_time_changed or posedge _check_t0_values )
    begin
        if (config_hold_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_hold_time_into_SystemVerilog( _interface_ref, config_hold_time ); // DPI call to imported task
            config_hold_time_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_reads_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_reads_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog( _interface_ref, config_num_max_outstanding_reads ); // DPI call to imported task
            config_num_max_outstanding_reads_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_writes_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_writes_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog( _interface_ref, config_num_max_outstanding_writes ); // DPI call to imported task
            config_num_max_outstanding_writes_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_wr_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_wr_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog( _interface_ref, config_max_outstanding_wr ); // DPI call to imported task
            config_max_outstanding_wr_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rd_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rd_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog( _interface_ref, config_max_outstanding_rd ); // DPI call to imported task
            config_max_outstanding_rd_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rw_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rw_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog( _interface_ref, config_max_outstanding_rw ); // DPI call to imported task
            config_max_outstanding_rw_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_wr_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_wr_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_wr_phase ); // DPI call to imported task
            config_num_outstanding_wr_phase_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_rd_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_rd_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_rd_phase ); // DPI call to imported task
            config_num_outstanding_rd_phase_changed = 1'b0;
        end
    end

    always @(posedge config_is_issuing_changed or posedge _check_t0_values )
    begin
        if (config_is_issuing_changed == 1'b1)
        begin
            dvc_axi4_get_config_is_issuing_into_SystemVerilog( _interface_ref, config_is_issuing ); // DPI call to imported task
            config_is_issuing_changed = 1'b0;
        end
    end



    // Sparse array of blocking control events
    event block_control[] = new[100];

    // Unblocks a blocked clock control thread by id
    function void unblock( int unsigned id );
    begin
        -> block_control[id];
    end
    endfunction
    export "DPI-C" dvc_axi4_unblock_SystemVerilog = function unblock;

    // Blocks a blocked clock control thread by id
    task automatic block( int unsigned id );
    begin
        event blocking_event ;
        if (id >= block_control.size())
        begin
            int newsize  = ( (  id / 100 ) + 1 ) * 100;
            block_control = new[newsize](block_control);
        end
        blocking_event = block_control[id];
        @ blocking_event;
    end
    endtask
    export "DPI-C" dvc_axi4_block_SystemVerilog = task block;


    function int is_call_back_registered(int cb_name);
        case( axi4_call_back_e'(cb_name) )
          AXI4_REPORTER_CB:
          begin
              return ( endPoint.size() > 0 ) ? 1 : 0;
          end
        endcase
    endfunction

    //--------------------------------------------------------------------------------
    // Task which blocks and outputs an error if the interface has not initialized properly
    //--------------------------------------------------------------------------------

    task _initialized();
        if (_interface_ref == 0)
        begin
            $display("Error: %m - Questa Verification IP failed to initialise. Please check questa_mvc.log for details");
            wait(_interface_ref!=0);
        end
    endtask

endinterface

`elsif VCS
// *****************************************************************************
//
// Copyright 2007-2022 Mentor Graphics Corporation
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY OF
// MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//
// *****************************************************************************
// SystemVerilog           Version: 20220701
// *****************************************************************************

`ifndef QVIP_MIX_AND_MATCH
(* cy_so="libaxi4_IN_SystemVerilog_MTI_full_DVC" *)
(* on_lib_load="axi4_IN_SystemVerilog_load" *)
`endif

interface mgc_common_axi4 #( int AXI4_ADDRESS_WIDTH = 64, int AXI4_RDATA_WIDTH = 1024, int AXI4_WDATA_WIDTH = 1024, int AXI4_ID_WIDTH = 18, int AXI4_USER_WIDTH = 8, int AXI4_REGION_MAP_SIZE = 16 )
    (input wire iACLK, input wire iARESETn);

import QUESTA_MVC::questa_mvc_reporter;
import QUESTA_MVC::questa_mvc_item_comms_semantic;
import QUESTA_MVC::questa_mvc_edge;
import QUESTA_MVC::QUESTA_MVC_POSEDGE;
import QUESTA_MVC::QUESTA_MVC_NEGEDGE;
import QUESTA_MVC::QUESTA_MVC_ANYEDGE;
import QUESTA_MVC::QUESTA_MVC_0_TO_1_EDGE;
import QUESTA_MVC::QUESTA_MVC_1_TO_0_EDGE;




    //-------------------------------------------------------------------------
    // Private wires
    //-------------------------------------------------------------------------
    wire ACLK;
    wire ARESETn;
    wire AWVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR;
    wire [2:0] AWPROT;
    wire [3:0] AWREGION;
    wire [7:0] AWLEN;
    wire [3:0] AWSIZE;
    wire [1:0] AWBURST;
    wire AWLOCK;
    wire [3:0] AWCACHE;
    wire [3:0] AWQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] AWID;
    wire [((AXI4_USER_WIDTH) - 1):0] AWUSER;
    wire AWREADY;
    wire ARVALID;
    wire [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR;
    wire [2:0] ARPROT;
    wire [3:0] ARREGION;
    wire [7:0] ARLEN;
    wire [3:0] ARSIZE;
    wire [1:0] ARBURST;
    wire ARLOCK;
    wire [3:0] ARCACHE;
    wire [3:0] ARQOS;
    wire [((AXI4_ID_WIDTH) - 1):0] ARID;
    wire [((AXI4_USER_WIDTH) - 1):0] ARUSER;
    wire ARREADY;
    wire RVALID;
    wire [((AXI4_RDATA_WIDTH) - 1):0] RDATA;
    wire [1:0] RRESP;
    wire RLAST;
    wire [((AXI4_ID_WIDTH) - 1):0] RID;
    wire [((AXI4_USER_WIDTH) - 1):0] RUSER;
    wire RREADY;
    wire WVALID;
    wire [((AXI4_WDATA_WIDTH) - 1):0] WDATA;
    wire [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB;
    wire WLAST;
    wire [((AXI4_USER_WIDTH) - 1):0] WUSER;
    wire WREADY;
    wire BVALID;
    wire [1:0] BRESP;
    wire [((AXI4_ID_WIDTH) - 1):0] BID;
    wire [((AXI4_USER_WIDTH) - 1):0] BUSER;
    wire BREADY;



    // Propagate global signals onto interface wires
    assign ACLK = iACLK;
    assign ARESETn = iARESETn;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Enables
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_rlast
    //
    //  
    // Enables use of optional RLAST signal.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable optional RLAST signal"
    bit config_enable_rlast;

    // Variable: config_enable_id
    //
    //  
    // Enables use of optional AWID, BID, ARID, RID signals.
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable ID signals"
    bit config_enable_id;

    // Variable: config_enable_all_assertions
    //
    // 
    // Enables all protocol assertions. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable all protocol assertions"
    bit config_enable_all_assertions;

    // Variable: config_enable_assertion
    //
    // 
    // Enables individual protocol assertion.
    // This variable controls whether specific assertion within QVIP (of type <axi4_assertion_e>) is enabled or disabled.
    // Individual assertion can be disabled as follows:-
    // //-----------------------------------------------------------------------
    // // < BFM interface>.config_enable_assertion[<name of assertion>] = 1'b0;
    // //-----------------------------------------------------------------------
    // 
    // For example, the assertion AXI4_READ_DATA_UNKN is disabled as follows:
    // <bfm>.config_enable_assertion[AXI4_READ_DATA_UNKN] =  1'b0; 
    // 
    // Here bfm is the AXI4 interface instance name for which the assertion to be disabled. 
    // 
    // Default: All assertions are enabled
    //   
    // 
    //
    // mentor configurator specification name "Enable individual protocol assertion" 
    bit [255:0] config_enable_assertion;

    // Variable: config_enable_user_sideband
    //
    // 
    // Enables user sideband signals. 
    // 
    // Default: 1
    // 
    //
    // mentor configurator specification name "Enable user sideband signals"
    bit config_enable_user_sideband;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Slave behavior control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_slave_exclusive
    //
    // 
    // Enables exclusive transactions support for slave.
    // If disabled, every exclusive read/write returns an OKAY response,
    // and exclusive write updates memory. 
    // 
    // Default: 1  
    // 
    //
    // mentor configurator specification name "Enable exclusive transaction support"
    bit config_enable_slave_exclusive;

    // Variable: config_read_data_reordering_depth
    //
    // 
    // Sets the maximum number of different read transaction addresses for which read 
    // data(response) can be sent in any order from slave. 
    // 
    // Default: 2 ** AXI4_ID_WIDTH
    // 
    //
    // mentor configurator specification name "Read data reordering depth"
    int unsigned config_read_data_reordering_depth;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Timeout control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_max_transaction_time_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) for any complete read or write transaction, which
    // includes time period for all individual phases of transaction. 
    // 
    // Default: 100000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout for complete Read/Write transaction"
    int unsigned config_max_transaction_time_factor;

    // Variable: config_burst_timeout_factor
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) between individual phases of a transaction. 
    // 
    // Default: 10000 clock cycles 
    // 
    //
    // mentor configurator specification name "Burst timeout between individual phases of a transaction"
    int unsigned config_burst_timeout_factor;

    // Variable: config_max_latency_AWVALID_assertion_to_AWREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of AWVALID to assertion of AWREADY.
    // An error message AXI4_AWREADY_NOT_ASSERTED_AFTER_AWVALID is generated if AWREADY is not asserted
    // after assertion of AWVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from AWVALID to AWREADY assertion"
    int unsigned config_max_latency_AWVALID_assertion_to_AWREADY;

    // Variable: config_max_latency_ARVALID_assertion_to_ARREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of ARVALID to assertion of ARREADY.
    // An error message AXI4_ARREADY_NOT_ASSERTED_AFTER_ARVALID is generated if ARREADY is not asserted
    // after assertion of ARVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from ARVALID to ARREADY assertion"
    int unsigned config_max_latency_ARVALID_assertion_to_ARREADY;

    // Variable: config_max_latency_RVALID_assertion_to_RREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of RVALID to assertion of RREADY.
    // An error message AXI4_RREADY_NOT_ASSERTED_AFTER_RVALID is generated if RREADY is not asserted
    // after assertion of RVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from RVALID to RREADY assertion"
    int unsigned config_max_latency_RVALID_assertion_to_RREADY;

    // Variable: config_max_latency_BVALID_assertion_to_BREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of BVALID to assertion of BREADY.
    // An error message AXI4_BREADY_NOT_ASSERTED_AFTER_BVALID is generated if BREADY is not asserted
    // after assertion of BVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from BVALID to BREADY assertion"
    int unsigned config_max_latency_BVALID_assertion_to_BREADY;

    // Variable: config_max_latency_WVALID_assertion_to_WREADY
    //
    //  
    // Sets maximum timeout period (in terms of ACLK) from assertion of WVALID to assertion of WREADY.
    // An error message AXI4_WREADY_NOT_ASSERTED_AFTER_WVALID is generated if WREADY is not asserted
    // after assertion of WVALID within this period. 
    // 
    // Default: 10000 clock cycles
    // 
    //
    // mentor configurator specification name "Timeout from WVALID to WREADY assertion"
    int unsigned config_max_latency_WVALID_assertion_to_WREADY;

    // 
    // //-----------------------------------------------------------------------------
    // Group: QOS and Region control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_enable_qos
    //
    //  
    // Enables quality-of-service support. 
    // 
    // Default: 1 
    // 
    //
    // mentor configurator specification name "Enable quality of service support"
    bit config_enable_qos;

    // Variable: config_setup_time
    //
    // 
    // Sets number of simulation time units from the setup time to the active 
    // clock edge of ACLK. The setup time will always be less than the time period
    // of the clock. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_setup_time;

    // Variable: config_hold_time
    //
    // 
    // Sets number of simulation time units from the hold time to the active 
    // clock edge of ACLK. 
    // 
    // Default: 0
    // 
    //
    // Note - This configuration variable is used in an expression involving time precision.
    //        To ensure its value is correct, use questa_mvc_sv_convert_to_precision API of QUESTA_MVC package.
    //
    longint unsigned config_hold_time;

    // 
    // //-----------------------------------------------------------------------------
    // Group: Master Outstanding Control
    // //-----------------------------------------------------------------------------
    // 


    // Variable: config_num_max_outstanding_reads
    //
    // 
    // Configures maximum number of read outstanding transfers allowed on the bus.
    // 
    // Default: -1
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding reads"
    int config_num_max_outstanding_reads;

    // Variable: config_num_max_outstanding_writes
    //
    //                                                                           
    // Configures maximum number of write outstanding transfers allowed on the bus. 
    //                                                                              
    // Default: -1                                                                 
    // 
    //
    // mentor configurator specification name "Configures maximum outstanding writes"
    int config_num_max_outstanding_writes;

    // Variable: config_max_outstanding_wr
    //
    // Configures maximum possible outstanding Write transactions
    //
    int config_max_outstanding_wr;

    // Variable: config_max_outstanding_rd
    //
    // Configures maximum possible outstanding Read transactions
    //
    int config_max_outstanding_rd;

    // Variable: config_max_outstanding_rw
    //
    // Configures maximum possible outstanding Combined (Read and Write) transactions
    //
    int config_max_outstanding_rw;

    // Variable: config_num_outstanding_wr_phase
    //
    // Provides current number of outstanding Write address phases 
    //
    int config_num_outstanding_wr_phase;

    // Variable: config_num_outstanding_rd_phase
    //
    // Provides current number of outstanding Read address phases
    //
    int config_num_outstanding_rd_phase;

    // Variable: config_is_issuing
    //
    // Enables Master component to use "config_max_outstanding_wr/config_max_outstanding_rd/config_max_outstanding_rw" variables for transaction issuing capability when set to true
    //
    bit config_is_issuing;


    //-------------------------------------------------------------------------
    // Deprecated variables - writing to these variables will cause a warning to be issued.
    //-------------------------------------------------------------------------
    bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr;


    //-------------------------------------------------------------------------
    // Private variables
    //-------------------------------------------------------------------------
    bit config_axi4lite_interface;
    axi4_interface_type_e config_interface_type;


    //-------------------------------------------------------------------------
    // Initialise local variables
    //-------------------------------------------------------------------------
    initial
    begin
         config_axi4lite_tr_id = 0;
    end

    //------------------------------------------------------------------------------
    // Group:- Interface ends
    //------------------------------------------------------------------------------
    //
    longint axi4_master_end;

    // Function:- get_axi4_master_end
    //
    // Returns a handle to the <master> end of this instance of the <axi4> interface.

    function longint get_axi4_master_end();
        return axi4_master_end;
    endfunction

    longint axi4_slave_end;

    // Function:- get_axi4_slave_end
    //
    // Returns a handle to the <slave> end of this instance of the <axi4> interface.

    function longint get_axi4_slave_end();
        return axi4_slave_end;
    endfunction

    longint axi4__monitor_end;

    // Function:- get_axi4__monitor_end
    //
    // Returns a handle to the <_monitor> end of this instance of the <axi4> interface.

    function longint get_axi4__monitor_end();
        return axi4__monitor_end;
    endfunction


    // Group:- Abstraction Levels
    // 
    // These functions are used set or get the abstraction levels of an interface end.
    // See <Abstraction Levels of Interface Ends> for more details on the meaning of
    // TLM or WLM connected and the valid combinations.


    //-------------------------------------------------------------------------
    // Function:- axi4_set_master_abstraction_level
    //
    //     Function to set whether the <master> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_master_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_master_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <master> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_master_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_master_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_set_slave_abstraction_level
    //
    //     Function to set whether the <slave> end of the interface is WLM
    //     or TLM connected. See <Abstraction Levels of Interface Ends> for a
    //     description of abstraction levels, how they affect the behavior of the
    //     QVIP, and guidelines for setting them.
    //
    // Arguments:
    //    wire_level - Set to 1 to be WLM connected.
    //    TLM_level -  Set to 1 to be TLM connected.
    //
    function void axi4_set_slave_abstraction_level
    (
        input bit          wire_level,
        input bit          TLM_level
    );
        axi4_set_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- axi4_get_slave_abstraction_level
    //
    //     Function to return the Abstraction level setting for the <slave> end.
    //     See <Abstraction Levels of Interface Ends> for a description of abstraction
    //     levels and how they affect the behavior of the Questa Verification IP.
    //
    // Arguments:
    //
    //    wire_level - Value = 1 if this end is WLM connected.
    //    TLM_level -  Value = 1 if this end is TLM connected.
    //------------------------------------------------------------------------------
    function void axi4_get_slave_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
        axi4_get_slave_end_abstraction_level( wire_level, TLM_level );
    endfunction

    import "DPI-C" context function longint dvc_axi4_initialise_SystemVerilog
    (
        int     usage_code,
        string  iface_version,
        longint generate_ver,
        int     qvip_mix_and_match,
        output longint master_end,
        output longint slave_end,
        output longint _monitor_end,
        input int AXI4_ADDRESS_WIDTH,
        input int AXI4_RDATA_WIDTH,
        input int AXI4_WDATA_WIDTH,
        input int AXI4_ID_WIDTH,
        input int AXI4_USER_WIDTH,
        input int AXI4_REGION_MAP_SIZE
    );

    `ifndef MVC_axi4_VERSION
    `define MVC_axi4_VERSION ""
    `endif

    // Handle to the linkage
    (* elab_init *) longint _interface_ref =
                                dvc_axi4_initialise_SystemVerilog
                                (
                                    18102076,
                                    `MVC_axi4_VERSION,
                                    20220701,
                                    `ifdef QVIP_MIX_AND_MATCH
                                    1
                                    `else
                                    0
                                    `endif
                                    ,
                                    axi4_master_end,
                                    axi4_slave_end,
                                    axi4__monitor_end,
                                    AXI4_ADDRESS_WIDTH,
                                    AXI4_RDATA_WIDTH,
                                    AXI4_WDATA_WIDTH,
                                    AXI4_ID_WIDTH,
                                    AXI4_USER_WIDTH,
                                    AXI4_REGION_MAP_SIZE
                                ); // DPI call to create transactor (called at
                                     // elaboration time as initialiser)

    questa_mvc_reporter endPoint[longint];
    export "DPI-C" dvc_axi4_process_reports = function process_reports;
    function void process_reports( input longint recipient, input string category, input string objectName, input string instanceName, input string error_no, input string severity, input string mess );
        if( endPoint.exists(recipient) )
            endPoint[recipient].report_message( category, "dvc_axi4", 0, objectName, instanceName, error_no, severity, mess );
        else
            $error("Invalid recipient (%d) when processing report", recipient);
    endfunction

    import "DPI-C" context dvc_axi4_register_end_point = function void axi4_register_end_point( input longint iface_ref, input longint as_end, input string name );

    // A function for registering a reporter to capture any reports coming from as_end
    function automatic void register_end_point( input longint as_end, input questa_mvc_reporter rep = null );
        if ( rep != null )
        begin
            if ( ( rep.name == "" ) || ( rep.name == "NULL" ) )
            begin
                $display("Error: %m: Reporter passed to register_end_point has a reserved name. Neither an empty string nor the string 'NULL' can be used.");
            end
            else
            begin
                axi4_register_end_point( _interface_ref, as_end, rep.name );
                endPoint[as_end] = rep;
            end
        end
        else
        begin
            axi4_register_end_point( _interface_ref, as_end, "NULL" );
            endPoint.delete( as_end );
        end
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group:- Registering Reports
    //
    //
    // The following methods are used to register a custom reporting object as
    // described in the Questa Verification IP base library section, <Customizing Error-Reporting>.
    // 
    //-------------------------------------------------------------------------

    function void register_interface_reporter( input questa_mvc_reporter _rep = null );
        register_end_point( _interface_ref, _rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_master_reporter
    //
    // Function used to register a reporter for the <master> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the master end.
    //
    function void register_master_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_master_end, rep );
    endfunction

    //-------------------------------------------------------------------------
    // Function:- register_slave_reporter
    //
    // Function used to register a reporter for the <slave> end of the
    // <axi4> interface. See <Customizing Error-Reporting> for a
    // description of creating, customising and using reporters.
    //
    // Arguments:
    //    rep - The reporter to be used for the slave end.
    //
    function void register_slave_reporter
    (
        input questa_mvc_reporter rep = null
    );
        register_end_point( axi4_slave_end, rep );
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_mvc_reporter
    //
    // Function used to get the handle for an already registered reporter.
    // By default returns the reporter associated with this interface. If an end handle is passed,
    // then the reporter for that end.
    //
    // Arguments:
    //    as_end - Optional, a handle for an end of this interface.
    //
    function questa_mvc_reporter get_mvc_reporter
    (
        input longint as_end = 0
    );
        if ( as_end == 0 )
            as_end = _interface_ref;
        if ( endPoint.exists( as_end ) )
            return endPoint[ as_end ];
        else
            return null;
    endfunction

    //-------------------------------------------------------------------------
    //
    // Group:- BFM Utility/Convenience Methods
    //
    // This is the group of utility functions provided by the QVIP BFM to
    // communicate from the SV world to the QVIP BFM.
    // This set of APIs can be used to either get status/statistics
    // information from the BFM or to set values in a particular database 
    // in the BFM. Please refer to individual functions for more information.
    //
    //-------------------------------------------------------------------------

    export "DPI-C" dvc_axi4_get_temp_static_fn_get_raddr_phase_info_addr = function axi4_get_temp_static_fn_get_raddr_phase_info_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_get_raddr_phase_info_addr = function axi4_set_temp_static_fn_get_raddr_phase_info_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_get_raddr_phase_info_addr;
    function void axi4_get_temp_static_fn_get_raddr_phase_info_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_get_raddr_phase_info_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_get_raddr_phase_info_addr( input int _d1, input bit  _value );
        temp_static_fn_get_raddr_phase_info_addr[_d1] = _value;
    endfunction
    function automatic void fn_get_raddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output int num_beats_rcvd
    );
         fn_get_raddr_phase_info_C(valid,id,num_beats_rcvd);
        addr = temp_static_fn_get_raddr_phase_info_addr;
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_get_waddr_phase_info_addr = function axi4_get_temp_static_fn_get_waddr_phase_info_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_get_waddr_phase_info_addr = function axi4_set_temp_static_fn_get_waddr_phase_info_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_get_waddr_phase_info_addr;
    function void axi4_get_temp_static_fn_get_waddr_phase_info_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_get_waddr_phase_info_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_get_waddr_phase_info_addr( input int _d1, input bit  _value );
        temp_static_fn_get_waddr_phase_info_addr[_d1] = _value;
    endfunction
    function automatic void fn_get_waddr_phase_info
    (
        inout bit valid,
        input bit [17:0] id,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr
    );
         fn_get_waddr_phase_info_C(valid,id);
        addr = temp_static_fn_get_waddr_phase_info_addr;
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_add_addr_map_entry_addr = function axi4_get_temp_static_fn_add_addr_map_entry_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_add_addr_map_entry_addr = function axi4_set_temp_static_fn_add_addr_map_entry_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_add_addr_map_entry_addr;
    function void axi4_get_temp_static_fn_add_addr_map_entry_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_add_addr_map_entry_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_add_addr_map_entry_addr( input int _d1, input bit  _value );
        temp_static_fn_add_addr_map_entry_addr[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_add_addr_map_entry
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets address map fields inside internal BFM. 
    // 
    // Inputs:
    // region - Region name 
    // addr - Start address of region
    // size - Size of address region
    function automatic void fn_add_addr_map_entry
    (
        input string region,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input longint unsigned size
    );
        temp_static_fn_add_addr_map_entry_addr = addr;
         fn_add_addr_map_entry_C(region,size);
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_add_wr_delay_data2data = function axi4_get_temp_static_fn_add_wr_delay_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_add_wr_delay_data2data = function axi4_set_temp_static_fn_add_wr_delay_data2data;
    int unsigned temp_static_fn_add_wr_delay_data2data[];
    function void axi4_get_temp_static_fn_add_wr_delay_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_add_wr_delay_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_add_wr_delay_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_add_wr_delay_data2data[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_add_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets delay values for write address-data pair and 
    // between write data beats, initiated by master end. 
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_add_wr_delay
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        const ref int unsigned data2data[]
    );
        int tmp_data2data_DIMS0;
        tmp_data2data_DIMS0 = data2data.size();
        temp_static_fn_add_wr_delay_data2data = data2data;
         fn_add_wr_delay_C(region,id,addr2data,tmp_data2data_DIMS0);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_delete_wr_delay
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function deletes delay values for a region-id pair
    // 
    // Inputs:
    // region - Region value for which write delays are to be inserted
    // id - Write transaction's id (AWID) for which delays are to be inserted
    // addr2data - Delays to be inserted between write address and data phase.
    // data2data - Delays to be inserted between data beats
    function automatic void fn_delete_wr_delay
    (
        input string region,
        input bit [17:0] id
    );
         fn_delete_wr_delay_C(region,id);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_drive_signals
    //-------------------------------------------------------------------------
    // A function to drive the required value by the user on the desired signal when VALID is 0. 
    //   It takes the name of the signal and value to be driven on that signal as an input:
    // 
    //   Note - In case, the user wants to retain the same value on the signal (irrespective of Valid signal), 
    //   the value needs to be passed as -1.
    function automatic void fn_drive_signals
    (
        input string signal,
        input longint value
    );
         fn_drive_signals_C(signal,value);
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_set_address_map_entry_start_addr = function axi4_get_temp_static_fn_set_address_map_entry_start_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_address_map_entry_start_addr = function axi4_set_temp_static_fn_set_address_map_entry_start_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_set_address_map_entry_start_addr;
    function void axi4_get_temp_static_fn_set_address_map_entry_start_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_set_address_map_entry_start_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_address_map_entry_start_addr( input int _d1, input bit  _value );
        temp_static_fn_set_address_map_entry_start_addr[_d1] = _value;
    endfunction
    export "DPI-C" dvc_axi4_get_temp_static_fn_set_address_map_entry_end_addr = function axi4_get_temp_static_fn_set_address_map_entry_end_addr;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_address_map_entry_end_addr = function axi4_set_temp_static_fn_set_address_map_entry_end_addr;
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_fn_set_address_map_entry_end_addr;
    function void axi4_get_temp_static_fn_set_address_map_entry_end_addr( input int _d1, output bit  _value );
        _value = temp_static_fn_set_address_map_entry_end_addr[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_address_map_entry_end_addr( input int _d1, input bit  _value );
        temp_static_fn_set_address_map_entry_end_addr[_d1] = _value;
    endfunction
    function automatic void fn_set_address_map_entry
    (
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] start_addr,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] end_addr
    );
        temp_static_fn_set_address_map_entry_start_addr = start_addr;
        temp_static_fn_set_address_map_entry_end_addr = end_addr;
         fn_set_address_map_entry_C();
    endfunction

    export "DPI-C" dvc_axi4_get_temp_static_fn_set_wr_def_delays_min_data2data = function axi4_get_temp_static_fn_set_wr_def_delays_min_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_wr_def_delays_min_data2data = function axi4_set_temp_static_fn_set_wr_def_delays_min_data2data;
    int unsigned temp_static_fn_set_wr_def_delays_min_data2data[];
    function void axi4_get_temp_static_fn_set_wr_def_delays_min_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_set_wr_def_delays_min_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_wr_def_delays_min_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_set_wr_def_delays_min_data2data[_d1] = _value;
    endfunction
    export "DPI-C" dvc_axi4_get_temp_static_fn_set_wr_def_delays_max_data2data = function axi4_get_temp_static_fn_set_wr_def_delays_max_data2data;
    export "DPI-C" dvc_axi4_set_temp_static_fn_set_wr_def_delays_max_data2data = function axi4_set_temp_static_fn_set_wr_def_delays_max_data2data;
    int unsigned temp_static_fn_set_wr_def_delays_max_data2data[];
    function void axi4_get_temp_static_fn_set_wr_def_delays_max_data2data( input int _d1, output int unsigned _value );
        _value = temp_static_fn_set_wr_def_delays_max_data2data[_d1];
    endfunction
    function void axi4_set_temp_static_fn_set_wr_def_delays_max_data2data( input int _d1, input int unsigned _value );
        temp_static_fn_set_wr_def_delays_max_data2data[_d1] = _value;
    endfunction
    //-------------------------------------------------------------------------
    // Function: fn_set_wr_def_delays
    //-------------------------------------------------------------------------
    // ---------------------------------------------------------------------------
    // This function sets default delay values for write address-data pair and data beats 
    // initiated by master end between input max. and min. values. In case non-randomized 
    // default value is desired, user can set max. and min. to same value.
    // 
    // Inputs:
    // min_addr2data - Minimum value of default delays to be inserted between write address and data phase.
    // min_data2data - Minimum value of default delays to be inserted between data beats
    // max_addr2data - Maximum value of default delays to be inserted between write address and data phase.
    // max_data2data - Maximum value of default delays to be inserted between data beats
    function automatic void fn_set_wr_def_delays
    (
        input int unsigned min_addr2data,
        const ref int unsigned min_data2data[],
        input int unsigned max_addr2data,
        const ref int unsigned max_data2data[]
    );
        int tmp_min_data2data_DIMS0;
        int tmp_max_data2data_DIMS0;
        tmp_min_data2data_DIMS0 = min_data2data.size();
        tmp_max_data2data_DIMS0 = max_data2data.size();
        temp_static_fn_set_wr_def_delays_min_data2data = min_data2data;
        temp_static_fn_set_wr_def_delays_max_data2data = max_data2data;
         fn_set_wr_def_delays_C(min_addr2data,tmp_min_data2data_DIMS0,max_addr2data,tmp_max_data2data_DIMS0);
    endfunction

    //-------------------------------------------------------------------------
    // Function: fn_print_perf_stats
    //-------------------------------------------------------------------------
    // API used to display the performance statistics of the current simulation. This
    // API shall be called only if axi4_vip_config variable 'cfg.agent_cfg.en_perf_stats.enable' has been configured to '1'.
    // 
    // Every statistics has 4 basic units:
    // minimum - Tells about the minimum metric within sliding window  
    // maximum - Tells about the maximum metric within sliding window 
    // mean    - Tells about the mean metric within a sliding window
    // sw      - Stores the calculation of the metrics over sliding window. 
    //           Statistics are collected for each of the sliding window segments 
    //           individually and then the actual statistic is calculated based 
    //           on all of the segments
    // 
    // The statistics are calculated for basic categories -
    // 
    // bus_usage: Stores the bus occupancy(as a percentage out of 100) for
    //   - Read transactions
    //   - Write transactions
    //   - Read Data burst
    //   - Write Data burst
    // See also, <axi4_bus_usage_s>
    // 
    // bandwidth: Stores the bandwidth value for
    //   - Read transactions
    //   - Write transactions
    // See also, <axi4_bandwidth_s>
    // 
    // latency: Stores the interphase delays as per the following categories
    //   - Write address to next write address delay
    //   - Write address to write data delay
    //   - Write data to write response delay
    //   - Read address to next read address delay
    //   - Read address to read data delay
    // See also, <axi4_latency_s>
    // 
    // waits: Stores the VALID-READY waits that are involved in each of the phases
    //   - Write address
    //   - Write data
    //   - Write response
    //   - Read address
    //   - Read data
    // See also, <axi4_waits_s>
    function automatic void fn_print_perf_stats
    (
    );
         fn_print_perf_stats_C();
    endfunction

    // Declare user visible wires variables, for non-continuous assignments.
    logic m_ACLK = 'z;
    logic m_ARESETn = 'z;
    logic m_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_AWADDR = 'z;
    logic [2:0] m_AWPROT = 'z;
    logic [3:0] m_AWREGION = 'z;
    logic [7:0] m_AWLEN = 'z;
    logic [3:0] m_AWSIZE = 'z;
    logic [1:0] m_AWBURST = 'z;
    logic m_AWLOCK = 'z;
    logic [3:0] m_AWCACHE = 'z;
    logic [3:0] m_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_AWUSER = 'z;
    logic m_AWREADY = 'z;
    logic m_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] m_ARADDR = 'z;
    logic [2:0] m_ARPROT = 'z;
    logic [3:0] m_ARREGION = 'z;
    logic [7:0] m_ARLEN = 'z;
    logic [3:0] m_ARSIZE = 'z;
    logic [1:0] m_ARBURST = 'z;
    logic m_ARLOCK = 'z;
    logic [3:0] m_ARCACHE = 'z;
    logic [3:0] m_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_ARUSER = 'z;
    logic m_ARREADY = 'z;
    logic m_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] m_RDATA = 'z;
    logic [1:0] m_RRESP = 'z;
    logic m_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_RUSER = 'z;
    logic m_RREADY = 'z;
    logic m_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] m_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] m_WSTRB = 'z;
    logic m_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_WUSER = 'z;
    logic m_WREADY = 'z;
    logic m_BVALID = 'z;
    logic [1:0] m_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] m_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] m_BUSER = 'z;
    logic m_BREADY = 'z;

    // Forces a sweep through the wire change checkers at time 0 to get around
    // process kick-off order unknowns
    bit _check_t0_values;
    always_comb _check_t0_values = 1;

    // handle control
    longint last_start_time = 0;

    longint last_end_time = 0;

    export "DPI-C" dvc_axi4_set_start_end_times = function set_start_end_times;

    function void set_start_end_times(longint _start, longint _end);
        last_start_time = _start;
        last_end_time = _end;
    endfunction


    function longint get_last_handle();
        return -1;
    endfunction


    function longint get_last_start_time();
        return last_start_time;
    endfunction


    function longint get_last_end_time();
        return last_end_time;
    endfunction


    //-------------------------------------------------------------------------
    // Tasks to wait for a number of specified edges on a wire
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_ACLK
    //     Wait for the specified change on wire <axi4::ACLK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ACLK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ACLK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ACLK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ACLK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ACLK === 0 );
                    @( ACLK );
                end
                while ( ACLK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ACLK === 1 );
                    @( ACLK );
                end
                while ( ACLK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARESETn
    //     Wait for the specified change on wire <axi4::ARESETn>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARESETn( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARESETn);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARESETn);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARESETn);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 0 );
                    @( ARESETn );
                end
                while ( ARESETn !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARESETn === 1 );
                    @( ARESETn );
                end
                while ( ARESETn !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWVALID
    //     Wait for the specified change on wire <axi4::AWVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 0 );
                    @( AWVALID );
                end
                while ( AWVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWVALID === 1 );
                    @( AWVALID );
                end
                while ( AWVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 0 );
                    @( AWADDR );
                end
                while ( AWADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR === 1 );
                    @( AWADDR );
                end
                while ( AWADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWADDR_index1
    //     Wait for the specified change on wire <axi4::AWADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 0 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWADDR[_this_dot_1] === 1 );
                    @( AWADDR[_this_dot_1] );
                end
                while ( AWADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 0 );
                    @( AWPROT );
                end
                while ( AWPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT === 1 );
                    @( AWPROT );
                end
                while ( AWPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWPROT_index1
    //     Wait for the specified change on wire <axi4::AWPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 0 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWPROT[_this_dot_1] === 1 );
                    @( AWPROT[_this_dot_1] );
                end
                while ( AWPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 0 );
                    @( AWREGION );
                end
                while ( AWREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION === 1 );
                    @( AWREGION );
                end
                while ( AWREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREGION_index1
    //     Wait for the specified change on wire <axi4::AWREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 0 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREGION[_this_dot_1] === 1 );
                    @( AWREGION[_this_dot_1] );
                end
                while ( AWREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 0 );
                    @( AWLEN );
                end
                while ( AWLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN === 1 );
                    @( AWLEN );
                end
                while ( AWLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLEN_index1
    //     Wait for the specified change on wire <axi4::AWLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 0 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLEN[_this_dot_1] === 1 );
                    @( AWLEN[_this_dot_1] );
                end
                while ( AWLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 0 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE === 1 );
                    @( AWSIZE );
                end
                while ( AWSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWSIZE_index1
    //     Wait for the specified change on wire <axi4::AWSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 0 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWSIZE[_this_dot_1] === 1 );
                    @( AWSIZE[_this_dot_1] );
                end
                while ( AWSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 0 );
                    @( AWBURST );
                end
                while ( AWBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST === 1 );
                    @( AWBURST );
                end
                while ( AWBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWBURST_index1
    //     Wait for the specified change on wire <axi4::AWBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 0 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWBURST[_this_dot_1] === 1 );
                    @( AWBURST[_this_dot_1] );
                end
                while ( AWBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWLOCK
    //     Wait for the specified change on wire <axi4::AWLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 0 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWLOCK === 1 );
                    @( AWLOCK );
                end
                while ( AWLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 0 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE === 1 );
                    @( AWCACHE );
                end
                while ( AWCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWCACHE_index1
    //     Wait for the specified change on wire <axi4::AWCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 0 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWCACHE[_this_dot_1] === 1 );
                    @( AWCACHE[_this_dot_1] );
                end
                while ( AWCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 0 );
                    @( AWQOS );
                end
                while ( AWQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS === 1 );
                    @( AWQOS );
                end
                while ( AWQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWQOS_index1
    //     Wait for the specified change on wire <axi4::AWQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 0 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWQOS[_this_dot_1] === 1 );
                    @( AWQOS[_this_dot_1] );
                end
                while ( AWQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID === 0 );
                    @( AWID );
                end
                while ( AWID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID === 1 );
                    @( AWID );
                end
                while ( AWID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWID_index1
    //     Wait for the specified change on wire <axi4::AWID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 0 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWID[_this_dot_1] === 1 );
                    @( AWID[_this_dot_1] );
                end
                while ( AWID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 0 );
                    @( AWUSER );
                end
                while ( AWUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER === 1 );
                    @( AWUSER );
                end
                while ( AWUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWUSER_index1
    //     Wait for the specified change on wire <axi4::AWUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 0 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWUSER[_this_dot_1] === 1 );
                    @( AWUSER[_this_dot_1] );
                end
                while ( AWUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_AWREADY
    //     Wait for the specified change on wire <axi4::AWREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_AWREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge AWREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge AWREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        AWREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 0 );
                    @( AWREADY );
                end
                while ( AWREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( AWREADY === 1 );
                    @( AWREADY );
                end
                while ( AWREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARVALID
    //     Wait for the specified change on wire <axi4::ARVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 0 );
                    @( ARVALID );
                end
                while ( ARVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARVALID === 1 );
                    @( ARVALID );
                end
                while ( ARVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 0 );
                    @( ARADDR );
                end
                while ( ARADDR !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR === 1 );
                    @( ARADDR );
                end
                while ( ARADDR !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARADDR_index1
    //     Wait for the specified change on wire <axi4::ARADDR>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARADDR_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARADDR[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 0 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARADDR[_this_dot_1] === 1 );
                    @( ARADDR[_this_dot_1] );
                end
                while ( ARADDR[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 0 );
                    @( ARPROT );
                end
                while ( ARPROT !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT === 1 );
                    @( ARPROT );
                end
                while ( ARPROT !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARPROT_index1
    //     Wait for the specified change on wire <axi4::ARPROT>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARPROT_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARPROT[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 0 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARPROT[_this_dot_1] === 1 );
                    @( ARPROT[_this_dot_1] );
                end
                while ( ARPROT[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 0 );
                    @( ARREGION );
                end
                while ( ARREGION !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION === 1 );
                    @( ARREGION );
                end
                while ( ARREGION !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREGION_index1
    //     Wait for the specified change on wire <axi4::ARREGION>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREGION_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREGION[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 0 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREGION[_this_dot_1] === 1 );
                    @( ARREGION[_this_dot_1] );
                end
                while ( ARREGION[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 0 );
                    @( ARLEN );
                end
                while ( ARLEN !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN === 1 );
                    @( ARLEN );
                end
                while ( ARLEN !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLEN_index1
    //     Wait for the specified change on wire <axi4::ARLEN>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLEN_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLEN[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 0 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLEN[_this_dot_1] === 1 );
                    @( ARLEN[_this_dot_1] );
                end
                while ( ARLEN[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 0 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE === 1 );
                    @( ARSIZE );
                end
                while ( ARSIZE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARSIZE_index1
    //     Wait for the specified change on wire <axi4::ARSIZE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARSIZE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARSIZE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 0 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARSIZE[_this_dot_1] === 1 );
                    @( ARSIZE[_this_dot_1] );
                end
                while ( ARSIZE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 0 );
                    @( ARBURST );
                end
                while ( ARBURST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST === 1 );
                    @( ARBURST );
                end
                while ( ARBURST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARBURST_index1
    //     Wait for the specified change on wire <axi4::ARBURST>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARBURST_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARBURST[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 0 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARBURST[_this_dot_1] === 1 );
                    @( ARBURST[_this_dot_1] );
                end
                while ( ARBURST[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARLOCK
    //     Wait for the specified change on wire <axi4::ARLOCK>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARLOCK( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARLOCK);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARLOCK);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 0 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARLOCK === 1 );
                    @( ARLOCK );
                end
                while ( ARLOCK !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 0 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE === 1 );
                    @( ARCACHE );
                end
                while ( ARCACHE !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARCACHE_index1
    //     Wait for the specified change on wire <axi4::ARCACHE>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARCACHE_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARCACHE[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 0 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARCACHE[_this_dot_1] === 1 );
                    @( ARCACHE[_this_dot_1] );
                end
                while ( ARCACHE[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 0 );
                    @( ARQOS );
                end
                while ( ARQOS !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS === 1 );
                    @( ARQOS );
                end
                while ( ARQOS !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARQOS_index1
    //     Wait for the specified change on wire <axi4::ARQOS>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARQOS_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARQOS[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 0 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARQOS[_this_dot_1] === 1 );
                    @( ARQOS[_this_dot_1] );
                end
                while ( ARQOS[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID === 0 );
                    @( ARID );
                end
                while ( ARID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID === 1 );
                    @( ARID );
                end
                while ( ARID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARID_index1
    //     Wait for the specified change on wire <axi4::ARID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 0 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARID[_this_dot_1] === 1 );
                    @( ARID[_this_dot_1] );
                end
                while ( ARID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 0 );
                    @( ARUSER );
                end
                while ( ARUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER === 1 );
                    @( ARUSER );
                end
                while ( ARUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARUSER_index1
    //     Wait for the specified change on wire <axi4::ARUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 0 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARUSER[_this_dot_1] === 1 );
                    @( ARUSER[_this_dot_1] );
                end
                while ( ARUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_ARREADY
    //     Wait for the specified change on wire <axi4::ARREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_ARREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge ARREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge ARREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        ARREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 0 );
                    @( ARREADY );
                end
                while ( ARREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( ARREADY === 1 );
                    @( ARREADY );
                end
                while ( ARREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RVALID
    //     Wait for the specified change on wire <axi4::RVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RVALID === 0 );
                    @( RVALID );
                end
                while ( RVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RVALID === 1 );
                    @( RVALID );
                end
                while ( RVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA === 0 );
                    @( RDATA );
                end
                while ( RDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA === 1 );
                    @( RDATA );
                end
                while ( RDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RDATA_index1
    //     Wait for the specified change on wire <axi4::RDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 0 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RDATA[_this_dot_1] === 1 );
                    @( RDATA[_this_dot_1] );
                end
                while ( RDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP === 0 );
                    @( RRESP );
                end
                while ( RRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP === 1 );
                    @( RRESP );
                end
                while ( RRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RRESP_index1
    //     Wait for the specified change on wire <axi4::RRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 0 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RRESP[_this_dot_1] === 1 );
                    @( RRESP[_this_dot_1] );
                end
                while ( RRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RLAST
    //     Wait for the specified change on wire <axi4::RLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RLAST === 0 );
                    @( RLAST );
                end
                while ( RLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RLAST === 1 );
                    @( RLAST );
                end
                while ( RLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID === 0 );
                    @( RID );
                end
                while ( RID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID === 1 );
                    @( RID );
                end
                while ( RID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RID_index1
    //     Wait for the specified change on wire <axi4::RID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 0 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RID[_this_dot_1] === 1 );
                    @( RID[_this_dot_1] );
                end
                while ( RID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER === 0 );
                    @( RUSER );
                end
                while ( RUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER === 1 );
                    @( RUSER );
                end
                while ( RUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RUSER_index1
    //     Wait for the specified change on wire <axi4::RUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 0 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RUSER[_this_dot_1] === 1 );
                    @( RUSER[_this_dot_1] );
                end
                while ( RUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_RREADY
    //     Wait for the specified change on wire <axi4::RREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_RREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge RREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge RREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        RREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( RREADY === 0 );
                    @( RREADY );
                end
                while ( RREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( RREADY === 1 );
                    @( RREADY );
                end
                while ( RREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WVALID
    //     Wait for the specified change on wire <axi4::WVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WVALID === 0 );
                    @( WVALID );
                end
                while ( WVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WVALID === 1 );
                    @( WVALID );
                end
                while ( WVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA === 0 );
                    @( WDATA );
                end
                while ( WDATA !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA === 1 );
                    @( WDATA );
                end
                while ( WDATA !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WDATA_index1
    //     Wait for the specified change on wire <axi4::WDATA>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WDATA_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WDATA[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 0 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WDATA[_this_dot_1] === 1 );
                    @( WDATA[_this_dot_1] );
                end
                while ( WDATA[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 0 );
                    @( WSTRB );
                end
                while ( WSTRB !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB === 1 );
                    @( WSTRB );
                end
                while ( WSTRB !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WSTRB_index1
    //     Wait for the specified change on wire <axi4::WSTRB>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WSTRB_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WSTRB[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 0 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WSTRB[_this_dot_1] === 1 );
                    @( WSTRB[_this_dot_1] );
                end
                while ( WSTRB[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WLAST
    //     Wait for the specified change on wire <axi4::WLAST>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WLAST( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WLAST);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WLAST);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WLAST);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WLAST === 0 );
                    @( WLAST );
                end
                while ( WLAST !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WLAST === 1 );
                    @( WLAST );
                end
                while ( WLAST !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER === 0 );
                    @( WUSER );
                end
                while ( WUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER === 1 );
                    @( WUSER );
                end
                while ( WUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WUSER_index1
    //     Wait for the specified change on wire <axi4::WUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 0 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WUSER[_this_dot_1] === 1 );
                    @( WUSER[_this_dot_1] );
                end
                while ( WUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_WREADY
    //     Wait for the specified change on wire <axi4::WREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_WREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge WREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge WREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        WREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( WREADY === 0 );
                    @( WREADY );
                end
                while ( WREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( WREADY === 1 );
                    @( WREADY );
                end
                while ( WREADY !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BVALID
    //     Wait for the specified change on wire <axi4::BVALID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BVALID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BVALID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BVALID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BVALID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BVALID === 0 );
                    @( BVALID );
                end
                while ( BVALID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BVALID === 1 );
                    @( BVALID );
                end
                while ( BVALID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP === 0 );
                    @( BRESP );
                end
                while ( BRESP !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP === 1 );
                    @( BRESP );
                end
                while ( BRESP !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BRESP_index1
    //     Wait for the specified change on wire <axi4::BRESP>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BRESP_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BRESP[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 0 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BRESP[_this_dot_1] === 1 );
                    @( BRESP[_this_dot_1] );
                end
                while ( BRESP[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID === 0 );
                    @( BID );
                end
                while ( BID !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID === 1 );
                    @( BID );
                end
                while ( BID !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BID_index1
    //     Wait for the specified change on wire <axi4::BID>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BID_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BID[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 0 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BID[_this_dot_1] === 1 );
                    @( BID[_this_dot_1] );
                end
                while ( BID[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER === 0 );
                    @( BUSER );
                end
                while ( BUSER !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER === 1 );
                    @( BUSER );
                end
                while ( BUSER !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BUSER_index1
    //     Wait for the specified change on wire <axi4::BUSER>.
    //
    // Arguments:
    //    _this_dot_1 - The array index for dimension 1.
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BUSER_index1( input int _this_dot_1, input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BUSER[_this_dot_1]);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 0 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BUSER[_this_dot_1] === 1 );
                    @( BUSER[_this_dot_1] );
                end
                while ( BUSER[_this_dot_1] !== 0 );
            end
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_BREADY
    //     Wait for the specified change on wire <axi4::BREADY>.
    //
    // Arguments:
    //     which_edge - The type of edge to wait for, one of <questa_mvc_edge>.
    //     count - The number of edges to wait for.
    //
    task automatic wait_for_BREADY( input questa_mvc_edge which_edge, input int count = 1 );
        int i;
        for ( i=0; i<count; i++ )
        begin
            if      ( which_edge == QUESTA_MVC_POSEDGE     ) @(posedge BREADY);
            else if ( which_edge == QUESTA_MVC_NEGEDGE     ) @(negedge BREADY);
            else if ( which_edge == QUESTA_MVC_ANYEDGE     ) @(        BREADY);
            else if ( which_edge == QUESTA_MVC_0_TO_1_EDGE )
            begin
                do
                begin
                    wait( BREADY === 0 );
                    @( BREADY );
                end
                while ( BREADY !== 1 );
            end
            else if ( which_edge == QUESTA_MVC_1_TO_0_EDGE )
            begin
                do
                begin
                    wait( BREADY === 1 );
                    @( BREADY );
                end
                while ( BREADY !== 0 );
            end
        end
    endtask

    //-------------------------------------------------------------------------
    // Tasks/functions to set/get wires
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_ACLK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ACLK>.
    //
    // Parameters:
    //     ACLK_param - The value to set onto wire <ACLK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ACLK( logic ACLK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ACLK = ACLK_param;
        else
            m_ACLK <= ACLK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ACLK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ACLK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ACLK>.
    //
    function automatic logic get_ACLK(  );
        return ACLK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARESETn
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARESETn>.
    //
    // Parameters:
    //     ARESETn_param - The value to set onto wire <ARESETn>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARESETn( logic ARESETn_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARESETn = ARESETn_param;
        else
            m_ARESETn <= ARESETn_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARESETn
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARESETn>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARESETn>.
    //
    function automatic logic get_ARESETn(  );
        return ARESETn;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWVALID>.
    //
    // Parameters:
    //     AWVALID_param - The value to set onto wire <AWVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWVALID( logic AWVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWVALID = AWVALID_param;
        else
            m_AWVALID <= AWVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWVALID>.
    //
    function automatic logic get_AWVALID(  );
        return AWVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWADDR>.
    //
    // Parameters:
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR = AWADDR_param;
        else
            m_AWADDR <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWADDR_param - The value to set onto wire <AWADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWADDR_index1( int _this_dot_1, logic  AWADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWADDR[_this_dot_1] = AWADDR_param;
        else
            m_AWADDR[_this_dot_1] <= AWADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_AWADDR(  );
        return AWADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWADDR>.
    //
    function automatic logic   get_AWADDR_index1( int _this_dot_1 );
        return AWADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWPROT>.
    //
    // Parameters:
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT( logic [2:0] AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT = AWPROT_param;
        else
            m_AWPROT <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWPROT_param - The value to set onto wire <AWPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWPROT_index1( int _this_dot_1, logic  AWPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWPROT[_this_dot_1] = AWPROT_param;
        else
            m_AWPROT[_this_dot_1] <= AWPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic [2:0]  get_AWPROT(  );
        return AWPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWPROT>.
    //
    function automatic logic   get_AWPROT_index1( int _this_dot_1 );
        return AWPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREGION>.
    //
    // Parameters:
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION( logic [3:0] AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION = AWREGION_param;
        else
            m_AWREGION <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWREGION_param - The value to set onto wire <AWREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREGION_index1( int _this_dot_1, logic  AWREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREGION[_this_dot_1] = AWREGION_param;
        else
            m_AWREGION[_this_dot_1] <= AWREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic [3:0]  get_AWREGION(  );
        return AWREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWREGION>.
    //
    function automatic logic   get_AWREGION_index1( int _this_dot_1 );
        return AWREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLEN>.
    //
    // Parameters:
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN( logic [7:0] AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN = AWLEN_param;
        else
            m_AWLEN <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWLEN_param - The value to set onto wire <AWLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLEN_index1( int _this_dot_1, logic  AWLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLEN[_this_dot_1] = AWLEN_param;
        else
            m_AWLEN[_this_dot_1] <= AWLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic [7:0]  get_AWLEN(  );
        return AWLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWLEN>.
    //
    function automatic logic   get_AWLEN_index1( int _this_dot_1 );
        return AWLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWSIZE>.
    //
    // Parameters:
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE( logic [3:0] AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE = AWSIZE_param;
        else
            m_AWSIZE <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWSIZE_param - The value to set onto wire <AWSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWSIZE_index1( int _this_dot_1, logic  AWSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWSIZE[_this_dot_1] = AWSIZE_param;
        else
            m_AWSIZE[_this_dot_1] <= AWSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic [3:0]  get_AWSIZE(  );
        return AWSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWSIZE>.
    //
    function automatic logic   get_AWSIZE_index1( int _this_dot_1 );
        return AWSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWBURST>.
    //
    // Parameters:
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST( logic [1:0] AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST = AWBURST_param;
        else
            m_AWBURST <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWBURST_param - The value to set onto wire <AWBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWBURST_index1( int _this_dot_1, logic  AWBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWBURST[_this_dot_1] = AWBURST_param;
        else
            m_AWBURST[_this_dot_1] <= AWBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic [1:0]  get_AWBURST(  );
        return AWBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWBURST>.
    //
    function automatic logic   get_AWBURST_index1( int _this_dot_1 );
        return AWBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWLOCK>.
    //
    // Parameters:
    //     AWLOCK_param - The value to set onto wire <AWLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWLOCK( logic AWLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWLOCK = AWLOCK_param;
        else
            m_AWLOCK <= AWLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWLOCK>.
    //
    function automatic logic get_AWLOCK(  );
        return AWLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWCACHE>.
    //
    // Parameters:
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE( logic [3:0] AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE = AWCACHE_param;
        else
            m_AWCACHE <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWCACHE_param - The value to set onto wire <AWCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWCACHE_index1( int _this_dot_1, logic  AWCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWCACHE[_this_dot_1] = AWCACHE_param;
        else
            m_AWCACHE[_this_dot_1] <= AWCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic [3:0]  get_AWCACHE(  );
        return AWCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWCACHE>.
    //
    function automatic logic   get_AWCACHE_index1( int _this_dot_1 );
        return AWCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWQOS>.
    //
    // Parameters:
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS( logic [3:0] AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS = AWQOS_param;
        else
            m_AWQOS <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWQOS_param - The value to set onto wire <AWQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWQOS_index1( int _this_dot_1, logic  AWQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWQOS[_this_dot_1] = AWQOS_param;
        else
            m_AWQOS[_this_dot_1] <= AWQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic [3:0]  get_AWQOS(  );
        return AWQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWQOS>.
    //
    function automatic logic   get_AWQOS_index1( int _this_dot_1 );
        return AWQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWID
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWID>.
    //
    // Parameters:
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID( logic [((AXI4_ID_WIDTH) - 1):0] AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID = AWID_param;
        else
            m_AWID <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWID_param - The value to set onto wire <AWID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWID_index1( int _this_dot_1, logic  AWID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWID[_this_dot_1] = AWID_param;
        else
            m_AWID[_this_dot_1] <= AWID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWID
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_AWID(  );
        return AWID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWID>.
    //
    function automatic logic   get_AWID_index1( int _this_dot_1 );
        return AWID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWUSER>.
    //
    // Parameters:
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER( logic [((AXI4_USER_WIDTH) - 1):0] AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER = AWUSER_param;
        else
            m_AWUSER <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     AWUSER_param - The value to set onto wire <AWUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWUSER_index1( int _this_dot_1, logic  AWUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWUSER[_this_dot_1] = AWUSER_param;
        else
            m_AWUSER[_this_dot_1] <= AWUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_AWUSER(  );
        return AWUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_AWUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <AWUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <AWUSER>.
    //
    function automatic logic   get_AWUSER_index1( int _this_dot_1 );
        return AWUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <AWREADY>.
    //
    // Parameters:
    //     AWREADY_param - The value to set onto wire <AWREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_AWREADY( logic AWREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_AWREADY = AWREADY_param;
        else
            m_AWREADY <= AWREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <AWREADY>.
    //
    function automatic logic get_AWREADY(  );
        return AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARVALID>.
    //
    // Parameters:
    //     ARVALID_param - The value to set onto wire <ARVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARVALID( logic ARVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARVALID = ARVALID_param;
        else
            m_ARVALID <= ARVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARVALID>.
    //
    function automatic logic get_ARVALID(  );
        return ARVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARADDR>.
    //
    // Parameters:
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR( logic [((AXI4_ADDRESS_WIDTH) - 1):0] ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR = ARADDR_param;
        else
            m_ARADDR <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARADDR_param - The value to set onto wire <ARADDR>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARADDR_index1( int _this_dot_1, logic  ARADDR_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARADDR[_this_dot_1] = ARADDR_param;
        else
            m_ARADDR[_this_dot_1] <= ARADDR_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARADDR>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic [((AXI4_ADDRESS_WIDTH) - 1):0]  get_ARADDR(  );
        return ARADDR;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARADDR_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARADDR>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARADDR>.
    //
    function automatic logic   get_ARADDR_index1( int _this_dot_1 );
        return ARADDR[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARPROT>.
    //
    // Parameters:
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT( logic [2:0] ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT = ARPROT_param;
        else
            m_ARPROT <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARPROT_param - The value to set onto wire <ARPROT>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARPROT_index1( int _this_dot_1, logic  ARPROT_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARPROT[_this_dot_1] = ARPROT_param;
        else
            m_ARPROT[_this_dot_1] <= ARPROT_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARPROT>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic [2:0]  get_ARPROT(  );
        return ARPROT;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARPROT_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARPROT>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARPROT>.
    //
    function automatic logic   get_ARPROT_index1( int _this_dot_1 );
        return ARPROT[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREGION>.
    //
    // Parameters:
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION( logic [3:0] ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION = ARREGION_param;
        else
            m_ARREGION <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARREGION_param - The value to set onto wire <ARREGION>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREGION_index1( int _this_dot_1, logic  ARREGION_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREGION[_this_dot_1] = ARREGION_param;
        else
            m_ARREGION[_this_dot_1] <= ARREGION_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREGION>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic [3:0]  get_ARREGION(  );
        return ARREGION;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARREGION_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARREGION>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARREGION>.
    //
    function automatic logic   get_ARREGION_index1( int _this_dot_1 );
        return ARREGION[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLEN>.
    //
    // Parameters:
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN( logic [7:0] ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN = ARLEN_param;
        else
            m_ARLEN <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARLEN_param - The value to set onto wire <ARLEN>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLEN_index1( int _this_dot_1, logic  ARLEN_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLEN[_this_dot_1] = ARLEN_param;
        else
            m_ARLEN[_this_dot_1] <= ARLEN_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLEN>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic [7:0]  get_ARLEN(  );
        return ARLEN;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARLEN_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARLEN>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARLEN>.
    //
    function automatic logic   get_ARLEN_index1( int _this_dot_1 );
        return ARLEN[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARSIZE>.
    //
    // Parameters:
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE( logic [3:0] ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE = ARSIZE_param;
        else
            m_ARSIZE <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARSIZE_param - The value to set onto wire <ARSIZE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARSIZE_index1( int _this_dot_1, logic  ARSIZE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARSIZE[_this_dot_1] = ARSIZE_param;
        else
            m_ARSIZE[_this_dot_1] <= ARSIZE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARSIZE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic [3:0]  get_ARSIZE(  );
        return ARSIZE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARSIZE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARSIZE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARSIZE>.
    //
    function automatic logic   get_ARSIZE_index1( int _this_dot_1 );
        return ARSIZE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARBURST>.
    //
    // Parameters:
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST( logic [1:0] ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST = ARBURST_param;
        else
            m_ARBURST <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARBURST_param - The value to set onto wire <ARBURST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARBURST_index1( int _this_dot_1, logic  ARBURST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARBURST[_this_dot_1] = ARBURST_param;
        else
            m_ARBURST[_this_dot_1] <= ARBURST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARBURST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic [1:0]  get_ARBURST(  );
        return ARBURST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARBURST_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARBURST>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARBURST>.
    //
    function automatic logic   get_ARBURST_index1( int _this_dot_1 );
        return ARBURST[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARLOCK
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARLOCK>.
    //
    // Parameters:
    //     ARLOCK_param - The value to set onto wire <ARLOCK>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARLOCK( logic ARLOCK_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARLOCK = ARLOCK_param;
        else
            m_ARLOCK <= ARLOCK_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARLOCK
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARLOCK>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARLOCK>.
    //
    function automatic logic get_ARLOCK(  );
        return ARLOCK;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARCACHE>.
    //
    // Parameters:
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE( logic [3:0] ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE = ARCACHE_param;
        else
            m_ARCACHE <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARCACHE_param - The value to set onto wire <ARCACHE>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARCACHE_index1( int _this_dot_1, logic  ARCACHE_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARCACHE[_this_dot_1] = ARCACHE_param;
        else
            m_ARCACHE[_this_dot_1] <= ARCACHE_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARCACHE>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic [3:0]  get_ARCACHE(  );
        return ARCACHE;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARCACHE_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARCACHE>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARCACHE>.
    //
    function automatic logic   get_ARCACHE_index1( int _this_dot_1 );
        return ARCACHE[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARQOS>.
    //
    // Parameters:
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS( logic [3:0] ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS = ARQOS_param;
        else
            m_ARQOS <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARQOS_param - The value to set onto wire <ARQOS>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARQOS_index1( int _this_dot_1, logic  ARQOS_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARQOS[_this_dot_1] = ARQOS_param;
        else
            m_ARQOS[_this_dot_1] <= ARQOS_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARQOS>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic [3:0]  get_ARQOS(  );
        return ARQOS;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARQOS_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARQOS>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARQOS>.
    //
    function automatic logic   get_ARQOS_index1( int _this_dot_1 );
        return ARQOS[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARID
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARID>.
    //
    // Parameters:
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID( logic [((AXI4_ID_WIDTH) - 1):0] ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID = ARID_param;
        else
            m_ARID <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARID_param - The value to set onto wire <ARID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARID_index1( int _this_dot_1, logic  ARID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARID[_this_dot_1] = ARID_param;
        else
            m_ARID[_this_dot_1] <= ARID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARID
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_ARID(  );
        return ARID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARID>.
    //
    function automatic logic   get_ARID_index1( int _this_dot_1 );
        return ARID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARUSER>.
    //
    // Parameters:
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER( logic [((AXI4_USER_WIDTH) - 1):0] ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER = ARUSER_param;
        else
            m_ARUSER <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     ARUSER_param - The value to set onto wire <ARUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARUSER_index1( int _this_dot_1, logic  ARUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARUSER[_this_dot_1] = ARUSER_param;
        else
            m_ARUSER[_this_dot_1] <= ARUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_ARUSER(  );
        return ARUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_ARUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <ARUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <ARUSER>.
    //
    function automatic logic   get_ARUSER_index1( int _this_dot_1 );
        return ARUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <ARREADY>.
    //
    // Parameters:
    //     ARREADY_param - The value to set onto wire <ARREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_ARREADY( logic ARREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_ARREADY = ARREADY_param;
        else
            m_ARREADY <= ARREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <ARREADY>.
    //
    function automatic logic get_ARREADY(  );
        return ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RVALID>.
    //
    // Parameters:
    //     RVALID_param - The value to set onto wire <RVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RVALID( logic RVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RVALID = RVALID_param;
        else
            m_RVALID <= RVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RVALID>.
    //
    function automatic logic get_RVALID(  );
        return RVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <RDATA>.
    //
    // Parameters:
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA( logic [((AXI4_RDATA_WIDTH) - 1):0] RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA = RDATA_param;
        else
            m_RDATA <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RDATA_param - The value to set onto wire <RDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RDATA_index1( int _this_dot_1, logic  RDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RDATA[_this_dot_1] = RDATA_param;
        else
            m_RDATA[_this_dot_1] <= RDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <RDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic [((AXI4_RDATA_WIDTH) - 1):0]  get_RDATA(  );
        return RDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RDATA>.
    //
    function automatic logic   get_RDATA_index1( int _this_dot_1 );
        return RDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <RRESP>.
    //
    // Parameters:
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP( logic [1:0] RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP = RRESP_param;
        else
            m_RRESP <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RRESP_param - The value to set onto wire <RRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RRESP_index1( int _this_dot_1, logic  RRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RRESP[_this_dot_1] = RRESP_param;
        else
            m_RRESP[_this_dot_1] <= RRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <RRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic [1:0]  get_RRESP(  );
        return RRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RRESP>.
    //
    function automatic logic   get_RRESP_index1( int _this_dot_1 );
        return RRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <RLAST>.
    //
    // Parameters:
    //     RLAST_param - The value to set onto wire <RLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RLAST( logic RLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RLAST = RLAST_param;
        else
            m_RLAST <= RLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <RLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RLAST>.
    //
    function automatic logic get_RLAST(  );
        return RLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RID
    //-------------------------------------------------------------------------
    //     Set the value of wire <RID>.
    //
    // Parameters:
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID( logic [((AXI4_ID_WIDTH) - 1):0] RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID = RID_param;
        else
            m_RID <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RID_param - The value to set onto wire <RID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RID_index1( int _this_dot_1, logic  RID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RID[_this_dot_1] = RID_param;
        else
            m_RID[_this_dot_1] <= RID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RID
    //-------------------------------------------------------------------------
    //     Get the value of wire <RID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_RID(  );
        return RID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RID>.
    //
    function automatic logic   get_RID_index1( int _this_dot_1 );
        return RID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <RUSER>.
    //
    // Parameters:
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER( logic [((AXI4_USER_WIDTH) - 1):0] RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER = RUSER_param;
        else
            m_RUSER <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_RUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     RUSER_param - The value to set onto wire <RUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RUSER_index1( int _this_dot_1, logic  RUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RUSER[_this_dot_1] = RUSER_param;
        else
            m_RUSER[_this_dot_1] <= RUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <RUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_RUSER(  );
        return RUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_RUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <RUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <RUSER>.
    //
    function automatic logic   get_RUSER_index1( int _this_dot_1 );
        return RUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <RREADY>.
    //
    // Parameters:
    //     RREADY_param - The value to set onto wire <RREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_RREADY( logic RREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_RREADY = RREADY_param;
        else
            m_RREADY <= RREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <RREADY>.
    //
    function automatic logic get_RREADY(  );
        return RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <WVALID>.
    //
    // Parameters:
    //     WVALID_param - The value to set onto wire <WVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WVALID( logic WVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WVALID = WVALID_param;
        else
            m_WVALID <= WVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <WVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WVALID>.
    //
    function automatic logic get_WVALID(  );
        return WVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WDATA
    //-------------------------------------------------------------------------
    //     Set the value of wire <WDATA>.
    //
    // Parameters:
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA( logic [((AXI4_WDATA_WIDTH) - 1):0] WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA = WDATA_param;
        else
            m_WDATA <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WDATA_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WDATA_param - The value to set onto wire <WDATA>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WDATA_index1( int _this_dot_1, logic  WDATA_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WDATA[_this_dot_1] = WDATA_param;
        else
            m_WDATA[_this_dot_1] <= WDATA_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WDATA
    //-------------------------------------------------------------------------
    //     Get the value of wire <WDATA>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic [((AXI4_WDATA_WIDTH) - 1):0]  get_WDATA(  );
        return WDATA;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WDATA_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WDATA>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WDATA>.
    //
    function automatic logic   get_WDATA_index1( int _this_dot_1 );
        return WDATA[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB
    //-------------------------------------------------------------------------
    //     Set the value of wire <WSTRB>.
    //
    // Parameters:
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB( logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB = WSTRB_param;
        else
            m_WSTRB <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WSTRB_param - The value to set onto wire <WSTRB>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WSTRB_index1( int _this_dot_1, logic  WSTRB_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WSTRB[_this_dot_1] = WSTRB_param;
        else
            m_WSTRB[_this_dot_1] <= WSTRB_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB
    //-------------------------------------------------------------------------
    //     Get the value of wire <WSTRB>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0]  get_WSTRB(  );
        return WSTRB;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WSTRB_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WSTRB>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WSTRB>.
    //
    function automatic logic   get_WSTRB_index1( int _this_dot_1 );
        return WSTRB[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WLAST
    //-------------------------------------------------------------------------
    //     Set the value of wire <WLAST>.
    //
    // Parameters:
    //     WLAST_param - The value to set onto wire <WLAST>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WLAST( logic WLAST_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WLAST = WLAST_param;
        else
            m_WLAST <= WLAST_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WLAST
    //-------------------------------------------------------------------------
    //     Get the value of wire <WLAST>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WLAST>.
    //
    function automatic logic get_WLAST(  );
        return WLAST;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <WUSER>.
    //
    // Parameters:
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER( logic [((AXI4_USER_WIDTH) - 1):0] WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER = WUSER_param;
        else
            m_WUSER <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_WUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     WUSER_param - The value to set onto wire <WUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WUSER_index1( int _this_dot_1, logic  WUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WUSER[_this_dot_1] = WUSER_param;
        else
            m_WUSER[_this_dot_1] <= WUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <WUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_WUSER(  );
        return WUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_WUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <WUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <WUSER>.
    //
    function automatic logic   get_WUSER_index1( int _this_dot_1 );
        return WUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <WREADY>.
    //
    // Parameters:
    //     WREADY_param - The value to set onto wire <WREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_WREADY( logic WREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_WREADY = WREADY_param;
        else
            m_WREADY <= WREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <WREADY>.
    //
    function automatic logic get_WREADY(  );
        return WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BVALID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BVALID>.
    //
    // Parameters:
    //     BVALID_param - The value to set onto wire <BVALID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BVALID( logic BVALID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BVALID = BVALID_param;
        else
            m_BVALID <= BVALID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BVALID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BVALID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BVALID>.
    //
    function automatic logic get_BVALID(  );
        return BVALID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BRESP
    //-------------------------------------------------------------------------
    //     Set the value of wire <BRESP>.
    //
    // Parameters:
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP( logic [1:0] BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP = BRESP_param;
        else
            m_BRESP <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BRESP_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BRESP_param - The value to set onto wire <BRESP>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BRESP_index1( int _this_dot_1, logic  BRESP_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BRESP[_this_dot_1] = BRESP_param;
        else
            m_BRESP[_this_dot_1] <= BRESP_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BRESP
    //-------------------------------------------------------------------------
    //     Get the value of wire <BRESP>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic [1:0]  get_BRESP(  );
        return BRESP;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BRESP_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BRESP>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BRESP>.
    //
    function automatic logic   get_BRESP_index1( int _this_dot_1 );
        return BRESP[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BID
    //-------------------------------------------------------------------------
    //     Set the value of wire <BID>.
    //
    // Parameters:
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID( logic [((AXI4_ID_WIDTH) - 1):0] BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID = BID_param;
        else
            m_BID <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BID_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BID_param - The value to set onto wire <BID>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BID_index1( int _this_dot_1, logic  BID_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BID[_this_dot_1] = BID_param;
        else
            m_BID[_this_dot_1] <= BID_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BID
    //-------------------------------------------------------------------------
    //     Get the value of wire <BID>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic [((AXI4_ID_WIDTH) - 1):0]  get_BID(  );
        return BID;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BID_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BID>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BID>.
    //
    function automatic logic   get_BID_index1( int _this_dot_1 );
        return BID[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BUSER
    //-------------------------------------------------------------------------
    //     Set the value of wire <BUSER>.
    //
    // Parameters:
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER( logic [((AXI4_USER_WIDTH) - 1):0] BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER = BUSER_param;
        else
            m_BUSER <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- set_BUSER_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     BUSER_param - The value to set onto wire <BUSER>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BUSER_index1( int _this_dot_1, logic  BUSER_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BUSER[_this_dot_1] = BUSER_param;
        else
            m_BUSER[_this_dot_1] <= BUSER_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BUSER
    //-------------------------------------------------------------------------
    //     Get the value of wire <BUSER>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic [((AXI4_USER_WIDTH) - 1):0]  get_BUSER(  );
        return BUSER;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_BUSER_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of wire <BUSER>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the wire <BUSER>.
    //
    function automatic logic   get_BUSER_index1( int _this_dot_1 );
        return BUSER[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of wire <BREADY>.
    //
    // Parameters:
    //     BREADY_param - The value to set onto wire <BREADY>.
    //     non_blocking - Set to 1 for a non-blocking assignment.
    //
    task automatic set_BREADY( logic BREADY_param = 'z, bit non_blocking = 1'b0 );
        if ( non_blocking == 1'b0 )
            m_BREADY = BREADY_param;
        else
            m_BREADY <= BREADY_param;
    endtask


    //-------------------------------------------------------------------------
    // Function:- get_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of wire <BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the wire <BREADY>.
    //
    function automatic logic get_BREADY(  );
        return BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id( bit [((AXI4_ID_WIDTH) - 1):0] config_axi4lite_tr_id_param );
        config_axi4lite_tr_id = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_axi4lite_tr_id_param - The value to assign to variable <config_axi4lite_tr_id>.
    //
    function automatic void set_config_axi4lite_tr_id_index1( int _this_dot_1, bit  config_axi4lite_tr_id_param );
        config_axi4lite_tr_id[_this_dot_1] = config_axi4lite_tr_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit [((AXI4_ID_WIDTH) - 1):0]  get_config_axi4lite_tr_id(  );
        return config_axi4lite_tr_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_tr_id_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_axi4lite_tr_id>.
    //
    function automatic bit   get_config_axi4lite_tr_id_index1( int _this_dot_1 );
        return config_axi4lite_tr_id[_this_dot_1];
    endfunction

    //-------------------------------------------------------------------------
    // Tasks to wait for a change to a global variable with read access
    //-------------------------------------------------------------------------


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_interface
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_interface>.
    //
    task automatic wait_for_config_axi4lite_interface(  );
        begin
            @( config_axi4lite_interface );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_interface_type
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_interface_type>.
    //
    task automatic wait_for_config_interface_type(  );
        begin
            @( config_interface_type );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_rlast
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_rlast>.
    //
    task automatic wait_for_config_enable_rlast(  );
        begin
            @( config_enable_rlast );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_id>.
    //
    task automatic wait_for_config_enable_id(  );
        begin
            @( config_enable_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_all_assertions
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_all_assertions>.
    //
    task automatic wait_for_config_enable_all_assertions(  );
        begin
            @( config_enable_all_assertions );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    task automatic wait_for_config_enable_assertion(  );
        begin
            @( config_enable_assertion );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_assertion_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_enable_assertion_index1( input int _this_dot_1 );
        begin
            @( config_enable_assertion[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_user_sideband
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_user_sideband>.
    //
    task automatic wait_for_config_enable_user_sideband(  );
        begin
            @( config_enable_user_sideband );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_slave_exclusive
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_slave_exclusive>.
    //
    task automatic wait_for_config_enable_slave_exclusive(  );
        begin
            @( config_enable_slave_exclusive );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_read_data_reordering_depth
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_read_data_reordering_depth>.
    //
    task automatic wait_for_config_read_data_reordering_depth(  );
        begin
            @( config_read_data_reordering_depth );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    task automatic wait_for_config_slave_start_addr(  );
        begin
            @( config_slave_start_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_start_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_start_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_start_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    task automatic wait_for_config_slave_end_addr(  );
        begin
            @( config_slave_end_addr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_slave_end_addr_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_slave_end_addr_index1( input int _this_dot_1 );
        begin
            @( config_slave_end_addr[_this_dot_1] );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_transaction_time_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_transaction_time_factor>.
    //
    task automatic wait_for_config_max_transaction_time_factor(  );
        begin
            @( config_max_transaction_time_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_burst_timeout_factor
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_burst_timeout_factor>.
    //
    task automatic wait_for_config_burst_timeout_factor(  );
        begin
            @( config_burst_timeout_factor );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_AWVALID_assertion_to_AWREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    task automatic wait_for_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        begin
            @( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_ARVALID_assertion_to_ARREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    task automatic wait_for_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        begin
            @( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_RVALID_assertion_to_RREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_RVALID_assertion_to_RREADY>.
    //
    task automatic wait_for_config_max_latency_RVALID_assertion_to_RREADY(  );
        begin
            @( config_max_latency_RVALID_assertion_to_RREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_BVALID_assertion_to_BREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_BVALID_assertion_to_BREADY>.
    //
    task automatic wait_for_config_max_latency_BVALID_assertion_to_BREADY(  );
        begin
            @( config_max_latency_BVALID_assertion_to_BREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_latency_WVALID_assertion_to_WREADY
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_latency_WVALID_assertion_to_WREADY>.
    //
    task automatic wait_for_config_max_latency_WVALID_assertion_to_WREADY(  );
        begin
            @( config_max_latency_WVALID_assertion_to_WREADY );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_enable_qos
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_enable_qos>.
    //
    task automatic wait_for_config_enable_qos(  );
        begin
            @( config_enable_qos );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_setup_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_setup_time>.
    //
    task automatic wait_for_config_setup_time(  );
        begin
            @( config_setup_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_hold_time
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_hold_time>.
    //
    task automatic wait_for_config_hold_time(  );
        begin
            @( config_hold_time );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_reads
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_reads>.
    //
    task automatic wait_for_config_num_max_outstanding_reads(  );
        begin
            @( config_num_max_outstanding_reads );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_max_outstanding_writes
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_max_outstanding_writes>.
    //
    task automatic wait_for_config_num_max_outstanding_writes(  );
        begin
            @( config_num_max_outstanding_writes );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_wr
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_wr>.
    //
    task automatic wait_for_config_max_outstanding_wr(  );
        begin
            @( config_max_outstanding_wr );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rd
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rd>.
    //
    task automatic wait_for_config_max_outstanding_rd(  );
        begin
            @( config_max_outstanding_rd );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_max_outstanding_rw
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_max_outstanding_rw>.
    //
    task automatic wait_for_config_max_outstanding_rw(  );
        begin
            @( config_max_outstanding_rw );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_wr_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_wr_phase>.
    //
    task automatic wait_for_config_num_outstanding_wr_phase(  );
        begin
            @( config_num_outstanding_wr_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_num_outstanding_rd_phase
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_num_outstanding_rd_phase>.
    //
    task automatic wait_for_config_num_outstanding_rd_phase(  );
        begin
            @( config_num_outstanding_rd_phase );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_is_issuing
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_is_issuing>.
    //
    task automatic wait_for_config_is_issuing(  );
        begin
            @( config_is_issuing );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    task automatic wait_for_config_axi4lite_tr_id(  );
        begin
            @( config_axi4lite_tr_id );
        end
    endtask


    //------------------------------------------------------------------------------
    // Function:- wait_for_config_axi4lite_tr_id_index1
    //------------------------------------------------------------------------------
    //     Wait for a change on variable <axi4::config_axi4lite_tr_id>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    task automatic wait_for_config_axi4lite_tr_id_index1( input int _this_dot_1 );
        begin
            @( config_axi4lite_tr_id[_this_dot_1] );
        end
    endtask


    //-------------------------------------------------------------------------
    // Functions to set global variables with write access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- set_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //     config_axi4lite_interface_param - The value to assign to variable <config_axi4lite_interface>.
    //
    function automatic void set_config_axi4lite_interface( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_interface_type
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_interface_type>.
    //
    // Parameters:
    //     config_interface_type_param - The value to assign to variable <config_interface_type>.
    //
    function automatic void set_config_interface_type( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //     config_enable_rlast_param - The value to assign to variable <config_enable_rlast>.
    //
    function automatic void set_config_enable_rlast( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_id
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_id>.
    //
    // Parameters:
    //     config_enable_id_param - The value to assign to variable <config_enable_id>.
    //
    function automatic void set_config_enable_id( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //     config_enable_all_assertions_param - The value to assign to variable <config_enable_all_assertions>.
    //
    function automatic void set_config_enable_all_assertions( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_enable_assertion_param - The value to assign to variable <config_enable_assertion>.
    //
    function automatic void set_config_enable_assertion_index1( int _this_dot_1, bit  config_enable_assertion_param );
        config_enable_assertion[_this_dot_1] = config_enable_assertion_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //     config_enable_user_sideband_param - The value to assign to variable <config_enable_user_sideband>.
    //
    function automatic void set_config_enable_user_sideband( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //     config_enable_slave_exclusive_param - The value to assign to variable <config_enable_slave_exclusive>.
    //
    function automatic void set_config_enable_slave_exclusive( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //     config_read_data_reordering_depth_param - The value to assign to variable <config_read_data_reordering_depth>.
    //
    function automatic void set_config_read_data_reordering_depth( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        config_slave_start_addr = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_start_addr_param - The value to assign to variable <config_slave_start_addr>.
    //
    function automatic void set_config_slave_start_addr_index1( int _this_dot_1, bit  config_slave_start_addr_param );
        config_slave_start_addr[_this_dot_1] = config_slave_start_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr( bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        config_slave_end_addr = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Set the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //     config_slave_end_addr_param - The value to assign to variable <config_slave_end_addr>.
    //
    function automatic void set_config_slave_end_addr_index1( int _this_dot_1, bit  config_slave_end_addr_param );
        config_slave_end_addr[_this_dot_1] = config_slave_end_addr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //     config_max_transaction_time_factor_param - The value to assign to variable <config_max_transaction_time_factor>.
    //
    function automatic void set_config_max_transaction_time_factor( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //     config_burst_timeout_factor_param - The value to assign to variable <config_burst_timeout_factor>.
    //
    function automatic void set_config_burst_timeout_factor( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //     config_max_latency_AWVALID_assertion_to_AWREADY_param - The value to assign to variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic void set_config_max_latency_AWVALID_assertion_to_AWREADY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //     config_max_latency_ARVALID_assertion_to_ARREADY_param - The value to assign to variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic void set_config_max_latency_ARVALID_assertion_to_ARREADY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //     config_max_latency_RVALID_assertion_to_RREADY_param - The value to assign to variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic void set_config_max_latency_RVALID_assertion_to_RREADY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //     config_max_latency_BVALID_assertion_to_BREADY_param - The value to assign to variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic void set_config_max_latency_BVALID_assertion_to_BREADY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //     config_max_latency_WVALID_assertion_to_WREADY_param - The value to assign to variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic void set_config_max_latency_WVALID_assertion_to_WREADY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_enable_qos
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_enable_qos>.
    //
    // Parameters:
    //     config_enable_qos_param - The value to assign to variable <config_enable_qos>.
    //
    function automatic void set_config_enable_qos( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_setup_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_setup_time>.
    //
    // Parameters:
    //     config_setup_time_param - The value to assign to variable <config_setup_time>.
    //
    function automatic void set_config_setup_time( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_hold_time
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_hold_time>.
    //
    // Parameters:
    //     config_hold_time_param - The value to assign to variable <config_hold_time>.
    //
    function automatic void set_config_hold_time( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //     config_num_max_outstanding_reads_param - The value to assign to variable <config_num_max_outstanding_reads>.
    //
    function automatic void set_config_num_max_outstanding_reads( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //     config_num_max_outstanding_writes_param - The value to assign to variable <config_num_max_outstanding_writes>.
    //
    function automatic void set_config_num_max_outstanding_writes( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //     config_max_outstanding_wr_param - The value to assign to variable <config_max_outstanding_wr>.
    //
    function automatic void set_config_max_outstanding_wr( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //     config_max_outstanding_rd_param - The value to assign to variable <config_max_outstanding_rd>.
    //
    function automatic void set_config_max_outstanding_rd( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //     config_max_outstanding_rw_param - The value to assign to variable <config_max_outstanding_rw>.
    //
    function automatic void set_config_max_outstanding_rw( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //     config_num_outstanding_wr_phase_param - The value to assign to variable <config_num_outstanding_wr_phase>.
    //
    function automatic void set_config_num_outstanding_wr_phase( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //     config_num_outstanding_rd_phase_param - The value to assign to variable <config_num_outstanding_rd_phase>.
    //
    function automatic void set_config_num_outstanding_rd_phase( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- set_config_is_issuing
    //-------------------------------------------------------------------------
    //     Set the value of variable <config_is_issuing>.
    //
    // Parameters:
    //     config_is_issuing_param - The value to assign to variable <config_is_issuing>.
    //
    function automatic void set_config_is_issuing( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to get global variables with read access
    //-------------------------------------------------------------------------


    //-------------------------------------------------------------------------
    // Function:- get_config_axi4lite_interface
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_axi4lite_interface>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_axi4lite_interface>.
    //
    function automatic bit get_config_axi4lite_interface(  );
        return config_axi4lite_interface;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_interface_type
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_interface_type>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_interface_type>.
    //
    function automatic axi4_interface_type_e get_config_interface_type(  );
        return config_interface_type;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_rlast
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_rlast>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_rlast>.
    //
    function automatic bit get_config_enable_rlast(  );
        return config_enable_rlast;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_id
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_id>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_id>.
    //
    function automatic bit get_config_enable_id(  );
        return config_enable_id;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_all_assertions
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_all_assertions>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_all_assertions>.
    //
    function automatic bit get_config_enable_all_assertions(  );
        return config_enable_all_assertions;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_assertion>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit [255:0]  get_config_enable_assertion(  );
        return config_enable_assertion;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_assertion_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_enable_assertion>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_enable_assertion>.
    //
    function automatic bit   get_config_enable_assertion_index1( int _this_dot_1 );
        return config_enable_assertion[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_user_sideband
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_user_sideband>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_user_sideband>.
    //
    function automatic bit get_config_enable_user_sideband(  );
        return config_enable_user_sideband;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_slave_exclusive
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_slave_exclusive>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_slave_exclusive>.
    //
    function automatic bit get_config_enable_slave_exclusive(  );
        return config_enable_slave_exclusive;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_read_data_reordering_depth
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_read_data_reordering_depth>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_read_data_reordering_depth>.
    //
    function automatic int unsigned get_config_read_data_reordering_depth(  );
        return config_read_data_reordering_depth;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_start_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_start_addr(  );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_start_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_start_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_start_addr>.
    //
    function automatic bit   get_config_slave_start_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_start_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_slave_end_addr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit [((AXI4_ADDRESS_WIDTH) - 1):0]  get_config_slave_end_addr(  );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_slave_end_addr_index1
    //-------------------------------------------------------------------------
    //     Get the value of one element of variable <config_slave_end_addr>.
    //
    // Parameters:
    //    _this_dot_1 - The array index for dimension 1.
    //
    // Returns the current value of the variable <config_slave_end_addr>.
    //
    function automatic bit   get_config_slave_end_addr_index1( int _this_dot_1 );
        dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
        return config_slave_end_addr[_this_dot_1];
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_transaction_time_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_transaction_time_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_transaction_time_factor>.
    //
    function automatic int unsigned get_config_max_transaction_time_factor(  );
        return config_max_transaction_time_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_burst_timeout_factor
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_burst_timeout_factor>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_burst_timeout_factor>.
    //
    function automatic int unsigned get_config_burst_timeout_factor(  );
        return config_burst_timeout_factor;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_AWVALID_assertion_to_AWREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_AWVALID_assertion_to_AWREADY>.
    //
    function automatic int unsigned get_config_max_latency_AWVALID_assertion_to_AWREADY(  );
        return config_max_latency_AWVALID_assertion_to_AWREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_ARVALID_assertion_to_ARREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_ARVALID_assertion_to_ARREADY>.
    //
    function automatic int unsigned get_config_max_latency_ARVALID_assertion_to_ARREADY(  );
        return config_max_latency_ARVALID_assertion_to_ARREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_RVALID_assertion_to_RREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_RVALID_assertion_to_RREADY>.
    //
    function automatic int unsigned get_config_max_latency_RVALID_assertion_to_RREADY(  );
        return config_max_latency_RVALID_assertion_to_RREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_BVALID_assertion_to_BREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_BVALID_assertion_to_BREADY>.
    //
    function automatic int unsigned get_config_max_latency_BVALID_assertion_to_BREADY(  );
        return config_max_latency_BVALID_assertion_to_BREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_latency_WVALID_assertion_to_WREADY
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_latency_WVALID_assertion_to_WREADY>.
    //
    function automatic int unsigned get_config_max_latency_WVALID_assertion_to_WREADY(  );
        return config_max_latency_WVALID_assertion_to_WREADY;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_enable_qos
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_enable_qos>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_enable_qos>.
    //
    function automatic bit get_config_enable_qos(  );
        return config_enable_qos;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_setup_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_setup_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_setup_time>.
    //
    function automatic longint unsigned get_config_setup_time(  );
        return config_setup_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_hold_time
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_hold_time>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_hold_time>.
    //
    function automatic longint unsigned get_config_hold_time(  );
        return config_hold_time;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_reads
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_reads>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_reads>.
    //
    function automatic int get_config_num_max_outstanding_reads(  );
        return config_num_max_outstanding_reads;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_max_outstanding_writes
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_max_outstanding_writes>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_max_outstanding_writes>.
    //
    function automatic int get_config_num_max_outstanding_writes(  );
        return config_num_max_outstanding_writes;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_wr
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_wr>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_wr>.
    //
    function automatic int get_config_max_outstanding_wr(  );
        return config_max_outstanding_wr;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rd
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rd>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rd>.
    //
    function automatic int get_config_max_outstanding_rd(  );
        return config_max_outstanding_rd;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_max_outstanding_rw
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_max_outstanding_rw>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_max_outstanding_rw>.
    //
    function automatic int get_config_max_outstanding_rw(  );
        return config_max_outstanding_rw;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_wr_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_wr_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_wr_phase>.
    //
    function automatic int get_config_num_outstanding_wr_phase(  );
        return config_num_outstanding_wr_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_num_outstanding_rd_phase
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_num_outstanding_rd_phase>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_num_outstanding_rd_phase>.
    //
    function automatic int get_config_num_outstanding_rd_phase(  );
        return config_num_outstanding_rd_phase;
    endfunction


    //-------------------------------------------------------------------------
    // Function:- get_config_is_issuing
    //-------------------------------------------------------------------------
    //     Get the value of variable <config_is_issuing>.
    //
    // Parameters:
    //
    // Returns the current value of the variable <config_is_issuing>.
    //
    function automatic bit get_config_is_issuing(  );
        return config_is_issuing;
    endfunction


    //-------------------------------------------------------------------------
    // Functions to set/get generic interface configuration
    //-------------------------------------------------------------------------

    function void set_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0,
        input int arg10 = 0
    );
        axi4_set_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10 );
    endfunction

    function int get_interface
    (
        input int what = 0,
        input int arg1 = 0,
        input int arg2 = 0,
        input int arg3 = 0,
        input int arg4 = 0,
        input int arg5 = 0,
        input int arg6 = 0,
        input int arg7 = 0,
        input int arg8 = 0,
        input int arg9 = 0
    );
        return axi4_get_interface( what, arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9 );
    endfunction

    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //-------------------------------------------------------------------------
    function string get_full_name();
        return axi4_get_full_name();
    endfunction

    //--------------------------------------------------------------------------
    //
    // Group:- Monitor Value Change on Variable
    //
    //--------------------------------------------------------------------------

    function automatic void axi4_local_set_config_axi4lite_interface_from_SystemVerilog( ref bit config_axi4lite_interface_param );
            dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog( _interface_ref,config_axi4lite_interface); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_axi4lite_interface_from_SystemVerilog( config_axi4lite_interface );
        end
    end

    function automatic void axi4_local_set_config_interface_type_from_SystemVerilog( ref axi4_interface_type_e config_interface_type_param );
            dvc_axi4_set_config_interface_type_from_SystemVerilog( _interface_ref,config_interface_type); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_interface_type_from_SystemVerilog( config_interface_type );
        end
    end

    function automatic void axi4_local_set_config_enable_rlast_from_SystemVerilog( ref bit config_enable_rlast_param );
            dvc_axi4_set_config_enable_rlast_from_SystemVerilog( _interface_ref,config_enable_rlast); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_rlast_from_SystemVerilog( config_enable_rlast );
        end
    end

    function automatic void axi4_local_set_config_enable_id_from_SystemVerilog( ref bit config_enable_id_param );
            dvc_axi4_set_config_enable_id_from_SystemVerilog( _interface_ref,config_enable_id); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_id_from_SystemVerilog( config_enable_id );
        end
    end

    function automatic void axi4_local_set_config_enable_all_assertions_from_SystemVerilog( ref bit config_enable_all_assertions_param );
            dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog( _interface_ref,config_enable_all_assertions); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_all_assertions_from_SystemVerilog( config_enable_all_assertions );
        end
    end

    function automatic void axi4_local_set_config_enable_assertion_from_SystemVerilog( ref bit [255:0] config_enable_assertion_param );
            dvc_axi4_set_config_enable_assertion_from_SystemVerilog( _interface_ref,config_enable_assertion); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_assertion_from_SystemVerilog( config_enable_assertion );
        end
    end

    function automatic void axi4_local_set_config_enable_user_sideband_from_SystemVerilog( ref bit config_enable_user_sideband_param );
            dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog( _interface_ref,config_enable_user_sideband); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_user_sideband_from_SystemVerilog( config_enable_user_sideband );
        end
    end

    function automatic void axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( ref bit config_enable_slave_exclusive_param );
            dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog( _interface_ref,config_enable_slave_exclusive); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_slave_exclusive_from_SystemVerilog( config_enable_slave_exclusive );
        end
    end

    function automatic void axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( ref int unsigned config_read_data_reordering_depth_param );
            dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog( _interface_ref,config_read_data_reordering_depth); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_read_data_reordering_depth_from_SystemVerilog( config_read_data_reordering_depth );
        end
    end

    function automatic void axi4_local_set_config_slave_start_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_start_addr_param );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1( _interface_ref,_this_dot_1,config_slave_start_addr[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
            dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog( _interface_ref ); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_start_addr_from_SystemVerilog( config_slave_start_addr );
        end
    end

    function automatic void axi4_local_set_config_slave_end_addr_from_SystemVerilog( ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] config_slave_end_addr_param );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1( _interface_ref,_this_dot_1,config_slave_end_addr[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
            dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog( _interface_ref ); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_slave_end_addr_from_SystemVerilog( config_slave_end_addr );
        end
    end

    function automatic void axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( ref int unsigned config_max_transaction_time_factor_param );
            dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog( _interface_ref,config_max_transaction_time_factor); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_transaction_time_factor_from_SystemVerilog( config_max_transaction_time_factor );
        end
    end

    function automatic void axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( ref int unsigned config_burst_timeout_factor_param );
            dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog( _interface_ref,config_burst_timeout_factor); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_burst_timeout_factor_from_SystemVerilog( config_burst_timeout_factor );
        end
    end

    function automatic void axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( ref int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
            dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( _interface_ref,config_max_latency_AWVALID_assertion_to_AWREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog( config_max_latency_AWVALID_assertion_to_AWREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( ref int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
            dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( _interface_ref,config_max_latency_ARVALID_assertion_to_ARREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog( config_max_latency_ARVALID_assertion_to_ARREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( ref int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
            dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( _interface_ref,config_max_latency_RVALID_assertion_to_RREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog( config_max_latency_RVALID_assertion_to_RREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( ref int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
            dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( _interface_ref,config_max_latency_BVALID_assertion_to_BREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog( config_max_latency_BVALID_assertion_to_BREADY );
        end
    end

    function automatic void axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( ref int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
            dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( _interface_ref,config_max_latency_WVALID_assertion_to_WREADY); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog( config_max_latency_WVALID_assertion_to_WREADY );
        end
    end

    function automatic void axi4_local_set_config_enable_qos_from_SystemVerilog( ref bit config_enable_qos_param );
            dvc_axi4_set_config_enable_qos_from_SystemVerilog( _interface_ref,config_enable_qos); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_enable_qos_from_SystemVerilog( config_enable_qos );
        end
    end

    function automatic void axi4_local_set_config_setup_time_from_SystemVerilog( ref longint unsigned config_setup_time_param );
            dvc_axi4_set_config_setup_time_from_SystemVerilog( _interface_ref,config_setup_time); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_setup_time_from_SystemVerilog( config_setup_time );
        end
    end

    function automatic void axi4_local_set_config_hold_time_from_SystemVerilog( ref longint unsigned config_hold_time_param );
            dvc_axi4_set_config_hold_time_from_SystemVerilog( _interface_ref,config_hold_time); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_hold_time_from_SystemVerilog( config_hold_time );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( ref int config_num_max_outstanding_reads_param );
            dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog( _interface_ref,config_num_max_outstanding_reads); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_reads_from_SystemVerilog( config_num_max_outstanding_reads );
        end
    end

    function automatic void axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( ref int config_num_max_outstanding_writes_param );
            dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog( _interface_ref,config_num_max_outstanding_writes); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_max_outstanding_writes_from_SystemVerilog( config_num_max_outstanding_writes );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( ref int config_max_outstanding_wr_param );
            dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog( _interface_ref,config_max_outstanding_wr); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_wr_from_SystemVerilog( config_max_outstanding_wr );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( ref int config_max_outstanding_rd_param );
            dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog( _interface_ref,config_max_outstanding_rd); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rd_from_SystemVerilog( config_max_outstanding_rd );
        end
    end

    function automatic void axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( ref int config_max_outstanding_rw_param );
            dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog( _interface_ref,config_max_outstanding_rw); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_max_outstanding_rw_from_SystemVerilog( config_max_outstanding_rw );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( ref int config_num_outstanding_wr_phase_param );
            dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog( _interface_ref,config_num_outstanding_wr_phase); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_wr_phase_from_SystemVerilog( config_num_outstanding_wr_phase );
        end
    end

    function automatic void axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( ref int config_num_outstanding_rd_phase_param );
            dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog( _interface_ref,config_num_outstanding_rd_phase); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_num_outstanding_rd_phase_from_SystemVerilog( config_num_outstanding_rd_phase );
        end
    end

    function automatic void axi4_local_set_config_is_issuing_from_SystemVerilog( ref bit config_is_issuing_param );
            dvc_axi4_set_config_is_issuing_from_SystemVerilog( _interface_ref,config_is_issuing); // DPI call to imported task
    endfunction

    initial
    begin
        wait(_interface_ref != 0);
        forever
        begin
            @( * ) axi4_local_set_config_is_issuing_from_SystemVerilog( config_is_issuing );
        end
    end

    //-------------------------------------------------------------------------
    // Transaction interface
    //-------------------------------------------------------------------------

    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_rw_transaction_addr;
    function void axi4_get_temp_static_rw_transaction_addr( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_addr[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_addr( input int _d1, input bit  _value );
        temp_static_rw_transaction_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_rw_transaction_id;
    function void axi4_get_temp_static_rw_transaction_id( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_id[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_id( input int _d1, input bit  _value );
        temp_static_rw_transaction_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_addr_user_data;
    function void axi4_get_temp_static_rw_transaction_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_rw_transaction_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_addr_user_data( input int _d1, input bit  _value );
        temp_static_rw_transaction_addr_user_data[_d1] = _value;
    endfunction
    bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] temp_static_rw_transaction_data_words [];
    function void axi4_get_temp_static_rw_transaction_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_rw_transaction_write_strobes [];
    function void axi4_get_temp_static_rw_transaction_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_wdata_user_data [];
    function void axi4_get_temp_static_rw_transaction_wdata_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_wdata_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_wdata_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_wdata_user_data[_d1][_d2] = _value;
    endfunction
    axi4_response_e temp_static_rw_transaction_resp[];
    function void axi4_get_temp_static_rw_transaction_resp( input int _d1, output axi4_response_e _value );
        _value = temp_static_rw_transaction_resp[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_resp( input int _d1, input axi4_response_e _value );
        temp_static_rw_transaction_resp[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_rw_transaction_resp_user_data [];
    function void axi4_get_temp_static_rw_transaction_resp_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_rw_transaction_resp_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_rw_transaction_resp_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_rw_transaction_resp_user_data[_d1][_d2] = _value;
    endfunction
    int temp_static_rw_transaction_data_valid_delay[];
    function void axi4_get_temp_static_rw_transaction_data_valid_delay( input int _d1, output int _value );
        _value = temp_static_rw_transaction_data_valid_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_valid_delay( input int _d1, input int _value );
        temp_static_rw_transaction_data_valid_delay[_d1] = _value;
    endfunction
    int temp_static_rw_transaction_data_ready_delay[];
    function void axi4_get_temp_static_rw_transaction_data_ready_delay( input int _d1, output int _value );
        _value = temp_static_rw_transaction_data_ready_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_data_ready_delay( input int _d1, input int _value );
        temp_static_rw_transaction_data_ready_delay[_d1] = _value;
    endfunction
    int unsigned temp_static_rw_transaction_write_data_beats_delay[];
    function void axi4_get_temp_static_rw_transaction_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_rw_transaction_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_rw_transaction_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_rw_transaction_write_data_beats_delay[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr;
    function void axi4_get_temp_static_read_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr( input int _d1, input bit  _value );
        temp_static_read_addr[_d1] = _value;
    endfunction
    int temp_static_read_rresp_delays[255:0];
    function void axi4_get_temp_static_read_rresp_delays( input int _d1, output int _value );
        _value = temp_static_read_rresp_delays[_d1];
    endfunction
    function void axi4_set_temp_static_read_rresp_delays( input int _d1, input int _value );
        temp_static_read_rresp_delays[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_id;
    function void axi4_get_temp_static_read_id( input int _d1, output bit  _value );
        _value = temp_static_read_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_id( input int _d1, input bit  _value );
        temp_static_read_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_user_data;
    function void axi4_get_temp_static_read_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_user_data( input int _d1, input bit  _value );
        temp_static_read_addr_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_data_words [];
    function void axi4_get_temp_static_read_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_read_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_read_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_read_data_words[_d1][_d2] = _value;
    endfunction
    axi4_response_e temp_static_read_resp[];
    function void axi4_get_temp_static_read_resp( input int _d1, output axi4_response_e _value );
        _value = temp_static_read_resp[_d1];
    endfunction
    function void axi4_set_temp_static_read_resp( input int _d1, input axi4_response_e _value );
        temp_static_read_resp[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_resp_user_data [];
    function void axi4_get_temp_static_read_resp_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_read_resp_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_read_resp_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_read_resp_user_data[_d1][_d2] = _value;
    endfunction
    longint temp_static_read_data_start_time[];
    function void axi4_get_temp_static_read_data_start_time( input int _d1, output longint _value );
        _value = temp_static_read_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_read_data_start_time( input int _d1, input longint _value );
        temp_static_read_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_read_data_end_time[];
    function void axi4_get_temp_static_read_data_end_time( input int _d1, output longint _value );
        _value = temp_static_read_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_read_data_end_time( input int _d1, input longint _value );
        temp_static_read_data_end_time[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr;
    function void axi4_get_temp_static_write_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr( input int _d1, input bit  _value );
        temp_static_write_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_id;
    function void axi4_get_temp_static_write_id( input int _d1, output bit  _value );
        _value = temp_static_write_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_id( input int _d1, input bit  _value );
        temp_static_write_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_user_data;
    function void axi4_get_temp_static_write_addr_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_addr_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_user_data( input int _d1, input bit  _value );
        temp_static_write_addr_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_data_words [];
    function void axi4_get_temp_static_write_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_write_strobes [];
    function void axi4_get_temp_static_write_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_write_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_wdata_user_data [];
    function void axi4_get_temp_static_write_wdata_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_wdata_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_wdata_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_write_wdata_user_data[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_user_data;
    function void axi4_get_temp_static_write_resp_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_user_data[_d1] = _value;
    endfunction
    longint temp_static_write_data_start_time[];
    function void axi4_get_temp_static_write_data_start_time( input int _d1, output longint _value );
        _value = temp_static_write_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_start_time( input int _d1, input longint _value );
        temp_static_write_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_write_data_end_time[];
    function void axi4_get_temp_static_write_data_end_time( input int _d1, output longint _value );
        _value = temp_static_write_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_end_time( input int _d1, input longint _value );
        temp_static_write_data_end_time[_d1] = _value;
    endfunction
    int unsigned temp_static_write_write_data_beats_delay[];
    function void axi4_get_temp_static_write_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_write_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_write_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_write_write_data_beats_delay[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_data_burst_data_words [];
    function void axi4_get_temp_static_write_data_burst_data_words( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_data_words[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_words( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_data_words[_d1][_d2] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_data_burst_write_strobes [];
    function void axi4_get_temp_static_write_data_burst_write_strobes( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_write_strobes[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_write_strobes( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_write_strobes[_d1][_d2] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_data_burst_user_data [];
    function void axi4_get_temp_static_write_data_burst_user_data( input int _d1, input int _d2, output bit _value );
        _value = temp_static_write_data_burst_user_data[_d1][_d2];
    endfunction
    function void axi4_set_temp_static_write_data_burst_user_data( input int _d1, input int _d2, input bit _value );
        temp_static_write_data_burst_user_data[_d1][_d2] = _value;
    endfunction
    longint temp_static_write_data_burst_data_start_time[];
    function void axi4_get_temp_static_write_data_burst_data_start_time( input int _d1, output longint _value );
        _value = temp_static_write_data_burst_data_start_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_start_time( input int _d1, input longint _value );
        temp_static_write_data_burst_data_start_time[_d1] = _value;
    endfunction
    longint temp_static_write_data_burst_data_end_time[];
    function void axi4_get_temp_static_write_data_burst_data_end_time( input int _d1, output longint _value );
        _value = temp_static_write_data_burst_data_end_time[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_end_time( input int _d1, input longint _value );
        temp_static_write_data_burst_data_end_time[_d1] = _value;
    endfunction
    int unsigned temp_static_write_data_burst_write_data_beats_delay[];
    function void axi4_get_temp_static_write_data_burst_write_data_beats_delay( input int _d1, output int unsigned _value );
        _value = temp_static_write_data_burst_write_data_beats_delay[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_write_data_beats_delay( input int _d1, input int unsigned _value );
        temp_static_write_data_burst_write_data_beats_delay[_d1] = _value;
    endfunction
    int temp_static_write_data_burst_data_valid2ready[];
    function void axi4_get_temp_static_write_data_burst_data_valid2ready( input int _d1, output int _value );
        _value = temp_static_write_data_burst_data_valid2ready[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data_valid2ready( input int _d1, input int _value );
        temp_static_write_data_burst_data_valid2ready[_d1] = _value;
    endfunction
    int temp_static_write_data_burst_data2data_clock[];
    function void axi4_get_temp_static_write_data_burst_data2data_clock( input int _d1, output int _value );
        _value = temp_static_write_data_burst_data2data_clock[_d1];
    endfunction
    function void axi4_set_temp_static_write_data_burst_data2data_clock( input int _d1, input int _value );
        temp_static_write_data_burst_data2data_clock[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr_channel_phase_addr;
    function void axi4_get_temp_static_read_addr_channel_phase_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_addr( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_addr_channel_phase_id;
    function void axi4_get_temp_static_read_addr_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_id( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_channel_phase_addr_user;
    function void axi4_get_temp_static_read_addr_channel_phase_addr_user( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_phase_addr_user[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_phase_addr_user( input int _d1, input bit  _value );
        temp_static_read_addr_channel_phase_addr_user[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_channel_phase_data;
    function void axi4_get_temp_static_read_channel_phase_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_data( input int _d1, input bit  _value );
        temp_static_read_channel_phase_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_channel_phase_id;
    function void axi4_get_temp_static_read_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_id( input int _d1, input bit  _value );
        temp_static_read_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_channel_phase_user_data;
    function void axi4_get_temp_static_read_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_read_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr_channel_phase_addr;
    function void axi4_get_temp_static_write_addr_channel_phase_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_addr( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_addr_channel_phase_id;
    function void axi4_get_temp_static_write_addr_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_id( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_channel_phase_addr_user;
    function void axi4_get_temp_static_write_addr_channel_phase_addr_user( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_phase_addr_user[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_phase_addr_user( input int _d1, input bit  _value );
        temp_static_write_addr_channel_phase_addr_user[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_channel_phase_data;
    function void axi4_get_temp_static_write_channel_phase_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_data( input int _d1, input bit  _value );
        temp_static_write_channel_phase_data[_d1] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_channel_phase_write_strobes;
    function void axi4_get_temp_static_write_channel_phase_write_strobes( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_write_strobes[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_write_strobes( input int _d1, input bit  _value );
        temp_static_write_channel_phase_write_strobes[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_channel_phase_user_data;
    function void axi4_get_temp_static_write_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_write_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_resp_channel_phase_id;
    function void axi4_get_temp_static_write_resp_channel_phase_id( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_phase_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_phase_id( input int _d1, input bit  _value );
        temp_static_write_resp_channel_phase_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_channel_phase_user_data;
    function void axi4_get_temp_static_write_resp_channel_phase_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_phase_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_phase_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_channel_phase_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_addr;
    function void axi4_get_temp_static_read_addr_channel_cycle_addr( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_addr[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_addr( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_id;
    function void axi4_get_temp_static_read_addr_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_addr_channel_cycle_user_data;
    function void axi4_get_temp_static_read_addr_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_addr_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_addr_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_read_addr_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_RDATA_WIDTH) - 1):0] temp_static_read_channel_cycle_data;
    function void axi4_get_temp_static_read_channel_cycle_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_data( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_read_channel_cycle_id;
    function void axi4_get_temp_static_read_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_read_channel_cycle_user_data;
    function void axi4_get_temp_static_read_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_read_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_read_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_read_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ADDRESS_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_addr;
    function void axi4_get_temp_static_write_addr_channel_cycle_addr( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_addr[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_addr( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_addr[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_id;
    function void axi4_get_temp_static_write_addr_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_addr_channel_cycle_user_data;
    function void axi4_get_temp_static_write_addr_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_addr_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_addr_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_addr_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_WDATA_WIDTH) - 1):0] temp_static_write_channel_cycle_data;
    function void axi4_get_temp_static_write_channel_cycle_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_data( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_data[_d1] = _value;
    endfunction
    bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] temp_static_write_channel_cycle_strb;
    function void axi4_get_temp_static_write_channel_cycle_strb( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_strb[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_strb( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_strb[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_channel_cycle_user_data;
    function void axi4_get_temp_static_write_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_channel_cycle_user_data[_d1] = _value;
    endfunction
    bit [((AXI4_ID_WIDTH) - 1):0] temp_static_write_resp_channel_cycle_id;
    function void axi4_get_temp_static_write_resp_channel_cycle_id( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_cycle_id[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_cycle_id( input int _d1, input bit  _value );
        temp_static_write_resp_channel_cycle_id[_d1] = _value;
    endfunction
    bit [((AXI4_USER_WIDTH) - 1):0] temp_static_write_resp_channel_cycle_user_data;
    function void axi4_get_temp_static_write_resp_channel_cycle_user_data( input int _d1, output bit  _value );
        _value = temp_static_write_resp_channel_cycle_user_data[_d1];
    endfunction
    function void axi4_set_temp_static_write_resp_channel_cycle_user_data( input int _d1, input bit  _value );
        temp_static_write_resp_channel_cycle_user_data[_d1] = _value;
    endfunction
    task automatic dvc_activate_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref axi4_rw_e read_or_write,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref int address_valid_delay,
        ref int data_valid_delay[],
        ref int write_response_valid_delay,
        ref int address_ready_delay,
        ref int data_ready_delay[],
        ref int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_rw_transaction_addr = addr;
            temp_static_rw_transaction_id = id;
            temp_static_rw_transaction_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_rw_transaction_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_rw_transaction_write_strobes = write_strobes;
            wdata_user_data_DIMS0 = wdata_user_data.size();
            temp_static_rw_transaction_wdata_user_data = wdata_user_data;
            resp_DIMS0 = resp.size();
            temp_static_rw_transaction_resp = resp;
            resp_user_data_DIMS0 = resp_user_data.size();
            temp_static_rw_transaction_resp_user_data = resp_user_data;
            data_valid_delay_DIMS0 = data_valid_delay.size();
            temp_static_rw_transaction_data_valid_delay = data_valid_delay;
            data_ready_delay_DIMS0 = data_ready_delay.size();
            temp_static_rw_transaction_data_ready_delay = data_ready_delay;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp_DIMS0, resp_user_data_DIMS0, address_valid_delay, data_valid_delay_DIMS0, write_response_valid_delay, address_ready_delay, data_ready_delay_DIMS0, write_response_ready_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_words.delete();
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_write_strobes.delete();
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_wdata_user_data.delete();
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_resp.delete();
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_resp_user_data.delete();
                    end
                    if (data_valid_delay_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_valid_delay = new [data_valid_delay_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_valid_delay.delete();
                    end
                    if (data_ready_delay_DIMS0 != 0)
                    begin
                        temp_static_rw_transaction_data_ready_delay = new [data_ready_delay_DIMS0];
                    end
                    else
                    begin
                        temp_static_rw_transaction_data_ready_delay.delete();
                    end
                    // Call function to get the sized params
                    axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, address_valid_delay, write_response_valid_delay, address_ready_delay, write_response_ready_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_rw_transaction_addr;
                    id = temp_static_rw_transaction_id;
                    addr_user_data = temp_static_rw_transaction_addr_user_data;
                    data_words = temp_static_rw_transaction_data_words;
                    temp_static_rw_transaction_data_words.delete();
                    write_strobes = temp_static_rw_transaction_write_strobes;
                    temp_static_rw_transaction_write_strobes.delete();
                    wdata_user_data = temp_static_rw_transaction_wdata_user_data;
                    temp_static_rw_transaction_wdata_user_data.delete();
                    resp = temp_static_rw_transaction_resp;
                    temp_static_rw_transaction_resp.delete();
                    resp_user_data = temp_static_rw_transaction_resp_user_data;
                    temp_static_rw_transaction_resp_user_data.delete();
                    data_valid_delay = temp_static_rw_transaction_data_valid_delay;
                    temp_static_rw_transaction_data_valid_delay.delete();
                    data_ready_delay = temp_static_rw_transaction_data_ready_delay;
                    temp_static_rw_transaction_data_ready_delay.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_rw_transaction
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_rw_e read_or_write,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [(((((AXI4_RDATA_WIDTH > AXI4_WDATA_WIDTH) ? AXI4_RDATA_WIDTH : AXI4_WDATA_WIDTH))) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output int address_valid_delay,
        ref int data_valid_delay[],
        output int write_response_valid_delay,
        output int address_ready_delay,
        ref int data_ready_delay[],
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_valid_delay_DIMS0;
                automatic int data_ready_delay_DIMS0;
                // Call function to get unsized params sizes.
                axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_valid_delay_DIMS0, data_ready_delay_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_write_strobes.delete();
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_wdata_user_data.delete();
                end
                if (resp_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_resp = new [resp_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_resp.delete();
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_resp_user_data.delete();
                end
                if (data_valid_delay_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_valid_delay = new [data_valid_delay_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_valid_delay.delete();
                end
                if (data_ready_delay_DIMS0 != 0)
                begin
                    temp_static_rw_transaction_data_ready_delay = new [data_ready_delay_DIMS0];
                end
                else
                begin
                    temp_static_rw_transaction_data_ready_delay.delete();
                end
                // Call function to get the sized params
                axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, read_or_write, prot, region, size, burst, lock, cache, qos, burst_length, address_valid_delay, write_response_valid_delay, address_ready_delay, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_rw_transaction_addr;
                id = temp_static_rw_transaction_id;
                addr_user_data = temp_static_rw_transaction_addr_user_data;
                data_words = temp_static_rw_transaction_data_words;
                temp_static_rw_transaction_data_words.delete();
                write_strobes = temp_static_rw_transaction_write_strobes;
                temp_static_rw_transaction_write_strobes.delete();
                wdata_user_data = temp_static_rw_transaction_wdata_user_data;
                temp_static_rw_transaction_wdata_user_data.delete();
                resp = temp_static_rw_transaction_resp;
                temp_static_rw_transaction_resp.delete();
                resp_user_data = temp_static_rw_transaction_resp_user_data;
                temp_static_rw_transaction_resp_user_data.delete();
                data_valid_delay = temp_static_rw_transaction_data_valid_delay;
                temp_static_rw_transaction_data_valid_delay.delete();
                data_ready_delay = temp_static_rw_transaction_data_ready_delay;
                temp_static_rw_transaction_data_ready_delay.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int rresp_delays[255:0],
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr = addr;
            temp_static_read_rresp_delays = rresp_delays;
            temp_static_read_id = id;
            temp_static_read_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_read_data_words = data_words;
            resp_DIMS0 = resp.size();
            temp_static_read_resp = resp;
            resp_user_data_DIMS0 = resp_user_data.size();
            temp_static_read_resp_user_data = resp_user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_read_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_read_data_end_time = data_end_time;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_read_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, resp_DIMS0, resp_user_data_DIMS0, addr_start_time, addr_end_time, data_start_time_DIMS0, data_end_time_DIMS0, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_read_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_words.delete();
                    end
                    if (resp_DIMS0 != 0)
                    begin
                        temp_static_read_resp = new [resp_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_resp.delete();
                    end
                    if (resp_user_data_DIMS0 != 0)
                    begin
                        temp_static_read_resp_user_data = new [resp_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_resp_user_data.delete();
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        temp_static_read_data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_start_time.delete();
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        temp_static_read_data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_read_data_end_time.delete();
                    end
                    // Call function to get the sized params
                    axi4_read_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, addr_start_time, addr_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_read_addr;
                    rresp_delays = temp_static_read_rresp_delays;
                    id = temp_static_read_id;
                    addr_user_data = temp_static_read_addr_user_data;
                    data_words = temp_static_read_data_words;
                    temp_static_read_data_words.delete();
                    resp = temp_static_read_resp;
                    temp_static_read_resp.delete();
                    resp_user_data = temp_static_read_resp_user_data;
                    temp_static_read_resp_user_data.delete();
                    data_start_time = temp_static_read_data_start_time;
                    temp_static_read_data_start_time.delete();
                    data_end_time = temp_static_read_data_end_time;
                    temp_static_read_data_end_time.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        ref int rresp_delays[255:0],
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_RDATA_WIDTH) - 1):0] data_words [],
        ref axi4_response_e resp[],
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data [],
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int resp_DIMS0;
                automatic int resp_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_read_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, resp_DIMS0, resp_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_read_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_read_data_words.delete();
                end
                if (resp_DIMS0 != 0)
                begin
                    temp_static_read_resp = new [resp_DIMS0];
                end
                else
                begin
                    temp_static_read_resp.delete();
                end
                if (resp_user_data_DIMS0 != 0)
                begin
                    temp_static_read_resp_user_data = new [resp_user_data_DIMS0];
                end
                else
                begin
                    temp_static_read_resp_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_read_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_read_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_read_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_read_data_end_time.delete();
                end
                // Call function to get the sized params
                axi4_read_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, size, burst, lock, cache, qos, burst_length, addr_start_time, addr_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr;
                rresp_delays = temp_static_read_rresp_delays;
                id = temp_static_read_id;
                addr_user_data = temp_static_read_addr_user_data;
                data_words = temp_static_read_data_words;
                temp_static_read_data_words.delete();
                resp = temp_static_read_resp;
                temp_static_read_resp.delete();
                resp_user_data = temp_static_read_resp_user_data;
                temp_static_read_resp_user_data.delete();
                data_start_time = temp_static_read_data_start_time;
                temp_static_read_data_start_time.delete();
                data_end_time = temp_static_read_data_end_time;
                temp_static_read_data_end_time.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_activate_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        ref bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        ref axi4_prot_e prot,
        ref int bresp_delay,
        ref bit [3:0] region,
        ref axi4_size_e size,
        ref axi4_burst_e burst,
        ref axi4_lock_e lock,
        ref axi4_cache_e cache,
        ref bit [3:0] qos,
        ref bit [((AXI4_ID_WIDTH) - 1):0] id,
        ref bit [7:0] burst_length,
        ref bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        ref axi4_response_e resp,
        ref bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        ref longint addr_start_time,
        ref longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref longint wr_resp_start_time,
        ref longint wr_resp_end_time,
        ref int address_valid_delay,
        input int _unit_id = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr = addr;
            temp_static_write_id = id;
            temp_static_write_addr_user_data = addr_user_data;
            data_words_DIMS0 = data_words.size();
            temp_static_write_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_write_write_strobes = write_strobes;
            wdata_user_data_DIMS0 = wdata_user_data.size();
            temp_static_write_wdata_user_data = wdata_user_data;
            temp_static_write_resp_user_data = resp_user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_write_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_write_data_end_time = data_end_time;
                // Call function to provide sized params and ingoing unsized params sizes.
                // In addition gets back updated sizes of unsized params.
                axi4_write_ActivatesActivatingActivate_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, resp, addr_start_time, addr_end_time, data_start_time_DIMS0, data_end_time_DIMS0, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                if ( ( _comms_semantic & QUESTA_MVC::QUESTA_MVC_COMMS_ACTIVATE ) != 0 )
                begin
                    // Create each unsized param
                    if (data_words_DIMS0 != 0)
                    begin
                        temp_static_write_data_words = new [data_words_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_words.delete();
                    end
                    if (write_strobes_DIMS0 != 0)
                    begin
                        temp_static_write_write_strobes = new [write_strobes_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_write_strobes.delete();
                    end
                    if (wdata_user_data_DIMS0 != 0)
                    begin
                        temp_static_write_wdata_user_data = new [wdata_user_data_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_wdata_user_data.delete();
                    end
                    if (data_start_time_DIMS0 != 0)
                    begin
                        temp_static_write_data_start_time = new [data_start_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_start_time.delete();
                    end
                    if (data_end_time_DIMS0 != 0)
                    begin
                        temp_static_write_data_end_time = new [data_end_time_DIMS0];
                    end
                    else
                    begin
                        temp_static_write_data_end_time.delete();
                    end
                    // Call function to get the sized params
                    axi4_write_ActivatesActivatingActivate_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, resp, addr_start_time, addr_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id); // DPI call to imported task
                    // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                    // In addition delete the storage allocated for the static variable(s)
                    addr = temp_static_write_addr;
                    id = temp_static_write_id;
                    addr_user_data = temp_static_write_addr_user_data;
                    data_words = temp_static_write_data_words;
                    temp_static_write_data_words.delete();
                    write_strobes = temp_static_write_write_strobes;
                    temp_static_write_write_strobes.delete();
                    wdata_user_data = temp_static_write_wdata_user_data;
                    temp_static_write_wdata_user_data.delete();
                    resp_user_data = temp_static_write_resp_user_data;
                    data_start_time = temp_static_write_data_start_time;
                    temp_static_write_data_start_time.delete();
                    data_end_time = temp_static_write_data_end_time;
                    temp_static_write_data_end_time.delete();
                end
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [3:0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [7:0] burst_length,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user_data,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] wdata_user_data [],
        output axi4_response_e resp,
        output bit [((AXI4_USER_WIDTH) - 1):0] resp_user_data,
        output longint addr_start_time,
        output longint addr_end_time,
        ref longint data_start_time[],
        ref longint data_end_time[],
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int wdata_user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, wdata_user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_write_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_write_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_write_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_write_write_strobes.delete();
                end
                if (wdata_user_data_DIMS0 != 0)
                begin
                    temp_static_write_wdata_user_data = new [wdata_user_data_DIMS0];
                end
                else
                begin
                    temp_static_write_wdata_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_write_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_write_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_end_time.delete();
                end
                // Call function to get the sized params
                axi4_write_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, bresp_delay, region, size, burst, lock, cache, qos, burst_length, resp, addr_start_time, addr_end_time, wr_resp_start_time, wr_resp_end_time, address_valid_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr;
                id = temp_static_write_id;
                addr_user_data = temp_static_write_addr_user_data;
                data_words = temp_static_write_data_words;
                temp_static_write_data_words.delete();
                write_strobes = temp_static_write_write_strobes;
                temp_static_write_write_strobes.delete();
                wdata_user_data = temp_static_write_wdata_user_data;
                temp_static_write_wdata_user_data.delete();
                resp_user_data = temp_static_write_resp_user_data;
                data_start_time = temp_static_write_data_start_time;
                temp_static_write_data_start_time.delete();
                data_end_time = temp_static_write_data_end_time;
                temp_static_write_data_end_time.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        input int wdata_burst_start_clock,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                automatic int data_valid2ready_DIMS0;
                automatic int data2data_clock_DIMS0;
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            data_words_DIMS0 = data_words.size();
            temp_static_write_data_burst_data_words = data_words;
            write_strobes_DIMS0 = write_strobes.size();
            temp_static_write_data_burst_write_strobes = write_strobes;
            user_data_DIMS0 = user_data.size();
            temp_static_write_data_burst_user_data = user_data;
            data_start_time_DIMS0 = data_start_time.size();
            temp_static_write_data_burst_data_start_time = data_start_time;
            data_end_time_DIMS0 = data_end_time.size();
            temp_static_write_data_burst_data_end_time = data_end_time;
            data_valid2ready_DIMS0 = data_valid2ready.size();
            temp_static_write_data_burst_data_valid2ready = data_valid2ready;
            data2data_clock_DIMS0 = data2data_clock.size();
            temp_static_write_data_burst_data2data_clock = data2data_clock;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, burst_length, data_words_DIMS0, write_strobes_DIMS0, user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, data_valid2ready_DIMS0, data2data_clock_DIMS0, wdata_burst_start_clock, _unit_id);
            // Delete the storage allocated for the static variable(s)
            temp_static_write_data_burst_data_words.delete();
            temp_static_write_data_burst_write_strobes.delete();
            temp_static_write_data_burst_user_data.delete();
            temp_static_write_data_burst_data_start_time.delete();
            temp_static_write_data_burst_data_end_time.delete();
            temp_static_write_data_burst_data_valid2ready.delete();
            temp_static_write_data_burst_data2data_clock.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output int burst_length,
        ref bit [((AXI4_WDATA_WIDTH) - 1):0] data_words [],
        ref bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes [],
        ref bit [((AXI4_USER_WIDTH) - 1):0] user_data [],
        ref longint data_start_time[],
        ref longint data_end_time[],
        ref int data_valid2ready[],
        ref int data2data_clock[],
        output int wdata_burst_start_clock,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                automatic int data_words_DIMS0;
                automatic int write_strobes_DIMS0;
                automatic int user_data_DIMS0;
                automatic int data_start_time_DIMS0;
                automatic int data_end_time_DIMS0;
                automatic int data_valid2ready_DIMS0;
                automatic int data2data_clock_DIMS0;
                // Call function to get unsized params sizes.
                axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, data_words_DIMS0, write_strobes_DIMS0, user_data_DIMS0, data_start_time_DIMS0, data_end_time_DIMS0, data_valid2ready_DIMS0, data2data_clock_DIMS0, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                if (data_words_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_words = new [data_words_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_words.delete();
                end
                if (write_strobes_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_write_strobes = new [write_strobes_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_write_strobes.delete();
                end
                if (user_data_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_user_data = new [user_data_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_user_data.delete();
                end
                if (data_start_time_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_start_time = new [data_start_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_start_time.delete();
                end
                if (data_end_time_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_end_time = new [data_end_time_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_end_time.delete();
                end
                if (data_valid2ready_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data_valid2ready = new [data_valid2ready_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data_valid2ready.delete();
                end
                if (data2data_clock_DIMS0 != 0)
                begin
                    temp_static_write_data_burst_data2data_clock = new [data2data_clock_DIMS0];
                end
                else
                begin
                    temp_static_write_data_burst_data2data_clock.delete();
                end
                // Call function to get the sized params
                axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, burst_length, wdata_burst_start_clock, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data_words = temp_static_write_data_burst_data_words;
                temp_static_write_data_burst_data_words.delete();
                write_strobes = temp_static_write_data_burst_write_strobes;
                temp_static_write_data_burst_write_strobes.delete();
                user_data = temp_static_write_data_burst_user_data;
                temp_static_write_data_burst_user_data.delete();
                data_start_time = temp_static_write_data_burst_data_start_time;
                temp_static_write_data_burst_data_start_time.delete();
                data_end_time = temp_static_write_data_burst_data_end_time;
                temp_static_write_data_burst_data_end_time.delete();
                data_valid2ready = temp_static_write_data_burst_data_valid2ready;
                temp_static_write_data_burst_data_valid2ready.delete();
                data2data_clock = temp_static_write_data_burst_data2data_clock;
                temp_static_write_data_burst_data2data_clock.delete();
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr_channel_phase_addr = addr;
            temp_static_read_addr_channel_phase_id = id;
            temp_static_read_addr_channel_phase_addr_user = addr_user;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr_channel_phase_addr;
                id = temp_static_read_addr_channel_phase_id;
                addr_user = temp_static_read_addr_channel_phase_addr_user;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, _unit_id);
        end
    endtask

    task automatic dvc_get_read_data_burst
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_channel_phase_data = data;
            temp_static_read_channel_phase_id = id;
            temp_static_read_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, last, data_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, last, data_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_read_channel_phase_data;
                id = temp_static_read_channel_phase_id;
                user_data = temp_static_read_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr_channel_phase_addr = addr;
            temp_static_write_addr_channel_phase_id = id;
            temp_static_write_addr_channel_phase_addr_user = addr_user;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_addr_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output axi4_prot_e prot,
        output bit [3:0] region,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] addr_user,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, prot, region, burst_length, size, burst, lock, cache, qos, address_valid_delay, address_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr_channel_phase_addr;
                id = temp_static_write_addr_channel_phase_id;
                addr_user = temp_static_write_addr_channel_phase_addr_user;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int data_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_channel_phase_data = data;
            temp_static_write_channel_phase_write_strobes = write_strobes;
            temp_static_write_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, last, data_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] write_strobes,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int data_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, last, data_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_write_channel_phase_data;
                write_strobes = temp_static_write_channel_phase_write_strobes;
                user_data = temp_static_write_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int write_response_ready_delay,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_resp_channel_phase_id = id;
            temp_static_write_resp_channel_phase_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, write_response_ready_delay, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_resp_channel_phase
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output int write_response_ready_delay,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, write_response_ready_delay, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                id = temp_static_write_resp_channel_phase_id;
                user_data = temp_static_write_resp_channel_phase_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_addr_channel_cycle_addr = addr;
            temp_static_read_addr_channel_cycle_id = id;
            temp_static_read_addr_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_read_addr_channel_cycle_addr;
                id = temp_static_read_addr_channel_cycle_id;
                user_data = temp_static_read_addr_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        input axi4_response_e resp,
        input bit last,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_read_channel_cycle_data = data;
            temp_static_read_channel_cycle_id = id;
            temp_static_read_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_read_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, last, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_read_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_RDATA_WIDTH) - 1):0] data,
        output axi4_response_e resp,
        output bit last,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, last, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_read_channel_cycle_data;
                id = temp_static_read_channel_cycle_id;
                user_data = temp_static_read_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_read_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input axi4_prot_e prot,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_addr_channel_cycle_addr = addr;
            temp_static_write_addr_channel_cycle_id = id;
            temp_static_write_addr_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_addr_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_ADDRESS_WIDTH) - 1):0] addr,
        output bit [3:0] region,
        output axi4_cache_e cache,
        output bit [7:0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [3:0] qos,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        output axi4_prot_e prot,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, region, cache, burst_length, size, burst, lock, qos, prot, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                addr = temp_static_write_addr_channel_cycle_addr;
                id = temp_static_write_addr_channel_cycle_id;
                user_data = temp_static_write_addr_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_addr_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        input bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        input bit last,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_channel_cycle_data = data;
            temp_static_write_channel_cycle_strb = strb;
            temp_static_write_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, last, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit [((AXI4_WDATA_WIDTH) - 1):0] data,
        output bit [(((AXI4_WDATA_WIDTH / 8)) - 1):0] strb,
        output bit last,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, last, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                data = temp_static_write_channel_cycle_data;
                strb = temp_static_write_channel_cycle_strb;
                user_data = temp_static_write_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask

    task automatic dvc_put_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit [((AXI4_ID_WIDTH) - 1):0] id,
        input bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
            // Pass to CY the size of each open dimension (assumes rectangular arrays)
            // In addition copy any unsized or flexibly sized parameters to a related static variable which will be accessed element by element from the C
            temp_static_write_resp_channel_cycle_id = id;
            temp_static_write_resp_channel_cycle_user_data = user_data;
            // Call function to provide sized params and ingoing unsized params sizes.
            axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, resp, _unit_id);
            // Delete the storage allocated for the static variable(s)
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_get_write_resp_channel_cycle
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output axi4_response_e resp,
        output bit [((AXI4_ID_WIDTH) - 1):0] id,
        output bit [((AXI4_USER_WIDTH) - 1):0] user_data,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin
            int _trans_id;

            wait(_interface_ref != 0);

            // the real code .....
            // Create an array to hold the unsized dims for each param (..._DIMS)
            begin // Block to create unsized data arrays
                // Call function to get unsized params sizes.
                axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, _trans_id, _unit_id, _using); // DPI call to imported task
                // Create each unsized param
                // Call function to get the sized params
                axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog(_comms_semantic,_as_end, _trans_id, resp, _unit_id, _using); // DPI call to imported task
                // Copy unsized data from static variable(s) which has/have been set element by element from the C++
                // In addition delete the storage allocated for the static variable(s)
                id = temp_static_write_resp_channel_cycle_id;
                user_data = temp_static_write_resp_channel_cycle_user_data;
            end // Block to create unsized data arrays
        end
    endtask

    task automatic dvc_put_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id = 0
    );
        begin

            wait(_interface_ref != 0);

            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id);
        end
    endtask

    task automatic dvc_get_write_resp_channel_ready
    (
        input questa_mvc_item_comms_semantic _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id = 0,
        input bit _using = 0
    );
        begin

            wait(_interface_ref != 0);

            // the real code .....
            // Call function to set/get the params, all are of known size
            axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog(_comms_semantic,_as_end, ready, _unit_id, _using); // DPI call to imported task
        end
    endtask


    //-------------------------------------------------------------------------
    // Generic Interface Configuration Support
    //

    import "DPI-C" context dvc_axi4_set_interface = function void axi4_set_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9,
        input int arg10
    );
    import "DPI-C" context dvc_axi4_get_interface = function int axi4_get_interface
    (
        input int what,
        input int arg1,
        input int arg2,
        input int arg3,
        input int arg4,
        input int arg5,
        input int arg6,
        input int arg7,
        input int arg8,
        input int arg9
    );


    //-------------------------------------------------------------------------
    // Functions to get the hierarchic name of this interface
    //
    import "DPI-C" context dvc_axi4_get_full_name = function string axi4_get_full_name();



    //-------------------------------------------------------------------------
    // Abstraction level Support
    //

    import "DPI-C" context dvc_axi4_set_master_end_abstraction_level =
    function void axi4_set_master_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_master_end_abstraction_level =
    function void axi4_get_master_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_set_slave_end_abstraction_level =
    function void axi4_set_slave_end_abstraction_level
    (
        input bit         wire_level,
        input bit         TLM_level
    );
    import "DPI-C" context dvc_axi4_get_slave_end_abstraction_level =
    function void axi4_get_slave_end_abstraction_level
    (
        output bit         wire_level,
        output bit         TLM_level
    );

    //-------------------------------------------------------------------------
    // Wire Level Interface Support
    //
    logic internal_ACLK = 'z;
    logic internal_ARESETn = 'z;
    logic internal_AWVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_AWADDR = 'z;
    logic [2:0] internal_AWPROT = 'z;
    logic [3:0] internal_AWREGION = 'z;
    logic [7:0] internal_AWLEN = 'z;
    logic [3:0] internal_AWSIZE = 'z;
    logic [1:0] internal_AWBURST = 'z;
    logic internal_AWLOCK = 'z;
    logic [3:0] internal_AWCACHE = 'z;
    logic [3:0] internal_AWQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_AWID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_AWUSER = 'z;
    logic internal_AWREADY = 'z;
    logic internal_ARVALID = 'z;
    logic [((AXI4_ADDRESS_WIDTH) - 1):0] internal_ARADDR = 'z;
    logic [2:0] internal_ARPROT = 'z;
    logic [3:0] internal_ARREGION = 'z;
    logic [7:0] internal_ARLEN = 'z;
    logic [3:0] internal_ARSIZE = 'z;
    logic [1:0] internal_ARBURST = 'z;
    logic internal_ARLOCK = 'z;
    logic [3:0] internal_ARCACHE = 'z;
    logic [3:0] internal_ARQOS = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_ARID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_ARUSER = 'z;
    logic internal_ARREADY = 'z;
    logic internal_RVALID = 'z;
    logic [((AXI4_RDATA_WIDTH) - 1):0] internal_RDATA = 'z;
    logic [1:0] internal_RRESP = 'z;
    logic internal_RLAST = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_RID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_RUSER = 'z;
    logic internal_RREADY = 'z;
    logic internal_WVALID = 'z;
    logic [((AXI4_WDATA_WIDTH) - 1):0] internal_WDATA = 'z;
    logic [(((AXI4_WDATA_WIDTH / 8)) - 1):0] internal_WSTRB = 'z;
    logic internal_WLAST = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_WUSER = 'z;
    logic internal_WREADY = 'z;
    logic internal_BVALID = 'z;
    logic [1:0] internal_BRESP = 'z;
    logic [((AXI4_ID_WIDTH) - 1):0] internal_BID = 'z;
    logic [((AXI4_USER_WIDTH) - 1):0] internal_BUSER = 'z;
    logic internal_BREADY = 'z;
    import "DPI-C" context dvc_axi4_set_ACLK_from_SystemVerilog = function void dvc_axi4_set_ACLK_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit ACLK_param
    );
    import "DPI-C" context dvc_axi4_get_ACLK_into_SystemVerilog = function void dvc_axi4_get_ACLK_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit ACLK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ACLK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARESETn_from_SystemVerilog = function void dvc_axi4_set_ARESETn_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARESETn_param
    );
    import "DPI-C" context dvc_axi4_get_ARESETn_into_SystemVerilog = function void dvc_axi4_get_ARESETn_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARESETn_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARESETn_from_CY;

    import "DPI-C" context dvc_axi4_set_AWVALID_from_SystemVerilog = function void dvc_axi4_set_AWVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWVALID_param
    );
    import "DPI-C" context dvc_axi4_get_AWVALID_into_SystemVerilog = function void dvc_axi4_get_AWVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWADDR_from_SystemVerilog_index1 = function void dvc_axi4_set_AWADDR_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWADDR_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWADDR_from_SystemVerilog = function void dvc_axi4_propagate_AWADDR_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWADDR_into_SystemVerilog = function void dvc_axi4_get_AWADDR_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWADDR_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_AWPROT_from_SystemVerilog = function void dvc_axi4_set_AWPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] AWPROT_param
    );
    import "DPI-C" context dvc_axi4_get_AWPROT_into_SystemVerilog = function void dvc_axi4_get_AWPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] AWPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREGION_from_SystemVerilog = function void dvc_axi4_set_AWREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWREGION_param
    );
    import "DPI-C" context dvc_axi4_get_AWREGION_into_SystemVerilog = function void dvc_axi4_get_AWREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLEN_from_SystemVerilog = function void dvc_axi4_set_AWLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] AWLEN_param
    );
    import "DPI-C" context dvc_axi4_get_AWLEN_into_SystemVerilog = function void dvc_axi4_get_AWLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] AWLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_AWSIZE_from_SystemVerilog = function void dvc_axi4_set_AWSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_AWSIZE_into_SystemVerilog = function void dvc_axi4_get_AWSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWBURST_from_SystemVerilog = function void dvc_axi4_set_AWBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] AWBURST_param
    );
    import "DPI-C" context dvc_axi4_get_AWBURST_into_SystemVerilog = function void dvc_axi4_get_AWBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] AWBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_AWLOCK_from_SystemVerilog = function void dvc_axi4_set_AWLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_AWLOCK_into_SystemVerilog = function void dvc_axi4_get_AWLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_AWCACHE_from_SystemVerilog = function void dvc_axi4_set_AWCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_AWCACHE_into_SystemVerilog = function void dvc_axi4_get_AWCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_AWQOS_from_SystemVerilog = function void dvc_axi4_set_AWQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] AWQOS_param
    );
    import "DPI-C" context dvc_axi4_get_AWQOS_into_SystemVerilog = function void dvc_axi4_get_AWQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] AWQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_AWID_from_SystemVerilog_index1 = function void dvc_axi4_set_AWID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWID_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWID_from_SystemVerilog = function void dvc_axi4_propagate_AWID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWID_into_SystemVerilog = function void dvc_axi4_get_AWID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWID_from_CY;

    import "DPI-C" context dvc_axi4_set_AWUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_AWUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  AWUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_AWUSER_from_SystemVerilog = function void dvc_axi4_propagate_AWUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_AWUSER_into_SystemVerilog = function void dvc_axi4_get_AWUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_AWUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_AWUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_AWREADY_from_SystemVerilog = function void dvc_axi4_set_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_AWREADY_into_SystemVerilog = function void dvc_axi4_get_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic AWREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_ARVALID_from_SystemVerilog = function void dvc_axi4_set_ARVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARVALID_param
    );
    import "DPI-C" context dvc_axi4_get_ARVALID_into_SystemVerilog = function void dvc_axi4_get_ARVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARADDR_from_SystemVerilog_index1 = function void dvc_axi4_set_ARADDR_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARADDR_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARADDR_from_SystemVerilog = function void dvc_axi4_propagate_ARADDR_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARADDR_into_SystemVerilog = function void dvc_axi4_get_ARADDR_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARADDR_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARADDR_from_CY;

    import "DPI-C" context dvc_axi4_set_ARPROT_from_SystemVerilog = function void dvc_axi4_set_ARPROT_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [2:0] ARPROT_param
    );
    import "DPI-C" context dvc_axi4_get_ARPROT_into_SystemVerilog = function void dvc_axi4_get_ARPROT_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [2:0] ARPROT_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARPROT_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREGION_from_SystemVerilog = function void dvc_axi4_set_ARREGION_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARREGION_param
    );
    import "DPI-C" context dvc_axi4_get_ARREGION_into_SystemVerilog = function void dvc_axi4_get_ARREGION_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARREGION_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREGION_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLEN_from_SystemVerilog = function void dvc_axi4_set_ARLEN_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [7:0] ARLEN_param
    );
    import "DPI-C" context dvc_axi4_get_ARLEN_into_SystemVerilog = function void dvc_axi4_get_ARLEN_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [7:0] ARLEN_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLEN_from_CY;

    import "DPI-C" context dvc_axi4_set_ARSIZE_from_SystemVerilog = function void dvc_axi4_set_ARSIZE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARSIZE_param
    );
    import "DPI-C" context dvc_axi4_get_ARSIZE_into_SystemVerilog = function void dvc_axi4_get_ARSIZE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARSIZE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARSIZE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARBURST_from_SystemVerilog = function void dvc_axi4_set_ARBURST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] ARBURST_param
    );
    import "DPI-C" context dvc_axi4_get_ARBURST_into_SystemVerilog = function void dvc_axi4_get_ARBURST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] ARBURST_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARBURST_from_CY;

    import "DPI-C" context dvc_axi4_set_ARLOCK_from_SystemVerilog = function void dvc_axi4_set_ARLOCK_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARLOCK_param
    );
    import "DPI-C" context dvc_axi4_get_ARLOCK_into_SystemVerilog = function void dvc_axi4_get_ARLOCK_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARLOCK_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARLOCK_from_CY;

    import "DPI-C" context dvc_axi4_set_ARCACHE_from_SystemVerilog = function void dvc_axi4_set_ARCACHE_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARCACHE_param
    );
    import "DPI-C" context dvc_axi4_get_ARCACHE_into_SystemVerilog = function void dvc_axi4_get_ARCACHE_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARCACHE_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARCACHE_from_CY;

    import "DPI-C" context dvc_axi4_set_ARQOS_from_SystemVerilog = function void dvc_axi4_set_ARQOS_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [3:0] ARQOS_param
    );
    import "DPI-C" context dvc_axi4_get_ARQOS_into_SystemVerilog = function void dvc_axi4_get_ARQOS_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [3:0] ARQOS_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARQOS_from_CY;

    import "DPI-C" context dvc_axi4_set_ARID_from_SystemVerilog_index1 = function void dvc_axi4_set_ARID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARID_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARID_from_SystemVerilog = function void dvc_axi4_propagate_ARID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARID_into_SystemVerilog = function void dvc_axi4_get_ARID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARID_from_CY;

    import "DPI-C" context dvc_axi4_set_ARUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_ARUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  ARUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_ARUSER_from_SystemVerilog = function void dvc_axi4_propagate_ARUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_ARUSER_into_SystemVerilog = function void dvc_axi4_get_ARUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_ARUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_ARUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_ARREADY_from_SystemVerilog = function void dvc_axi4_set_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_ARREADY_into_SystemVerilog = function void dvc_axi4_get_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic ARREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_RVALID_from_SystemVerilog = function void dvc_axi4_set_RVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RVALID_param
    );
    import "DPI-C" context dvc_axi4_get_RVALID_into_SystemVerilog = function void dvc_axi4_get_RVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_RVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_RDATA_from_SystemVerilog_index1 = function void dvc_axi4_set_RDATA_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RDATA_param
    );
    import "DPI-C" context dvc_axi4_propagate_RDATA_from_SystemVerilog = function void dvc_axi4_propagate_RDATA_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RDATA_into_SystemVerilog = function void dvc_axi4_get_RDATA_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RDATA_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_RRESP_from_SystemVerilog = function void dvc_axi4_set_RRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] RRESP_param
    );
    import "DPI-C" context dvc_axi4_get_RRESP_into_SystemVerilog = function void dvc_axi4_get_RRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] RRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_RRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_RLAST_from_SystemVerilog = function void dvc_axi4_set_RLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RLAST_param
    );
    import "DPI-C" context dvc_axi4_get_RLAST_into_SystemVerilog = function void dvc_axi4_get_RLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_RLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_RID_from_SystemVerilog_index1 = function void dvc_axi4_set_RID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RID_param
    );
    import "DPI-C" context dvc_axi4_propagate_RID_from_SystemVerilog = function void dvc_axi4_propagate_RID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RID_into_SystemVerilog = function void dvc_axi4_get_RID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RID_from_CY;

    import "DPI-C" context dvc_axi4_set_RUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_RUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  RUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_RUSER_from_SystemVerilog = function void dvc_axi4_propagate_RUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_RUSER_into_SystemVerilog = function void dvc_axi4_get_RUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_RUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_RUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_RREADY_from_SystemVerilog = function void dvc_axi4_set_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_RREADY_into_SystemVerilog = function void dvc_axi4_get_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic RREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_WVALID_from_SystemVerilog = function void dvc_axi4_set_WVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WVALID_param
    );
    import "DPI-C" context dvc_axi4_get_WVALID_into_SystemVerilog = function void dvc_axi4_get_WVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_WVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_WDATA_from_SystemVerilog_index1 = function void dvc_axi4_set_WDATA_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WDATA_param
    );
    import "DPI-C" context dvc_axi4_propagate_WDATA_from_SystemVerilog = function void dvc_axi4_propagate_WDATA_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WDATA_into_SystemVerilog = function void dvc_axi4_get_WDATA_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WDATA_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WDATA_from_CY;

    import "DPI-C" context dvc_axi4_set_WSTRB_from_SystemVerilog_index1 = function void dvc_axi4_set_WSTRB_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WSTRB_param
    );
    import "DPI-C" context dvc_axi4_propagate_WSTRB_from_SystemVerilog = function void dvc_axi4_propagate_WSTRB_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WSTRB_into_SystemVerilog = function void dvc_axi4_get_WSTRB_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WSTRB_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WSTRB_from_CY;

    import "DPI-C" context dvc_axi4_set_WLAST_from_SystemVerilog = function void dvc_axi4_set_WLAST_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WLAST_param
    );
    import "DPI-C" context dvc_axi4_get_WLAST_into_SystemVerilog = function void dvc_axi4_get_WLAST_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WLAST_param

    );
    export "DPI-C" function dvc_axi4_initialise_WLAST_from_CY;

    import "DPI-C" context dvc_axi4_set_WUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_WUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  WUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_WUSER_from_SystemVerilog = function void dvc_axi4_propagate_WUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_WUSER_into_SystemVerilog = function void dvc_axi4_get_WUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_WUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_WUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_WREADY_from_SystemVerilog = function void dvc_axi4_set_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_WREADY_into_SystemVerilog = function void dvc_axi4_get_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic WREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_BVALID_from_SystemVerilog = function void dvc_axi4_set_BVALID_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BVALID_param
    );
    import "DPI-C" context dvc_axi4_get_BVALID_into_SystemVerilog = function void dvc_axi4_get_BVALID_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BVALID_param

    );
    export "DPI-C" function dvc_axi4_initialise_BVALID_from_CY;

    import "DPI-C" context dvc_axi4_set_BRESP_from_SystemVerilog = function void dvc_axi4_set_BRESP_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic [1:0] BRESP_param
    );
    import "DPI-C" context dvc_axi4_get_BRESP_into_SystemVerilog = function void dvc_axi4_get_BRESP_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic [1:0] BRESP_param

    );
    export "DPI-C" function dvc_axi4_initialise_BRESP_from_CY;

    import "DPI-C" context dvc_axi4_set_BID_from_SystemVerilog_index1 = function void dvc_axi4_set_BID_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  BID_param
    );
    import "DPI-C" context dvc_axi4_propagate_BID_from_SystemVerilog = function void dvc_axi4_propagate_BID_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_BID_into_SystemVerilog = function void dvc_axi4_get_BID_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_BID_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_BID_from_CY;

    import "DPI-C" context dvc_axi4_set_BUSER_from_SystemVerilog_index1 = function void dvc_axi4_set_BUSER_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input logic  BUSER_param
    );
    import "DPI-C" context dvc_axi4_propagate_BUSER_from_SystemVerilog = function void dvc_axi4_propagate_BUSER_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_BUSER_into_SystemVerilog = function void dvc_axi4_get_BUSER_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_BUSER_from_CY_index1;
    export "DPI-C" function dvc_axi4_initialise_BUSER_from_CY;

    import "DPI-C" context dvc_axi4_set_BREADY_from_SystemVerilog = function void dvc_axi4_set_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input logic BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_BREADY_into_SystemVerilog = function void dvc_axi4_get_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output logic BREADY_param

    );
    export "DPI-C" function dvc_axi4_initialise_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog = function void dvc_axi4_set_config_axi4lite_interface_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_axi4lite_interface_param
    );
    import "DPI-C" context dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog = function void dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_axi4lite_interface_param

    );
    export "DPI-C" function dvc_axi4_set_config_axi4lite_interface_from_CY;

    import "DPI-C" context dvc_axi4_set_config_interface_type_from_SystemVerilog = function void dvc_axi4_set_config_interface_type_from_SystemVerilog
    (
        input longint _iface_ref,
        input axi4_interface_type_e config_interface_type_param
    );
    import "DPI-C" context dvc_axi4_get_config_interface_type_into_SystemVerilog = function void dvc_axi4_get_config_interface_type_into_SystemVerilog
    (
        input longint _iface_ref,
        output axi4_interface_type_e config_interface_type_param

    );
    export "DPI-C" function dvc_axi4_set_config_interface_type_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_rlast_from_SystemVerilog = function void dvc_axi4_set_config_enable_rlast_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_rlast_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_rlast_into_SystemVerilog = function void dvc_axi4_get_config_enable_rlast_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_rlast_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_rlast_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_id_from_SystemVerilog = function void dvc_axi4_set_config_enable_id_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_id_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_id_into_SystemVerilog = function void dvc_axi4_get_config_enable_id_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_id_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_id_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog = function void dvc_axi4_set_config_enable_all_assertions_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_all_assertions_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog = function void dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_all_assertions_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_all_assertions_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_assertion_from_SystemVerilog = function void dvc_axi4_set_config_enable_assertion_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit [255:0] config_enable_assertion_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_assertion_into_SystemVerilog = function void dvc_axi4_get_config_enable_assertion_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit [255:0] config_enable_assertion_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_assertion_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog = function void dvc_axi4_set_config_enable_user_sideband_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_user_sideband_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog = function void dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_user_sideband_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_user_sideband_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog = function void dvc_axi4_set_config_enable_slave_exclusive_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_slave_exclusive_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog = function void dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_slave_exclusive_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_slave_exclusive_from_CY;

    import "DPI-C" context dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog = function void dvc_axi4_set_config_read_data_reordering_depth_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_read_data_reordering_depth_param
    );
    import "DPI-C" context dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog = function void dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_read_data_reordering_depth_param

    );
    export "DPI-C" function dvc_axi4_set_config_read_data_reordering_depth_from_CY;

    import "DPI-C" context dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1 = function void dvc_axi4_set_config_slave_start_addr_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input bit  config_slave_start_addr_param
    );
    import "DPI-C" context dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog = function void dvc_axi4_propagate_config_slave_start_addr_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_start_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_start_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_config_slave_start_addr_from_CY_index1;

    import "DPI-C" context dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1 = function void dvc_axi4_set_config_slave_end_addr_from_SystemVerilog_index1
    (
        input longint _iface_ref,
        input int unsigned _this_dot_1,
        input bit  config_slave_end_addr_param
    );
    import "DPI-C" context dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog = function void dvc_axi4_propagate_config_slave_end_addr_from_SystemVerilog
    (
        input longint _iface_ref    );
    import "DPI-C" context dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_deprecated_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    import "DPI-C" context dvc_axi4_get_config_slave_end_addr_into_SystemVerilog = function void dvc_axi4_get_config_slave_end_addr_into_SystemVerilog
    (
        input longint _iface_ref
    );
    export "DPI-C" function dvc_axi4_set_config_slave_end_addr_from_CY_index1;

    import "DPI-C" context dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog = function void dvc_axi4_set_config_max_transaction_time_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_transaction_time_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog = function void dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_transaction_time_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_transaction_time_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog = function void dvc_axi4_set_config_burst_timeout_factor_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_burst_timeout_factor_param
    );
    import "DPI-C" context dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog = function void dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_burst_timeout_factor_param

    );
    export "DPI-C" function dvc_axi4_set_config_burst_timeout_factor_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_RVALID_assertion_to_RREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_RVALID_assertion_to_RREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_BVALID_assertion_to_BREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_BVALID_assertion_to_BREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog = function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_SystemVerilog
    (
        input longint _iface_ref,
        input int unsigned config_max_latency_WVALID_assertion_to_WREADY_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog = function void dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog
    (
        input longint _iface_ref,
        output int unsigned config_max_latency_WVALID_assertion_to_WREADY_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY;

    import "DPI-C" context dvc_axi4_set_config_enable_qos_from_SystemVerilog = function void dvc_axi4_set_config_enable_qos_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_enable_qos_param
    );
    import "DPI-C" context dvc_axi4_get_config_enable_qos_into_SystemVerilog = function void dvc_axi4_get_config_enable_qos_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_enable_qos_param

    );
    export "DPI-C" function dvc_axi4_set_config_enable_qos_from_CY;

    import "DPI-C" context dvc_axi4_set_config_setup_time_from_SystemVerilog = function void dvc_axi4_set_config_setup_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_setup_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_setup_time_into_SystemVerilog = function void dvc_axi4_get_config_setup_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_setup_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_setup_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_hold_time_from_SystemVerilog = function void dvc_axi4_set_config_hold_time_from_SystemVerilog
    (
        input longint _iface_ref,
        input longint unsigned config_hold_time_param
    );
    import "DPI-C" context dvc_axi4_get_config_hold_time_into_SystemVerilog = function void dvc_axi4_get_config_hold_time_into_SystemVerilog
    (
        input longint _iface_ref,
        output longint unsigned config_hold_time_param

    );
    export "DPI-C" function dvc_axi4_set_config_hold_time_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_reads_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_reads_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_reads_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_reads_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog = function void dvc_axi4_set_config_num_max_outstanding_writes_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_max_outstanding_writes_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog = function void dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_max_outstanding_writes_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_max_outstanding_writes_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_wr_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_wr_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_wr_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_wr_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rd_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rd_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rd_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rd_from_CY;

    import "DPI-C" context dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog = function void dvc_axi4_set_config_max_outstanding_rw_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_max_outstanding_rw_param
    );
    import "DPI-C" context dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog = function void dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_max_outstanding_rw_param

    );
    export "DPI-C" function dvc_axi4_set_config_max_outstanding_rw_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_wr_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_wr_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_wr_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_wr_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog = function void dvc_axi4_set_config_num_outstanding_rd_phase_from_SystemVerilog
    (
        input longint _iface_ref,
        input int config_num_outstanding_rd_phase_param
    );
    import "DPI-C" context dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog = function void dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog
    (
        input longint _iface_ref,
        output int config_num_outstanding_rd_phase_param

    );
    export "DPI-C" function dvc_axi4_set_config_num_outstanding_rd_phase_from_CY;

    import "DPI-C" context dvc_axi4_set_config_is_issuing_from_SystemVerilog = function void dvc_axi4_set_config_is_issuing_from_SystemVerilog
    (
        input longint _iface_ref,
        input bit config_is_issuing_param
    );
    import "DPI-C" context dvc_axi4_get_config_is_issuing_into_SystemVerilog = function void dvc_axi4_get_config_is_issuing_into_SystemVerilog
    (
        input longint _iface_ref,
        output bit config_is_issuing_param

    );
    export "DPI-C" function dvc_axi4_set_config_is_issuing_from_CY;

    function void dvc_axi4_initialise_ACLK_from_CY();
        internal_ACLK = 'z;
        m_ACLK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARESETn_from_CY();
        internal_ARESETn = 'z;
        m_ARESETn = 'z;
    endfunction

    function void dvc_axi4_initialise_AWVALID_from_CY();
        internal_AWVALID = 'z;
        m_AWVALID = 'z;
    endfunction

    function void dvc_axi4_set_AWADDR_from_CY_index1( int _this_dot_1, logic  AWADDR_param );
        internal_AWADDR[_this_dot_1] = AWADDR_param;
    endfunction

    function void dvc_axi4_initialise_AWADDR_from_CY();
        internal_AWADDR = 'z;
        m_AWADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_AWPROT_from_CY();
        internal_AWPROT = 'z;
        m_AWPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREGION_from_CY();
        internal_AWREGION = 'z;
        m_AWREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLEN_from_CY();
        internal_AWLEN = 'z;
        m_AWLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_AWSIZE_from_CY();
        internal_AWSIZE = 'z;
        m_AWSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWBURST_from_CY();
        internal_AWBURST = 'z;
        m_AWBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_AWLOCK_from_CY();
        internal_AWLOCK = 'z;
        m_AWLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_AWCACHE_from_CY();
        internal_AWCACHE = 'z;
        m_AWCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_AWQOS_from_CY();
        internal_AWQOS = 'z;
        m_AWQOS = 'z;
    endfunction

    function void dvc_axi4_set_AWID_from_CY_index1( int _this_dot_1, logic  AWID_param );
        internal_AWID[_this_dot_1] = AWID_param;
    endfunction

    function void dvc_axi4_initialise_AWID_from_CY();
        internal_AWID = 'z;
        m_AWID = 'z;
    endfunction

    function void dvc_axi4_set_AWUSER_from_CY_index1( int _this_dot_1, logic  AWUSER_param );
        internal_AWUSER[_this_dot_1] = AWUSER_param;
    endfunction

    function void dvc_axi4_initialise_AWUSER_from_CY();
        internal_AWUSER = 'z;
        m_AWUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_AWREADY_from_CY();
        internal_AWREADY = 'z;
        m_AWREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_ARVALID_from_CY();
        internal_ARVALID = 'z;
        m_ARVALID = 'z;
    endfunction

    function void dvc_axi4_set_ARADDR_from_CY_index1( int _this_dot_1, logic  ARADDR_param );
        internal_ARADDR[_this_dot_1] = ARADDR_param;
    endfunction

    function void dvc_axi4_initialise_ARADDR_from_CY();
        internal_ARADDR = 'z;
        m_ARADDR = 'z;
    endfunction

    function void dvc_axi4_initialise_ARPROT_from_CY();
        internal_ARPROT = 'z;
        m_ARPROT = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREGION_from_CY();
        internal_ARREGION = 'z;
        m_ARREGION = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLEN_from_CY();
        internal_ARLEN = 'z;
        m_ARLEN = 'z;
    endfunction

    function void dvc_axi4_initialise_ARSIZE_from_CY();
        internal_ARSIZE = 'z;
        m_ARSIZE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARBURST_from_CY();
        internal_ARBURST = 'z;
        m_ARBURST = 'z;
    endfunction

    function void dvc_axi4_initialise_ARLOCK_from_CY();
        internal_ARLOCK = 'z;
        m_ARLOCK = 'z;
    endfunction

    function void dvc_axi4_initialise_ARCACHE_from_CY();
        internal_ARCACHE = 'z;
        m_ARCACHE = 'z;
    endfunction

    function void dvc_axi4_initialise_ARQOS_from_CY();
        internal_ARQOS = 'z;
        m_ARQOS = 'z;
    endfunction

    function void dvc_axi4_set_ARID_from_CY_index1( int _this_dot_1, logic  ARID_param );
        internal_ARID[_this_dot_1] = ARID_param;
    endfunction

    function void dvc_axi4_initialise_ARID_from_CY();
        internal_ARID = 'z;
        m_ARID = 'z;
    endfunction

    function void dvc_axi4_set_ARUSER_from_CY_index1( int _this_dot_1, logic  ARUSER_param );
        internal_ARUSER[_this_dot_1] = ARUSER_param;
    endfunction

    function void dvc_axi4_initialise_ARUSER_from_CY();
        internal_ARUSER = 'z;
        m_ARUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_ARREADY_from_CY();
        internal_ARREADY = 'z;
        m_ARREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_RVALID_from_CY();
        internal_RVALID = 'z;
        m_RVALID = 'z;
    endfunction

    function void dvc_axi4_set_RDATA_from_CY_index1( int _this_dot_1, logic  RDATA_param );
        internal_RDATA[_this_dot_1] = RDATA_param;
    endfunction

    function void dvc_axi4_initialise_RDATA_from_CY();
        internal_RDATA = 'z;
        m_RDATA = 'z;
    endfunction

    function void dvc_axi4_initialise_RRESP_from_CY();
        internal_RRESP = 'z;
        m_RRESP = 'z;
    endfunction

    function void dvc_axi4_initialise_RLAST_from_CY();
        internal_RLAST = 'z;
        m_RLAST = 'z;
    endfunction

    function void dvc_axi4_set_RID_from_CY_index1( int _this_dot_1, logic  RID_param );
        internal_RID[_this_dot_1] = RID_param;
    endfunction

    function void dvc_axi4_initialise_RID_from_CY();
        internal_RID = 'z;
        m_RID = 'z;
    endfunction

    function void dvc_axi4_set_RUSER_from_CY_index1( int _this_dot_1, logic  RUSER_param );
        internal_RUSER[_this_dot_1] = RUSER_param;
    endfunction

    function void dvc_axi4_initialise_RUSER_from_CY();
        internal_RUSER = 'z;
        m_RUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_RREADY_from_CY();
        internal_RREADY = 'z;
        m_RREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_WVALID_from_CY();
        internal_WVALID = 'z;
        m_WVALID = 'z;
    endfunction

    function void dvc_axi4_set_WDATA_from_CY_index1( int _this_dot_1, logic  WDATA_param );
        internal_WDATA[_this_dot_1] = WDATA_param;
    endfunction

    function void dvc_axi4_initialise_WDATA_from_CY();
        internal_WDATA = 'z;
        m_WDATA = 'z;
    endfunction

    function void dvc_axi4_set_WSTRB_from_CY_index1( int _this_dot_1, logic  WSTRB_param );
        internal_WSTRB[_this_dot_1] = WSTRB_param;
    endfunction

    function void dvc_axi4_initialise_WSTRB_from_CY();
        internal_WSTRB = 'z;
        m_WSTRB = 'z;
    endfunction

    function void dvc_axi4_initialise_WLAST_from_CY();
        internal_WLAST = 'z;
        m_WLAST = 'z;
    endfunction

    function void dvc_axi4_set_WUSER_from_CY_index1( int _this_dot_1, logic  WUSER_param );
        internal_WUSER[_this_dot_1] = WUSER_param;
    endfunction

    function void dvc_axi4_initialise_WUSER_from_CY();
        internal_WUSER = 'z;
        m_WUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_WREADY_from_CY();
        internal_WREADY = 'z;
        m_WREADY = 'z;
    endfunction

    function void dvc_axi4_initialise_BVALID_from_CY();
        internal_BVALID = 'z;
        m_BVALID = 'z;
    endfunction

    function void dvc_axi4_initialise_BRESP_from_CY();
        internal_BRESP = 'z;
        m_BRESP = 'z;
    endfunction

    function void dvc_axi4_set_BID_from_CY_index1( int _this_dot_1, logic  BID_param );
        internal_BID[_this_dot_1] = BID_param;
    endfunction

    function void dvc_axi4_initialise_BID_from_CY();
        internal_BID = 'z;
        m_BID = 'z;
    endfunction

    function void dvc_axi4_set_BUSER_from_CY_index1( int _this_dot_1, logic  BUSER_param );
        internal_BUSER[_this_dot_1] = BUSER_param;
    endfunction

    function void dvc_axi4_initialise_BUSER_from_CY();
        internal_BUSER = 'z;
        m_BUSER = 'z;
    endfunction

    function void dvc_axi4_initialise_BREADY_from_CY();
        internal_BREADY = 'z;
        m_BREADY = 'z;
    endfunction

    function void dvc_axi4_set_config_axi4lite_interface_from_CY( bit config_axi4lite_interface_param );
        config_axi4lite_interface = config_axi4lite_interface_param;
    endfunction

    function void dvc_axi4_set_config_interface_type_from_CY( axi4_interface_type_e config_interface_type_param );
        config_interface_type = config_interface_type_param;
    endfunction

    function void dvc_axi4_set_config_enable_rlast_from_CY( bit config_enable_rlast_param );
        config_enable_rlast = config_enable_rlast_param;
    endfunction

    function void dvc_axi4_set_config_enable_id_from_CY( bit config_enable_id_param );
        config_enable_id = config_enable_id_param;
    endfunction

    function void dvc_axi4_set_config_enable_all_assertions_from_CY( bit config_enable_all_assertions_param );
        config_enable_all_assertions = config_enable_all_assertions_param;
    endfunction

    function void dvc_axi4_set_config_enable_assertion_from_CY( bit [255:0] config_enable_assertion_param );
        config_enable_assertion = config_enable_assertion_param;
    endfunction

    function void dvc_axi4_set_config_enable_user_sideband_from_CY( bit config_enable_user_sideband_param );
        config_enable_user_sideband = config_enable_user_sideband_param;
    endfunction

    function void dvc_axi4_set_config_enable_slave_exclusive_from_CY( bit config_enable_slave_exclusive_param );
        config_enable_slave_exclusive = config_enable_slave_exclusive_param;
    endfunction

    function void dvc_axi4_set_config_read_data_reordering_depth_from_CY( int unsigned config_read_data_reordering_depth_param );
        config_read_data_reordering_depth = config_read_data_reordering_depth_param;
    endfunction

    function void dvc_axi4_set_config_slave_start_addr_from_CY_index1( int _this_dot_1, bit  config_slave_start_addr_param );
        config_slave_start_addr[_this_dot_1] = config_slave_start_addr_param;
    endfunction

    function void dvc_axi4_set_config_slave_end_addr_from_CY_index1( int _this_dot_1, bit  config_slave_end_addr_param );
        config_slave_end_addr[_this_dot_1] = config_slave_end_addr_param;
    endfunction

    function void dvc_axi4_set_config_max_transaction_time_factor_from_CY( int unsigned config_max_transaction_time_factor_param );
        config_max_transaction_time_factor = config_max_transaction_time_factor_param;
    endfunction

    function void dvc_axi4_set_config_burst_timeout_factor_from_CY( int unsigned config_burst_timeout_factor_param );
        config_burst_timeout_factor = config_burst_timeout_factor_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_AWVALID_assertion_to_AWREADY_from_CY( int unsigned config_max_latency_AWVALID_assertion_to_AWREADY_param );
        config_max_latency_AWVALID_assertion_to_AWREADY = config_max_latency_AWVALID_assertion_to_AWREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_ARVALID_assertion_to_ARREADY_from_CY( int unsigned config_max_latency_ARVALID_assertion_to_ARREADY_param );
        config_max_latency_ARVALID_assertion_to_ARREADY = config_max_latency_ARVALID_assertion_to_ARREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_RVALID_assertion_to_RREADY_from_CY( int unsigned config_max_latency_RVALID_assertion_to_RREADY_param );
        config_max_latency_RVALID_assertion_to_RREADY = config_max_latency_RVALID_assertion_to_RREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_BVALID_assertion_to_BREADY_from_CY( int unsigned config_max_latency_BVALID_assertion_to_BREADY_param );
        config_max_latency_BVALID_assertion_to_BREADY = config_max_latency_BVALID_assertion_to_BREADY_param;
    endfunction

    function void dvc_axi4_set_config_max_latency_WVALID_assertion_to_WREADY_from_CY( int unsigned config_max_latency_WVALID_assertion_to_WREADY_param );
        config_max_latency_WVALID_assertion_to_WREADY = config_max_latency_WVALID_assertion_to_WREADY_param;
    endfunction

    function void dvc_axi4_set_config_enable_qos_from_CY( bit config_enable_qos_param );
        config_enable_qos = config_enable_qos_param;
    endfunction

    function void dvc_axi4_set_config_setup_time_from_CY( longint unsigned config_setup_time_param );
        config_setup_time = config_setup_time_param;
    endfunction

    function void dvc_axi4_set_config_hold_time_from_CY( longint unsigned config_hold_time_param );
        config_hold_time = config_hold_time_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_reads_from_CY( int config_num_max_outstanding_reads_param );
        config_num_max_outstanding_reads = config_num_max_outstanding_reads_param;
    endfunction

    function void dvc_axi4_set_config_num_max_outstanding_writes_from_CY( int config_num_max_outstanding_writes_param );
        config_num_max_outstanding_writes = config_num_max_outstanding_writes_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_wr_from_CY( int config_max_outstanding_wr_param );
        config_max_outstanding_wr = config_max_outstanding_wr_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rd_from_CY( int config_max_outstanding_rd_param );
        config_max_outstanding_rd = config_max_outstanding_rd_param;
    endfunction

    function void dvc_axi4_set_config_max_outstanding_rw_from_CY( int config_max_outstanding_rw_param );
        config_max_outstanding_rw = config_max_outstanding_rw_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_wr_phase_from_CY( int config_num_outstanding_wr_phase_param );
        config_num_outstanding_wr_phase = config_num_outstanding_wr_phase_param;
    endfunction

    function void dvc_axi4_set_config_num_outstanding_rd_phase_from_CY( int config_num_outstanding_rd_phase_param );
        config_num_outstanding_rd_phase = config_num_outstanding_rd_phase_param;
    endfunction

    function void dvc_axi4_set_config_is_issuing_from_CY( bit config_is_issuing_param );
        config_is_issuing = config_is_issuing_param;
    endfunction


    //--------------------------------------------------------------------------
    //
    // Group:- TLM Interface Support
    //
    //--------------------------------------------------------------------------
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_addr = function axi4_get_temp_static_rw_transaction_addr;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_addr = function axi4_set_temp_static_rw_transaction_addr;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_id = function axi4_get_temp_static_rw_transaction_id;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_id = function axi4_set_temp_static_rw_transaction_id;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_addr_user_data = function axi4_get_temp_static_rw_transaction_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_addr_user_data = function axi4_set_temp_static_rw_transaction_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_words = function axi4_get_temp_static_rw_transaction_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_words = function axi4_set_temp_static_rw_transaction_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_write_strobes = function axi4_get_temp_static_rw_transaction_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_write_strobes = function axi4_set_temp_static_rw_transaction_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_wdata_user_data = function axi4_get_temp_static_rw_transaction_wdata_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_wdata_user_data = function axi4_set_temp_static_rw_transaction_wdata_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_resp = function axi4_get_temp_static_rw_transaction_resp;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_resp = function axi4_set_temp_static_rw_transaction_resp;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_resp_user_data = function axi4_get_temp_static_rw_transaction_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_resp_user_data = function axi4_set_temp_static_rw_transaction_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_valid_delay = function axi4_get_temp_static_rw_transaction_data_valid_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_valid_delay = function axi4_set_temp_static_rw_transaction_data_valid_delay;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_data_ready_delay = function axi4_get_temp_static_rw_transaction_data_ready_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_data_ready_delay = function axi4_set_temp_static_rw_transaction_data_ready_delay;
    export "DPI-C" dvc_axi4_get_temp_static_rw_transaction_write_data_beats_delay = function axi4_get_temp_static_rw_transaction_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_rw_transaction_write_data_beats_delay = function axi4_set_temp_static_rw_transaction_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr = function axi4_get_temp_static_read_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr = function axi4_set_temp_static_read_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_rresp_delays = function axi4_get_temp_static_read_rresp_delays;
    export "DPI-C" dvc_axi4_set_temp_static_read_rresp_delays = function axi4_set_temp_static_read_rresp_delays;
    export "DPI-C" dvc_axi4_get_temp_static_read_id = function axi4_get_temp_static_read_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_id = function axi4_set_temp_static_read_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_user_data = function axi4_get_temp_static_read_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_user_data = function axi4_set_temp_static_read_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_words = function axi4_get_temp_static_read_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_words = function axi4_set_temp_static_read_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_read_resp = function axi4_get_temp_static_read_resp;
    export "DPI-C" dvc_axi4_set_temp_static_read_resp = function axi4_set_temp_static_read_resp;
    export "DPI-C" dvc_axi4_get_temp_static_read_resp_user_data = function axi4_get_temp_static_read_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_resp_user_data = function axi4_set_temp_static_read_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_start_time = function axi4_get_temp_static_read_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_start_time = function axi4_set_temp_static_read_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_read_data_end_time = function axi4_get_temp_static_read_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_read_data_end_time = function axi4_set_temp_static_read_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr = function axi4_get_temp_static_write_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr = function axi4_set_temp_static_write_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_id = function axi4_get_temp_static_write_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_id = function axi4_set_temp_static_write_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_user_data = function axi4_get_temp_static_write_addr_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_user_data = function axi4_set_temp_static_write_addr_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_words = function axi4_get_temp_static_write_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_words = function axi4_set_temp_static_write_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_write_write_strobes = function axi4_get_temp_static_write_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_write_strobes = function axi4_set_temp_static_write_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_wdata_user_data = function axi4_get_temp_static_write_wdata_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_wdata_user_data = function axi4_set_temp_static_write_wdata_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_user_data = function axi4_get_temp_static_write_resp_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_user_data = function axi4_set_temp_static_write_resp_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_start_time = function axi4_get_temp_static_write_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_start_time = function axi4_set_temp_static_write_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_end_time = function axi4_get_temp_static_write_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_end_time = function axi4_set_temp_static_write_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_write_data_beats_delay = function axi4_get_temp_static_write_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_write_write_data_beats_delay = function axi4_set_temp_static_write_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_words = function axi4_get_temp_static_write_data_burst_data_words;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_words = function axi4_set_temp_static_write_data_burst_data_words;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_write_strobes = function axi4_get_temp_static_write_data_burst_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_write_strobes = function axi4_set_temp_static_write_data_burst_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_user_data = function axi4_get_temp_static_write_data_burst_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_user_data = function axi4_set_temp_static_write_data_burst_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_start_time = function axi4_get_temp_static_write_data_burst_data_start_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_start_time = function axi4_set_temp_static_write_data_burst_data_start_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_end_time = function axi4_get_temp_static_write_data_burst_data_end_time;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_end_time = function axi4_set_temp_static_write_data_burst_data_end_time;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_write_data_beats_delay = function axi4_get_temp_static_write_data_burst_write_data_beats_delay;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_write_data_beats_delay = function axi4_set_temp_static_write_data_burst_write_data_beats_delay;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data_valid2ready = function axi4_get_temp_static_write_data_burst_data_valid2ready;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data_valid2ready = function axi4_set_temp_static_write_data_burst_data_valid2ready;
    export "DPI-C" dvc_axi4_get_temp_static_write_data_burst_data2data_clock = function axi4_get_temp_static_write_data_burst_data2data_clock;
    export "DPI-C" dvc_axi4_set_temp_static_write_data_burst_data2data_clock = function axi4_set_temp_static_write_data_burst_data2data_clock;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_addr = function axi4_get_temp_static_read_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_addr = function axi4_set_temp_static_read_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_id = function axi4_get_temp_static_read_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_id = function axi4_set_temp_static_read_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_phase_addr_user = function axi4_get_temp_static_read_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_phase_addr_user = function axi4_set_temp_static_read_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_data = function axi4_get_temp_static_read_channel_phase_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_data = function axi4_set_temp_static_read_channel_phase_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_id = function axi4_get_temp_static_read_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_id = function axi4_set_temp_static_read_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_phase_user_data = function axi4_get_temp_static_read_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_phase_user_data = function axi4_set_temp_static_read_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_addr = function axi4_get_temp_static_write_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_addr = function axi4_set_temp_static_write_addr_channel_phase_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_id = function axi4_get_temp_static_write_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_id = function axi4_set_temp_static_write_addr_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_phase_addr_user = function axi4_get_temp_static_write_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_phase_addr_user = function axi4_set_temp_static_write_addr_channel_phase_addr_user;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_data = function axi4_get_temp_static_write_channel_phase_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_data = function axi4_set_temp_static_write_channel_phase_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_write_strobes = function axi4_get_temp_static_write_channel_phase_write_strobes;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_write_strobes = function axi4_set_temp_static_write_channel_phase_write_strobes;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_phase_user_data = function axi4_get_temp_static_write_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_phase_user_data = function axi4_set_temp_static_write_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_phase_id = function axi4_get_temp_static_write_resp_channel_phase_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_phase_id = function axi4_set_temp_static_write_resp_channel_phase_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_phase_user_data = function axi4_get_temp_static_write_resp_channel_phase_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_phase_user_data = function axi4_set_temp_static_write_resp_channel_phase_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_addr = function axi4_get_temp_static_read_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_addr = function axi4_set_temp_static_read_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_id = function axi4_get_temp_static_read_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_id = function axi4_set_temp_static_read_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_addr_channel_cycle_user_data = function axi4_get_temp_static_read_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_addr_channel_cycle_user_data = function axi4_set_temp_static_read_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_data = function axi4_get_temp_static_read_channel_cycle_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_data = function axi4_set_temp_static_read_channel_cycle_data;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_id = function axi4_get_temp_static_read_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_id = function axi4_set_temp_static_read_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_read_channel_cycle_user_data = function axi4_get_temp_static_read_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_read_channel_cycle_user_data = function axi4_set_temp_static_read_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_addr = function axi4_get_temp_static_write_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_addr = function axi4_set_temp_static_write_addr_channel_cycle_addr;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_id = function axi4_get_temp_static_write_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_id = function axi4_set_temp_static_write_addr_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_addr_channel_cycle_user_data = function axi4_get_temp_static_write_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_addr_channel_cycle_user_data = function axi4_set_temp_static_write_addr_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_data = function axi4_get_temp_static_write_channel_cycle_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_data = function axi4_set_temp_static_write_channel_cycle_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_strb = function axi4_get_temp_static_write_channel_cycle_strb;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_strb = function axi4_set_temp_static_write_channel_cycle_strb;
    export "DPI-C" dvc_axi4_get_temp_static_write_channel_cycle_user_data = function axi4_get_temp_static_write_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_channel_cycle_user_data = function axi4_set_temp_static_write_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_cycle_id = function axi4_get_temp_static_write_resp_channel_cycle_id;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_cycle_id = function axi4_set_temp_static_write_resp_channel_cycle_id;
    export "DPI-C" dvc_axi4_get_temp_static_write_resp_channel_cycle_user_data = function axi4_get_temp_static_write_resp_channel_cycle_user_data;
    export "DPI-C" dvc_axi4_set_temp_static_write_resp_channel_cycle_user_data = function axi4_set_temp_static_write_resp_channel_cycle_user_data;
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_rw_e read_or_write,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_rw_transaction_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_rw_e read_or_write,
        inout axi4_prot_e prot,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout int address_valid_delay,
        inout int write_response_valid_delay,
        inout int address_ready_delay,
        inout int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_ready_delay_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_rw_transaction_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_rw_e read_or_write,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output int address_valid_delay,
        output int write_response_valid_delay,
        output int address_ready_delay,
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_prot_e prot,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_read_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_prot_e prot,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int resp_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output longint addr_start_time,
        output longint addr_end_time,
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [3:0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [3:0] qos,
        inout bit [7:0] burst_length,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout axi4_response_e resp,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ActivatesActivatingActivate_open_SystemVerilog =
    task axi4_write_ActivatesActivatingActivate_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        inout axi4_prot_e prot,
        inout int bresp_delay,
        inout bit [((4) - 1):0] region,
        inout axi4_size_e size,
        inout axi4_burst_e burst,
        inout axi4_lock_e lock,
        inout axi4_cache_e cache,
        inout bit [((4) - 1):0] qos,
        inout bit [((8) - 1):0] burst_length,
        inout axi4_response_e resp,
        inout longint addr_start_time,
        inout longint addr_end_time,
        inout longint wr_resp_start_time,
        inout longint wr_resp_end_time,
        inout int address_valid_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int wdata_user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output int bresp_delay,
        output bit [((4) - 1):0] region,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output bit [((8) - 1):0] burst_length,
        output axi4_response_e resp,
        output longint addr_start_time,
        output longint addr_end_time,
        output longint wr_resp_start_time,
        output longint wr_resp_end_time,
        output int address_valid_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_SendSendingSent_SystemVerilog =
    task axi4_write_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int burst_length,
        input int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data_valid2ready_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int data2data_clock_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int wdata_burst_start_clock,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        inout int data_words_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int write_strobes_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int user_data_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_start_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_end_time_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data_valid2ready_DIMS0, // Array to pass in and/or out the unsized dims of param
        inout int data2data_clock_DIMS0, // Array to pass in and/or out the unsized dims of param
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_data_burst_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output int burst_length,
        output int wdata_burst_start_clock,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_data_burst_SendSendingSent_SystemVerilog =
    task axi4_read_data_burst_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_data_burst_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_read_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit last,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output bit last,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_prot_e prot,
        input bit [3:0] region,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input axi4_cache_e cache,
        input bit [3:0] qos,
        input int address_valid_delay,
        input int address_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_addr_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_prot_e prot,
        output bit [((4) - 1):0] region,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output axi4_cache_e cache,
        output bit [((4) - 1):0] qos,
        output int address_valid_delay,
        output int address_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit last,
        input int data_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit last,
        output int data_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_phase_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input int write_response_ready_delay,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_phase_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_resp_channel_phase_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output int write_response_ready_delay,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((4) - 1):0] region,
        output axi4_cache_e cache,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [((4) - 1):0] qos,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_read_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input bit last,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_read_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        output bit last,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_read_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_read_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit [3:0] region,
        input axi4_cache_e cache,
        input bit [7:0] burst_length,
        input axi4_size_e size,
        input axi4_burst_e burst,
        input axi4_lock_e lock,
        input bit [3:0] qos,
        input axi4_prot_e prot,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_addr_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit [((4) - 1):0] region,
        output axi4_cache_e cache,
        output bit [((8) - 1):0] burst_length,
        output axi4_size_e size,
        output axi4_burst_e burst,
        output axi4_lock_e lock,
        output bit [((4) - 1):0] qos,
        output axi4_prot_e prot,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_addr_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_addr_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit last,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output bit last,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_cycle_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input axi4_response_e resp,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_cycle_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output int _trans_id,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog =
    task axi4_write_resp_channel_cycle_ReceivedReceivingReceive_open_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input int _trans_id,
        output axi4_response_e resp,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog =
    task axi4_write_resp_channel_ready_SendSendingSent_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        input bit ready,
        input int _unit_id
    );
    import "DPI-C" context dvc_axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog =
    task axi4_write_resp_channel_ready_ReceivedReceivingReceive_SystemVerilog
    (
        input int _comms_semantic,
        input longint _as_end,
        output bit ready,
        input int _unit_id,
        input bit _using
    );
    import "DPI-C" context dvc_axi4_fn_get_raddr_phase_info = function void fn_get_raddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id,
        output int num_beats_rcvd
    );

    import "DPI-C" context dvc_axi4_fn_get_waddr_phase_info = function void fn_get_waddr_phase_info_C
    (
        inout bit valid,
        input bit [17:0] id
    );

    import "DPI-C" context dvc_axi4_fn_add_addr_map_entry = function void fn_add_addr_map_entry_C
    (
        input string region,
        input longint unsigned size
    );

    import "DPI-C" context dvc_axi4_fn_add_wr_delay = function void fn_add_wr_delay_C
    (
        input string region,
        input bit [17:0] id,
        input int unsigned addr2data,
        input int data2data_DIMS0
    );

    import "DPI-C" context dvc_axi4_fn_delete_wr_delay = function void fn_delete_wr_delay_C
    (
        input string region,
        input bit [17:0] id
    );

    import "DPI-C" context dvc_axi4_fn_drive_signals = function void fn_drive_signals_C
    (
        input string signal,
        input longint value
    );

    import "DPI-C" context dvc_axi4_fn_set_address_map_entry = function void fn_set_address_map_entry_C
    (
    );

    import "DPI-C" context dvc_axi4_fn_set_wr_def_delays = function void fn_set_wr_def_delays_C
    (
        input int unsigned min_addr2data,
        input int min_data2data_DIMS0,
        input int unsigned max_addr2data,
        input int max_data2data_DIMS0
    );

    import "DPI-C" context dvc_axi4_fn_print_perf_stats = function void fn_print_perf_stats_C
    (
    );

    // Waiter task and control
    reg sim_wait_for_control = 0;

    always @(posedge sim_wait_for_control)
    begin
        disable wait_for;
        sim_wait_for_control = 0;
    end

    export "DPI-C" dvc_axi4_wait_for = task wait_for;

    task wait_for();
        begin
            wait(0 == 1);
        end
    endtask

    // Drive wires (from Cohesive) 
    assign ACLK = internal_ACLK;
    assign ARESETn = internal_ARESETn;
    assign AWVALID = internal_AWVALID;
    assign AWADDR = internal_AWADDR;
    assign AWPROT = internal_AWPROT;
    assign AWREGION = internal_AWREGION;
    assign AWLEN = internal_AWLEN;
    assign AWSIZE = internal_AWSIZE;
    assign AWBURST = internal_AWBURST;
    assign AWLOCK = internal_AWLOCK;
    assign AWCACHE = internal_AWCACHE;
    assign AWQOS = internal_AWQOS;
    assign AWID = internal_AWID;
    assign AWUSER = internal_AWUSER;
    assign AWREADY = internal_AWREADY;
    assign ARVALID = internal_ARVALID;
    assign ARADDR = internal_ARADDR;
    assign ARPROT = internal_ARPROT;
    assign ARREGION = internal_ARREGION;
    assign ARLEN = internal_ARLEN;
    assign ARSIZE = internal_ARSIZE;
    assign ARBURST = internal_ARBURST;
    assign ARLOCK = internal_ARLOCK;
    assign ARCACHE = internal_ARCACHE;
    assign ARQOS = internal_ARQOS;
    assign ARID = internal_ARID;
    assign ARUSER = internal_ARUSER;
    assign ARREADY = internal_ARREADY;
    assign RVALID = internal_RVALID;
    assign RDATA = internal_RDATA;
    assign RRESP = internal_RRESP;
    assign RLAST = internal_RLAST;
    assign RID = internal_RID;
    assign RUSER = internal_RUSER;
    assign RREADY = internal_RREADY;
    assign WVALID = internal_WVALID;
    assign WDATA = internal_WDATA;
    assign WSTRB = internal_WSTRB;
    assign WLAST = internal_WLAST;
    assign WUSER = internal_WUSER;
    assign WREADY = internal_WREADY;
    assign BVALID = internal_BVALID;
    assign BRESP = internal_BRESP;
    assign BID = internal_BID;
    assign BUSER = internal_BUSER;
    assign BREADY = internal_BREADY;
    // Drive wires (from User) 
    assign ACLK = m_ACLK;
    assign ARESETn = m_ARESETn;
    assign AWVALID = m_AWVALID;
    assign AWADDR = m_AWADDR;
    assign AWPROT = m_AWPROT;
    assign AWREGION = m_AWREGION;
    assign AWLEN = m_AWLEN;
    assign AWSIZE = m_AWSIZE;
    assign AWBURST = m_AWBURST;
    assign AWLOCK = m_AWLOCK;
    assign AWCACHE = m_AWCACHE;
    assign AWQOS = m_AWQOS;
    assign AWID = m_AWID;
    assign AWUSER = m_AWUSER;
    assign AWREADY = m_AWREADY;
    assign ARVALID = m_ARVALID;
    assign ARADDR = m_ARADDR;
    assign ARPROT = m_ARPROT;
    assign ARREGION = m_ARREGION;
    assign ARLEN = m_ARLEN;
    assign ARSIZE = m_ARSIZE;
    assign ARBURST = m_ARBURST;
    assign ARLOCK = m_ARLOCK;
    assign ARCACHE = m_ARCACHE;
    assign ARQOS = m_ARQOS;
    assign ARID = m_ARID;
    assign ARUSER = m_ARUSER;
    assign ARREADY = m_ARREADY;
    assign RVALID = m_RVALID;
    assign RDATA = m_RDATA;
    assign RRESP = m_RRESP;
    assign RLAST = m_RLAST;
    assign RID = m_RID;
    assign RUSER = m_RUSER;
    assign RREADY = m_RREADY;
    assign WVALID = m_WVALID;
    assign WDATA = m_WDATA;
    assign WSTRB = m_WSTRB;
    assign WLAST = m_WLAST;
    assign WUSER = m_WUSER;
    assign WREADY = m_WREADY;
    assign BVALID = m_BVALID;
    assign BRESP = m_BRESP;
    assign BID = m_BID;
    assign BUSER = m_BUSER;
    assign BREADY = m_BREADY;

    reg ACLK_changed = 0;
    reg ARESETn_changed = 0;
    reg AWVALID_changed = 0;
    reg AWADDR_changed = 0;
    reg AWPROT_changed = 0;
    reg AWREGION_changed = 0;
    reg AWLEN_changed = 0;
    reg AWSIZE_changed = 0;
    reg AWBURST_changed = 0;
    reg AWLOCK_changed = 0;
    reg AWCACHE_changed = 0;
    reg AWQOS_changed = 0;
    reg AWID_changed = 0;
    reg AWUSER_changed = 0;
    reg AWREADY_changed = 0;
    reg ARVALID_changed = 0;
    reg ARADDR_changed = 0;
    reg ARPROT_changed = 0;
    reg ARREGION_changed = 0;
    reg ARLEN_changed = 0;
    reg ARSIZE_changed = 0;
    reg ARBURST_changed = 0;
    reg ARLOCK_changed = 0;
    reg ARCACHE_changed = 0;
    reg ARQOS_changed = 0;
    reg ARID_changed = 0;
    reg ARUSER_changed = 0;
    reg ARREADY_changed = 0;
    reg RVALID_changed = 0;
    reg RDATA_changed = 0;
    reg RRESP_changed = 0;
    reg RLAST_changed = 0;
    reg RID_changed = 0;
    reg RUSER_changed = 0;
    reg RREADY_changed = 0;
    reg WVALID_changed = 0;
    reg WDATA_changed = 0;
    reg WSTRB_changed = 0;
    reg WLAST_changed = 0;
    reg WUSER_changed = 0;
    reg WREADY_changed = 0;
    reg BVALID_changed = 0;
    reg BRESP_changed = 0;
    reg BID_changed = 0;
    reg BUSER_changed = 0;
    reg BREADY_changed = 0;
    reg config_axi4lite_interface_changed = 0;
    reg config_interface_type_changed = 0;
    reg config_enable_rlast_changed = 0;
    reg config_enable_id_changed = 0;
    reg config_enable_all_assertions_changed = 0;
    reg config_enable_assertion_changed = 0;
    reg config_enable_user_sideband_changed = 0;
    reg config_enable_slave_exclusive_changed = 0;
    reg config_read_data_reordering_depth_changed = 0;
    reg config_slave_start_addr_changed = 0;
    reg config_slave_end_addr_changed = 0;
    reg config_max_transaction_time_factor_changed = 0;
    reg config_burst_timeout_factor_changed = 0;
    reg config_max_latency_AWVALID_assertion_to_AWREADY_changed = 0;
    reg config_max_latency_ARVALID_assertion_to_ARREADY_changed = 0;
    reg config_max_latency_RVALID_assertion_to_RREADY_changed = 0;
    reg config_max_latency_BVALID_assertion_to_BREADY_changed = 0;
    reg config_max_latency_WVALID_assertion_to_WREADY_changed = 0;
    reg config_enable_qos_changed = 0;
    reg config_setup_time_changed = 0;
    reg config_hold_time_changed = 0;
    reg config_num_max_outstanding_reads_changed = 0;
    reg config_num_max_outstanding_writes_changed = 0;
    reg config_max_outstanding_wr_changed = 0;
    reg config_max_outstanding_rd_changed = 0;
    reg config_max_outstanding_rw_changed = 0;
    reg config_num_outstanding_wr_phase_changed = 0;
    reg config_num_outstanding_rd_phase_changed = 0;
    reg config_is_issuing_changed = 0;

    // SV wire change monitors

    function automatic void axi4_local_set_ACLK_from_SystemVerilog(  );
        dvc_axi4_set_ACLK_from_SystemVerilog( _interface_ref, ACLK); // DPI call to imported task
    endfunction

    always @( ACLK or posedge _check_t0_values )
    begin
        axi4_local_set_ACLK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARESETn_from_SystemVerilog(  );
        dvc_axi4_set_ARESETn_from_SystemVerilog( _interface_ref, ARESETn); // DPI call to imported task
    endfunction

    always @( ARESETn or posedge _check_t0_values )
    begin
        axi4_local_set_ARESETn_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWVALID_from_SystemVerilog(  );
        dvc_axi4_set_AWVALID_from_SystemVerilog( _interface_ref, AWVALID); // DPI call to imported task
    endfunction

    always @( AWVALID or posedge _check_t0_values )
    begin
        axi4_local_set_AWVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWADDR_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWADDR_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWADDR[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWADDR_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWADDR or posedge _check_t0_values )
    begin
        axi4_local_set_AWADDR_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWPROT_from_SystemVerilog(  );
        dvc_axi4_set_AWPROT_from_SystemVerilog( _interface_ref, AWPROT); // DPI call to imported task
    endfunction

    always @( AWPROT or posedge _check_t0_values )
    begin
        axi4_local_set_AWPROT_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWREGION_from_SystemVerilog(  );
        dvc_axi4_set_AWREGION_from_SystemVerilog( _interface_ref, AWREGION); // DPI call to imported task
    endfunction

    always @( AWREGION or posedge _check_t0_values )
    begin
        axi4_local_set_AWREGION_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWLEN_from_SystemVerilog(  );
        dvc_axi4_set_AWLEN_from_SystemVerilog( _interface_ref, AWLEN); // DPI call to imported task
    endfunction

    always @( AWLEN or posedge _check_t0_values )
    begin
        axi4_local_set_AWLEN_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWSIZE_from_SystemVerilog(  );
        dvc_axi4_set_AWSIZE_from_SystemVerilog( _interface_ref, AWSIZE); // DPI call to imported task
    endfunction

    always @( AWSIZE or posedge _check_t0_values )
    begin
        axi4_local_set_AWSIZE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWBURST_from_SystemVerilog(  );
        dvc_axi4_set_AWBURST_from_SystemVerilog( _interface_ref, AWBURST); // DPI call to imported task
    endfunction

    always @( AWBURST or posedge _check_t0_values )
    begin
        axi4_local_set_AWBURST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWLOCK_from_SystemVerilog(  );
        dvc_axi4_set_AWLOCK_from_SystemVerilog( _interface_ref, AWLOCK); // DPI call to imported task
    endfunction

    always @( AWLOCK or posedge _check_t0_values )
    begin
        axi4_local_set_AWLOCK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWCACHE_from_SystemVerilog(  );
        dvc_axi4_set_AWCACHE_from_SystemVerilog( _interface_ref, AWCACHE); // DPI call to imported task
    endfunction

    always @( AWCACHE or posedge _check_t0_values )
    begin
        axi4_local_set_AWCACHE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWQOS_from_SystemVerilog(  );
        dvc_axi4_set_AWQOS_from_SystemVerilog( _interface_ref, AWQOS); // DPI call to imported task
    endfunction

    always @( AWQOS or posedge _check_t0_values )
    begin
        axi4_local_set_AWQOS_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWID or posedge _check_t0_values )
    begin
        axi4_local_set_AWID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_AWUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,AWUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_AWUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( AWUSER or posedge _check_t0_values )
    begin
        axi4_local_set_AWUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_AWREADY_from_SystemVerilog(  );
        dvc_axi4_set_AWREADY_from_SystemVerilog( _interface_ref, AWREADY); // DPI call to imported task
    endfunction

    always @( AWREADY or posedge _check_t0_values )
    begin
        axi4_local_set_AWREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARVALID_from_SystemVerilog(  );
        dvc_axi4_set_ARVALID_from_SystemVerilog( _interface_ref, ARVALID); // DPI call to imported task
    endfunction

    always @( ARVALID or posedge _check_t0_values )
    begin
        axi4_local_set_ARVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARADDR_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ADDRESS_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARADDR_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARADDR[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARADDR_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARADDR or posedge _check_t0_values )
    begin
        axi4_local_set_ARADDR_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARPROT_from_SystemVerilog(  );
        dvc_axi4_set_ARPROT_from_SystemVerilog( _interface_ref, ARPROT); // DPI call to imported task
    endfunction

    always @( ARPROT or posedge _check_t0_values )
    begin
        axi4_local_set_ARPROT_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARREGION_from_SystemVerilog(  );
        dvc_axi4_set_ARREGION_from_SystemVerilog( _interface_ref, ARREGION); // DPI call to imported task
    endfunction

    always @( ARREGION or posedge _check_t0_values )
    begin
        axi4_local_set_ARREGION_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARLEN_from_SystemVerilog(  );
        dvc_axi4_set_ARLEN_from_SystemVerilog( _interface_ref, ARLEN); // DPI call to imported task
    endfunction

    always @( ARLEN or posedge _check_t0_values )
    begin
        axi4_local_set_ARLEN_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARSIZE_from_SystemVerilog(  );
        dvc_axi4_set_ARSIZE_from_SystemVerilog( _interface_ref, ARSIZE); // DPI call to imported task
    endfunction

    always @( ARSIZE or posedge _check_t0_values )
    begin
        axi4_local_set_ARSIZE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARBURST_from_SystemVerilog(  );
        dvc_axi4_set_ARBURST_from_SystemVerilog( _interface_ref, ARBURST); // DPI call to imported task
    endfunction

    always @( ARBURST or posedge _check_t0_values )
    begin
        axi4_local_set_ARBURST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARLOCK_from_SystemVerilog(  );
        dvc_axi4_set_ARLOCK_from_SystemVerilog( _interface_ref, ARLOCK); // DPI call to imported task
    endfunction

    always @( ARLOCK or posedge _check_t0_values )
    begin
        axi4_local_set_ARLOCK_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARCACHE_from_SystemVerilog(  );
        dvc_axi4_set_ARCACHE_from_SystemVerilog( _interface_ref, ARCACHE); // DPI call to imported task
    endfunction

    always @( ARCACHE or posedge _check_t0_values )
    begin
        axi4_local_set_ARCACHE_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARQOS_from_SystemVerilog(  );
        dvc_axi4_set_ARQOS_from_SystemVerilog( _interface_ref, ARQOS); // DPI call to imported task
    endfunction

    always @( ARQOS or posedge _check_t0_values )
    begin
        axi4_local_set_ARQOS_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARID or posedge _check_t0_values )
    begin
        axi4_local_set_ARID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_ARUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,ARUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_ARUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( ARUSER or posedge _check_t0_values )
    begin
        axi4_local_set_ARUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_ARREADY_from_SystemVerilog(  );
        dvc_axi4_set_ARREADY_from_SystemVerilog( _interface_ref, ARREADY); // DPI call to imported task
    endfunction

    always @( ARREADY or posedge _check_t0_values )
    begin
        axi4_local_set_ARREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RVALID_from_SystemVerilog(  );
        dvc_axi4_set_RVALID_from_SystemVerilog( _interface_ref, RVALID); // DPI call to imported task
    endfunction

    always @( RVALID or posedge _check_t0_values )
    begin
        axi4_local_set_RVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RDATA_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_RDATA_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RDATA_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RDATA[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RDATA_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RDATA or posedge _check_t0_values )
    begin
        axi4_local_set_RDATA_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RRESP_from_SystemVerilog(  );
        dvc_axi4_set_RRESP_from_SystemVerilog( _interface_ref, RRESP); // DPI call to imported task
    endfunction

    always @( RRESP or posedge _check_t0_values )
    begin
        axi4_local_set_RRESP_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RLAST_from_SystemVerilog(  );
        dvc_axi4_set_RLAST_from_SystemVerilog( _interface_ref, RLAST); // DPI call to imported task
    endfunction

    always @( RLAST or posedge _check_t0_values )
    begin
        axi4_local_set_RLAST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RID or posedge _check_t0_values )
    begin
        axi4_local_set_RID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_RUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,RUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_RUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( RUSER or posedge _check_t0_values )
    begin
        axi4_local_set_RUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_RREADY_from_SystemVerilog(  );
        dvc_axi4_set_RREADY_from_SystemVerilog( _interface_ref, RREADY); // DPI call to imported task
    endfunction

    always @( RREADY or posedge _check_t0_values )
    begin
        axi4_local_set_RREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WVALID_from_SystemVerilog(  );
        dvc_axi4_set_WVALID_from_SystemVerilog( _interface_ref, WVALID); // DPI call to imported task
    endfunction

    always @( WVALID or posedge _check_t0_values )
    begin
        axi4_local_set_WVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WDATA_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_WDATA_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_WDATA_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WDATA[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WDATA_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WDATA or posedge _check_t0_values )
    begin
        axi4_local_set_WDATA_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WSTRB_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( (AXI4_WDATA_WIDTH / 8) ); _this_dot_1++)
        begin
            dvc_axi4_set_WSTRB_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WSTRB[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WSTRB_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WSTRB or posedge _check_t0_values )
    begin
        axi4_local_set_WSTRB_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WLAST_from_SystemVerilog(  );
        dvc_axi4_set_WLAST_from_SystemVerilog( _interface_ref, WLAST); // DPI call to imported task
    endfunction

    always @( WLAST or posedge _check_t0_values )
    begin
        axi4_local_set_WLAST_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_WUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,WUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_WUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( WUSER or posedge _check_t0_values )
    begin
        axi4_local_set_WUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_WREADY_from_SystemVerilog(  );
        dvc_axi4_set_WREADY_from_SystemVerilog( _interface_ref, WREADY); // DPI call to imported task
    endfunction

    always @( WREADY or posedge _check_t0_values )
    begin
        axi4_local_set_WREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BVALID_from_SystemVerilog(  );
        dvc_axi4_set_BVALID_from_SystemVerilog( _interface_ref, BVALID); // DPI call to imported task
    endfunction

    always @( BVALID or posedge _check_t0_values )
    begin
        axi4_local_set_BVALID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BRESP_from_SystemVerilog(  );
        dvc_axi4_set_BRESP_from_SystemVerilog( _interface_ref, BRESP); // DPI call to imported task
    endfunction

    always @( BRESP or posedge _check_t0_values )
    begin
        axi4_local_set_BRESP_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BID_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_ID_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_BID_from_SystemVerilog_index1( _interface_ref, _this_dot_1,BID[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_BID_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( BID or posedge _check_t0_values )
    begin
        axi4_local_set_BID_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BUSER_from_SystemVerilog(  );
        begin
        for (int _this_dot_1= 0; _this_dot_1 < ( AXI4_USER_WIDTH ); _this_dot_1++)
        begin
            dvc_axi4_set_BUSER_from_SystemVerilog_index1( _interface_ref, _this_dot_1,BUSER[_this_dot_1]); // DPI call to imported task
          end/* 1 */
        end
        dvc_axi4_propagate_BUSER_from_SystemVerilog( _interface_ref); // DPI call to imported task
    endfunction

    always @( BUSER or posedge _check_t0_values )
    begin
        axi4_local_set_BUSER_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end

    function automatic void axi4_local_set_BREADY_from_SystemVerilog(  );
        dvc_axi4_set_BREADY_from_SystemVerilog( _interface_ref, BREADY); // DPI call to imported task
    endfunction

    always @( BREADY or posedge _check_t0_values )
    begin
        axi4_local_set_BREADY_from_SystemVerilog(); // Call to local task which flattens data as necessary
    end


    // CY wire and variable changed flag monitors

    always @(posedge ACLK_changed or posedge _check_t0_values )
    begin
        while (ACLK_changed == 1'b1)
        begin
            dvc_axi4_get_ACLK_into_SystemVerilog( _interface_ref, internal_ACLK ); // DPI call to imported task
            ACLK_changed = 1'b0;
            #0  #0 if ( ACLK !== internal_ACLK )
            begin
                axi4_local_set_ACLK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARESETn_changed or posedge _check_t0_values )
    begin
        while (ARESETn_changed == 1'b1)
        begin
            logic temp_ARESETn;

            dvc_axi4_get_ARESETn_into_SystemVerilog( _interface_ref, temp_ARESETn ); // DPI call to imported task
            internal_ARESETn = temp_ARESETn;
            ARESETn_changed = 1'b0;
            #0  #0 if ( ARESETn !== internal_ARESETn )
            begin
                axi4_local_set_ARESETn_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWVALID_changed or posedge _check_t0_values )
    begin
        while (AWVALID_changed == 1'b1)
        begin
            logic temp_AWVALID;

            dvc_axi4_get_AWVALID_into_SystemVerilog( _interface_ref, temp_AWVALID ); // DPI call to imported task
            internal_AWVALID = temp_AWVALID;
            AWVALID_changed = 1'b0;
            #0  #0 if ( AWVALID !== internal_AWVALID )
            begin
                axi4_local_set_AWVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWADDR_changed or posedge _check_t0_values )
    begin
        while (AWADDR_changed == 1'b1)
        begin
            dvc_axi4_get_AWADDR_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWADDR_changed = 1'b0;
            #0  #0 if ( AWADDR !== internal_AWADDR )
            begin
                axi4_local_set_AWADDR_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWPROT_changed or posedge _check_t0_values )
    begin
        while (AWPROT_changed == 1'b1)
        begin
            dvc_axi4_get_AWPROT_into_SystemVerilog( _interface_ref, internal_AWPROT ); // DPI call to imported task
            AWPROT_changed = 1'b0;
            #0  #0 if ( AWPROT !== internal_AWPROT )
            begin
                axi4_local_set_AWPROT_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWREGION_changed or posedge _check_t0_values )
    begin
        while (AWREGION_changed == 1'b1)
        begin
            dvc_axi4_get_AWREGION_into_SystemVerilog( _interface_ref, internal_AWREGION ); // DPI call to imported task
            AWREGION_changed = 1'b0;
            #0  #0 if ( AWREGION !== internal_AWREGION )
            begin
                axi4_local_set_AWREGION_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWLEN_changed or posedge _check_t0_values )
    begin
        while (AWLEN_changed == 1'b1)
        begin
            dvc_axi4_get_AWLEN_into_SystemVerilog( _interface_ref, internal_AWLEN ); // DPI call to imported task
            AWLEN_changed = 1'b0;
            #0  #0 if ( AWLEN !== internal_AWLEN )
            begin
                axi4_local_set_AWLEN_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWSIZE_changed or posedge _check_t0_values )
    begin
        while (AWSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_AWSIZE_into_SystemVerilog( _interface_ref, internal_AWSIZE ); // DPI call to imported task
            AWSIZE_changed = 1'b0;
            #0  #0 if ( AWSIZE !== internal_AWSIZE )
            begin
                axi4_local_set_AWSIZE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWBURST_changed or posedge _check_t0_values )
    begin
        while (AWBURST_changed == 1'b1)
        begin
            dvc_axi4_get_AWBURST_into_SystemVerilog( _interface_ref, internal_AWBURST ); // DPI call to imported task
            AWBURST_changed = 1'b0;
            #0  #0 if ( AWBURST !== internal_AWBURST )
            begin
                axi4_local_set_AWBURST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWLOCK_changed or posedge _check_t0_values )
    begin
        while (AWLOCK_changed == 1'b1)
        begin
            logic temp_AWLOCK;

            dvc_axi4_get_AWLOCK_into_SystemVerilog( _interface_ref, temp_AWLOCK ); // DPI call to imported task
            internal_AWLOCK = temp_AWLOCK;
            AWLOCK_changed = 1'b0;
            #0  #0 if ( AWLOCK !== internal_AWLOCK )
            begin
                axi4_local_set_AWLOCK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWCACHE_changed or posedge _check_t0_values )
    begin
        while (AWCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_AWCACHE_into_SystemVerilog( _interface_ref, internal_AWCACHE ); // DPI call to imported task
            AWCACHE_changed = 1'b0;
            #0  #0 if ( AWCACHE !== internal_AWCACHE )
            begin
                axi4_local_set_AWCACHE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWQOS_changed or posedge _check_t0_values )
    begin
        while (AWQOS_changed == 1'b1)
        begin
            dvc_axi4_get_AWQOS_into_SystemVerilog( _interface_ref, internal_AWQOS ); // DPI call to imported task
            AWQOS_changed = 1'b0;
            #0  #0 if ( AWQOS !== internal_AWQOS )
            begin
                axi4_local_set_AWQOS_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWID_changed or posedge _check_t0_values )
    begin
        while (AWID_changed == 1'b1)
        begin
            dvc_axi4_get_AWID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWID_changed = 1'b0;
            #0  #0 if ( AWID !== internal_AWID )
            begin
                axi4_local_set_AWID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWUSER_changed or posedge _check_t0_values )
    begin
        while (AWUSER_changed == 1'b1)
        begin
            dvc_axi4_get_AWUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            AWUSER_changed = 1'b0;
            #0  #0 if ( AWUSER !== internal_AWUSER )
            begin
                axi4_local_set_AWUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge AWREADY_changed or posedge _check_t0_values )
    begin
        while (AWREADY_changed == 1'b1)
        begin
            logic temp_AWREADY;

            dvc_axi4_get_AWREADY_into_SystemVerilog( _interface_ref, temp_AWREADY ); // DPI call to imported task
            internal_AWREADY = temp_AWREADY;
            AWREADY_changed = 1'b0;
            #0  #0 if ( AWREADY !== internal_AWREADY )
            begin
                axi4_local_set_AWREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARVALID_changed or posedge _check_t0_values )
    begin
        while (ARVALID_changed == 1'b1)
        begin
            logic temp_ARVALID;

            dvc_axi4_get_ARVALID_into_SystemVerilog( _interface_ref, temp_ARVALID ); // DPI call to imported task
            internal_ARVALID = temp_ARVALID;
            ARVALID_changed = 1'b0;
            #0  #0 if ( ARVALID !== internal_ARVALID )
            begin
                axi4_local_set_ARVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARADDR_changed or posedge _check_t0_values )
    begin
        while (ARADDR_changed == 1'b1)
        begin
            dvc_axi4_get_ARADDR_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARADDR_changed = 1'b0;
            #0  #0 if ( ARADDR !== internal_ARADDR )
            begin
                axi4_local_set_ARADDR_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARPROT_changed or posedge _check_t0_values )
    begin
        while (ARPROT_changed == 1'b1)
        begin
            dvc_axi4_get_ARPROT_into_SystemVerilog( _interface_ref, internal_ARPROT ); // DPI call to imported task
            ARPROT_changed = 1'b0;
            #0  #0 if ( ARPROT !== internal_ARPROT )
            begin
                axi4_local_set_ARPROT_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARREGION_changed or posedge _check_t0_values )
    begin
        while (ARREGION_changed == 1'b1)
        begin
            dvc_axi4_get_ARREGION_into_SystemVerilog( _interface_ref, internal_ARREGION ); // DPI call to imported task
            ARREGION_changed = 1'b0;
            #0  #0 if ( ARREGION !== internal_ARREGION )
            begin
                axi4_local_set_ARREGION_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARLEN_changed or posedge _check_t0_values )
    begin
        while (ARLEN_changed == 1'b1)
        begin
            dvc_axi4_get_ARLEN_into_SystemVerilog( _interface_ref, internal_ARLEN ); // DPI call to imported task
            ARLEN_changed = 1'b0;
            #0  #0 if ( ARLEN !== internal_ARLEN )
            begin
                axi4_local_set_ARLEN_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARSIZE_changed or posedge _check_t0_values )
    begin
        while (ARSIZE_changed == 1'b1)
        begin
            dvc_axi4_get_ARSIZE_into_SystemVerilog( _interface_ref, internal_ARSIZE ); // DPI call to imported task
            ARSIZE_changed = 1'b0;
            #0  #0 if ( ARSIZE !== internal_ARSIZE )
            begin
                axi4_local_set_ARSIZE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARBURST_changed or posedge _check_t0_values )
    begin
        while (ARBURST_changed == 1'b1)
        begin
            dvc_axi4_get_ARBURST_into_SystemVerilog( _interface_ref, internal_ARBURST ); // DPI call to imported task
            ARBURST_changed = 1'b0;
            #0  #0 if ( ARBURST !== internal_ARBURST )
            begin
                axi4_local_set_ARBURST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARLOCK_changed or posedge _check_t0_values )
    begin
        while (ARLOCK_changed == 1'b1)
        begin
            logic temp_ARLOCK;

            dvc_axi4_get_ARLOCK_into_SystemVerilog( _interface_ref, temp_ARLOCK ); // DPI call to imported task
            internal_ARLOCK = temp_ARLOCK;
            ARLOCK_changed = 1'b0;
            #0  #0 if ( ARLOCK !== internal_ARLOCK )
            begin
                axi4_local_set_ARLOCK_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARCACHE_changed or posedge _check_t0_values )
    begin
        while (ARCACHE_changed == 1'b1)
        begin
            dvc_axi4_get_ARCACHE_into_SystemVerilog( _interface_ref, internal_ARCACHE ); // DPI call to imported task
            ARCACHE_changed = 1'b0;
            #0  #0 if ( ARCACHE !== internal_ARCACHE )
            begin
                axi4_local_set_ARCACHE_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARQOS_changed or posedge _check_t0_values )
    begin
        while (ARQOS_changed == 1'b1)
        begin
            dvc_axi4_get_ARQOS_into_SystemVerilog( _interface_ref, internal_ARQOS ); // DPI call to imported task
            ARQOS_changed = 1'b0;
            #0  #0 if ( ARQOS !== internal_ARQOS )
            begin
                axi4_local_set_ARQOS_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARID_changed or posedge _check_t0_values )
    begin
        while (ARID_changed == 1'b1)
        begin
            dvc_axi4_get_ARID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARID_changed = 1'b0;
            #0  #0 if ( ARID !== internal_ARID )
            begin
                axi4_local_set_ARID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARUSER_changed or posedge _check_t0_values )
    begin
        while (ARUSER_changed == 1'b1)
        begin
            dvc_axi4_get_ARUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            ARUSER_changed = 1'b0;
            #0  #0 if ( ARUSER !== internal_ARUSER )
            begin
                axi4_local_set_ARUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge ARREADY_changed or posedge _check_t0_values )
    begin
        while (ARREADY_changed == 1'b1)
        begin
            logic temp_ARREADY;

            dvc_axi4_get_ARREADY_into_SystemVerilog( _interface_ref, temp_ARREADY ); // DPI call to imported task
            internal_ARREADY = temp_ARREADY;
            ARREADY_changed = 1'b0;
            #0  #0 if ( ARREADY !== internal_ARREADY )
            begin
                axi4_local_set_ARREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RVALID_changed or posedge _check_t0_values )
    begin
        while (RVALID_changed == 1'b1)
        begin
            logic temp_RVALID;

            dvc_axi4_get_RVALID_into_SystemVerilog( _interface_ref, temp_RVALID ); // DPI call to imported task
            internal_RVALID = temp_RVALID;
            RVALID_changed = 1'b0;
            #0  #0 if ( RVALID !== internal_RVALID )
            begin
                axi4_local_set_RVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RDATA_changed or posedge _check_t0_values )
    begin
        while (RDATA_changed == 1'b1)
        begin
            dvc_axi4_get_RDATA_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RDATA_changed = 1'b0;
            #0  #0 if ( RDATA !== internal_RDATA )
            begin
                axi4_local_set_RDATA_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RRESP_changed or posedge _check_t0_values )
    begin
        while (RRESP_changed == 1'b1)
        begin
            dvc_axi4_get_RRESP_into_SystemVerilog( _interface_ref, internal_RRESP ); // DPI call to imported task
            RRESP_changed = 1'b0;
            #0  #0 if ( RRESP !== internal_RRESP )
            begin
                axi4_local_set_RRESP_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RLAST_changed or posedge _check_t0_values )
    begin
        while (RLAST_changed == 1'b1)
        begin
            logic temp_RLAST;

            dvc_axi4_get_RLAST_into_SystemVerilog( _interface_ref, temp_RLAST ); // DPI call to imported task
            internal_RLAST = temp_RLAST;
            RLAST_changed = 1'b0;
            #0  #0 if ( RLAST !== internal_RLAST )
            begin
                axi4_local_set_RLAST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RID_changed or posedge _check_t0_values )
    begin
        while (RID_changed == 1'b1)
        begin
            dvc_axi4_get_RID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RID_changed = 1'b0;
            #0  #0 if ( RID !== internal_RID )
            begin
                axi4_local_set_RID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RUSER_changed or posedge _check_t0_values )
    begin
        while (RUSER_changed == 1'b1)
        begin
            dvc_axi4_get_RUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            RUSER_changed = 1'b0;
            #0  #0 if ( RUSER !== internal_RUSER )
            begin
                axi4_local_set_RUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge RREADY_changed or posedge _check_t0_values )
    begin
        while (RREADY_changed == 1'b1)
        begin
            logic temp_RREADY;

            dvc_axi4_get_RREADY_into_SystemVerilog( _interface_ref, temp_RREADY ); // DPI call to imported task
            internal_RREADY = temp_RREADY;
            RREADY_changed = 1'b0;
            #0  #0 if ( RREADY !== internal_RREADY )
            begin
                axi4_local_set_RREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WVALID_changed or posedge _check_t0_values )
    begin
        while (WVALID_changed == 1'b1)
        begin
            logic temp_WVALID;

            dvc_axi4_get_WVALID_into_SystemVerilog( _interface_ref, temp_WVALID ); // DPI call to imported task
            internal_WVALID = temp_WVALID;
            WVALID_changed = 1'b0;
            #0  #0 if ( WVALID !== internal_WVALID )
            begin
                axi4_local_set_WVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WDATA_changed or posedge _check_t0_values )
    begin
        while (WDATA_changed == 1'b1)
        begin
            dvc_axi4_get_WDATA_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WDATA_changed = 1'b0;
            #0  #0 if ( WDATA !== internal_WDATA )
            begin
                axi4_local_set_WDATA_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WSTRB_changed or posedge _check_t0_values )
    begin
        while (WSTRB_changed == 1'b1)
        begin
            dvc_axi4_get_WSTRB_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WSTRB_changed = 1'b0;
            #0  #0 if ( WSTRB !== internal_WSTRB )
            begin
                axi4_local_set_WSTRB_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WLAST_changed or posedge _check_t0_values )
    begin
        while (WLAST_changed == 1'b1)
        begin
            logic temp_WLAST;

            dvc_axi4_get_WLAST_into_SystemVerilog( _interface_ref, temp_WLAST ); // DPI call to imported task
            internal_WLAST = temp_WLAST;
            WLAST_changed = 1'b0;
            #0  #0 if ( WLAST !== internal_WLAST )
            begin
                axi4_local_set_WLAST_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WUSER_changed or posedge _check_t0_values )
    begin
        while (WUSER_changed == 1'b1)
        begin
            dvc_axi4_get_WUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            WUSER_changed = 1'b0;
            #0  #0 if ( WUSER !== internal_WUSER )
            begin
                axi4_local_set_WUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge WREADY_changed or posedge _check_t0_values )
    begin
        while (WREADY_changed == 1'b1)
        begin
            logic temp_WREADY;

            dvc_axi4_get_WREADY_into_SystemVerilog( _interface_ref, temp_WREADY ); // DPI call to imported task
            internal_WREADY = temp_WREADY;
            WREADY_changed = 1'b0;
            #0  #0 if ( WREADY !== internal_WREADY )
            begin
                axi4_local_set_WREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BVALID_changed or posedge _check_t0_values )
    begin
        while (BVALID_changed == 1'b1)
        begin
            logic temp_BVALID;

            dvc_axi4_get_BVALID_into_SystemVerilog( _interface_ref, temp_BVALID ); // DPI call to imported task
            internal_BVALID = temp_BVALID;
            BVALID_changed = 1'b0;
            #0  #0 if ( BVALID !== internal_BVALID )
            begin
                axi4_local_set_BVALID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BRESP_changed or posedge _check_t0_values )
    begin
        while (BRESP_changed == 1'b1)
        begin
            dvc_axi4_get_BRESP_into_SystemVerilog( _interface_ref, internal_BRESP ); // DPI call to imported task
            BRESP_changed = 1'b0;
            #0  #0 if ( BRESP !== internal_BRESP )
            begin
                axi4_local_set_BRESP_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BID_changed or posedge _check_t0_values )
    begin
        while (BID_changed == 1'b1)
        begin
            dvc_axi4_get_BID_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            BID_changed = 1'b0;
            #0  #0 if ( BID !== internal_BID )
            begin
                axi4_local_set_BID_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BUSER_changed or posedge _check_t0_values )
    begin
        while (BUSER_changed == 1'b1)
        begin
            dvc_axi4_get_BUSER_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            BUSER_changed = 1'b0;
            #0  #0 if ( BUSER !== internal_BUSER )
            begin
                axi4_local_set_BUSER_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge BREADY_changed or posedge _check_t0_values )
    begin
        while (BREADY_changed == 1'b1)
        begin
            logic temp_BREADY;

            dvc_axi4_get_BREADY_into_SystemVerilog( _interface_ref, temp_BREADY ); // DPI call to imported task
            internal_BREADY = temp_BREADY;
            BREADY_changed = 1'b0;
            #0  #0 if ( BREADY !== internal_BREADY )
            begin
                axi4_local_set_BREADY_from_SystemVerilog(  );
            end
        end
    end

    always @(posedge config_axi4lite_interface_changed or posedge _check_t0_values )
    begin
        if (config_axi4lite_interface_changed == 1'b1)
        begin
            dvc_axi4_get_config_axi4lite_interface_into_SystemVerilog( _interface_ref, config_axi4lite_interface ); // DPI call to imported task
            config_axi4lite_interface_changed = 1'b0;
        end
    end

    always @(posedge config_interface_type_changed or posedge _check_t0_values )
    begin
        if (config_interface_type_changed == 1'b1)
        begin
            dvc_axi4_get_config_interface_type_into_SystemVerilog( _interface_ref, config_interface_type ); // DPI call to imported task
            config_interface_type_changed = 1'b0;
        end
    end

    always @(posedge config_enable_rlast_changed or posedge _check_t0_values )
    begin
        if (config_enable_rlast_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_rlast_into_SystemVerilog( _interface_ref, config_enable_rlast ); // DPI call to imported task
            config_enable_rlast_changed = 1'b0;
        end
    end

    always @(posedge config_enable_id_changed or posedge _check_t0_values )
    begin
        if (config_enable_id_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_id_into_SystemVerilog( _interface_ref, config_enable_id ); // DPI call to imported task
            config_enable_id_changed = 1'b0;
        end
    end

    always @(posedge config_enable_all_assertions_changed or posedge _check_t0_values )
    begin
        if (config_enable_all_assertions_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_all_assertions_into_SystemVerilog( _interface_ref, config_enable_all_assertions ); // DPI call to imported task
            config_enable_all_assertions_changed = 1'b0;
        end
    end

    always @(posedge config_enable_assertion_changed or posedge _check_t0_values )
    begin
        if (config_enable_assertion_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_assertion_into_SystemVerilog( _interface_ref, config_enable_assertion ); // DPI call to imported task
            config_enable_assertion_changed = 1'b0;
        end
    end

    always @(posedge config_enable_user_sideband_changed or posedge _check_t0_values )
    begin
        if (config_enable_user_sideband_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_user_sideband_into_SystemVerilog( _interface_ref, config_enable_user_sideband ); // DPI call to imported task
            config_enable_user_sideband_changed = 1'b0;
        end
    end

    always @(posedge config_enable_slave_exclusive_changed or posedge _check_t0_values )
    begin
        if (config_enable_slave_exclusive_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_slave_exclusive_into_SystemVerilog( _interface_ref, config_enable_slave_exclusive ); // DPI call to imported task
            config_enable_slave_exclusive_changed = 1'b0;
        end
    end

    always @(posedge config_read_data_reordering_depth_changed or posedge _check_t0_values )
    begin
        if (config_read_data_reordering_depth_changed == 1'b1)
        begin
            dvc_axi4_get_config_read_data_reordering_depth_into_SystemVerilog( _interface_ref, config_read_data_reordering_depth ); // DPI call to imported task
            config_read_data_reordering_depth_changed = 1'b0;
        end
    end

    always @(posedge config_slave_start_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_start_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_start_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            config_slave_start_addr_changed = 1'b0;
        end
    end

    always @(posedge config_slave_end_addr_changed or posedge _check_t0_values )
    begin
        if (config_slave_end_addr_changed == 1'b1)
        begin
            dvc_axi4_get_config_slave_end_addr_into_SystemVerilog( _interface_ref ); // DPI call to imported task
            config_slave_end_addr_changed = 1'b0;
        end
    end

    always @(posedge config_max_transaction_time_factor_changed or posedge _check_t0_values )
    begin
        if (config_max_transaction_time_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_transaction_time_factor_into_SystemVerilog( _interface_ref, config_max_transaction_time_factor ); // DPI call to imported task
            config_max_transaction_time_factor_changed = 1'b0;
        end
    end

    always @(posedge config_burst_timeout_factor_changed or posedge _check_t0_values )
    begin
        if (config_burst_timeout_factor_changed == 1'b1)
        begin
            dvc_axi4_get_config_burst_timeout_factor_into_SystemVerilog( _interface_ref, config_burst_timeout_factor ); // DPI call to imported task
            config_burst_timeout_factor_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_AWVALID_assertion_to_AWREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_AWVALID_assertion_to_AWREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_AWVALID_assertion_to_AWREADY_into_SystemVerilog( _interface_ref, config_max_latency_AWVALID_assertion_to_AWREADY ); // DPI call to imported task
            config_max_latency_AWVALID_assertion_to_AWREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_ARVALID_assertion_to_ARREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_ARVALID_assertion_to_ARREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_ARVALID_assertion_to_ARREADY_into_SystemVerilog( _interface_ref, config_max_latency_ARVALID_assertion_to_ARREADY ); // DPI call to imported task
            config_max_latency_ARVALID_assertion_to_ARREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_RVALID_assertion_to_RREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_RVALID_assertion_to_RREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_RVALID_assertion_to_RREADY_into_SystemVerilog( _interface_ref, config_max_latency_RVALID_assertion_to_RREADY ); // DPI call to imported task
            config_max_latency_RVALID_assertion_to_RREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_BVALID_assertion_to_BREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_BVALID_assertion_to_BREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_BVALID_assertion_to_BREADY_into_SystemVerilog( _interface_ref, config_max_latency_BVALID_assertion_to_BREADY ); // DPI call to imported task
            config_max_latency_BVALID_assertion_to_BREADY_changed = 1'b0;
        end
    end

    always @(posedge config_max_latency_WVALID_assertion_to_WREADY_changed or posedge _check_t0_values )
    begin
        if (config_max_latency_WVALID_assertion_to_WREADY_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_latency_WVALID_assertion_to_WREADY_into_SystemVerilog( _interface_ref, config_max_latency_WVALID_assertion_to_WREADY ); // DPI call to imported task
            config_max_latency_WVALID_assertion_to_WREADY_changed = 1'b0;
        end
    end

    always @(posedge config_enable_qos_changed or posedge _check_t0_values )
    begin
        if (config_enable_qos_changed == 1'b1)
        begin
            dvc_axi4_get_config_enable_qos_into_SystemVerilog( _interface_ref, config_enable_qos ); // DPI call to imported task
            config_enable_qos_changed = 1'b0;
        end
    end

    always @(posedge config_setup_time_changed or posedge _check_t0_values )
    begin
        if (config_setup_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_setup_time_into_SystemVerilog( _interface_ref, config_setup_time ); // DPI call to imported task
            config_setup_time_changed = 1'b0;
        end
    end

    always @(posedge config_hold_time_changed or posedge _check_t0_values )
    begin
        if (config_hold_time_changed == 1'b1)
        begin
            dvc_axi4_get_config_hold_time_into_SystemVerilog( _interface_ref, config_hold_time ); // DPI call to imported task
            config_hold_time_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_reads_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_reads_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_reads_into_SystemVerilog( _interface_ref, config_num_max_outstanding_reads ); // DPI call to imported task
            config_num_max_outstanding_reads_changed = 1'b0;
        end
    end

    always @(posedge config_num_max_outstanding_writes_changed or posedge _check_t0_values )
    begin
        if (config_num_max_outstanding_writes_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_max_outstanding_writes_into_SystemVerilog( _interface_ref, config_num_max_outstanding_writes ); // DPI call to imported task
            config_num_max_outstanding_writes_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_wr_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_wr_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_wr_into_SystemVerilog( _interface_ref, config_max_outstanding_wr ); // DPI call to imported task
            config_max_outstanding_wr_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rd_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rd_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rd_into_SystemVerilog( _interface_ref, config_max_outstanding_rd ); // DPI call to imported task
            config_max_outstanding_rd_changed = 1'b0;
        end
    end

    always @(posedge config_max_outstanding_rw_changed or posedge _check_t0_values )
    begin
        if (config_max_outstanding_rw_changed == 1'b1)
        begin
            dvc_axi4_get_config_max_outstanding_rw_into_SystemVerilog( _interface_ref, config_max_outstanding_rw ); // DPI call to imported task
            config_max_outstanding_rw_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_wr_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_wr_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_wr_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_wr_phase ); // DPI call to imported task
            config_num_outstanding_wr_phase_changed = 1'b0;
        end
    end

    always @(posedge config_num_outstanding_rd_phase_changed or posedge _check_t0_values )
    begin
        if (config_num_outstanding_rd_phase_changed == 1'b1)
        begin
            dvc_axi4_get_config_num_outstanding_rd_phase_into_SystemVerilog( _interface_ref, config_num_outstanding_rd_phase ); // DPI call to imported task
            config_num_outstanding_rd_phase_changed = 1'b0;
        end
    end

    always @(posedge config_is_issuing_changed or posedge _check_t0_values )
    begin
        if (config_is_issuing_changed == 1'b1)
        begin
            dvc_axi4_get_config_is_issuing_into_SystemVerilog( _interface_ref, config_is_issuing ); // DPI call to imported task
            config_is_issuing_changed = 1'b0;
        end
    end



    // Sparse array of blocking control events
    event block_control[] = new[100];

    // Unblocks a blocked clock control thread by id
    function void unblock( int unsigned id );
    begin
        -> block_control[id];
    end
    endfunction
    export "DPI-C" dvc_axi4_unblock_SystemVerilog = function unblock;

    // Blocks a blocked clock control thread by id
    task automatic block( int unsigned id );
    begin
        event blocking_event ;
        if (id >= block_control.size())
        begin
            int newsize  = ( (  id / 100 ) + 1 ) * 100;
            block_control = new[newsize](block_control);
        end
        blocking_event = block_control[id];
        @ blocking_event;
    end
    endtask
    export "DPI-C" dvc_axi4_block_SystemVerilog = task block;


    function int is_call_back_registered(int cb_name);
        case( axi4_call_back_e'(cb_name) )
          AXI4_REPORTER_CB:
          begin
              return ( endPoint.size() > 0 ) ? 1 : 0;
          end
        endcase
    endfunction

    //--------------------------------------------------------------------------------
    // Task which blocks and outputs an error if the interface has not initialized properly
    //--------------------------------------------------------------------------------

    task _initialized();
        if (_interface_ref == 0)
        begin
            $display("Error: %m - Questa Verification IP failed to initialise. Please check questa_mvc.log for details");
            wait(_interface_ref!=0);
        end
    endtask

endinterface

`endif
